/*
Copyright (c)  2013   The Regents of the University of California
All Rights Reserved
 
Permission to copy, modify and distribute any part of this FlowMax for educational, research and non-profit purposes, without fee, and without a written agreement is hereby granted, provided that the above copyright notice, this paragraph and the following three paragraphs appear in all copies.
Those desiring to incorporate this FlowMax into commercial products or use for commercial purposes should contact the Technology Transfer Office, University of California, San Diego, 9500 Gilman Drive, Mail Code 0910, La Jolla, CA 92093-0910, Ph: (858) 534-5815, FAX: (858) 534-7345, E-MAIL:invent@ucsd.edu.
IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES, INCLUDING LOST PROFITS, ARISING OUT OF THE USE OF THIS FlowMax, EVEN IF THE UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
THE FlowMax PROVIDED HEREIN IS ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.  THE UNIVERSITY OF CALIFORNIA MAKES NO REPRESENTATIONS AND EXTENDS NO WARRANTIES OF ANY KIND, EITHER IMPLIED OR EXPRESS, 
INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE, OR THAT THE USE OF THE FlowMax WILL NOT INFRINGE ANY PATENT, TRADEMARK OR OTHER RIGHTS.
*/
package FlowMax;


import java.awt.BasicStroke;
import java.awt.Color;
import java.awt.Cursor;
import java.awt.Font;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Rectangle;
import java.awt.Shape;
import java.awt.image.BufferedImage;
import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Random;

import javax.imageio.ImageIO;
import javax.swing.JCheckBox;
import javax.swing.JPanel;

import common.Methods;

public class Counter extends JPanel{

	private FlowMax owner;
	
	private static final long serialVersionUID = 1L;
	ArrayList<Double> scores = new ArrayList<Double>();
	ArrayList<String> names = new ArrayList<String>();
	ArrayList<double[]> times = new ArrayList<double[]>();
	ArrayList<double[]> averages = new ArrayList<double[]>();
	ArrayList<double[]> sds = new ArrayList<double[]>();
	ArrayList<double[]> paramList = new ArrayList<double[]>();  //this includes the sensitivities
	ArrayList<double[]> paramList2 = new ArrayList<double[]>();  //this includes the sensitivities and is in hrs
	ArrayList<ArrayList<double[]>> clist = new ArrayList<ArrayList<double[]>>(); 
	int gens = 10;
	int size = 100;
	int time = 150;
	double max = -1;
	int v = 100; //number of curves to draw based on sampling parameters from the sensitivity range
	boolean transparency = true;
	boolean nameLbs = true;
	boolean axes = true;
	boolean totalOnly = false;
	boolean fcyton = true;
	BufferedImage img;
	boolean redraw = true;
	Random gen = new Random();
	private final int nparams;
	public Color dot_color = Color.red;
	boolean realSolutions = false;
	BufferedImage counterpic = null;
	
	public Counter(FlowMax owner)
	{
		this.owner = owner;
		nparams= owner.getNParams();

		counterpic = Methods.getImageFromResource("VisualizeCounts.png");

		
	}
	
	public void addParameters(double[] params2, String name, double score, ArrayList<double[]> candidates)
	{
		double[] params = Methods.duplicate(params2);
		double[] oparams = Methods.duplicate(params2);
		for(int i = 0; i < 8; i+=2)
		{
			double mu = Math.log(params[i])-0.5*Math.log(1+Math.pow(params[i+1]/params[i],2));
			double sigma = Math.sqrt(Math.log(1+Math.pow(params[i+1]/params[i],2)));
			double mu2 = Math.log(params[i]-params[i+nparams])-0.5*Math.log(1+Math.pow((params[i+1]-params[i+1+nparams])/(params[i]-params[i+nparams]),2));
			double sigma2 = Math.sqrt(Math.log(1+Math.pow((params[i+1]-params[i+1+nparams])/(params[i]-params[i+nparams]),2)));
			double mu3 = Math.log(params[i]+params[i+nparams*2])-0.5*Math.log(1+Math.pow((params[i+1]+params[i+1+nparams*2])/(params[i]+params[i+nparams*2]),2));
			double sigma3 = Math.sqrt(Math.log(1+Math.pow((params[i+1]+params[i+1+nparams*2])/(params[i]+params[i+nparams*2]),2)));
			params[i] = mu;
			params[i+1]  = sigma;
			params[i+nparams]= mu-mu2;
			params[i+1+nparams]= sigma-sigma2;
			params[i+nparams*2]= mu3-mu;
			params[i+1+nparams*2]= sigma3-sigma;
		}
		ArrayList<double[]> newcandidates = new ArrayList<double[]>();
		for(double[] p : candidates)
		{
			double[] p2 = Methods.duplicate(p);
			for(int i = 0; i < 8; i+=2)
			{
				double mu = Math.log(p[i])-0.5*Math.log(1+Math.pow(p[i+1]/p[i],2));
				double sigma = Math.sqrt(Math.log(1+Math.pow(p[i+1]/p[i],2)));
				p2[i] = mu;
				p2[i+1]=sigma;
			}
			newcandidates.add(p2);
		}
		names.add(name);
		paramList.add(params);
		paramList2.add(oparams);
		scores.add(score);
		clist.add(newcandidates);
		times.add(new double[0]);
		averages.add(new double[0]);
		sds.add(new double[0]);
	}
	
	public void addParameters(double[] params2, String name, double score,
			double[] averages2, double[] sds2, double[] times2,ArrayList<double[]> candidates) {
		double[] params = Methods.duplicate(params2);
		double[] oparams = Methods.duplicate(params2);
		for(int i = 0; i < 8; i+=2)
		{
			double mu = Math.log(params[i])-0.5*Math.log(1+Math.pow(params[i+1]/params[i],2));
			double sigma = Math.sqrt(Math.log(1+Math.pow(params[i+1]/params[i],2)));
			double mu2 = Math.log(params[i]-params[i+nparams])-0.5*Math.log(1+Math.pow((params[i+1]-params[i+1+nparams])/(params[i]-params[i+nparams]),2));
			double sigma2 = Math.sqrt(Math.log(1+Math.pow((params[i+1]-params[i+1+nparams])/(params[i]-params[i+nparams]),2)));
			double mu3 = Math.log(params[i]+params[i+nparams*2])-0.5*Math.log(1+Math.pow((params[i+1]+params[i+1+nparams*2])/(params[i]+params[i+nparams*2]),2));
			double sigma3 = Math.sqrt(Math.log(1+Math.pow((params[i+1]+params[i+1+nparams*2])/(params[i]+params[i+nparams*2]),2)));
			params[i] = mu;
			params[i+1]  = sigma;
			params[i+nparams]= mu-mu2;
			params[i+1+nparams]= sigma-sigma2;
			params[i+nparams*2]= mu3-mu;
			params[i+1+nparams*2]= sigma3-sigma;
		}
		ArrayList<double[]> newcandidates = new ArrayList<double[]>();
		for(double[] p : candidates)
		{
			double[] p2 = Methods.duplicate(p);
			for(int i = 0; i < 8; i+=2)
			{
				double mu = Math.log(p[i])-0.5*Math.log(1+Math.pow(p[i+1]/p[i],2));
				double sigma = Math.sqrt(Math.log(1+Math.pow(p[i+1]/p[i],2)));
				p2[i] = mu;
				p2[i+1]=sigma;
			}
			newcandidates.add(p2);
		}
		clist.add(newcandidates);
		names.add(name);
		paramList.add(params);
		paramList2.add(oparams);
		scores.add(score);
		times.add(times2);
		averages.add(averages2);
		sds.add(sds2);
	}
	
	/**
	 * Deprecated
	 * @param output
	 */
	public void saveImage(String output)
	{
		updateUI();
		try {
			ImageIO.write(img,"png", new File(output));
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
	}

	public void paintComponent(Graphics g)
	{
		super.paintComponent(g);
		Graphics2D g3 = (Graphics2D)g;
		if(names.size() == 0)
		{
			g3.drawImage(counterpic,0,0,this);
			return;
		}
		if(img==null || redraw)
		{
			int yspacer = size/50+20;
			int xoff = size/5+10;
			int yoff = yspacer;

			int textheight = 7+size/50;
			if(transparency)
				img = new BufferedImage(xoff+size+20,yoff+((size+yspacer)*paramList.size()+textheight*2),BufferedImage.TYPE_INT_ARGB);
			else
				img = new BufferedImage(xoff+size+20,yoff+((size+yspacer)*paramList.size()+textheight*2),BufferedImage.TYPE_INT_RGB);
			Graphics2D g2 = (Graphics2D) img.getGraphics();

			g2.setFont(new Font("Calibri",Font.PLAIN,textheight));
		//	g2.fillRect(0, 0, img.getWidth(), img.getHeight());
			Color[] colors = {Color.red,Color.blue,Color.green,Color.cyan,Color.darkGray,Color.yellow.darker().darker(),Color.magenta,Color.lightGray,Color.pink,Color.orange};
			Color[] colors2 = {Color.red,Color.blue,Color.green,Color.cyan,Color.darkGray,Color.yellow.darker().darker(),Color.magenta,Color.lightGray,Color.pink,Color.orange};

			for(int i = 0; i < paramList.size(); i++)
			{
				if(realSolutions)
					v=clist.get(i).size();
				if(transparency)
				{
					for(int j = 0; j < colors.length; j++)
					{
						int r = colors2[j].getRed();
						int gr = colors2[j].getGreen();
						int b = colors2[j].getBlue();
						colors[j] = new Color(r,gr,b,Math.min(Math.max(2,15*size/v),255));
					}
				}
				double[][][] allsolutions = new double[v][gens][size];
				for(int trial = 0; trial < v; trial++)
				{

					double[] ps = paramList.get(i);
					double[] params = new double[ps.length];
					if(realSolutions)
						params = clist.get(i).get(trial);
					else
					{
						for(int j = 0; j < nparams; j++)
						{
							params[j] = (ps[j]-ps[j+nparams])+gen.nextDouble()*(ps[j+nparams*2]+ps[j+nparams]);
						}
					}
					if(fcyton)
						allsolutions[trial]=Methods.computeCytons(params, size, gens, time)[0];
					else
						allsolutions[trial]=Methods.computeCytonsOld(params, size, gens, time)[0];
				}
				if(max <= 0)
				{
					for(int trial = 0; trial < v; trial++)
					{
						double[][] solution = allsolutions[trial];
						for(int t = 0; t < size; t++)
						{
							double lmax = 0;
							for(int j = 0; j < gens; j++)
								lmax+=solution[j][t];
							if(lmax > max)
								max=lmax;
						}
					}
					for(int j = 0; j < averages.get(i).length; j++)
					{
						double lmax = averages.get(i)[j]+sds.get(i)[j];
						if(lmax > max) max = lmax;
					}
				}
				Shape clip = g2.getClip();
				g2.setClip(new Rectangle(xoff,yoff,size,size));
				for(int trial = 0; trial < v; trial++)
				{
					double[][] solution = allsolutions[trial];
					for(int t = 0; t < size-1; t++)
					{
						//lets draw line segments for each
						double sum1 = 0;
						double sum2 = 0;
						for(int j = 0; j < gens; j++)
						{
							g2.setColor(colors[j%colors.length]);
							int x1 = xoff+t;
							int x2 = xoff+t+1;
							double c = solution[j][t];
							double c2 = solution[j][t+1];
							if(!totalOnly)
							{
								int y1 = yoff+(int)(size*(1.0-((double)c)/max));
								int y2 = yoff+(int)(size*(1.0-((double)c2)/max));
								g2.drawLine(x1, y1, x2, y2);
							}
							sum1+=c;
							sum2+=c2;
						}
						int x1 = xoff+t;
						int x2 = xoff+t+1;
						int y1 = yoff+(int)(size*(1.0-(sum1)/max));
						int y2 = yoff+(int)(size*(1.0-(sum2)/max));
						if(transparency)
							g2.setColor(new Color(0,0,0,Math.min(255,Math.max(2,15*size/v))));
						else
							g2.setColor(Color.black);
						g2.drawLine(x1, y1, x2, y2);
					}
				}
				g2.setClip(clip);
				//lets draw y axis and labels
				if(axes)
				{
					g2.setColor(Color.black);
					g2.drawLine(xoff,yoff,xoff,yoff+size);
					int reps = Math.min(10,3+size/80);
					
					for(int j = 0; j <= reps; j++)
					{
						//lets draw the y ticks
						int y = yoff+size-(j*size)/reps;
						g2.drawLine(xoff-2, y, xoff+2, y);
						g2.drawString(String.format("%2.2f",(j*(max/1000000.0))/reps), xoff-textheight*2, y);
					}
					drawRotatedText(-Math.PI/2,"Cells x10^6",12,yoff+size/2+textheight*3,g2);
				}
				if(nameLbs)g2.drawString(names.get(i),xoff,yoff-textheight);
				//lets draw the counts and error bars
				g2.setClip(new Rectangle(xoff,yoff,size,size));
				for(int t = 0; t < averages.get(i).length; t++)
				{
					double avg = averages.get(i)[t];
					double sd = sds.get(i)[t];
					double timepoint = times.get(i)[t];
					g2.setColor(dot_color.darker());
					g2.drawLine(xoff+(int)((size*timepoint)/time),yoff+(int)(size*(1.0-(avg/max))),xoff+(int)((size*timepoint)/time),yoff+(int)(size*(1.0-((avg+sd)/max))));
					g2.drawLine(xoff+(int)((size*timepoint)/time)-2,yoff+(int)(size*(1.0-((avg+sd)/max))),xoff+(int)((size*timepoint)/time)+2,yoff+(int)(size*(1.0-((avg+sd)/max))));
					g2.drawLine(xoff+(int)((size*timepoint)/time),yoff+(int)(size*(1.0-(avg/max))),xoff+(int)((size*timepoint)/time),yoff+(int)(size*(1.0-((avg-sd)/max))));
					g2.drawLine(xoff+(int)((size*timepoint)/time)-2,yoff+(int)(size*(1.0-((avg-sd)/max))),xoff+(int)((size*timepoint)/time)+2,yoff+(int)(size*(1.0-((avg-sd)/max))));
					g2.setColor(dot_color.brighter());
					g2.fillOval(xoff+(int)((size*timepoint)/time)-2, yoff+(int)(size*(1.0-(avg/max)))-2, 4, 4);
					g2.setColor(dot_color);
					g2.drawOval(xoff+(int)((size*timepoint)/time)-2, yoff+(int)(size*(1.0-(avg/max)))-2, 4, 4);
				}
				g2.setColor(Color.BLACK);
				g2.setClip(clip);
				yoff+=size+yspacer;
			}
			yoff-=yspacer;
			//lets draw x axis and labels
			if(axes)
			{
				g2.setColor(Color.black);
				
				g2.drawLine(xoff,yoff,xoff+size,yoff);
				int reps = Math.min(10,3+size/80);
				
				for(int j = 0; j <= reps; j++)
				{
					//lets draw the x ticks
					int x = xoff+(j*size)/reps;
					g2.drawLine(x, yoff-2, x, yoff+2);
					g2.drawString(String.format("%2.1f",(((double)(j*time))/reps)), x-10, yoff+textheight+2);
				}
				
				g2.drawString("Time (h)",xoff+size/2-textheight*2,yoff+textheight*2+1);
			}
			redraw = false;
		}
		g3.drawImage(img,0,0,this);
	}

	private void drawRotatedText(double theta, String string, int i, int j,
			Graphics2D g2) {
		g2.translate(i,j);
		g2.rotate(theta);
		g2.drawString(string,0,0);
		g2.rotate(-theta);
		g2.translate(-i, -j);
		
	}


}
