/*
Copyright (c)  2013   The Regents of the University of California
All Rights Reserved
 
Permission to copy, modify and distribute any part of this FlowMax for educational, research and non-profit purposes, without fee, and without a written agreement is hereby granted, provided that the above copyright notice, this paragraph and the following three paragraphs appear in all copies.
Those desiring to incorporate this FlowMax into commercial products or use for commercial purposes should contact the Technology Transfer Office, University of California, San Diego, 9500 Gilman Drive, Mail Code 0910, La Jolla, CA 92093-0910, Ph: (858) 534-5815, FAX: (858) 534-7345, E-MAIL:invent@ucsd.edu.
IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES, INCLUDING LOST PROFITS, ARISING OUT OF THE USE OF THIS FlowMax, EVEN IF THE UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
THE FlowMax PROVIDED HEREIN IS ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.  THE UNIVERSITY OF CALIFORNIA MAKES NO REPRESENTATIONS AND EXTENDS NO WARRANTIES OF ANY KIND, EITHER IMPLIED OR EXPRESS, 
INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE, OR THAT THE USE OF THE FlowMax WILL NOT INFRINGE ANY PATENT, TRADEMARK OR OTHER RIGHTS.
*/

package FlowMax;


import java.awt.Color;
import java.awt.Cursor;
import java.awt.Font;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.image.BufferedImage;
import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Random;

import javax.imageio.ImageIO;
import javax.swing.JCheckBox;
import javax.swing.JComboBox;
import javax.swing.JPanel;

import common.Methods;

public class Drawer extends JPanel{

	private FlowMax owner;
	
	private static final long serialVersionUID = 1L;
	ArrayList<Double> scores = new ArrayList<Double>();
	ArrayList<String> names = new ArrayList<String>();
	ArrayList<double[]> paramList = new ArrayList<double[]>();  //this includes the sensitivities
	ArrayList<double[]> paramList2 = new ArrayList<double[]>();  //this includes the sensitivities and is in hrs
	ArrayList<ArrayList<double[]>> clist = new ArrayList<ArrayList<double[]>>(); 
	int gens = 10;
	int size = 100;
	int time = 150;
	int maxPercent = 100;
	int v = 10; //number of curves to draw based on sampling parameters from the sensitivity range
	Color[] overlayColors = null;
	BufferedImage img;
	boolean redraw = true;
	Random gen = new Random();
	private final int nparams;
	private JCheckBox pCytonBox,lblbox,namebox,tbox, overlayBox,maxBox;
	private JComboBox drawBox;
	boolean realSolutions = true;
	BufferedImage drawerpic = null;
	
	public Drawer(FlowMax owner)
	{
		this.nparams = owner.getNParams();
		this.owner = owner;

		drawerpic = Methods.getImageFromResource("VisualizeParameters.png");
	}
	
	
	public void addParameters(double[] params2, String name, double score, ArrayList<double[]> candidates)
	{
		double[] params = Methods.duplicate(params2);
		double[] oparams = Methods.duplicate(params2);
		for(int i = 0; i < 8; i+=2)
		{
			double mu = Math.log(params[i])-0.5*Math.log(1+Math.pow(params[i+1]/params[i],2));
			double sigma = Math.sqrt(Math.log(1+Math.pow(params[i+1]/params[i],2)));
			double mu2 = Math.log(params[i]-params[i+nparams])-0.5*Math.log(1+Math.pow((params[i+1]-params[i+1+nparams])/(params[i]-params[i+nparams]),2));
			double sigma2 = Math.sqrt(Math.log(1+Math.pow((params[i+1]-params[i+1+nparams])/(params[i]-params[i+nparams]),2)));
			double mu3 = Math.log(params[i]+params[i+nparams*2])-0.5*Math.log(1+Math.pow((params[i+1]+params[i+1+nparams*2])/(params[i]+params[i+nparams*2]),2));
			double sigma3 = Math.sqrt(Math.log(1+Math.pow((params[i+1]+params[i+1+nparams*2])/(params[i]+params[i+nparams*2]),2)));
			params[i] = mu;
			params[i+1]  = sigma;
			params[i+nparams]= mu-mu2;
			params[i+1+nparams]= sigma-sigma2;
			params[i+nparams*2]= mu3-mu;
			params[i+1+nparams*2]= sigma3-sigma;
		}
		ArrayList<double[]> newcandidates = new ArrayList<double[]>();
		for(double[] p : candidates)
		{
			double[] p2 = Methods.duplicate(p);
			for(int i = 0; i < 8; i+=2)
			{
				double mu = Math.log(p[i])-0.5*Math.log(1+Math.pow(p[i+1]/p[i],2));
				double sigma = Math.sqrt(Math.log(1+Math.pow(p[i+1]/p[i],2)));
				p2[i] = mu;
				p2[i+1]=sigma;
			}
			newcandidates.add(p2);
		}
		names.add(name);
		paramList.add(params);
		paramList2.add(oparams);
		scores.add(score);
		clist.add(newcandidates);
	}
	
	
	
	
	public void setLblbox(JCheckBox lblbox) {
		this.lblbox = lblbox;
	}


	public void setNamebox(JCheckBox namebox) {
		this.namebox = namebox;
	}


	public void setTbox(JCheckBox tbox) {
		this.tbox = tbox;
	}
	
	public void setpCytonBox(JCheckBox box) {
		this.pCytonBox=box;
	}
	
	public void setOverlayBox(JCheckBox overlayBoxP) {
		this.overlayBox = overlayBoxP;
		
	}
	
	public void setDrawBox(JComboBox drawBox) {
		this.drawBox = drawBox;
		
	}
	
	public void setMaxPercent(int mp)
	{

		this.maxPercent = mp;
		if(maxPercent < 1)
		{
			maxPercent=1;
		}
	}
	
	public void setOverlayColors(Color[] colors)
	{
		this.overlayColors = colors;
	}
	


	public void setMaxBox(JCheckBox maxBoxP) {
		// TODO Auto-generated method stub
		maxBox=maxBoxP;
	}
	/**
	 * Deprecated
	 * @param output
	 */
	public void saveImage(String output)
	{
		updateUI();
		try {
			ImageIO.write(img,"png", new File(output));
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
	}
	
	private int a(int x)
	{
		return 20+x+size/10;
	}
	
	public void paintComponent(Graphics g)
	{
		super.paintComponent(g);
		Graphics2D g3 = (Graphics2D)g;
		if(names.size() == 0)
		{
			g3.drawImage(drawerpic,0,0,this);
			return;
		}
		
	
		
		if(img==null || redraw)
		{
			try {
				owner.setCursor(Cursor.getPredefinedCursor(Cursor.WAIT_CURSOR));
				redraw = false;
				int textheight = 8+size/50;
				int ybuffer = textheight*3;
				int gspace = textheight+2;  //space between graphs
				int vertical_multiplier = size+textheight;
				if(!lblbox.isSelected()) ybuffer = 1;
				int rightBuffer = Math.max(10, 150-size);
				img = new BufferedImage(size*7+rightBuffer+gspace*2,(paramList.size())*(vertical_multiplier)+ybuffer+textheight+5,BufferedImage.TYPE_INT_ARGB);
				Graphics2D g2 = (Graphics2D) img.getGraphics();
				//g2.setColor(Color.WHITE);
				//g2.fillRect(0,0,img.getWidth(),img.getHeight());

				if(paramList == null || names == null ||paramList.size() != names.size())
					return;
				//for each set of params
					double[][][]saved = new double[paramList.size()][v][];
					
					double m_div0 = 0;
					double m_div1 = 0;
					double m_die0 = 0;
					double m_die1 = 0;
					
					for(int ind = 0;  ind< paramList.size(); ind++)
					{
						double[] ps = paramList.get(ind);
						for(int rep = 0; rep < v; rep++)
						{
							double[] params = new double[ps.length];
							if(realSolutions)
								params = Methods.prefix(clist.get(ind).get(rep),Constants.nparams);
							else
							{
								for(int j = 0; j < nparams; j++)
								{
									params[j] = (ps[j]-ps[j+nparams])+gen.nextDouble()*(ps[j+nparams*2]+ps[j+nparams]);
								}
							}
							saved[ind][rep]=params;
							m_div0 = Math.max(m_div0,Methods.getLogNormalPDF(params[0],params[1],Math.exp(params[0]-params[1]*params[1])));
							m_div1 = Math.max(m_div1,Methods.getLogNormalPDF(params[2],params[3],Math.exp(params[2]-params[3]*params[3])));
							m_die0 = Math.max(m_die0,Methods.getLogNormalPDF(params[4],params[5],Math.exp(params[4]-params[5]*params[5])));
							m_die1 = Math.max(m_die1,Methods.getLogNormalPDF(params[6],params[7],Math.exp(params[6]-params[7]*params[7])));
						}
					}
				
				for(int ind = 0;  ind< paramList.size(); ind++)
				{
					int i = ind;
					if(overlayBox.isSelected())
						i = 0;
					int tb = textheight+5;  //top buffer
					g2.setColor(Color.BLACK);
					g2.setFont(new Font("Calibri",Font.PLAIN,textheight));
					g2.setColor(Color.black);
					//Drawleftmost box for Pfs
					g2.drawLine(a(0), i*(vertical_multiplier)+size+tb, a(0)+(size), i*(vertical_multiplier)+size+tb);  //bottom 
					g2.drawLine(a(0), i*(vertical_multiplier)+tb, a(0), i*(vertical_multiplier)+size+tb);  //left
					g2.drawLine(a(0), i*(vertical_multiplier)+tb, a(0)+(size), i*(vertical_multiplier)+tb); //top 
					g2.drawLine(a(0)+size, i*(vertical_multiplier)+tb, a(0)+size, i*(vertical_multiplier)+size+tb);  //right
					Color c1,c2,c3,c4,c5,c6;
					if(realSolutions)
						v = clist.get(i).size();
					if(!overlayBox.isSelected())
					{
						if(tbox.isSelected() && drawBox.getSelectedIndex()<2)
						{
							float a = Math.min(Math.max(2,30*size/v),255)/255f;
							c1 =new Color(0,0.5f,0,a);
							c2 = new Color(1f,0,0,a);
							c3 = new Color(0,0,1f,a);
							c4 = new Color(0.3f,0.3f,0.3f,a);
							c5 = new Color(0.1f,0.1f,0.1f,a);
							c6 = new Color(0.05f,0.05f,0.05f,a);
						}
						else
						{
							c1 =new Color(0,0.5f,0);
							c2 = new Color(1f,0,0);
							c3 = new Color(0,0,1f);
							c4 = new Color(0.3f,0.3f,0.3f);
							c5 = new Color(0.1f,0.1f,0.1f);
							c6 = new Color(0.05f,0.05f,0.05f);
						}
					}
					else
					{
						//we want to pick the colors such that they have high alpha and that each set of parameters has its own color instead of the default colors
						Color col = overlayColors[ind%overlayColors.length];
						float a =255;
						if(tbox.isSelected())
						{
							a= (Math.min(Math.max(2,(20*size)/v),255));
							c6 = new Color(col.getRed(),col.getGreen(),col.getBlue(),(int)Math.max(1, Math.min(255,a)));
						}
						else
							c6 =col;
						c1=c2=c3=c4=c5=c6;
					}
					
					double[][] Hdiv0 = new double[size*2][v];
					double[][] Hdiv1 = new double[size*2][v];
					double[][] Hdie0 = new double[size*2][v];
					double[][] Hdie1 = new double[size*2][v];
					double[][] oFs = new double[gens][v];
					for(int rep = 0; rep < v; rep++)
					{
						System.err.println("Drawing rep: "+rep);
						double[] params = saved[ind][rep];
						double[] pF = new double[gens];
						double pF0 = params[10];
						double[] ddcurve = new double[gens+1];
						ddcurve[0] = 1-pF0;
						double[]ddcdf = new double[gens+1];
						ddcdf[0] = 1-pF0;
						if(!pCytonBox.isSelected())
						{
							for(int j = 1; j <= gens; j++)
							{
								ddcurve[j] = pF0*(Methods.getGaussianCDF(params[8],params[9],j,j+1
										
										)/(1-Methods.getGaussianCDF(params[8],params[9],1)));
								ddcdf[j] = ddcdf[j-1]+ddcurve[j];
							}
						//	ddcdf[gens] = ddcdf[gens-1]+pF0*(Methods.getGaussianCDF(params[8],params[9],gens-1,gens)/(1-Methods.getGaussianCDF(params[8],params[9],1)));
							pF[0] = params[10];
							for(int j = 1; j< gens;j++)
							{
								if((1-ddcdf[j+1])/(1-ddcdf[j]) > 0)
									pF[j] = (1-ddcdf[j+1])/(1-ddcdf[j]);
								else
									pF[j] = 0;
							}
						}
						else
						{
							for(int j = 1; j <= gens; j++)
							{
								ddcurve[j] = pF0*(Methods.getGaussianCDF(params[8],params[9],j-1,j)/(1-Methods.getGaussianCDF(params[8],params[9],0)));
								ddcdf[j] = ddcdf[j-1]+ddcurve[j];
							}
							//ddcdf[gens] = ddcdf[gens-1]+pF0*(Methods.getGaussianCDF(params[8],params[9],gens-1,gens)/(1-Methods.getGaussianCDF(params[8],params[9],1)));
							pF[0] = params[10];
							for(int j = 1; j < gens; j++)
							{
								if((1-ddcdf[j])/(1-ddcdf[j-1]) > 0)
									pF[j] = (1-ddcdf[j])/(1-ddcdf[j-1]);
								else
									pF[j] = 0;
							}
						}
						if(drawBox.getSelectedIndex() >0)
						{
							//we will sample the model using a monte carlo approach
							//now we know how much to sample each

							int currentGen = 0;
							int cellCount = 10000;

							if(maxBox.isSelected())
							{
								m_div0 =m_div1=m_die0=m_die1=0;
							}
							while(currentGen < gens)
							{
								int divCount = 0;
								for(int cell = 0; cell < cellCount; cell++)
								{
									boolean respond = gen.nextDouble()<pF[currentGen];
									double Tdiv = Double.POSITIVE_INFINITY;
									double Tdie = Double.POSITIVE_INFINITY;
									
									if(respond)
									{
										if(currentGen ==0) Tdiv=Math.exp(params[0]+params[1]*gen.nextGaussian());
										else Tdiv=Math.exp(params[2]+params[3]*gen.nextGaussian());
									}
									if(!respond||!pCytonBox.isSelected())
									{
										if(currentGen ==0) Tdie=Math.exp(params[4]+params[5]*gen.nextGaussian());
										else Tdie=Math.exp(params[6]+params[7]*gen.nextGaussian());
									}
									
									if(Tdiv < Tdie)
									{
										oFs[currentGen][rep]++;
										divCount++;
										if(currentGen==0)
										{
											try{
											Hdiv0[(int)Math.round((Tdiv*(size*2))/(time))][rep]++;
											m_div0= Math.max(Hdiv0[(int)Math.round((Tdiv*(size*2))/(time))][rep],m_div0);
											}catch(Exception e){}
										}
										else
										{
											try{
											Hdiv1[(int)Math.round((Tdiv*(size*2))/(time))][rep]++;
											m_div1 = Math.max(Hdiv1[(int)Math.round((Tdiv*(size*2))/(time))][rep],m_div1);
											}catch(Exception e){}
										}

									}
									else
									{
										if(currentGen==0)
										{
											try{
											Hdie0[(int)Math.round((Tdie*(size*2))/(time))][rep]++;
											m_die0 = Math.max(Hdie0[(int)Math.round((Tdie*(size*2))/(time))][rep],m_die0);
											}catch(Exception e){}
										}
										else
										{
											try{
											Hdie1[(int)Math.round((Tdie*(size*2))/(time))][rep]++;
											m_die1 = Math.max(Hdie1[(int)Math.round((Tdie*(size*2))/(time))][rep],m_die1);
											}catch(Exception e){}
										}
									}
								}
								if(maxBox.isSelected() && drawBox.getSelectedIndex() == 2)
								{
									for(int j = 0; j< size*2; j++)
									{
										if(m_div0 > 0)
										Hdiv0[j][rep]/=m_div0;
										if(m_div1 > 0)
										Hdiv1[j][rep]/=m_div1;
										if(m_die0 > 0)
										Hdie0[j][rep]/=m_die0;
										if(m_die1 > 0)
										Hdie1[j][rep]/=m_die1;
									}
								}
								oFs[currentGen][rep]/=cellCount;
								currentGen++;
								cellCount=divCount*2;
							}
							
							if(drawBox.getSelectedIndex() == 1)
							{
								double odiv0,odiv1,odie0,odie1;
								odiv0=odiv1=odie0=odie1=0;
			
								for(int j = 1; j < size*2; j++)
								{
									double xdiv0 = Hdiv0[j][rep];
									double xdiv1 = Hdiv1[j][rep];
									double xdie0 = Hdie0[j][rep];
									double xdie1 = Hdie1[j][rep];
									g2.setColor(c1);
									g2.drawLine((int)(a(2)+j+size+gspace), (int)(i*(vertical_multiplier)+size/2-(xdiv0/m_div0)*size/2)+tb,(int)(a(2)+j-1+size)+gspace, (int)(i*(vertical_multiplier)+size/2-(odiv0/m_div0)*size/2)+tb);
									g2.setColor(c2);
									g2.drawLine((int)(a(2)+j+size)+gspace, (int)(i*(vertical_multiplier)+size/2+(xdie0/m_die0)*size/2)+tb,(int)(a(2)+j-1+size)+gspace, (int)(i*(vertical_multiplier)+size/2+(odie0/m_die0)*size/2)+tb);
									g2.setColor(c3);
									g2.drawLine((int)(a(4)+j+size*3)+gspace*2, (int)(i*(vertical_multiplier)+size/2-(xdiv1/m_div1)*size/2)+tb,(int)(a(4)+j-1+size*3)+gspace*2, (int)(i*(vertical_multiplier)+size/2-(odiv1/m_div1)*size/2)+tb);
									g2.setColor(c4);					
									g2.drawLine((int)(a(4)+j+size*3)+gspace*2, (int)(i*(vertical_multiplier)+size/2+(xdie1/m_die1)*size/2)+tb,(int)(a(4)+j-1+size*3)+gspace*2, (int)(i*(vertical_multiplier)+size/2+(odie1/m_die1)*size/2)+tb);
									odiv0=xdiv0;
									odiv1=xdiv1;
									odie0=xdie0;
									odie1=xdie1;
								}
								for(int j = 0; j < gens; j++)
								{
									//double val = pF[j];
									//g2.drawLine(20, i*(vertical_multiplier0), 20, i*(vertical_multiplier0)+size);
									g2.setColor(c5);			
									if(j > 0) g2.drawLine((int)(a(0)+((j-1)*size)/(gens-1)-1),(int)( i*(vertical_multiplier)+size-(size*(((double)oFs[j-1][rep]))*(100.0/maxPercent))-1+tb),(int)(a(0)+((j)*size)/(gens-1)-1),(int)( i*(vertical_multiplier)+size-(size*(((double)oFs[j][rep]))*(100.0/maxPercent))-1)+tb);
									if(j > 0)
									{
										g2.setColor(c5);
										g2.fillOval((int)(a(0)+((j)*size)/(gens-1)-2),(int)( i*(vertical_multiplier)+size-(size*(((double)oFs[j][rep]))*(100.0/maxPercent))-2)+tb, 5, 5);
									}
									else
									{
										g2.setColor(c5);			
										g2.fillOval((int)(a(0)+((j)*size)/(gens-1)-3),(int)( i*(vertical_multiplier)+size-(size*(((double)oFs[j][rep]))*(100.0/maxPercent))-3)+tb, 7, 7);
									}
								}
							}
							else if (rep == v-1)
							{
								double[][] aoFs = new double[gens][2];
								m_div0 =m_die0=m_div1=m_die1 = 0;
								for(int j = 0; j < size*2; j++)
								{
									double xdiv0std = Methods.std(Hdiv0[j]);
									double xdiv1std = Methods.std(Hdiv1[j]);
									double xdie0std =  Methods.std(Hdie0[j]);
									double xdie1std =  Methods.std(Hdie1[j]);
									double xdiv0 =  Methods.average(Hdiv0[j]);
									double xdiv1 = Methods.average(Hdiv1[j]);
									double xdie0 = Methods.average(Hdie0[j]); 
									double xdie1 = Methods.average(Hdie1[j]); 
									if(xdiv0std+xdiv0 > m_div0) m_div0 = xdiv0std+xdiv0;
									if(xdiv1std+xdiv1 > m_div1) m_div1 = xdiv1std+xdiv1;
									if(xdie0std+xdie0 > m_die0) m_die0 = xdie0std+xdie0;
									if(xdie1std+xdie1 > m_die1) m_die1 = xdie1std+xdie1;
								}
								for(int j = 0; j < size*2; j++)
								{
									double xdiv0std = Methods.std(Hdiv0[j]);
									double xdiv1std = Methods.std(Hdiv1[j]);
									double xdie0std =  Methods.std(Hdie0[j]);
									double xdie1std =  Methods.std(Hdie1[j]);

									double xdiv0 =  Methods.average(Hdiv0[j]);
									double xdiv1 = Methods.average(Hdiv1[j]);
									double xdie0 = Methods.average(Hdie0[j]); 
									double xdie1 = Methods.average(Hdie1[j]); 
									g2.setColor(c1);
									g2.drawLine((int)(a(2)+j+size+gspace), (int)(i*(vertical_multiplier)+size/2-(Math.max(0,xdiv0-xdiv0std)/m_div0)*size/2)+tb,(int)(a(2)+j+size+gspace), (int)(i*(vertical_multiplier)+size/2-((xdiv0+xdiv0std)/m_div0)*size/2)+tb);
									g2.setColor(c5);
									g2.drawLine((int)(a(2)+j+size+gspace), (int)(i*(vertical_multiplier)+size/2-((xdiv0)/m_div0)*size/2)+tb,(int)(a(2)+j+size+gspace), (int)(i*(vertical_multiplier)+size/2-((xdiv0)/m_div0)*size/2)+tb);
									g2.setColor(c2);
									g2.drawLine((int)(a(2)+j+size)+gspace, (int)(i*(vertical_multiplier)+size/2+(Math.max(0,xdie0-xdie0std)/m_die0)*size/2)+tb,(int)(a(2)+j+size)+gspace, (int)(i*(vertical_multiplier)+size/2+((xdie0+xdie0std)/m_die0)*size/2)+tb);
									g2.setColor(c5);
									g2.drawLine((int)(a(2)+j+size)+gspace, (int)(i*(vertical_multiplier)+size/2+((xdie0)/m_die0)*size/2)+tb,(int)(a(2)+j+size)+gspace, (int)(i*(vertical_multiplier)+size/2+((xdie0)/m_die0)*size/2)+tb);
									g2.setColor(c3);
									g2.drawLine((int)(a(4)+j+size*3)+gspace*2, (int)(i*(vertical_multiplier)+size/2-(Math.max(0,xdiv1-xdiv1std)/m_div1)*size/2)+tb,(int)(a(4)+j+size*3)+gspace*2, (int)(i*(vertical_multiplier)+size/2-((xdiv1+xdiv1std)/m_div1)*size/2)+tb);
									g2.setColor(c5);
									g2.drawLine((int)(a(4)+j+size*3)+gspace*2, (int)(i*(vertical_multiplier)+size/2-((xdiv1)/m_div1)*size/2)+tb,(int)(a(4)+j+size*3)+gspace*2, (int)(i*(vertical_multiplier)+size/2-((xdiv1)/m_div1)*size/2)+tb);
									g2.setColor(c4);					
									g2.drawLine((int)(a(4)+j+size*3)+gspace*2, (int)(i*(vertical_multiplier)+size/2+(Math.max(0,xdie1-xdie1std)/m_die1)*size/2)+tb,(int)(a(4)+j+size*3)+gspace*2, (int)(i*(vertical_multiplier)+size/2+((xdie1+xdie1std)/m_die1)*size/2)+tb);
									g2.setColor(c5);
									g2.drawLine((int)(a(4)+j+size*3)+gspace*2, (int)(i*(vertical_multiplier)+size/2+((xdie1)/m_die1)*size/2)+tb,(int)(a(4)+j+size*3)+gspace*2, (int)(i*(vertical_multiplier)+size/2+((xdie1)/m_die1)*size/2)+tb);
								}
								
								for(int j = 0; j < gens; j++)
								{
									aoFs[j][0] = Methods.average(oFs[j]);
									aoFs[j][1] = Methods.std(oFs[j]);
								}
								g2.setColor(c5);			

								for(int j = 0; j < gens; j++)
								{
									//double val = pF[j];
									//g2.drawLine(20, i*(vertical_multiplier0), 20, i*(vertical_multiplier0)+size);
									if(j > 0)
									{
										
										g2.fillOval((int)(a(0)+((j)*size)/(gens-1)-1)-2,(int)( i*(vertical_multiplier)+size-(size*((aoFs[j][0]))*(100.0/maxPercent))-1+tb)-2,5,5);
										g2.drawLine((int)(a(0)+((j)*size)/(gens-1)-1),(int)( i*(vertical_multiplier)+size-(size*((aoFs[j][0]-aoFs[j][1]))*(100.0/maxPercent))-1+tb),(int)(a(0)+((j-1)*size)/(gens-1)-1),(int)( i*(vertical_multiplier)+size-(size*((aoFs[j-1][0]-aoFs[j-1][1]))*(100.0/maxPercent))-1)+tb);
										g2.drawLine((int)(a(0)+((j)*size)/(gens-1)-1),(int)( i*(vertical_multiplier)+size-(size*((aoFs[j][0]))*(100.0/maxPercent))-1+tb),(int)(a(0)+((j-1)*size)/(gens-1)-1),(int)( i*(vertical_multiplier)+size-(size*((aoFs[j-1][0]))*(100.0/maxPercent))-1)+tb);
										g2.drawLine((int)(a(0)+((j)*size)/(gens-1)-1),(int)( i*(vertical_multiplier)+size-(size*((aoFs[j][0]+aoFs[j][1]))*(100.0/maxPercent))-1+tb),(int)(a(0)+((j-1)*size)/(gens-1)-1),(int)( i*(vertical_multiplier)+size-(size*((aoFs[j-1][0]+aoFs[j-1][1]))*(100.0/maxPercent))-1)+tb);

									}
									else
									{

										g2.fillOval((int)(a(0)+((j)*size)/(gens-1)-1)-3,(int)( i*(vertical_multiplier)+size-(size*((aoFs[j][0]))*(100.0/maxPercent))-1+tb)-3,7,7);
										g2.drawLine((int)(a(0)+((j)*size)/(gens-1)-1),(int)( i*(vertical_multiplier)+size-(size*((aoFs[j][0]-aoFs[j][1]))*(100.0/maxPercent))-1+tb),(int)(a(0)+((j)*size)/(gens-1)-1),(int)( i*(vertical_multiplier)+size-(size*((aoFs[j][0]+aoFs[j][1]))*(100.0/maxPercent))-1)+tb);
									}
								}
							}
						} //drawing the distributions directly
						else if(drawBox.getSelectedIndex() == 0)
						{
							if(maxBox.isSelected())
							{
								m_div0 = Methods.getLogNormalPDF(params[0],params[1],Math.exp(params[0]-params[1]*params[1]));
								m_div1 = Methods.getLogNormalPDF(params[2],params[3],Math.exp(params[2]-params[3]*params[3]));
								m_die0 = Methods.getLogNormalPDF(params[4],params[5],Math.exp(params[4]-params[5]*params[5]));
								m_die1 = Methods.getLogNormalPDF(params[6],params[7],Math.exp(params[6]-params[7]*params[7]));
							}
		
							double odiv0,odiv1,odie0,odie1;
							odiv0=odiv1=odie0=odie1=0;
		
							for(int j = 1; j < size*2; j++)
							{
								double xdiv0 = Methods.getLogNormalPDF(params[0],params[1],(((double)j)/(size*2))*time);
								double xdiv1 = Methods.getLogNormalPDF(params[2],params[3],(((double)j)/(size*2))*time);
								double xdie0 = Methods.getLogNormalPDF(params[4],params[5],(((double)j)/(size*2))*time);
								double xdie1 = Methods.getLogNormalPDF(params[6],params[7],(((double)j)/(size*2))*time);
								g2.setColor(c1);
								g2.drawLine((int)(a(2)+j+size+gspace), (int)(i*(vertical_multiplier)+size/2-(xdiv0/m_div0)*size/2)+tb,(int)(a(2)+j-1+size)+gspace, (int)(i*(vertical_multiplier)+size/2-(odiv0/m_div0)*size/2)+tb);
								g2.setColor(c2);
								g2.drawLine((int)(a(2)+j+size)+gspace, (int)(i*(vertical_multiplier)+size/2+(xdie0/m_die0)*size/2)+tb,(int)(a(2)+j-1+size)+gspace, (int)(i*(vertical_multiplier)+size/2+(odie0/m_die0)*size/2)+tb);
								g2.setColor(c3);
								g2.drawLine((int)(a(4)+j+size*3)+gspace*2, (int)(i*(vertical_multiplier)+size/2-(xdiv1/m_div1)*size/2)+tb,(int)(a(4)+j-1+size*3)+gspace*2, (int)(i*(vertical_multiplier)+size/2-(odiv1/m_div1)*size/2)+tb);
								g2.setColor(c4);					
								g2.drawLine((int)(a(4)+j+size*3)+gspace*2, (int)(i*(vertical_multiplier)+size/2+(xdie1/m_die1)*size/2)+tb,(int)(a(4)+j-1+size*3)+gspace*2, (int)(i*(vertical_multiplier)+size/2+(odie1/m_die1)*size/2)+tb);
								odiv0=xdiv0;
								odiv1=xdiv1;
								odie0=xdie0;
								odie1=xdie1;
							}
							for(int j = 0; j < gens; j++)
							{
								//double val = pF[j];
								//g2.drawLine(20, i*(vertical_multiplier0), 20, i*(vertical_multiplier0)+size);
								g2.setColor(c5);			
								if(j > 0) g2.drawLine((int)(a(0)+((j-1)*size)/(gens-1)-1),(int)( i*(vertical_multiplier)+size-(size*pF[j-1]*(100.0/maxPercent))-1+tb),(int)(a(0)+((j)*size)/(gens-1)-1),(int)( i*(vertical_multiplier)+size-(size*pF[j]*(100.0/maxPercent))-1)+tb);
								if(j > 0)
								{
									g2.setColor(c5);
									g2.fillOval((int)(a(0)+((j)*size)/(gens-1)-2),(int)( i*(vertical_multiplier)+size-(size*pF[j]*(100.0/maxPercent))-2)+tb, 5, 5);
								}
								else
								{
									g2.setColor(c5);			
									g2.fillOval((int)(a(0)+((j)*size)/(gens-1)-3),(int)( i*(vertical_multiplier)+size-(size*pF[j]*(100.0/maxPercent))-3)+tb, 7, 7);
								}
							}
						}
					}
					g2.setColor(Color.black);

					//Draw vertical axes
					//left
					g2.drawLine(a(2)+size+gspace,i*(vertical_multiplier)+tb,a(2)+size+gspace,i*(vertical_multiplier)+size+tb);
					g2.drawLine(a(4)+size*3+gspace*2,i*(vertical_multiplier)+tb,a(4)+size*3+gspace*2,i*(vertical_multiplier)+size+tb);
					//right
					g2.drawLine(a(2)+3*size+gspace,i*(vertical_multiplier)+tb,a(2)+3*size+gspace,i*(vertical_multiplier)+size+tb);
					g2.drawLine(a(4)+5*size+gspace*2,i*(vertical_multiplier)+tb,a(4)+size*5+gspace*2,i*(vertical_multiplier)+size+tb);
					//top
					g2.drawLine(a(2)+size+gspace,i*(vertical_multiplier)+tb,a(2)+3*size+gspace,i*(vertical_multiplier)+tb);
					g2.drawLine(a(4)+size*3+gspace*2,i*(vertical_multiplier)+tb,a(4)+5*size+2*gspace,i*(vertical_multiplier)+tb);
					//bottom
					g2.drawLine(a(2)+size+gspace,i*(vertical_multiplier)+tb+size,a(2)+3*size+gspace,i*(vertical_multiplier)+tb+size);
					g2.drawLine(a(4)+size*3+gspace*2,i*(vertical_multiplier)+tb+size,a(4)+5*size+2*gspace,i*(vertical_multiplier)+tb+size);
//					g2.drawLine(20+i*(vertical_multiplier0),40,20+i*(vertical_multiplier0)+size,40);
//					g2.drawLine(20+i*(vertical_multiplier0),60+size*2,20+i*(vertical_multiplier0)+size,60+size*2);
					//draw horizontal ticks
					if(ind==paramList.size()-1 && lblbox.isSelected())  //draw ticks and lbls
					{
						int marks = Math.min(4+size/100,10);
						g2.setFont(new Font("Calibri",Font.PLAIN,(textheight*3)/2));
						for(int j = 0; j < marks; j++)
						{
							g2.drawString(""+(j*time)/marks,a(-3)+size+j*(2*size/marks)+gspace, i*(vertical_multiplier)+size+textheight*2+tb);
							g2.drawString(""+(j*time)/marks,a(-3)+size*3+j*(2*size/marks)+gspace*2, i*(vertical_multiplier)+size+textheight*2+tb);
						}
						g2.drawString(" h",size*3-size/100+7+gspace,i*(vertical_multiplier)+size+textheight*2+tb);
						g2.drawString(" h",size*5-size/95+7+gspace*2,i*(vertical_multiplier)+size+textheight*2+tb);
						g2.drawString(" h",size*3-size/100+7+gspace,i*(vertical_multiplier)+size+textheight*2+tb);
						g2.drawString(" h",size*5-size/95+7+gspace*2,i*(vertical_multiplier)+size+textheight*2+tb);
						g2.setFont(new Font("Calibri",Font.PLAIN,textheight));


					}
					//draw pf axes lbls?
					if(ind==paramList.size()-1 && lblbox.isSelected())
					{
						g2.setFont(new Font("Calibri",Font.PLAIN,(textheight*3)/2));
						for(int j = 0; j < gens; j++)
						{
							g2.drawString(""+j,a(-2)+j*(size/(gens-1)),i*(vertical_multiplier)+size+textheight*2+tb);
							g2.drawLine(a(0)+j*size/(gens-1), i*(vertical_multiplier)+size-2+tb,a(0)+j*size/(gens-1), i*(vertical_multiplier)+size+4+tb);
						}
						g2.setFont(new Font("Calibri",Font.PLAIN,textheight));
					}
					if(lblbox.isSelected())
					{
						g2.setFont(new Font("Calibri",Font.PLAIN,(textheight*3)/2));
						g2.drawString(String.format("%3d",(maxPercent))+"%", 0, i*(vertical_multiplier)+tb+textheight);
						g2.drawString(String.format("%3d",((maxPercent*3)/4))+"%", 1, i*(vertical_multiplier)+5+(int)(size*0.25)+tb);
						g2.drawString(String.format("%3d",((maxPercent)/2))+"%", 0, i*(vertical_multiplier)+textheight/2+(int)(size*0.5+tb));
						g2.drawString(String.format("%3d",((maxPercent)/4))+"%", 1, i*(vertical_multiplier)+5+(int)(size*0.75)+tb);
						g2.drawString("0%", 0, i*(vertical_multiplier)+5+(int)(size*1)-5+tb);

						g2.setFont(new Font("Calibri",Font.PLAIN,textheight));
						g2.drawLine(a(-1),i*(vertical_multiplier)+tb,a(2),i*(vertical_multiplier)+tb);
						g2.drawLine(a(-1),i*(vertical_multiplier)+(int)(size*0.25)+tb,a(2),i*(vertical_multiplier)+(int)(size*0.25)+tb);
						g2.drawLine(a(-1),i*(vertical_multiplier)+(int)(size*0.5)+tb,a(2),i*(vertical_multiplier)+(int)(size*0.5)+tb);
						g2.drawLine(a(-1),i*(vertical_multiplier)+(int)(size*0.75)+tb,a(2),i*(vertical_multiplier)+(int)(size*0.75)+tb);
						
						int marks = Math.min(4+size/100,10);
						for(int j = 0; j < marks; j++)
						{
							g2.drawLine(a(2)+size+j*(2*size/marks)+gspace, i*(vertical_multiplier)+size-2+tb, a(2)+size+j*(2*size/marks)+gspace, i*(vertical_multiplier)+size+2+tb);
							g2.drawLine(a(4)+size*3+j*(2*size/marks)+gspace*2, i*(vertical_multiplier)+size-2+tb, a(4)+size*3+j*(2*size/marks)+2*gspace, i*(vertical_multiplier)+size+2+tb);
						}
						g2.setColor(Color.BLACK);
						g2.setFont(new Font("Calibri",Font.PLAIN,(textheight*3)/2));
						g2.drawString("Tdie0",size*3-size/6+gspace,i*(vertical_multiplier)+size-5+tb);
						g2.drawString("Tdiv0",size*3-size/6+gspace,i*(vertical_multiplier)+5+size/50+5+tb);
						g2.drawString("Tdie1+",size*5-size/6+gspace*2,i*(vertical_multiplier)+size-5+tb);
						g2.drawString("Tdiv1+",size*5-size/6+gspace*2,i*(vertical_multiplier)+5+size/50+5+tb);

						//g2.drawLine(2+i*(vertical_multiplier0),90+size*5,(i+1)*(vertical_multiplier0)-2,90+size*5);
						//lets also draw the Pr-> arrows
						drawRotatedText(-Math.PI/2,"<-Pr",a(1)+size+gspace,tb+i*(vertical_multiplier)+size,g2);
						drawRotatedText(-Math.PI/2,"Pr->",a(1)+size+gspace,tb+i*(vertical_multiplier)+textheight*3-5,g2);
						drawRotatedText(-Math.PI/2,"<-Pr",a(3)+size*3+2*gspace,tb+i*(vertical_multiplier)+size,g2);
						drawRotatedText(-Math.PI/2,"Pr->",a(3)+size*3+2*gspace,tb+i*(vertical_multiplier)+textheight*3-5,g2);
						g2.setFont(new Font("Calibri",Font.PLAIN,textheight));
					}
					if(!overlayBox.isSelected())
					{
						g2.setFont(new Font("Calibri",Font.PLAIN,8+size/50));
						double[] ps2 = Methods.duplicate(paramList2.get(i));
						int spacing = g2.getFont().getSize()+1;
						g2.setFont(new Font("Calibri",Font.PLAIN,textheight));
						g2.setColor(Color.RED);
						g2.drawString("Phenotype:",a(25)+size*6+gspace*2,i*(vertical_multiplier)+spacing+tb);
						g2.setColor(Color.BLACK);
						g2.drawString(names.get(i),a(25)+size*6+gspace*2,i*(vertical_multiplier)+spacing*2+tb);
						g2.setColor(Color.RED);
						g2.drawString("Draw Date:",a(25)+size*6+gspace*2,i*(vertical_multiplier)+spacing*3+tb);
						g2.setColor(Color.BLACK);
						g2.drawString(Methods.now(),a(25)+size*6+gspace*2,i*(vertical_multiplier)+spacing*4+tb);
						if(i%2==0)g2.setColor(Color.BLACK);
						else g2.setColor(Color.BLUE);
						g2.drawString(String.format("Tdiv0 = %3.3f|%3.3f", ps2[0],ps2[1]),a(10)+size*5+gspace*2,i*(vertical_multiplier)+spacing+tb);
						g2.drawString(String.format("Tdiv1+ = %3.3f|%3.3f", ps2[2],ps2[3]),a(10)+size*5+gspace*2,i*(vertical_multiplier)+spacing*2+tb);
						g2.drawString(String.format("Tdie0 = %3.3f|%3.3f", ps2[4],ps2[5]),a(10)+size*5+gspace*2,i*(vertical_multiplier)+spacing*3+tb);
						g2.drawString(String.format("Tdie1+ = %3.3f|%3.3f", ps2[6],ps2[7]),a(10)+size*5+gspace*2,i*(vertical_multiplier)+spacing*4+tb);
						g2.drawString(String.format("Prol. Cap. = %3.2f|%3.2f", ps2[8],ps2[9]),a(10)+size*5+gspace*2,i*(vertical_multiplier)+spacing*5+tb);
						g2.drawString(String.format("F0 = %3.3f", ps2[10]),a(10)+size*5+gspace*2,i*(vertical_multiplier)+spacing*6+tb);
						g2.drawString(String.format("Start Cells = %,d", (int)ps2[11]),a(10)+size*5+gspace*2,i*(vertical_multiplier)+spacing*7+tb);
						g2.drawString(String.format("Model Diff = %3.2f %%",scores.get(i)),a(10)+size*5+gspace*2,i*(vertical_multiplier)+spacing*8+tb);
					}
					else
					{
						//lets draw the legend
						g2.setFont(new Font("Calibri",Font.PLAIN,8+size/50));
						g2.setColor(overlayColors[ind%overlayColors.length]);
						g2.fillRect(a(10)+size*5+gspace*2, tb+ind*(size/10), (size/10)-2, (size/10)-2);
						g2.setColor(Color.black);
						g2.drawRect(a(10)+size*5+gspace*2, tb+ind*(size/10), (size/10)-2, (size/10)-2);
						g2.drawString(names.get(ind),a(15)+size*5+gspace*2+(size/10),tb+ind*(size/10)-2+g2.getFont().getSize());
					}
					//Experiment name
					if(namebox.isSelected()&& !overlayBox.isSelected())
					{
						g2.setColor(Color.BLACK);
						g2.setFont(new Font("Calibri",Font.PLAIN,8+size/50));
						g2.drawString(names.get(i),a(4),i*(vertical_multiplier)+8+size/50+tb);
						g2.drawString(names.get(i),a(6)+size+gspace,i*(vertical_multiplier)+8+size/50+tb);
						g2.drawString(names.get(i),a(8)+size*3+gspace*2,i*(vertical_multiplier)+8+size/50+tb);
					}
					
				}
				//erase the top part sticking out
				g2.setBackground(new Color(255,255,255,0));
				g2.clearRect(0, 0, getWidth(),textheight+5);
				//draw TOP labels
				if(lblbox.isSelected())
				{
					g2.setFont(new Font("Calibri",Font.PLAIN,(textheight*3)/2));
					g2.setColor(Color.BLACK);
					g2.drawString("Fs", a(0), textheight+1);
					g2.drawString("Undivided Cells", a(1)+size+gspace, textheight+1);
					g2.drawString("Dividing Cells", a(3)+size*3+gspace*2, textheight+1);

				}
			}
			catch(Exception e)
			{
				//do nothing
			}
			finally{
				owner.setCursor(Cursor.getDefaultCursor());
			}
		}
		else
		{
			g3.setColor(Color.BLACK);
			g3.drawString("Click on the \"Visualize\" button to draw the best fit parameter distributions.",10,10);
		}
		g3.drawImage(img,0,0,this);
	}

	private void drawRotatedText(double theta, String string, int i, int j,
			Graphics2D g2) {
		g2.translate(i,j);
		g2.rotate(theta);
		g2.drawString(string,0,0);
		g2.rotate(-theta);
		g2.translate(-i, -j);
		
	}







}
