/*
Copyright (c)  2013   The Regents of the University of California
All Rights Reserved
 
Permission to copy, modify and distribute any part of this FlowMax for educational, research and non-profit purposes, without fee, and without a written agreement is hereby granted, provided that the above copyright notice, this paragraph and the following three paragraphs appear in all copies.
Those desiring to incorporate this FlowMax into commercial products or use for commercial purposes should contact the Technology Transfer Office, University of California, San Diego, 9500 Gilman Drive, Mail Code 0910, La Jolla, CA 92093-0910, Ph: (858) 534-5815, FAX: (858) 534-7345, E-MAIL:invent@ucsd.edu.
IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES, INCLUDING LOST PROFITS, ARISING OUT OF THE USE OF THIS FlowMax, EVEN IF THE UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
THE FlowMax PROVIDED HEREIN IS ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.  THE UNIVERSITY OF CALIFORNIA MAKES NO REPRESENTATIONS AND EXTENDS NO WARRANTIES OF ANY KIND, EITHER IMPLIED OR EXPRESS, 
INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE, OR THAT THE USE OF THE FlowMax WILL NOT INFRINGE ANY PATENT, TRADEMARK OR OTHER RIGHTS.
*/
package FlowMax;

import java.awt.Color;
import java.util.ArrayList;
import java.util.Random;


import common.Methods;

public class Experiment
{
	ArrayList<TimePoint> timePoints = new ArrayList<TimePoint>();
	public int gens;
	public int bincount = 200;
	String name = "";
	double[][] sensitivities;
	double cytonPValue = 1;
	double bestScore = Double.POSITIVE_INFINITY;
	double fluorescencePValue = 1;
	Random gen = new Random();
	int totalCells = 0;
	int repeats =1; //the number of times to fit
	boolean abort_flag = false; //set this to true while fitting to illicit an exit from the loop
	double[] abestparams;
	
	double maxT = 0;
	
	public double[] getAbestparams() {
		return abestparams;
	}

	private ArrayList<FitListener> listeners = new ArrayList<FitListener>();
	private ArrayList<FluorescenceFitListener> flisteners = new ArrayList<FluorescenceFitListener>();
	
	public ArrayList<FluorescenceFitListener> getFlisteners() {
		return flisteners;
	}

	public void registerFitListener(FitListener l)
	{
		listeners.add(l);
	}
	
	public void registerFluorescenceFitListener(FluorescenceFitListener l)
	{
		flisteners.add(l);
	}
	
	public Experiment clone()
	{
		Experiment clone = new Experiment(gens,bincount,timePoints);
		ArrayList<TimePoint> newtp = new ArrayList<TimePoint>();
		for(int i = 0; i < timePoints.size(); i++)
		{
			TimePoint tp =timePoints.get(i).clone();
			tp.setParent(clone);
			newtp.add(tp);
		}
		clone.timePoints = newtp;
		clone.name = name;
		clone.maxT = maxT;
		return clone;
	}
	
	public Experiment(int gens, int bincount, ArrayList<TimePoint> timePoints)
	{
		this.gens = gens;
		this.timePoints = timePoints;
		this.bincount = bincount;
		for(int i = 0; i < timePoints.size(); i++)
			if(timePoints.get(i).getTime() > maxT) maxT = timePoints.get(i).getTime();
	}
	
	
	public Experiment(ArrayList<ArrayList<ArrayList<Double>>> raw,ArrayList<ArrayList<Double>> undivided, int gens,ArrayList<ArrayList<Integer>> volumes,int num_bins, ArrayList<Double> times, String name )
	{
		this.name = name;
		this.bincount = num_bins;
		this.gens = gens;
		for(int i = 0; i < times.size(); i++)
		{
			TimePoint tp = new TimePoint(times.get(i),undivided.get(i),volumes.get(i),raw.get(i),this);
			timePoints.add(tp);
			if(times.get(i) > maxT) maxT = times.get(i);
			totalCells+=tp.getTotalCount();
		}
	}
	
	public boolean fitFluorescence(double aq, double[][] lims)
	{
		long fft = System.nanoTime();
		//lets start with time point 1
		double[] starts = new double[4];
		starts[0] = lims[0][0];
		starts[1] = lims[1][1];
		starts[2] = lims[2][1];
		starts[3] = 0;
		for(int i = 0; i < timePoints.size(); i++)
		{
			//fit this time point using the current ratio, CV, and background

		//	System.err.println("Started fitting time point "+(i+1));
		//	timePoints.get(i).fitFluorescence(starts[0],starts[1],starts[2],starts[3],aq);
			boolean finished = timePoints.get(i).fitFluorescence2(lims, starts);
			if(!finished) return false;
			starts[0] = timePoints.get(i).CV;
			starts[1] = lims[1][1];
			starts[2] = lims[2][1];
			//starts[1] = timePoints.get(i).ratio;
			//starts[2] = timePoints.get(i).background;
			//starts[3] = timePoints.get(i).shift;
			double tollerance = Math.pow(0.5,(i+1));
			lims[0][0] = Math.max(starts[0]*(1-tollerance),lims[0][0]);
			lims[0][1] = Math.min(starts[0]*(1+tollerance),lims[0][1]);
			
		}
		double[] averages = new double[5];
		boolean[] worked = new boolean[timePoints.size()];
		int count_good = 0;
		for(int i = 0; i < timePoints.size(); i++)
		{
			double pvalue = timePoints.get(i).getPValue();
			if(pvalue < 0.1)
			{
				averages[0]+=timePoints.get(i).CV;
				averages[1]+=timePoints.get(i).ratio;
				averages[2]+=timePoints.get(i).background;
				averages[4]+=pvalue;
				count_good++;
				worked[i]=true;
			}
		}
		for(int i = 0; i < timePoints.size(); i++)
		{
			for(int j = 0; j < timePoints.get(i).getReplicates().size(); j++)
				timePoints.get(i).getReplicates().get(j).computeWeights();
			if(!worked[i] && count_good > 2)
			{
				timePoints.get(i).CV = averages[0]/count_good;
				timePoints.get(i).ratio = averages[1]/count_good;
			//	timePoints.get(i).background = averages[2]/count_good;  //this is not a good thing to average in
		//		timePoints.get(i).shift = averages[3]/count_good;

				double newPvalue = timePoints.get(i).getPValue();
				averages[4]+=newPvalue;
				if(newPvalue >0.1) System.err.println("Time Point "+(i+1)+" failed to fit the fluorescence utterly! Using average fluorescence values and hoping for the best");
			}
		}
		fluorescencePValue = getPValue();
		System.err.println("Fluorescence fitting took : "+((System.nanoTime()-fft)/(1000000000))+" seconds. P-value for fit is "+fluorescencePValue);
		return true;
	}

	public ArrayList<double[]> fitCyton(int gens, double[][] lims, double dt, boolean pcyton, int trials_to_run)
	{
		return fitCyton(gens,lims,dt,pcyton,trials_to_run,0);
	}
	
	/**
	 * 
	 * @param gens
	 * @param lims
	 * @param dt
	 * @param pcyton
	 * @param trials_to_run
	 * @param fit_mode 0= full model and scoring function: weighted correlated RMSD,1 no-correlation weighted RMSD full model, 2= no gen dependence mixed scoring fxn, 3= full model avg RMSD
	 * @return
	 */
	public ArrayList<double[]> fitCyton(int gens, double[][] lims, double dt, boolean pcyton, int trials_to_run,int fit_mode)
	{
		
		repeats = trials_to_run;
		//double mean1 = Math.exp(limits[i][0]+0.5*limits[i+1][0]*limits[i+1][0]);
		//double std1 = mean1*Math.sqrt(Math.exp(limits[i+1][0]*limits[i+1][0])-1);
		double maxT = 0;
		double hmin = Double.POSITIVE_INFINITY;
		double hmax = Double.NEGATIVE_INFINITY;
		for(int i = 0; i < timePoints.size(); i++)
		{
			if(timePoints.get(i).getTime() > maxT) maxT=timePoints.get(i).getTime();
			for(int j = 0; j < timePoints.get(i).getReplicates().size(); j++)
			{
				for(int k = 0; k < timePoints.get(i).getReplicates().get(j).data.raw.size(); k++)
				{
					if(timePoints.get(i).getReplicates().get(j).data.raw.get(k) < hmin) hmin =timePoints.get(i).getReplicates().get(j).data.raw.get(k);
					if(timePoints.get(i).getReplicates().get(j).data.raw.get(k) > hmax) hmax =timePoints.get(i).getReplicates().get(j).data.raw.get(k);
				}
			}
		}
	//	System.err.println("Min Max are: "+hmin+","+hmax);
		for(int i = 0; i < timePoints.size(); i++)
		{
			for(int j = 0; j < timePoints.get(i).getReplicates().size(); j++)
			{
				timePoints.get(i).getReplicates().get(j).data.resetMinMax(gens*10,Math.max(Constants.minimum,hmin),Math.min(Constants.maximum,hmax));
				timePoints.get(i).getReplicates().get(j).gaussians = null;
			}
		}
		abestparams = new double[Constants.nparams];
		double absolute_best = Double.POSITIVE_INFINITY;

		ArrayList<double[]> solutions = new ArrayList<double[]>();
		for(int repeat = 0; repeat < repeats; repeat++)
		{
			long fittime = System.nanoTime();
			double[] solution = new double[Constants.nparams*3+2];
			System.err.println("Starting Repeat: "+repeat);
//			double best = Double.POSITIVE_INFINITY;  //best for this round
			double verybest =Double.POSITIVE_INFINITY;
			double last = Double.POSITIVE_INFINITY;
			double[] vbestparams = new double[Constants.nparams];
			double Tstart = 1000;  
			double T = Tstart;
			int counter  = 0;
			boolean changing = true;
		//	double[] bestparams = new double[nparams];
			double[] params = new double[Constants.nparams];
			double[] lastparams = Methods.duplicate(params);
			int trial = 0;
			//gens=2;
			while(trial < 4000)
			{

				for(int i = 0; i < Constants.nparams; i++)
				{
					params[i] = gen.nextDouble()*(lims[i][1]-lims[i][0])+lims[i][0];
				}
				if(fit_mode == 2)
				{
					params[2] = params[0];
					params[3] = params[1];
					params[6] = params[4];
					params[7] = params[5];
					params[8] = 999;
					params[9] = 0.1;
				}
				cytonToFluorescence(params,pcyton);
				double score = 0;
				if(fit_mode==0 || fit_mode==2)
					score =getCytonScore2();
				else if(fit_mode == 1)
					score = getCytonScoreWeightedRMSD();
				else if(fit_mode == 3)
					score = getCytonScoreUnweightedRMSD();
				else if(fit_mode == 4)
					score = getCytonScoreWeightedTrueAreaDiff();
				else if(fit_mode == 5)
					score = getCytonScoreUnweightedSD();
				
				if(score < verybest)
				{
					updateListeners(vbestparams,repeat,counter,score);
					verybest = score;
					vbestparams = Methods.duplicate(params);
				}
				trial++;
			}
			///////////////////////////////////////////////////////////////////////////-=-=-=-=-=-=-=-=PREFIT FINISHED-=-=-=-=-=-=-=-=-=-=-
			cytonToFluorescence(vbestparams,pcyton);
			if(fit_mode==0|| fit_mode==2)
				last =getCytonScore2();
			else if(fit_mode == 1)
				last = getCytonScoreWeightedRMSD();
			else if(fit_mode == 3)
				last= getCytonScoreUnweightedRMSD();
			else if(fit_mode == 4)
				last = getCytonScoreWeightedTrueAreaDiff();
			else if(fit_mode == 5)
				last = getCytonScoreUnweightedSD();
			lastparams = Methods.duplicate(vbestparams);
			int rep = 0;
			counter = 0;
			T=last/10;
			double Tmin = T/10000;
			while(changing)
			{
				if(abort_flag)
				{
					abort_flag = false;
					return new ArrayList<double[]>();
				}
				rep++;
				counter++;
				for(int t = 0; t < timePoints.size(); t++)
				{
					for(int r = 0; r < timePoints.get(t).getReplicates().size(); r++)
					{
						timePoints.get(t).getReplicates().get(r).backUpWeights();
					}
				}
				//lets make some changes to the parameters
				for(int i = 0; i < Constants.nparams; i++ )
				{
					double chance = gen.nextDouble();
					if(chance > 0.05)
						params[i] = Methods.reflectWithin(lastparams[i]+gen.nextGaussian()*lastparams[i]*0.01,lims[i][0],lims[i][1]);
					else
						params[i] = Methods.reflectWithin(lastparams[i]+gen.nextGaussian()*lastparams[i]*3,lims[i][0],lims[i][1]);
				}
				if(fit_mode == 2)
				{
					params[2] = params[0];
					params[3] = params[1];
					params[6] = params[4];
					params[7] = params[5];
					params[8] = 999;
					params[9] = 0.1;
				}
				//if(rep > 2000 && gens < 10) gens++;
				cytonToFluorescence(params,pcyton);  //this updates the weights for all runs in this condition
				double score =0;
				if(fit_mode==0|| fit_mode==2)
					score =getCytonScore2();
				else if(fit_mode == 1)
					score = getCytonScoreWeightedRMSD();
				else if(fit_mode == 3)
					score = getCytonScoreUnweightedRMSD();
				else if(fit_mode == 4)
					score = getCytonScoreWeightedTrueAreaDiff();
				else if(fit_mode == 5)
					score = getCytonScoreUnweightedSD();
				double diff = (score-last);
				double chance =  Math.exp(-(diff)/(T));
				if(gen.nextDouble() < chance)
				{
					if(score <verybest)
					{
						rep = 0;
						updateListeners(params,repeat,counter,score);
						verybest = score;
						vbestparams = Methods.duplicate(params);
						T=Math.min(Tmin*10000,T*1.1);
					} //verybest
				}//chance
				else
				{
					//revert to the previous
					params = Methods.duplicate(lastparams);
					score = last;
					for(int t = 0; t < timePoints.size(); t++)
					{
						for(int r = 0; r < timePoints.get(t).getReplicates().size(); r++)
						{
							timePoints.get(t).getReplicates().get(r).restoreWeights();
						}
					}
				}
				//define last params
				last = score;
				lastparams = Methods.duplicate(params);
				//current = score;
				T = T*0.9995;

				if(counter%1000==0)System.err.println("Count = "+counter+" T = "+T+" rep "+rep+" best score thus far "+verybest);
				if((counter >1000 && T < Tmin)) changing=false;
			}//changing
			changing=true;
			rep = 0;
			counter=0;
			System.err.println("Refining the best fit");
///////////////////////////////////////////////////////////////////////////-=-=-=-=-=-=-=-=REFINE TIME=-=-=-=-=-=-=-=-=-=-klsdjflkasjdfklasjdklfjaskldjfklsjadflk
			params = Methods.duplicate(vbestparams);
			cytonToFluorescence(params,pcyton);
			if(fit_mode==0|| fit_mode==2)
				last =getCytonScore2();
			else if(fit_mode == 1)
				last = getCytonScoreWeightedRMSD();
			else if(fit_mode == 3)
				last = getCytonScoreUnweightedRMSD();
			else if(fit_mode == 4)
				last = getCytonScoreWeightedTrueAreaDiff();
			else if(fit_mode == 5)
				last = getCytonScoreUnweightedSD();
			T= last/10;
			Tmin = T/1000;
			while(changing)
			{
				if(abort_flag)
				{
					abort_flag = false;
					return new ArrayList<double[]>();
				}
				rep++;
				counter++;
				for(int t = 0; t < timePoints.size(); t++)
				{
					for(int r = 0; r < timePoints.get(t).getReplicates().size(); r++)
					{
						timePoints.get(t).getReplicates().get(r).backUpWeights();
					}
				}
				//lets make some changes to the parameters
				for(int i = 0; i < Constants.nparams; i++ )
				{
					double chance = gen.nextDouble();
					if(chance > 0.05)
						params[i] = Methods.reflectWithin(lastparams[i]+gen.nextGaussian()*lastparams[i]*0.001,lims[i][0],lims[i][1]);
					else
						params[i] = Methods.reflectWithin(lastparams[i]+gen.nextGaussian()*lastparams[i]*0.1,lims[i][0],lims[i][1]);
				}
				if(fit_mode == 2)
				{
					params[2] = params[0];
					params[3] = params[1];
					params[6] = params[4];
					params[7] = params[5];
					params[8] = 999;
					params[9] = 0.1;
				}
				cytonToFluorescence(params,pcyton);  //this updates the weights for all runs in this condition
				double score =0;
				if(fit_mode==0|| fit_mode==2)
					score =getCytonScore2();
				else if(fit_mode == 1)
					score = getCytonScoreWeightedRMSD();
				else if(fit_mode == 3)
					score = getCytonScoreUnweightedRMSD();
				else if(fit_mode == 4)
					score = getCytonScoreWeightedTrueAreaDiff();
				else if(fit_mode == 5)
					score = getCytonScoreUnweightedSD();
				double diff = (score-last);
				double chance =  Math.exp(-(diff)/(T));
				if(gen.nextDouble() < chance)
				{
					if(score <verybest)
					{
						rep = 0;
						updateListeners(params,repeat,trial,score);
						verybest = score;
						vbestparams = Methods.duplicate(params);
						T=Math.min(Tmin*1000,T*3);
					} //verybest
				}//chance
				else
				{
					//revert to the previous
					params = Methods.duplicate(lastparams);
					score = last;
					for(int t = 0; t < timePoints.size(); t++)
					{
						for(int r = 0; r < timePoints.get(t).getReplicates().size(); r++)
						{
							timePoints.get(t).getReplicates().get(r).restoreWeights();
						}
					}
				}
				//define last params
				last = score;
				lastparams = Methods.duplicate(params);
				//current = score;
				T = T*0.995;

				if(counter%1000==0)System.err.println("Count = "+counter+" T = "+T+" rep "+rep+" best score thus far "+verybest);
				if((counter >1000 && T < Tmin) ) changing=false;
			}//changing
			if(verybest < absolute_best)
			{
				absolute_best = verybest;
				abestparams = Methods.duplicate(vbestparams);
			}
///////////////////////////////////////////////////////////////////////////-=-=-=-=-=-=-=-=SENSITIIVITY ANALYSIS START-=-=-=-=-=-=-=-=-=-=-
			cytonToFluorescence(vbestparams,pcyton);
			bestScore = 0;
			if(fit_mode==0|| fit_mode==2)
				bestScore =getCytonScore2();
			else if(fit_mode == 1)
				bestScore = getCytonScoreWeightedRMSD();
			else if(fit_mode == 3)
				bestScore = getCytonScoreUnweightedRMSD();
			else if(fit_mode == 4)
				bestScore = getCytonScoreWeightedTrueAreaDiff();
			else if(fit_mode == 5)
				bestScore = getCytonScoreUnweightedSD();
			params = abestparams;
			cytonPValue = getPValue();
			updateListeners(abestparams,repeat,trial,bestScore);
			//--------------------------------------------------------------------------  SENSITIVITY!!!!!!!!!!!!!!
			double[] cytonParamsAfterFit = new double[Constants.nparams];
			//cytonPValue = getPValue();
			System.err.println("P value = "+cytonPValue);
			System.err.println("Performing Sensitivity Analysis");
			double[][] result = getSensitivities(Constants.acceptable_error, vbestparams,lims,pcyton);
			sensitivities = new double[Constants.nparams][2];
			double mean1 = Math.exp(0.5*Math.pow(result[1][0],2)+result[0][0]);
			double std1 = mean1*Math.sqrt(Math.exp(Math.pow(result[1][0],2))-1);
			double mean3 = Math.exp(0.5*Math.pow(result[1][1],2)+result[0][1]);
			double std3 = mean3*Math.sqrt(Math.exp(Math.pow(result[1][1],2))-1);	
			double mean2 = Math.exp(0.5*Math.pow(vbestparams[1],2)+vbestparams[0]);
			double std2 = mean2*Math.sqrt(Math.exp(Math.pow(vbestparams[1],2))-1);
			System.err.println(String.format("Tdiv0u\t%f\t%f\t%f",(mean2-mean1),mean2,(mean3-mean2)));
			System.err.println(String.format("Tdiv0s\t%f\t%f\t%f",(std2-std1),std2,Math.abs(std3-std2)));
			sensitivities[0][0] = (mean2-mean1);
			sensitivities[0][1] = (mean3-mean2);
			cytonParamsAfterFit[0] = mean2;
			sensitivities[1][0] = (std2-std1);
			sensitivities[1][1] = (std3-std2);
			cytonParamsAfterFit[1] = std2;
			mean1 = Math.exp(0.5*Math.pow(result[3][0],2)+result[2][0]);
			std1 = mean1*Math.sqrt(Math.exp(Math.pow(result[3][0],2))-1);
			mean3 = Math.exp(0.5*Math.pow(result[3][1],2)+result[2][1]);
			std3 = mean3*Math.sqrt(Math.exp(Math.pow(result[3][1],2))-1);	
			mean2 = Math.exp(0.5*Math.pow(vbestparams[3],2)+vbestparams[2]);
			std2 = mean2*Math.sqrt(Math.exp(Math.pow(vbestparams[3],2))-1);
			System.err.println(String.format("Tdiv1u\t%f\t%f\t%f",(mean2-mean1),mean2,(mean3-mean2)));
			System.err.println(String.format("Tdiv1s\t%f\t%f\t%f",(std2-std1),std2,Math.abs(std3-std2)));
			sensitivities[2][0] = (mean2-mean1);
			sensitivities[2][1] = (mean3-mean2);
			cytonParamsAfterFit[2] = mean2;
			sensitivities[3][0] = (std2-std1);
			sensitivities[3][1] = (std3-std2);
			cytonParamsAfterFit[3] = std2;
			mean1 = Math.exp(0.5*Math.pow(result[5][0],2)+result[4][0]);
			std1 = mean1*Math.sqrt(Math.exp(Math.pow(result[5][0],2))-1);
			mean3 = Math.exp(0.5*Math.pow(result[5][1],2)+result[4][1]);
			std3 = mean3*Math.sqrt(Math.exp(Math.pow(result[5][1],2))-1);	
			mean2 = Math.exp(0.5*Math.pow(vbestparams[5],2)+vbestparams[4]);
			std2 = mean2*Math.sqrt(Math.exp(Math.pow(vbestparams[5],2))-1);
			System.err.println(String.format("Tdie0u\t%f\t%f\t%f",(mean2-mean1),mean2,(mean3-mean2)));
			System.err.println(String.format("Tdie0s\t%f\t%f\t%f",(std2-std1),std2,Math.abs(std3-std2)));
			sensitivities[4][0] = (mean2-mean1);
			sensitivities[4][1] = (mean3-mean2);
			cytonParamsAfterFit[4] = mean2;
			sensitivities[5][0] = (std2-std1);
			sensitivities[5][1] = (std3-std2);
			cytonParamsAfterFit[5] = std2;
			mean1 = Math.exp(0.5*Math.pow(result[7][0],2)+result[6][0]);
			std1 = mean1*Math.sqrt(Math.exp(Math.pow(result[7][0],2))-1);
			mean3 = Math.exp(0.5*Math.pow(result[7][1],2)+result[6][1]);
			std3 = mean3*Math.sqrt(Math.exp(Math.pow(result[7][1],2))-1);	
			mean2 = Math.exp(0.5*Math.pow(vbestparams[7],2)+vbestparams[6]);
			std2 = mean2*Math.sqrt(Math.exp(Math.pow(vbestparams[7],2))-1);
			System.err.println(String.format("Tdie1u\t%f\t%f\t%f",(mean2-mean1),mean2,(mean3-mean2)));
			System.err.println(String.format("Tdie1s\t%f\t%f\t%f",(std2-std1),std2,Math.abs(std3-std2)));
			sensitivities[6][0] = (mean2-mean1);
			sensitivities[6][1] = (mean3-mean2);
			cytonParamsAfterFit[6] = mean2;
			sensitivities[7][0] = (std2-std1);
			sensitivities[7][1] = (std3-std2);
			cytonParamsAfterFit[7] = std2;
			System.err.println(String.format("Du\t%f\t%f\t%f", vbestparams[8]-result[8][0],vbestparams[8],result[8][1]-vbestparams[8]));
			System.err.println(String.format("Dstd\t%f\t%f\t%f", vbestparams[9]-result[9][0],vbestparams[9],result[9][1]-vbestparams[9]));
			System.err.println(String.format("pF0\t%f\t%f\t%f", vbestparams[10]-result[10][0],vbestparams[10],result[10][1]-vbestparams[10]));
			System.err.println(String.format("Start C\t%f\t%f\t%f", vbestparams[11]-result[11][0],vbestparams[11],result[11][1]-vbestparams[11]));
			sensitivities[8][0] = vbestparams[8]-result[8][0];
			sensitivities[8][1] = result[8][1]-vbestparams[8];
			cytonParamsAfterFit[8] = vbestparams[8];
			sensitivities[9][0] = vbestparams[9]-result[9][0];
			sensitivities[9][1] = result[9][1]-vbestparams[9];
			cytonParamsAfterFit[9] = vbestparams[9];
			sensitivities[10][0] = vbestparams[10]-result[10][0];
			sensitivities[10][1] = result[10][1]-vbestparams[10];
			cytonParamsAfterFit[10] = vbestparams[10];
			sensitivities[11][0] = vbestparams[11]-result[11][0];
			sensitivities[11][1] = result[11][1]-vbestparams[11];
			cytonParamsAfterFit[11] = vbestparams[11];
			solution[0] = bestScore; //cytonPValue;
			solution[1] = getAreaDiff();
			for(int i = 0; i < Constants.nparams; i++)
				solution[i+2] = cytonParamsAfterFit[i];
			for(int i = 0; i < Constants.nparams; i++)
				solution[i+Constants.nparams+2] = sensitivities[i][0];
			for(int i = 0; i < Constants.nparams; i++)
				solution[i+2*Constants.nparams+2] = sensitivities[i][1];
			int index = 0;
			while(index < solutions.size() && solutions.get(index)[1] < solution[1])
				index++;
			solutions.add(index,solution);
			System.err.println("PCyton fit took : "+((System.nanoTime()-fittime)/1000000000)+" seconds.");
		}
		cytonToFluorescence(abestparams,pcyton);
		cytonPValue = getPValue();
		System.err.println("Best p-value after fitting: "+cytonPValue);
		updateListeners(abestparams,repeats,0,absolute_best);
		return solutions;
	}

	private void updateListeners(double[] params,int repeat,int trial, double score) {
		
		for(int i = 0; i < listeners.size(); i++)
			listeners.get(i).processCytonUpdate(params,repeat,trial,score,this);
	}

	/**
	 * Method used to solve the Pcyton model using the current parameters and update the weights for all model histograms...
	 * @param params (PCyton Parameters)
	 * @param steps (How many steps to solve for)
	 * @param Generations (The number of divisions expected)
	 * @param T (Total time, should be about half of the steps)
	 * @param dt (step size in hours)
	 */
	public void cytonToFluorescence(double[] params, boolean pcyton)
	{
		int steps = (int)((maxT+1.0)/Constants.dt);
		
		double[][] result;
		if (pcyton) result = Methods.computeCytons(params,steps,gens, (int)(maxT+1.0))[0];
		else result = Methods.computeCytonsOld(params,steps,gens, (int)(maxT+1.0))[0];
		//lets interpret this result in terms of fluorescences and histograms
		for(int t = 0; t < timePoints.size(); t++)
		{
			int index = (int)(timePoints.get(t).getTime()/Constants.dt);
			//lets compute the total cells in all divisions at this time
			double totalCells = 0;
			double[] weights = new double[gens];
			for(int g = 0; g < gens; g++)
			{
				totalCells+= result[g][index];
				weights[g] = result[g][index];
				if(Double.isNaN(weights[g])) weights[g] = 0;
			}
			if(totalCells == 0)
			{
				totalCells = 1;
				weights[0]=1;
			}
			for(int g = 0; g < gens; g++)
			{
				if(totalCells > 0)
					weights[g] = weights[g]/totalCells;
				else
					weights[g] = 0;
			}
			for(int r = 0; r < timePoints.get(t).getReplicates().size() ; r++)
			{
				for(int w = 0; w < timePoints.get(t).getReplicates().get(r).weights.length; w++)
				{
					if(w < weights.length)
						timePoints.get(t).getReplicates().get(r).weights[w] = weights[w];
					else
						timePoints.get(t).getReplicates().get(r).weights[w] = 0;
				}
				timePoints.get(t).getReplicates().get(r).totalCyton = totalCells;
			}
		}
		
	}
	

	
	/**
	 * Finds the acceptable param ranges around a given set of parameters for a given tollerance value (i.e. 0.1 or 10%).
	 * Uses the area difference between histograms as the difference function
	 * @param tollerance
	 * @param params
	 * @param lims
	 * @param steps
	 * @param Generations
	 * @param T
	 * @param dt
	 * @return
	 */
	private double[][] getSensitivities(double tollerance, double[] params, double[][] lims,boolean pcyton)
	{
		double[][] sensitivities = new double[Constants.nparams][2];
		cytonToFluorescence(params,pcyton);
		double original = getAreaDiff(); //our reference score

		for(int i = 0; i < Constants.nparams; i++)
		{
			if(params[i] > lims[i][1]) params[i] = lims[i][1];
			if(params[i] < lims[i][0]) params[i] = lims[i][0];
			System.err.println("Sensing param "+i);
			double current = original;
			double max_step = lims[i][1]-params[i];
			double step = max_step/1000;
			double[] ps= Methods.duplicate(params);
			while(ps[i] < lims[i][1] && step > max_step/10000)
			{
				ps[i]+=step;
				cytonToFluorescence(ps,pcyton);
				current = getAreaDiff();
				if(current-original > tollerance)
				{
					ps[i]-= step; //step back 
					step=step/2; //and reduce step size
				}
				else
				{
					step=Math.min(step*2,lims[i][1]-(ps[i]));
				}
				//System.err.println("step is "+step+" param is "+ps[i]);
			}
			sensitivities[i][1] = ps[i];
			max_step =params[i]-lims[i][0];
			step = max_step/1000;
			ps= Methods.duplicate(params);
			while(ps[i] > lims[i][0] && step > max_step/10000)
			{
				ps[i]-=step;
				cytonToFluorescence(ps,pcyton);
				current = getAreaDiff();
				if(current-original > tollerance)
				{
					ps[i]+= step; //step back 
					step=step/2; //and reduce step size
				}
				else
				{
					step=Math.min(step*2,(ps[i]-lims[i][0]));
				}
			//	System.err.println("step is "+step+" param is "+ps[i]);
			}
			sensitivities[i][0] = ps[i];
		}
		return sensitivities;
	}
	

	public double getAreaDiff() {
		double total_diff=0;
		double total_cells=0;
		for(int i = 0; i < timePoints.size(); i++)
		{
			for(int j = 0; j < timePoints.get(i).getReplicates().size(); j++)
			{
				double diff = timePoints.get(i).getReplicates().get(j).getAreaDiff();
				double cells = timePoints.get(i).getReplicates().get(j).getCellsPerMl();
				total_diff+=diff*cells;
				total_cells+=cells;
			}
		}
		return(total_diff/total_cells);
	}
	

	
	
	
	public double getCytonScoreUnweighted() {
		double score = 0;
		for(int i = 0; i < timePoints.size(); i++)
		{
			double tscores = 0;
			for(int j = 0; j < timePoints.get(i).getReplicates().size(); j++)
			{
				tscores+=timePoints.get(i).getReplicates().get(j).getCytonScore2();
			}
			double weightedScores = tscores/timePoints.get(i).getReplicates().size();
			score+=weightedScores;
		}
		return score/timePoints.size();  //weighted weighted score
	}

	
	public double getCytonScore2() {
		double totalCells = 0;
		double score = 0;
		for(int i = 0; i < timePoints.size(); i++)
		{
			double timeCells = 0;
			double tscores = 0;
			for(int j = 0; j < timePoints.get(i).getReplicates().size(); j++)
			{
				double rCount = timePoints.get(i).getReplicates().get(j).getCellsPerMl();
				tscores+=rCount*timePoints.get(i).getReplicates().get(j).getCytonScore2();
				timeCells+=rCount;
			}
			double weightedScores = tscores/timeCells;
			timeCells=timeCells/timePoints.get(i).getReplicates().size(); //this is now the average number of cells in these replicates
			totalCells+=timeCells;
			score+=timeCells*weightedScores;
		}
		return score/totalCells;  //weighted weighted score
	}
	
	
	public double getCytonScoreWeightedRMSD() {
		double totalCells = 0;
		double score = 0;
		for(int i = 0; i < timePoints.size(); i++)
		{
			double timeCells = 0;
			double tscores = 0;
			for(int j = 0; j < timePoints.get(i).getReplicates().size(); j++)
			{
				double rCount = timePoints.get(i).getReplicates().get(j).getCellsPerMl();
				tscores+=rCount*timePoints.get(i).getReplicates().get(j).getCytonScoreWeightedRMSD();
				timeCells+=rCount;
			}
			double weightedScores = tscores/timeCells;
			timeCells=timeCells/timePoints.get(i).getReplicates().size(); //this is now the average number of cells in these replicates
			totalCells+=timeCells;
			score+=timeCells*weightedScores;
		}
		return score/totalCells;  //weighted weighted score
	}
	
	public double getCytonScoreUnweightedRMSD() {

		double score = 0;
		int count = 0;
		for(int i = 0; i < timePoints.size(); i++)
		{
			for(int j = 0; j < timePoints.get(i).getReplicates().size(); j++)
			{
				score+=timePoints.get(i).getReplicates().get(j).getCytonScoreWeightedRMSD();
				count++;
			}
		}
		return score/count;  //weighted weighted score
	}
	
	public double getCytonScoreUnweightedSD() {

		double score = 0;
		int count = 0;
		for(int i = 0; i < timePoints.size(); i++)
		{
			for(int j = 0; j < timePoints.get(i).getReplicates().size(); j++)
			{
				score+=timePoints.get(i).getReplicates().get(j).getCytonScoreWeightedRMSD();
				count++;
			}
		}
		return score/count;  //weighted weighted score
	}
	
	public double getCytonScoreWeightedTrueAreaDiff() {
		double totalCells = 0;
		double score = 0;
		for(int i = 0; i < timePoints.size(); i++)
		{
			double timeCells = 0;
			double tscores = 0;
			for(int j = 0; j < timePoints.get(i).getReplicates().size(); j++)
			{
				double rCount = timePoints.get(i).getReplicates().get(j).getCellsPerMl();
				tscores+=rCount*timePoints.get(i).getReplicates().get(j).getTrueAreaDiff();
				timeCells+=rCount;
			}
			double weightedScores = tscores/timeCells;
			timeCells=timeCells/timePoints.get(i).getReplicates().size(); //this is now the average number of cells in these replicates
			totalCells+=timeCells;
			score+=timeCells*weightedScores;
		}
		return score/totalCells;  //weighted weighted score
	}
	
	public double getPValue()
	{
		ArrayList<Double> vals = new ArrayList<Double>();
		for(int t= 0; t < timePoints.size(); t++)
		{
			double val = timePoints.get(t).getPValue();
			int i = 0;
			for(i = 0; i < vals.size(); i++)
			{
				if(vals.get(i)> val)
					break;
			}
			vals.add(i,val);
		}
		if(vals.size()%2 == 1)
			return vals.get(vals.size()/2);
		else
			return  (vals.get(vals.size()/2)+vals.get(vals.size()/2-1))/2;
	}


	
	
}
