/*
Copyright (c)  2013   The Regents of the University of California
All Rights Reserved
 
Permission to copy, modify and distribute any part of this FlowMax for educational, research and non-profit purposes, without fee, and without a written agreement is hereby granted, provided that the above copyright notice, this paragraph and the following three paragraphs appear in all copies.
Those desiring to incorporate this FlowMax into commercial products or use for commercial purposes should contact the Technology Transfer Office, University of California, San Diego, 9500 Gilman Drive, Mail Code 0910, La Jolla, CA 92093-0910, Ph: (858) 534-5815, FAX: (858) 534-7345, E-MAIL:invent@ucsd.edu.
IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES, INCLUDING LOST PROFITS, ARISING OUT OF THE USE OF THIS FlowMax, EVEN IF THE UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
THE FlowMax PROVIDED HEREIN IS ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.  THE UNIVERSITY OF CALIFORNIA MAKES NO REPRESENTATIONS AND EXTENDS NO WARRANTIES OF ANY KIND, EITHER IMPLIED OR EXPRESS, 
INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE, OR THAT THE USE OF THE FlowMax WILL NOT INFRINGE ANY PATENT, TRADEMARK OR OTHER RIGHTS.
*/
package FlowMax;

import java.awt.Color;
import java.awt.Font;
import java.awt.Graphics2D;
import java.awt.image.BufferedImage;
import java.io.IOException;

import javax.imageio.ImageIO;

import common.Dendogramer;
import common.GUIGuts;
import common.Gallery;
import common.Methods;
import common.Plotter;
import common.TreeNode;

public class FMGallery extends Gallery {

	Dendogramer tree;
	BufferedImage mainpic = null;
	
	public FMGallery(GUIGuts owner, Dendogramer tree) {
		super(owner);
		this.tree = tree;
		
		mainpic = Methods.getImageFromResource("MainPage.png");

	}
	
	public void makeCurrentPlot(FMTreeNode tn) {
		boolean found = false;
		for(int i = 0; i < plots.size(); i++)
		{
			if(plots.get(i).getNode() != null && plots.get(i).getNode().equals(tn))
			{
				current_plot = plots.get(i);
				found = true;
				break;
			}
		}
		if(!found)
		{
			FMPlotter p = new FMPlotter(this,tn,tree);
			addNewGraph(p);
			current_plot = p;
		}
		current_plot.updateDrawParams();

	}

	@Override
	public void drawInstructions(Graphics2D g2) {
		/*
		g2.setFont(new Font("Calibri",Font.PLAIN,20));
		g2.setColor( Color.BLACK);
		drawShadowText(g2,"Welcome to FlowMax V"+owner.version,2,15);
		g2.setFont(new Font("Calibri",Font.ITALIC,16));
		drawShadowText(g2,"a phenotyping tool for the analysis of cell population dynamics in terms of cellular parameters",2,30);
		g2.setFont(new Font("Calibri",Font.PLAIN,13));
		drawShadowText(g2,"FlowMax was created at the Signaling Systems Laboratory in San Diego, CA",2,55);
		drawShadowText(g2,"http://signalingsystems.ucsd.edu 2011-2012",2,70);
		drawShadowText(g2,"For a suscinct tutorial please refer to the accompanying pdf document: FlowMax Tutorial.pdf",2,85);
		g2.setFont(new Font("Calibri",Font.PLAIN,13));
		g2.setColor(Color.blue);
		drawShadowText(g2,"To begin, load in some FCS 3.0 files (time points) then create 1D CFSE live cell histograms.",2,110);
		drawShadowText(g2,"For each CFSE histogram: ",12,125);
		drawShadowText(g2,"1) Don't forget to click on the \"CHECK\" button",22,140);
		drawShadowText(g2,"2) Type in the experimental condition (e.g. \"WT\") and time after stimulation (hrs.)",22,155);
		drawShadowText(g2,"3) Click on the gen0 peak for each histogram",22,170);
		drawShadowText(g2,"Click on the PROLIFERATION tab, and push \"Phenotype Pcyton\" to begin model parameterization.",12,185);
		drawShadowText(g2,"Results will be accessible as files in the \"Phenotypes\" directory and as generated analyses under",12,200);
		drawShadowText(g2,"the PHENOTYPES tab after fitting finishes.",22,215);
		drawShadowText(g2,"If you are unsure about the number of times to fit, try running overnight with the maximum number",12,230);
		*/
		g2.drawImage(mainpic,0,0,this);
	}

	private void drawShadowText(Graphics2D g2, String text, int x, int y)
	{
		Color prev = g2.getColor();
		Color shadow = new Color(prev.getRed(),prev.getGreen(),prev.getBlue(),50);
		g2.setColor(shadow);
		g2.drawString(text, x+1, y+1);
		g2.setColor(prev);
		g2.drawString(text, x, y);
	}
}
