/*
Copyright (c)  2013   The Regents of the University of California
All Rights Reserved
 
Permission to copy, modify and distribute any part of this FlowMax for educational, research and non-profit purposes, without fee, and without a written agreement is hereby granted, provided that the above copyright notice, this paragraph and the following three paragraphs appear in all copies.
Those desiring to incorporate this FlowMax into commercial products or use for commercial purposes should contact the Technology Transfer Office, University of California, San Diego, 9500 Gilman Drive, Mail Code 0910, La Jolla, CA 92093-0910, Ph: (858) 534-5815, FAX: (858) 534-7345, E-MAIL:invent@ucsd.edu.
IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES, INCLUDING LOST PROFITS, ARISING OUT OF THE USE OF THIS FlowMax, EVEN IF THE UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
THE FlowMax PROVIDED HEREIN IS ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.  THE UNIVERSITY OF CALIFORNIA MAKES NO REPRESENTATIONS AND EXTENDS NO WARRANTIES OF ANY KIND, EITHER IMPLIED OR EXPRESS, 
INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE, OR THAT THE USE OF THE FlowMax WILL NOT INFRINGE ANY PATENT, TRADEMARK OR OTHER RIGHTS.
*/
package FlowMax;

import java.awt.Color;
import java.awt.Dimension;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.MouseEvent;
import java.io.IOException;

import javax.imageio.ImageIO;
import javax.swing.ImageIcon;
import javax.swing.JButton;
import javax.swing.JOptionPane;

import common.Dendogramer;
import common.Dialogs;
import common.Gallery;
import common.Graph;
import common.Methods;
import common.Plotter;
import common.TreeNode;

public class FMPlotter extends Plotter{

	Dendogramer tree;
	
	public FMPlotter(FMGallery parent, final FMTreeNode tn, final Dendogramer tree) {
		super(parent, tn);
		this.tree = tree;
		JButton includeBtn = new JButton(Methods.getImageIcon(("check.png")));
		buttons.put("check",includeBtn);
		if(tn.yindex != -1)includeBtn.setEnabled(false);
		includeBtn.addActionListener(new ActionListener(){
			
			public void actionPerformed(ActionEvent arg0) {
				// TODO Auto-generated method stub
				if(node.yindex == -1)
				{
					String cond = null;
					while(cond == null)
					{
						cond = JOptionPane.showInputDialog("Experimental condtion across time?","WT");
					}
					double dtime = Dialogs.getDouble("Timepoint (hrs):", 0, 99999, 0.5);
					node.getSourceFCS().setTime(dtime);
					node.getSourceFCS().setCondition(cond);
					mode = 6;
					getButton("check").setEnabled(false);
					grf.updateUI();
				}

			}
			
		});
		includeBtn.setPreferredSize(new Dimension(40,30));
		if(tn != null && tn.yindex != -1) includeBtn.setVisible(false);
		
		JButton setVolumeBtn = new JButton(Methods.getImageIcon(("CellConc.png")));
		buttons.put("volume",setVolumeBtn);
		setVolumeBtn.setEnabled(false);
		setVolumeBtn.addActionListener(new ActionListener(){

			public void actionPerformed(ActionEvent arg0) {
				// TODO Auto-generated method stub
				try{
				int currentCellCountperMl =(int)( node.getPoints().size()/(node.getSourceFCS().getVolume()/1000000.0));
				String userInput =JOptionPane.showInputDialog(self,"Cell count per ml",""+currentCellCountperMl);
				while(userInput == null)
					userInput =JOptionPane.showInputDialog(self,"Cell count per ml in current gate",""+currentCellCountperMl);
				double adjustedCellCount =Double.parseDouble(userInput);
				node.getSourceFCS().setVolume((int)(node.getSourceFCS().getVolume()*(currentCellCountperMl/adjustedCellCount)));
				}
				catch(Exception e)
				{
					JOptionPane.showMessageDialog(self, "There was an error trying to set the cell count. Try again with some valid numerical input.");
				}
			}
		});
		setVolumeBtn.setPreferredSize(new Dimension(40,30));
		if(tn != null && tn.yindex != -1) setVolumeBtn.setVisible(false);
		
		bottom.add(setVolumeBtn);
		bottom.add(includeBtn);
		invalidate();
		updateUI();
	}
	
	protected void initializeGraph()
	{
		grf = new FMGraph((FMTreeNode) node, this);
	}
	
	protected void pmouseReleased(MouseEvent evt)
	{
		super.pmouseReleased(evt);
		//System.err.println("Mouse released!");
		//now lets add another check for mode 6 which SETTING INITIAL PEAK MODE!
		if(mode == 6 && evt.getButton()==MouseEvent.BUTTON1)
    	{
    		//set new initial peak
    		mode = 0;
			((FMFCS)node.getSourceFCS()).setReady((FMTreeNode) node);
			getButton("check").setEnabled(true);
    		//need to recompute new bounds for the plot
    		if(node.xlog)
    		{
    			double minpower = Math.log10(node.xmin); 
    			double maxpower = Math.log10(node.xmax);
    			double f1 = ((double)evt.getX()-grf.getYMargin())/(grf.getWidth()-grf.getYMargin());
    			f1 = Math.min(1,Math.max(0,f1));
    			((FMTreeNode)node).first_peak = (int)Math.pow(10,f1*(maxpower-minpower)+minpower);
    		}
    		else
    		{
    			double minpower = node.xmin; 
    			double maxpower = node.xmax;
    			double f1 = ((double)evt.getX()-grf.getYMargin())/(grf.getWidth()-grf.getYMargin());
    			f1 = Math.min(1,Math.max(0,f1));
    			((FMTreeNode)node).first_peak = (int)(f1*(maxpower-minpower)+minpower);
				
    		}
			getButton("check").setEnabled(true);
			tree.setState(node.getSourceFCS().getRoot(),Color.green.getRGB());
			tree.updateUI();
    		grf.updateUI();
    	}
	}
	
	public void updateDrawParams()
	{
		super.updateDrawParams();
		if(node.yindex==-1)
		{
			getButton("check").setVisible(true);
			getButton("volume").setVisible(true);
		}
		else
		{
			getButton("check").setVisible(false);
			getButton("volume").setVisible(false);
		}
		updateUI();
	}

	
}
