/*
Copyright (c)  2013   The Regents of the University of California
All Rights Reserved
 
Permission to copy, modify and distribute any part of this FlowMax for educational, research and non-profit purposes, without fee, and without a written agreement is hereby granted, provided that the above copyright notice, this paragraph and the following three paragraphs appear in all copies.
Those desiring to incorporate this FlowMax into commercial products or use for commercial purposes should contact the Technology Transfer Office, University of California, San Diego, 9500 Gilman Drive, Mail Code 0910, La Jolla, CA 92093-0910, Ph: (858) 534-5815, FAX: (858) 534-7345, E-MAIL:invent@ucsd.edu.
IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES, INCLUDING LOST PROFITS, ARISING OUT OF THE USE OF THIS FlowMax, EVEN IF THE UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
THE FlowMax PROVIDED HEREIN IS ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.  THE UNIVERSITY OF CALIFORNIA MAKES NO REPRESENTATIONS AND EXTENDS NO WARRANTIES OF ANY KIND, EITHER IMPLIED OR EXPRESS, 
INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE, OR THAT THE USE OF THE FlowMax WILL NOT INFRINGE ANY PATENT, TRADEMARK OR OTHER RIGHTS.
*/
package FlowMax;
import common.*;

import java.awt.Color;
import java.io.PrintWriter;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.Scanner;

import common.FCS;
import common.Gate;
import common.Matrix;
import common.Point;
import common.TreeNode;

public class FMTreeNode extends TreeNode {

	public int first_peak = 0;
	
	public FMTreeNode(LinkedList<Point> points, FMFCS source) {

		this.points = points;
		this.parent = null;
		sourceFCS = source;
		changeXIndex(xindex);
		changeYIndex(yindex);
	}
	
	public FMFCS getSourceFCS()
	{
		return (FMFCS)sourceFCS;
	}
	
	public FMTreeNode clone() {
		FMTreeNode clone = new FMTreeNode(gate.clone());
		clone.recalculatePoints = true;
		clone.xlog = xlog;
		clone.ylog = ylog;
		clone.xindex = xindex;
		clone.yindex = yindex;
		clone.xmin = xmin;
		clone.ymin = ymin;
		clone.margin = margin;
		clone.parent = parent;
		clone.sourceFCS = sourceFCS;
		clone.xmax = xmax;
		clone.ymax = ymax;
		clone.first_peak = first_peak;
		clone.changeXIndex(xindex);
		clone.changeYIndex(yindex);
		Iterator<TreeNode> itr = leaves.iterator();
		while (itr.hasNext()) {
			clone.leaves.add(itr.next().clone());
		}
		return clone;
	}
	
	public FMTreeNode(Scanner s, FMFCS source)
	{
		this.sourceFCS = source;
		xmax = source.getMaxes()[0];
		ymax = source.getMaxes()[1];
		if(s.hasNextLine())
		{
			String next = s.nextLine();
			if(!next.startsWith("TREENODE:"))
				return;
			System.err.println("Loading a new treenode");
			String gates = s.nextLine();
			if(gates.startsWith("GATE:"))
			{
				String[] vals = gates.substring(5).split(",");
				if(vals.length < 6 || vals[5].compareTo("#")==0)
					gate = new Gate(Integer.parseInt(vals[1]),Integer.parseInt(vals[2]),new Boolean(vals[3]), new Boolean(vals[4]),vals[0],0);
				else
					gate = new Gate(Integer.parseInt(vals[1]),Integer.parseInt(vals[2]),new Boolean(vals[3]), new Boolean(vals[4]),vals[0],Integer.parseInt(vals[5]));
				System.err.println("Gate loaded: "+gate.name);
				next = s.nextLine();
				int points_added = 0;
				while (!next.startsWith("ENDGATE")) {
					String[] spliter = next.split(",");
					if (vals.length == 6
							|| vals.length == 7
							|| (vals.length == 5 && gate.logx == false && gate.logy == false))
					{
							gate.addPoint(Double.parseDouble(spliter[0]),
									Double.parseDouble(spliter[1]));
							points_added++;
					}
					else if (vals.length == 5 && gate.logx == true
							&& gate.logy == false)
					{
						gate.addPoint(
								Math.pow(10, Double.parseDouble(spliter[0])),
								Double.parseDouble(spliter[1]));
						points_added++;
					}
					else if (vals.length == 5 && gate.logx == false
							&& gate.logy == true)
					{
						gate.addPoint(Double.parseDouble(spliter[0]),
								Math.pow(10, Double.parseDouble(spliter[1])));
						points_added++;
					}
					else if (vals.length == 5 && gate.logx == true
							&& gate.logy == true)
					{
						gate.addPoint(
								Math.pow(10, Double.parseDouble(spliter[0])),
								Math.pow(10, Double.parseDouble(spliter[1])));
						points_added++;
					}

					// System.err.println(spliter[0]+","+spliter[1]);
					next = s.nextLine();
				}
				if(points_added == 1) //this is actually a 1D gate!
				{
					gate.gatetype  =2;
				}
			}
			else
				next = gates;
			
			{
				System.err.println("Endgate");
				String details = next;
				if(details.compareTo("ENDGATE")==0) details = s.nextLine(); //otherwise details is already loaded since we didn't have gate information.
				//if we are dealing with the old way of saving (matrix for each TreeNode), then we have to check if the matrix is non identity and if so assign to FCS... otherwise skip
				if(details.startsWith("INVERSE:"))
				{
					String[] es = details.substring(9).split(" ");
					int dim = (int)Math.sqrt(es.length);
					double[][] arr = new double[source.getNames().length][source.getNames().length];
					for(int i = 0; i < source.getNames().length; i++)
						arr[i][i] = 1.0;
					boolean identity = true;
					for(int i = 0; i < es.length; i++)
					{
						double value = Double.parseDouble(es[i]);
						int row = i/dim+Math.min(source.getNames().length-dim,2);
						int col = i%dim+Math.min(source.getNames().length-dim,2);
						arr[row][col] = value;
						if((value != 1 && row == col )||(value != 0 && row !=col))
							identity =false;
					}
					if(!identity)  //we only want to replace the matrix if its different from identity
						source.setInverse(new Matrix(arr));
					details = s.nextLine();
				}
				String[] split = details.split(",");
				if(split.length==10)
				{
					first_peak = Integer.parseInt(split[0]);
					xlog = new Boolean(split[1]);
					ylog = new Boolean(split[2]);
					xindex = Integer.parseInt(split[3]);
					yindex = Integer.parseInt(split[4]);
					xmin = Integer.parseInt(split[5]);
					ymin = Integer.parseInt(split[6]);
					xmax = Integer.parseInt(split[7]); 
					ymax = Integer.parseInt(split[8]);
					margin = Integer.parseInt(split[9]);
				}
				else
				{
					xlog = new Boolean(split[0]);
					ylog = new Boolean(split[1]);
					xindex = Integer.parseInt(split[2]);
					yindex = Integer.parseInt(split[3]);
					xmin = Integer.parseInt(split[4]);
					ymin = Integer.parseInt(split[5]);
					xmax = Integer.parseInt(split[6]); 
					ymax = Integer.parseInt(split[7]);
					margin = Integer.parseInt(split[8]);
				}
				if(first_peak > 0)
				{
					source.setReady(this);
				}
				String children = s.nextLine();
				if(children.startsWith("ENDTREENODE:"))
				{
					int count = Integer.parseInt(children.substring(12));
					System.err.println("Loading "+count+" children");
					for(int i = 0; i < count; i++)
					{
						FMTreeNode ntn = new FMTreeNode(s,source);
						ntn.parent = this;
						leaves.add(ntn);
					}
				}
				else
				{
					System.err.println("NO ENDTREENODE");
				}
			}
		}
	}
	
	public void saveSelf(PrintWriter pw) {
		// TODO Auto-generated method stub
		pw.println("TREENODE:");
		if (gate != null) {
			pw.println("GATE:" + gate.name + "," + gate.x + "," + gate.y + ","
					+ gate.logx + "," + gate.logy + "," + gate.c.getRGB()
					+ ",#");
			for (int i = 0; i < gate.p.xpoints.length; i++)
				pw.println(gate.p.xpoints[i] + "," + gate.p.ypoints[i]);
			pw.println("ENDGATE");
		}

		pw.println(first_peak+","+xlog + "," + ylog + "," + xindex + "," + yindex + "," + xmin
				+ "," + ymin + "," + xmax + "," + ymax + "," + margin);
		pw.println("ENDTREENODE:" + leaves.size());
		Iterator<TreeNode> itr = leaves.iterator();
		while (itr.hasNext())
			itr.next().saveSelf(pw);
	}

	public FMTreeNode(Gate g) {
		super(g);
	}

	public FMTreeNode addAnd(Gate g) {
		FMTreeNode newNode = new FMTreeNode(g);
		newNode.initialize(this);
		leaves.add(newNode);
		return newNode;
	}

	public FMTreeNode getReady() {
		if(first_peak != 0)
			return this;
		else
		{
			for(TreeNode n : leaves)
			{
				FMTreeNode result = ((FMTreeNode)n).getReady();
				if(result != null) return result;
			}
		}
		return null;
	}

}
