/*
Copyright (c)  2013  The Regents of the University of California
All Rights Reserved
 
Permission to copy, modify and distribute any part of this FlowMax for educational, research and non-profit purposes, without fee, and without a written agreement is hereby granted, provided that the above copyright notice, this paragraph and the following three paragraphs appear in all copies.
Those desiring to incorporate this FlowMax into commercial products or use for commercial purposes should contact the Technology Transfer Office, University of California, San Diego, 9500 Gilman Drive, Mail Code 0910, La Jolla, CA 92093-0910, Ph: (858) 534-5815, FAX: (858) 534-7345, E-MAIL:invent@ucsd.edu.
IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES, INCLUDING LOST PROFITS, ARISING OUT OF THE USE OF THIS FlowMax, EVEN IF THE UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
THE FlowMax PROVIDED HEREIN IS ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.  THE UNIVERSITY OF CALIFORNIA MAKES NO REPRESENTATIONS AND EXTENDS NO WARRANTIES OF ANY KIND, EITHER IMPLIED OR EXPRESS, 
INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE, OR THAT THE USE OF THE FlowMax WILL NOT INFRINGE ANY PATENT, TRADEMARK OR OTHER RIGHTS.
*/
package FlowMax;
import java.awt.BasicStroke;
import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Container;
import java.awt.Cursor;
import java.awt.Dimension;
import java.awt.FlowLayout;
import java.awt.Font;
import java.awt.Graphics2D;
import java.awt.GridLayout;
import java.awt.Insets;
import java.awt.Toolkit;
import java.awt.datatransfer.DataFlavor;
import java.awt.datatransfer.Transferable;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.ComponentEvent;
import java.awt.event.ComponentListener;
import java.awt.event.ItemEvent;
import java.awt.event.ItemListener;
import java.awt.event.KeyEvent;
import java.awt.event.KeyListener;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.awt.image.BufferedImage;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileWriter;
import java.io.IOException;
import java.io.PrintWriter;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.Random;
import java.util.Scanner;

import javax.imageio.ImageIO;
import javax.swing.DefaultListModel;
import javax.swing.JButton;
import javax.swing.JCheckBox;
import javax.swing.JComboBox;
import javax.swing.JFileChooser;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JList;
import javax.swing.JSeparator;
import javax.swing.SwingConstants;
import javax.swing.TransferHandler;
import javax.swing.UIManager;
import javax.swing.UnsupportedLookAndFeelException;

import javax.swing.JMenuItem;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JPopupMenu;
import javax.swing.JScrollPane;

import javax.swing.JSpinner;
import javax.swing.JTabbedPane;
import javax.swing.JTextArea;

import javax.swing.SpinnerNumberModel;

import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;
import javax.swing.filechooser.FileFilter;

import common.*;

/**
 * Main class and GUI for FlowMax, a phenotyping tool described in:
 * 
 * Shokhirev MN and Hoffmann A. Quantitative deconvolution of CFSE cell population time course data with confidence. 2012. 
 *  
 * @author Max Shokhirev
 *
 */

public class FlowMax extends GUIGuts implements FluorescenceFitListener, FitListener{

	private static final long serialVersionUID = 1L;

	private static final int nparams = 14;

	double[][] limits = {{1.7916,4.95174345},{0.01667,0.832555},{1.7916,4.95174345},{0.01667,0.832555},{1.7916,4.95174345},{0.01667,0.832555},{1.7916,4.95174345},{0.01667,0.832555},{-3,8},{0.001,10},{0.001,1},{20000,3000000},{0,0},{0,0}};
	double[][] cspinners = {{6,24,200,200},{0.1,1,200,200},{6,6,200,200},{0.1,1,200,200},{6,12,200,200},{0.1,1,200,200},{6,6,200,200},{0.1,1,200,200},{-3,-3,8,8},{0.001,0.001,10,10},{0.001,0.001,1,1},{10000,100000,1000000,5000000},{0,0,0,1},{0,0,0,1}};

	
	public static final double dt = 0.5;
	
	public static final int maximum = 16777216;
	public static final int minimum = 1;
	String separator = System.getProperty("line.separator");
	
	//SELF
	FlowMax self;
	
	//Memory
	ArrayList<FMFCS> runs = new ArrayList<FMFCS>();
	HashMap<String, ArrayList<FMFCS>> map = new HashMap<String,ArrayList<FMFCS>>() ;
	
	//GUI
	private Dendogramer tree;  // where you choose data to display
	private JScrollPane scroller;
	private FMGallery plot;  // draws the graphs
	private Drawer draw;  // the actual params from solutions
	private Counter counter; //solution generational and total counts
	private JButton saveBtn;
	private JButton loadBtn;
	private Compensator compensator;
	private JFrame colorFrame;
	private Colorer colorer;

	//NON MAXLIB GUI STUFF
	private JTabbedPane tabber;
	private JPanel proliferator;
	private ProliferationDisplay profDisp;
	private JScrollPane scroller2;
	private JComboBox exptCombo;
	private JSpinner genSpinner;
	private JSpinner sizeSpinner;
	private JSpinner countsSpinner;
	private JSpinner[] fluorescenceSpinners;
	private JSpinner repeatsSpinner;
	private JSpinner[] cytonSpinners = new JSpinner[nparams*2];
	private JCheckBox drawBox;
	private JCheckBox colorBox;
	private JTextArea stats;
	private JCheckBox advancedMode;
	private JPanel Pright;
	private ClickResponder cr;
	private JPanel advancedPanel;
	private JButton PhenfcytonBtn;
	private JButton PhenCytonBtn;
	private JButton fitFluorescenceBtn;
	private JButton CancelBtn;
	private JButton LoadParams,SaveParams;

	
	private ClickList slist;
	private ArrayList<Double> answerScores = new ArrayList<Double>();
	private ArrayList<double[]> answers = new ArrayList<double[]>();
	private ArrayList<double[]> answerTimes = new ArrayList<double[]>();
	private ArrayList<double[][][]> answerCounts = new ArrayList<double[][][]>();  //timepoint, replicate, gen!!!
	private ArrayList<ArrayList<double[]>> answerCandidates = new ArrayList<ArrayList<double[]>>();  //the real solutions
	
	private ArrayList<String> answerNames = new ArrayList<String>();
	private ArrayList<String> conditionNames = new ArrayList<String>();
	
	private JSpinner timeSpinnerP;
	private JSpinner gensSpinnerP;
	private JSpinner vSpinnerP;
	private JSpinner drawsizeSpinnerP;
	private JSpinner timeSpinnerC;
	private JSpinner gensSpinnerC;
	private JSpinner vSpinnerC;
	private JSpinner drawsizeSpinnerC;
	private JSpinner maxHeightSpinnerC;
	private JSpinner fsmaxSpinner;
	
	private JSpinner gensSpinnerComp;
	
	private JCheckBox tboxP;
	private JCheckBox nameboxP;
	private JCheckBox lblboxP;
	private JCheckBox overlayBoxP;
	private JCheckBox maxBoxP;
	private JCheckBox fcytonboxP;
	private JComboBox drawBoxP;
	private JCheckBox tboxC;
	private JCheckBox nameboxC;
	private JCheckBox lblboxC;
	private JCheckBox totalsBox;
	private JCheckBox fcytonboxC;
	private JCheckBox actualSP;
	private JCheckBox actualSC;
	private JButton[] overlayColorButtons;

	
	
	private JButton vButton;  //visualize params 
	private JButton lButton;  //load phens
	private JButton cButton;  //compare button!
	private JButton mButton; //morph
	private JCheckBox[] morphParams;  //the params to be morphed
	private JButton colButton; //color chooser for counter
	private JButton countsButton;  //draw counts
	
	
	private JButton loadFiles;
	private JTabbedPane drawTabs;
	
	private JPopupMenu tmenu,dmenu,cmenu; //tree menu and Draw menu
	private JMenuItem copyGateMI, pasteGateMI;
	private JFrame optFrame;

	public Dendogramer getTree() {
		return tree;
	}

	public HashMap<String, ArrayList<FMFCS>> getMap() {
		return map;
	}

	public FMGallery getPlot() {
		return plot;
	}

	public Drawer getDraw() {
		return draw;
	}

	public Compensator getCompensator() {
		return compensator;
	}

	public FlowMax()
	{
		self = this;
		setUpGUI();

	}
	


	private void loadEverything(File f)
	{
		try {
			setCursor(Cursor.WAIT_CURSOR);

			System.err.println("Loading workspace from file: "+f.getAbsolutePath());
			runs.clear();
			if(tree.getRs() != null)
				tree.getRs().clear();
			tree.getList().clear();
			Scanner s = new Scanner(f);
			FMFCS fcs = null;
			File root = null;
			while(s.hasNextLine())
			{
				String next = s.nextLine();
				if(next.compareTo("TREE")==0)
					break;
				else if(next.compareTo("COMPENSATION")==0)
				{
					String[] first_row = s.nextLine().split(",");
					Matrix m = Matrix.identityMatrix(first_row.length);
					for(int i = 0; i < m.getNcol(); i++)
					{
						m.setElement(0, i, Double.parseDouble(first_row[i]));
					}
					for(int r = 0; r < m.getNrow()-1; r++) //the rest of the rows need to be loaded in!
					{
						String[] row = s.nextLine().split(",");
						for(int i = 0; i < m.getNcol(); i++)
						{
							m.setElement(r+1, i, Double.parseDouble(row[i]));
						}
					}
					//at this point we have a fully defined matrix!
					if(fcs != null) 
					{
						fcs.setInverse(m);
					}
				}
				else
				{
					String[] ss = next.split(";");
					File ff = new File(ss[0].replace('|', Constants.fseparator.charAt(0)));
					while(!ff.exists())
					{
						if(root == null)
						{
							root =Dialogs.getFolderFile("Couldn't find the files, please locate a real folder in the path:"+ff.getAbsolutePath());


						}
						//lets go back until we find a valid file
						File prefix = ff.getParentFile();
						while(true)
						{

							String suffix = ff.getAbsolutePath().substring(prefix.getAbsolutePath().length());
							File hybrid = new File(root.getAbsolutePath()+suffix);
							if(hybrid.exists())
							{
								ff = hybrid;
								break;
							}
							else
								prefix=prefix.getParentFile();
						}
					}
					fcs = new FMFCS(Methods.processFCSFileWIthNoData(ff));
					fcs.setSourceFile(ff);
					fcs.setCondition(ss[1]);
					fcs.setTime(Double.parseDouble(ss[2]));
					if(ss.length > 3)
						fcs.setVolume(Integer.parseInt(ss[3]));
					runs.add(fcs);
				}
			}
			tree.setRs(runs);
			//s.nextLine();
			for(int i = 0; i < tree.getRs().size(); i++)
			{
				FMTreeNode newnode = new FMTreeNode(s,(FMFCS)tree.getRs().get(i));
				tree.getRs().get(i).setRoot(newnode);
				if(((FMFCS)tree.getRs().get(i)).getReady() != null)
				{
				//	tree.addIcon(newnode,ImageIO.read(getClass().getResource("check.png")));
					tree.setState(newnode,false,Color.green.getRGB()); 
				}
				//tree.getRs().get(i).getRoot().reloadDataFromFile();
				//tree.getRs().get(i).getRoot().propagateFromRoot();
				//tree.getRs().get(i).propagateColors();
			}
			setCursor(Cursor.DEFAULT_CURSOR);
		} catch (Exception E) {
			// TODO Auto-generated catch block
			
			JOptionPane.showMessageDialog(this, "ERROR:"+E.getMessage());
			setCursor(Cursor.DEFAULT_CURSOR);
		}
		
	}
		
	/** 
	 * Runs whenever someone clicks... allows for non-blocking handling of commands from the GUI
	 * @author Max
	 *
	 */
	private class ClickResponder extends Thread
	{
		int mode = 0;
		
		public ClickResponder(int mode)
		{
			super();
			this.mode = mode;
		}
		
		public void run() {
		 if(mode == 7)  //If you want to visualize a phenotype
			{
				//drawTabs.setSelectedIndex(0);
				
				draw.paramList.clear();
				draw.paramList2.clear();
				draw.names.clear();
				draw.clist.clear();
				invalidate();
				//lets go through and update our drawer based on what is selected
				for(int i = 0; i < slist.getSelectedIndices().length; i++)
				{
					draw.addParameters(answers.get(slist.getSelectedIndices()[i]), answerNames.get(slist.getSelectedIndices()[i]),answerScores.get(slist.getSelectedIndices()[i]),answerCandidates.get(slist.getSelectedIndices()[i]));
				}
				
				draw.redraw = true;
				//now lets try to visualize this
				draw.gens = (Integer)gensSpinnerP.getValue();
				draw.size = (Integer)drawsizeSpinnerP.getValue();
				draw.v = (Integer)vSpinnerP.getValue();
				draw.time = (Integer)timeSpinnerP.getValue();
				
				draw.setPreferredSize(new Dimension(draw.size+100, slist.getSelectedIndices().length*(draw.size+30)));
				drawTabs.getRootPane().getContentPane().invalidate();
				vButton.setEnabled(false);
				vButton.updateUI();
				draw.setNamebox(nameboxP);
				draw.setLblbox(lblboxP);
				draw.setTbox(tboxP);
				draw.setOverlayBox(overlayBoxP);			
				draw.setpCytonBox(fcytonboxP);
				draw.setDrawBox(drawBoxP);
				draw.setMaxBox(maxBoxP);
				Color[] ocolors = new Color[overlayColorButtons.length];
				for(int i = 0; i < overlayColorButtons.length; i++)
					ocolors[i]= overlayColorButtons[i].getForeground();
				draw.setOverlayColors(ocolors);
				draw.setMaxPercent((Integer)fsmaxSpinner.getValue());
				draw.realSolutions = !actualSP.isSelected();
				draw.paintComponent(draw.getGraphics());
				vButton.setEnabled(true);
				vButton.updateUI();
				
			}
			else if(mode == 9)  //If you want to visualize a phen counts
			{
				setCursor(Cursor.WAIT_CURSOR);
				drawTabs.setSelectedIndex(1);
				counter.paramList.clear();
				counter.paramList2.clear();
				counter.names.clear();
				counter.averages.clear();
				counter.sds.clear();
				counter.times.clear();
				counter.clist.clear();
				invalidate();
				//lets go through and update our drawer based on what is selected
				for(int i = 0; i < slist.getSelectedIndices().length; i++)
				{
					if(answerTimes.get(slist.getSelectedIndices()[i]) != null)
					{
						double[] averages = new double[answerCounts.get(slist.getSelectedIndices()[i]).length];
						double[] sds = new double[answerCounts.get(slist.getSelectedIndices()[i]).length];
						double[] times = new double[answerCounts.get(slist.getSelectedIndices()[i]).length];
						for(int t = 0; t < answerCounts.get(slist.getSelectedIndices()[i]).length; t++)
						{
							double[] sums = new double[answerCounts.get(slist.getSelectedIndices()[i])[t].length];
							for(int g = 0; g < answerCounts.get(slist.getSelectedIndices()[i])[t][0].length; g++)
							{
								for(int r = 0; r < sums.length; r++)
									sums[r]+=answerCounts.get(slist.getSelectedIndices()[i])[t][r][g];
							}
							averages[t]= Methods.average(sums);
							sds[t] = Methods.std(sums);
							times[t] = answerTimes.get(slist.getSelectedIndices()[i])[t];
						}
						counter.addParameters(answers.get(slist.getSelectedIndices()[i]), answerNames.get(slist.getSelectedIndices()[i]),answerScores.get(slist.getSelectedIndices()[i]),averages,sds,times, answerCandidates.get(slist.getSelectedIndices()[i]));
					}
					else
						counter.addParameters(answers.get(slist.getSelectedIndices()[i]), answerNames.get(slist.getSelectedIndices()[i]),answerScores.get(slist.getSelectedIndices()[i]), answerCandidates.get(slist.getSelectedIndices()[i]));
				}
				counter.redraw = true;
				//now lets try to visualize this
				counter.gens = (Integer)gensSpinnerC.getValue();
				counter.size = (Integer)(drawsizeSpinnerC.getValue());
				counter.v = (Integer)vSpinnerC.getValue();
				counter.time = (Integer)timeSpinnerC.getValue();
				counter.max = ((Integer)maxHeightSpinnerC.getValue())*(10000.0);
				counter.transparency = tboxC.isSelected();
				counter.nameLbs = nameboxC.isSelected();
				counter.axes = lblboxC.isSelected();
				counter.totalOnly = totalsBox.isSelected();
				counter.fcyton = fcytonboxC.isSelected();
				counter.realSolutions = !actualSC.isSelected();
				counter.setPreferredSize(new Dimension(counter.size+100, slist.getSelectedIndices().length*(counter.size+40)));
				drawTabs.getRootPane().getContentPane().invalidate();
				
				countsButton.setEnabled(false);
				countsButton.updateUI();
				countsButton.setEnabled(false);
				counter.paintComponent(counter.getGraphics());
				countsButton.setEnabled(true);
				countsButton.updateUI();
				setCursor(Cursor.DEFAULT_CURSOR);
				
			}
			else if(mode == 8)  //If you want to compare a pair of phenotypes
			{
				//first we need to get the params and ranges
				
				if(slist.getSelectedIndices().length < 2 )
				{
					JOptionPane.showMessageDialog(null, "You need to select two phenotypes!");
					return;
				}
				PhenfcytonBtn.setVisible(false);
				PhenCytonBtn.setVisible(false);
				CancelBtn.setVisible(true);
				//drawfcyton.setVisible(false);
				//drawCyton.setVisible(false);
				updateFCSMapping();
				if(map.get(conditionNames.get(slist.getSelectedIndices()[0]))== null)
				{
					while(map.get(conditionNames.get(slist.getSelectedIndices()[0])) == null)
					{
						if(map.keySet().size() > 1)
						{
							Iterator<String> itr = map.keySet().iterator();
							String newname = Dialogs.getSelection(itr,"Which CFSE time course corresponds to "+answerNames.get(slist.getSelectedIndices()[0]));
							conditionNames.set(slist.getSelectedIndices()[0], newname);
						}
						else if(map.keySet().size() ==1)
							conditionNames.set(slist.getSelectedIndices()[0], map.keySet().iterator().next());
						else
						{
							JOptionPane.showMessageDialog(null,"Please load the data sets that you want to test for significance with these phenotypes.");
							tabber.setSelectedIndex(0);
							return;
						}
					}
				}
				if(map.get(conditionNames.get(slist.getSelectedIndices()[1]))== null)
				{
					while(map.get(conditionNames.get(slist.getSelectedIndices()[1])) == null)
					{
						if(map.keySet().size() > 1)
						{
							Iterator<String> itr = map.keySet().iterator();
							String newname = Dialogs.getSelection(itr,"Which CFSE time course corresponds to "+answerNames.get(slist.getSelectedIndices()[1]));
							conditionNames.set(slist.getSelectedIndices()[1], newname);
						}
						else if(map.keySet().size() ==1)
							conditionNames.set(slist.getSelectedIndices()[1], map.keySet().iterator().next());
						else
						{
							JOptionPane.showMessageDialog(null,"Please load the data sets that you want to test for significance with these phenotypes.");
							tabber.setSelectedIndex(0);
						}
					}
				}
				boolean sameData = conditionNames.get(slist.getSelectedIndices()[0]).compareTo(conditionNames.get(slist.getSelectedIndices()[1]))==0;
				double[] params1r = answers.get(slist.getSelectedIndices()[0]);
				double[] params2r = answers.get(slist.getSelectedIndices()[1]);
				double[] params1 = Methods.getLogParams(params1r,nparams);
				double[] params2 = Methods.getLogParams(params2r,nparams);
				//we need to now load in the first data set an perform a cell fluroescence fit
				ArrayList<ArrayList<Double>> data = new ArrayList<ArrayList<Double>>();
				ArrayList<Double> undivided = new ArrayList<Double>();
				ArrayList<Double> times = new ArrayList<Double>();
				ArrayList<Integer> volumes = new ArrayList<Integer>();
				int gens = (Integer)gensSpinnerComp.getValue();
				System.err.println("Preparing the first dataset");
				ArrayList<FMFCS> list = map.get(conditionNames.get(slist.getSelectedIndices()[0]));
				for(int i = 0; i < list.size(); i++)
				{
					list.get(i).reloadFromFCS();
					LinkedList<Point> dat = list.get(i).getReady().getPoints();
					ArrayList<Double> logdat = new ArrayList<Double>();
					Iterator<Point> itr = dat.iterator();
					while(itr.hasNext())
						logdat.add(Math.log10(Math.max(itr.next().ps[ list.get(i).getReady().xindex],1)));
					data.add(logdat);
					undivided.add(Math.log10(list.get(i).getReady().first_peak));
					times.add(list.get(i).getTime());
					volumes.add(list.get(i).getVolume());
					list.get(i).getReady().clearAllData();
				}
				tabber.setSelectedIndex(1);
				System.err.println("Fitting the cell fluorescence parameters");
				setCursor(Cursor.WAIT_CURSOR);
				proliferationAnalysis(gens, undivided, times, volumes,data, 2,conditionNames.get(slist.getSelectedIndices()[0]));
				//now lets save the experiment for later!
				Experiment A = profDisp.getExp().clone();
				ArrayList<ArrayList<Double>> data2 = new ArrayList<ArrayList<Double>>();
				ArrayList<Double> undivided2 = new ArrayList<Double>();
				ArrayList<Double> times2 = new ArrayList<Double>();
				ArrayList<Integer> volumes2 = new ArrayList<Integer>();
				list = map.get(conditionNames.get(slist.getSelectedIndices()[1]));
				System.err.println("Loading the other data set");
				for(int i = 0; i < list.size(); i++)
				{
					LinkedList<Point> dat = list.get(i).getReady().getPoints();
					ArrayList<Double> logdat = new ArrayList<Double>();
					Iterator<Point> itr = dat.iterator();
					while(itr.hasNext())
						logdat.add(Math.log10(Math.max(itr.next().ps[ list.get(i).getReady().xindex],1)));
					data.add(logdat);
					undivided.add(Math.log10(list.get(i).getReady().first_peak));
					times.add(list.get(i).getTime());
					volumes.add(list.get(i).getVolume());
					list.get(i).getReady().clearAllData();
				}
				if(!sameData)
				{
					System.err.println("Fitting the cell fluorescence parameters");
					
					proliferationAnalysis(gens, undivided2, times2, volumes2,data2, 2,conditionNames.get(slist.getSelectedIndices()[1]));
				}
				Experiment B = profDisp.getExp().clone();
				//now we continuously generate params from the acceptable distribution and calculate the ds:
				int num_trials = 10000;
				double[] A1,A2,B1,B2;
				A1 = new double[num_trials];  //list of differences between model 1 and dataset A
				A2 = new double[num_trials];  //list of differences between model 2 and dataset A
				B1 = new double[num_trials];  //list of differences between model 1 and dataset B
				B2 = new double[num_trials];  // "
				//lets do m1 vs A,B
				Random gen = new Random();
				
				System.err.println("Generating "+num_trials+" parameter sets from the first solution and calculating differences.");
				for(int i = 0; i < num_trials; i++)
				{

					double[] params = new double[nparams];
					for(int p = 0; p < params.length; p++)
					{
						params[p] = gen.nextDouble()*(params1[p+nparams*2]+params1[p+nparams])+params1[p]-params1[p+nparams];
					}
					//now lets calculate D for A
					A.cytonToFluorescence(params, true);
					double DA = A.getAreaDiff();
					//now lets calculate D for B
					double DB = DA;
					if(!sameData)
					{
						B.cytonToFluorescence(params, true);
						DB = B.getAreaDiff();
					}
					//System.err.println("1DA = "+DA+" but DB = "+DB);
					A1[i]=DA; B1[i] = DB;
					if(i%(num_trials/100)==0) System.err.println("1DA = "+DA+" but DB = "+DB);//System.err.print(".");
				}
				System.err.println();
				System.err.println("Generating "+num_trials+" parameter sets from the second solution and calculating differences.");
				for(int i = 0; i < num_trials; i++)
				{
					//if(i%(num_trials/100)==0) System.err.print(".");
					double[] params = new double[nparams];
					for(int p = 0; p < params.length; p++)
					{
						params[p] = gen.nextDouble()*(params2[p+nparams*2]+params2[p+nparams])+params2[p]-params2[p+nparams];
					}
					//now lets calculate D for A
					A.cytonToFluorescence(params, true);
					double DA = A.getAreaDiff();
					//now lets calculate D for B
					double DB = DA;
					if(!sameData)
					{
						B.cytonToFluorescence(params,true);
						DB = B.getAreaDiff();
					}
					//System.err.println("2DA = "+DA+" but DB = "+DB);
					A2[i]=DA; B2[i] = DB;
					if(i%(num_trials/100)==0) System.err.println("1DA = "+DA+" but DB = "+DB);//System.err.print(".");
				}
				System.err.println("Sorting...");
				Arrays.sort(A1);
				Arrays.sort(A2);
				Arrays.sort(B1);
				Arrays.sort(B2);
				System.err.println("Calculating probabilities...");
				String m1 = answerNames.get(slist.getSelectedIndices()[0]);
				String m2 = answerNames.get(slist.getSelectedIndices()[1]);
				String c1 = conditionNames.get(slist.getSelectedIndices()[0]);
				String c2 = conditionNames.get(slist.getSelectedIndices()[1]);
				setCursor(Cursor.DEFAULT_CURSOR);
				File savefile = Dialogs.getSaveFile("csv", "Save the probability comparison of "+m1+" and "+m2+" ");
				try {
					PrintWriter pw = new PrintWriter(savefile);
					pw.println("Comparing phenotype:,"+m1+", and phenotype:,"+m2+", in the context of datasets:,"+c1+",and,"+c2+", on date:,"+Methods.now());
					pw.println("Param,"+m1+","+m2+",low "+m1+",low "+m2+",high "+m1+",high "+m2);
					int p = 0;
					pw.println("E[Tdiv0],"+params1r[p]+","+params2r[p]+","+params1r[nparams+p]+","+params2r[nparams+p]+","+params1r[2*nparams+p]+","+params2r[2*nparams+p]);
					p = 1;
					pw.println("sd[Tdiv0],"+params1r[p]+","+params2r[p]+","+params1r[nparams+p]+","+params2r[nparams+p]+","+params1r[2*nparams+p]+","+params2r[2*nparams+p]);
					p = 2;
					pw.println("E[Tdiv1+],"+params1r[p]+","+params2r[p]+","+params1r[nparams+p]+","+params2r[nparams+p]+","+params1r[2*nparams+p]+","+params2r[2*nparams+p]);
					p = 3;
					pw.println("sd[Tdiv1+],"+params1r[p]+","+params2r[p]+","+params1r[nparams+p]+","+params2r[nparams+p]+","+params1r[2*nparams+p]+","+params2r[2*nparams+p]);
					p = 4;
					pw.println("E[Tdie0],"+params1r[p]+","+params2r[p]+","+params1r[nparams+p]+","+params2r[nparams+p]+","+params1r[2*nparams+p]+","+params2r[2*nparams+p]);
					p = 5;
					pw.println("sd[Tdie0],"+params1r[p]+","+params2r[p]+","+params1r[nparams+p]+","+params2r[nparams+p]+","+params1r[2*nparams+p]+","+params2r[2*nparams+p]);
					p = 6;
					pw.println("E[Tdie1+],"+params1r[p]+","+params2r[p]+","+params1r[nparams+p]+","+params2r[nparams+p]+","+params1r[2*nparams+p]+","+params2r[2*nparams+p]);
					p = 7;
					pw.println("sd[Tdie1+],"+params1r[p]+","+params2r[p]+","+params1r[nparams+p]+","+params2r[nparams+p]+","+params1r[2*nparams+p]+","+params2r[2*nparams+p]);
					p = 8;
					pw.println("Du,"+params1r[p]+","+params2r[p]+","+params1r[nparams+p]+","+params2r[nparams+p]+","+params1r[2*nparams+p]+","+params2r[2*nparams+p]);
					p = 9;
					pw.println("Ds,"+params1r[p]+","+params2r[p]+","+params1r[nparams+p]+","+params2r[nparams+p]+","+params1r[2*nparams+p]+","+params2r[2*nparams+p]);
					p = 10;
					pw.println("F0,"+params1r[p]+","+params2r[p]+","+params1r[nparams+p]+","+params2r[nparams+p]+","+params1r[2*nparams+p]+","+params2r[2*nparams+p]);
					p = 11;
					pw.println("Starting cells,"+params1r[p]+","+params2r[p]+","+params1r[nparams+p]+","+params2r[nparams+p]+","+params1r[2*nparams+p]+","+params2r[2*nparams+p]);

					pw.println("Average %difference between the phenotype: "+m1+" and the dataset "+c1+" is,"+Methods.average(A1)+", s.d.,"+Methods.std(A1));
					if(!sameData)pw.println("Average %difference between the phenotype: "+m1+" and the dataset "+c2+" is,"+Methods.average(B1)+", s.d.,"+Methods.std(B1));
					pw.println("Average %difference between the phenotype: "+m2+" and the dataset "+c1+" is,"+Methods.average(A2)+", s.d.,"+Methods.std(A2));
					if(!sameData)pw.println("Average %difference between the phenotype: "+m2+" and the dataset "+c2+" is,"+Methods.average(B2)+", s.d.,"+Methods.std(B2));
					double probA1B1 =Methods.getProbabilityLess(A1, B1);
					if(!sameData)pw.println("Probability that ,"+m1+", describes ,"+c1+", better than ,"+c2+", = ,"+probA1B1);
					double probA2B2 =Methods.getProbabilityLess(A2, B2);
					if(!sameData)pw.println("Probability that ,"+m2+", describes ,"+c1+", better than ,"+c2+", = ,"+probA2B2);
					double probA1A2 =Methods.getProbabilityLess(A1, A2);
					pw.println("Probability that ,"+m1+", describes ,"+c1+", better than ,"+m2+", does = ,"+probA1A2);
					double probB1B2 =Methods.getProbabilityLess(B1, B2);
					if(!sameData)pw.println("Probability that ,"+m1+", describes ,"+c2+", better than ,"+m2+", does = ,"+probB1B2);				
					//probability that model one best describes A
					double prob1A = probA1B1*probA1A2;  //the model does better with A than B AND the other model does worse with A
					//probability that model one best describes B
					double prob1B = (1-probA1B1)*(probB1B2);//the model 1 does better with B than A AND the other model does worse with B
					//probability that model two best describes A
					double prob2A = probA2B2*(1-probA1A2);//the model does better with A than B AND the other model does worse with A
					//probability that model two best describes B
					double prob2B = (1-probA2B2)*(1-probB1B2);
					//probability that model one fits A better than model two fits B
					double prob1differs2 = Math.max(Math.min(probA1A2,1-probA1A2),Math.min(probB1B2,1-probB1B2));
					//System.err.println(String.format("M1_A=%d, M1 B = %d, M2 A=%d, M2 B=%d" , m1_A,m1_B,m2_A,m2_B));
					pw.println("Probability that,"+answerNames.get(slist.getSelectedIndices()[0])+",describes,"+conditionNames.get(slist.getSelectedIndices()[0])+",is,"+prob1A);
					if(!sameData)pw.println("Probability that,"+answerNames.get(slist.getSelectedIndices()[0])+",describes,"+conditionNames.get(slist.getSelectedIndices()[1])+",is,"+prob1B);
					pw.println("Probability that,"+answerNames.get(slist.getSelectedIndices()[1])+",describes,"+conditionNames.get(slist.getSelectedIndices()[0])+",is,"+prob2A);
					if(!sameData)pw.println("Probability that,"+answerNames.get(slist.getSelectedIndices()[1])+",describes,"+conditionNames.get(slist.getSelectedIndices()[1])+",is,"+prob2B);
					if(!sameData)pw.println("Conservative probability that,"+c1+",differs from,"+c2+",is,"+Math.max(probA1B1,probA2B2));
					pw.println("Conservative probability that,"+answerNames.get(slist.getSelectedIndices()[0])+",is identical to,"+answerNames.get(slist.getSelectedIndices()[1])+",given the data sets is,"+prob1differs2);
					pw.close();
					JOptionPane.showMessageDialog(null, "Analysis saved in "+savefile+". Conservative p-value for difference in phenotypes is: "+prob1differs2);
				} catch (FileNotFoundException e) {
					// TODO Auto-generated catch block
					e.printStackTrace();
				}
				draw.paramList.clear();
				draw.paramList2.clear();
				draw.names.clear();
				CancelBtn.setVisible(false);
				tabber.setSelectedIndex(2);
			}
			else if(mode==10) // intermediate phen gen
			{
				if(slist.getSelectedIndices().length < 2 )
				{
					JOptionPane.showMessageDialog(null, "You need to select two phenotypes!");
					return;
				}
				//morph
				//we need to morph the parameters in groups
				double[] p1 = answers.get(slist.getSelectedIndices()[0]);
				double[] p2 = answers.get(slist.getSelectedIndices()[1]);
				String n1 = answerNames.get(slist.getSelectedIndices()[0]);
				String n2 = answerNames.get(slist.getSelectedIndices()[1]);
				String c1 = conditionNames.get(slist.getSelectedIndices()[0]);
				
				double[] m1 = Methods.duplicate(p1);
				for(int i = 0; i < 12; i++)
				{
					if(morphParams[i].isSelected())
					{
						m1[i] = p2[i];
						m1[i+Constants.nparams]=p2[i+Constants.nparams];
						m1[i+Constants.nparams*2]=p2[i+Constants.nparams*2];
					}
				}
				answers.add(m1);
				conditionNames.add(c1);
				answerNames.add(n1+"<-"+n2);
				answerScores.add(0.0);
				answerTimes.add(answerTimes.get(slist.getSelectedIndices()[1]));
				answerCounts.add(answerCounts.get(slist.getSelectedIndices()[1]));
				answerCandidates.add(answerCandidates.get(slist.getSelectedIndices()[0]));
				
				((DefaultListModel)slist.getModel()).addElement(n1+"<-"+n2);
				

				
			}
			else  //proliferation with fcyton or cyton
			{
				// TODO Auto-generated method stub
				//drawfcyton.setVisible(false);
				//drawCyton.setVisible(false);
				PhenCytonBtn.setVisible(false);
				PhenfcytonBtn.setVisible(false);
				CancelBtn.setVisible(true);
				for(int i = 0; i < fluorescenceSpinners.length; i++)
					fluorescenceSpinners[i].setEnabled(false);
				for(int i = 0; i < cytonSpinners.length; i++)
					cytonSpinners[i].setEnabled(false);
				if(exptCombo.getSelectedIndex() == -1) return;
				ArrayList<ArrayList<Double>> data = new ArrayList<ArrayList<Double>>();
				ArrayList<Double> undivided = new ArrayList<Double>();
				ArrayList<Double> times = new ArrayList<Double>();
				ArrayList<Integer> volumes = new ArrayList<Integer>();
				int gens = (Integer)genSpinner.getValue();
				ArrayList<FMFCS> list = map.get(exptCombo.getSelectedItem());
				try {
					 setCursor(Cursor.getPredefinedCursor(Cursor.WAIT_CURSOR));
					for(int i = 0; i < list.size(); i++)
					{
						if(list.get(i).getReady().getPoints().size()==0) list.get(i).getReady().reloadDataFromFile();
						LinkedList<Point> dat = list.get(i).getReady().getPoints();
						ArrayList<Double> logdat = new ArrayList<Double>();
						Iterator<Point> itr = dat.iterator();
						while(itr.hasNext())
							logdat.add(Math.log10(Math.max(itr.next().ps[ list.get(i).getReady().xindex],1)));
						data.add(logdat);
						undivided.add(Math.log10(list.get(i).getReady().first_peak));
						times.add(list.get(i).getTime());
						volumes.add(list.get(i).getVolume());
						list.get(i).getReady().clearAllData();
					}
				} finally {
				setCursor(Cursor.getDefaultCursor());
				}
				
				proliferationAnalysis(gens, undivided, times, volumes,data, mode,(String)exptCombo.getSelectedItem());
				PhenCytonBtn.setVisible(true);
				PhenfcytonBtn.setVisible(true);
				CancelBtn.setVisible(false);
				for(int i = 0; i < fluorescenceSpinners.length; i++)
					fluorescenceSpinners[i].setEnabled(true);
				for(int i = 0; i < cytonSpinners.length; i++)
					cytonSpinners[i].setEnabled(true);
			}
		}
	}
	
	

	/**
	 * Used to setup the graphical user interface.
	 */
	private void setUpGUI() {
		Container cp = this.getContentPane();
		JPanel mainPanel = new JPanel();
		mainPanel.setLayout(new BorderLayout(1,1));

		this.addComponentListener(new ComponentListener(){

			public void componentHidden(ComponentEvent arg0) {
				// TODO Auto-generated method stub
				
			}

			public void componentMoved(ComponentEvent arg0) {
				// TODO Auto-generated method stub
				
			}

			public void componentResized(ComponentEvent arg0) {
				scroller.setPreferredSize(new Dimension(198,self.getHeight()-116));
				tree.updateUI();
				self.invalidate();
				plot.repackGraphs();
			}

			public void componentShown(ComponentEvent e) {
				// TODO Auto-generated method stub
				
			}
			
		});
		
		tmenu = new JPopupMenu();
		copyGateMI = new JMenuItem("Copy Gate");
		pasteGateMI = new JMenuItem("Paste Gate");
		
		copyGateMI.setFont(new Font("Calibri",Font.PLAIN,11));
		pasteGateMI.setFont(new Font("Calibri",Font.PLAIN,11));
		
		copyGateMI.addActionListener(new ActionListener(){

			public void actionPerformed(ActionEvent arg0) {
				// TODO Auto-generated method stub
				if(tree.getSelected() != null && tree.getSelected().getParent() != null)
				{
					tree.setCopy(tree.getSelected().clone());
					System.err.println("copied!");
					tree.updateUI();
				}

			}
		});
		
		pasteGateMI.addActionListener(new ActionListener(){

			public void actionPerformed(ActionEvent arg0) {
				// TODO Auto-generated method stub
				if(tree.getCopy() != null && tree.getSelected() != null)
				{
					FMTreeNode newcopy = (FMTreeNode)tree.getCopy().clone();
					tree.getCopy().propagateNewParent(tree.getSelected());
					tree.getSelected().getLeaves().add(tree.getCopy());
					tree.setCopy(newcopy);
					tree.updateUI();
					plot.updateUI();
				}


			}
			
		});
		
		tmenu.add(copyGateMI);
		tmenu.add(pasteGateMI);
		
		tree = new Dendogramer();
		tree.setToolTipText("Tree view of all FCS data files and gates. Click on a data set to display it at right.");
		tree.setPreferredSize(new Dimension(198,100));
		tree.setFocusable(true);
		
		tree.addMouseListener(new MouseListener(){

			public void mouseClicked(MouseEvent arg0) {
				// TODO Auto-generated method stub
				//lets figure out where you clicked and show that

			}

			public void mouseEntered(MouseEvent arg0) {
				// TODO Auto-generated method stub
				
			}

			public void mouseExited(MouseEvent arg0) {
				// TODO Auto-generated method stub
				
			}

			public void mousePressed(MouseEvent arg0) {
				try {
					tree.Click(arg0.getX(), arg0.getY());
					FMTreeNode tn = (FMTreeNode)tree.getClicked(arg0.getX(),arg0.getY());
					if(tn != null)
					{
						tree.setSelected(tn);
						tree.updateUI();
						if(arg0.getButton()==MouseEvent.BUTTON3&& tree.getSelected() != null && tree.getSelected().getParent() != null)
						{
							tmenu.show(arg0.getComponent(), arg0.getX(), arg0.getY());
						}	
						if(arg0.getClickCount()==2)
						{
							setCursor(Cursor.getPredefinedCursor(Cursor.WAIT_CURSOR));
			
							if(tn == null) return;
							if(tree.getSelected() != null &&plot.getCurrent_plot() != null && plot.getCurrent_plot().getNode().equals(tn)) return;
							/*
							if(tree.getSelected() != null && tree.getSelected().getPoints().size() ==0)
							{
								tree.getSelected().clearAllData();
								tn.reloadDataFromFile();
							}
							*/
							if(!tn.getSourceFCS().isLoaded())
							{
								tn.reloadDataFromFile();
							}
							else
							{
								tn.getSourceFCS().propagateColors();
								tn.propagateFromRoot();
							}
							
							//if(tn.points == null)tn.getPoints();  //makes you recalculate your points!~
							if(tn!=null)plot.makeCurrentPlot(tn);
							plot.updateAll();
						}
						else if(arg0.getClickCount() == 1)
						{

							if(arg0.getX() > tree.getWidth()-35)
							{
								//clicked on either the compensation or color BTN!
								if(tn != null && tn.getGate() == null) //clicked on compensation
								{
									
									if(compensator == null) compensator = new Compensator(tn.getSourceFCS(),plot);
									else compensator.setCompensation(tn.getSourceFCS());
								}
								else
								{
									//clicked on color!
									colorFrame = new JFrame("Pick a color for: "+tn.getGate().name); 
									colorFrame.getContentPane().setLayout(new GridLayout(1,1));
									colorer = new FMColorer(self,tn,colorFrame);
									colorFrame.setDefaultCloseOperation(JFrame.DISPOSE_ON_CLOSE);
									colorFrame.getContentPane().add(colorer);
									colorFrame.pack();
									colorFrame.setSize(250, 250);
									colorFrame.setVisible(true);
								}
							}
						}
						requestFocusInWindow();
					}
				} finally {
					setCursor(Cursor.getDefaultCursor());
				}
				
			}

			public void mouseReleased(MouseEvent arg0) {
				// TODO Auto-generated method stub

			}
			
		});
		loadFiles = new JButton("Load FCS");
		loadFiles.setToolTipText("Load one or more FCS3.0 data files from the flow cytometer.");
		loadFiles.setPreferredSize(new Dimension(198,20));
		loadFiles.addActionListener(new ActionListener(){
			public void actionPerformed(ActionEvent arg0) {
					File[] files = Dialogs.getOpenFiles("fcs", "Load one or more FCS3.0 data files from the flow cytometer.");
					setCursor(Cursor.getPredefinedCursor(Cursor.WAIT_CURSOR));
					for(int i = 0; i < files.length; i++)
					{
						System.err.println("Loading: "+files[i].getAbsolutePath());
						FMFCS newfcs = new FMFCS(Methods.processFCSFileWIthNoData(files[i]));
						runs.add(newfcs);	
					}
					Collections.sort(runs, new Comparator<FMFCS>(){

						@Override
						public int compare(FMFCS o1, FMFCS o2) {
							// TODO Auto-generated method stub
							int result =  o1.getExperiment().split(" ")[0].compareTo(o2.getExperiment().split(" ")[0]);
							if( result == 0)
							{
								if(o1.getSourceFile().lastModified() < o2.getSourceFile().lastModified())
									return -1;
								else if (o1.getSourceFile().lastModified()> o2.getSourceFile().lastModified())
									return 1;
								else
									return 0;
							}
							else
								return result;
						}
						
					});
					//now lets update the dendogramer
					tree.setRs(runs);
					tree.updateUI();
					setCursor(Cursor.getDefaultCursor());
				}
			}
		);
		loadBtn = new JButton("Load Workspace");
		loadBtn.setToolTipText("Load a previously saved workspace including all data sets, gates, and settings.");
		loadBtn.setPreferredSize(new Dimension(198,20));
		loadBtn.addActionListener(new ActionListener(){
			public void actionPerformed(ActionEvent arg0) {
				File loaded = Dialogs.getOpenFile("wrk","Load a previously saved workspace including all data sets, gates, and settings.");
				if(loaded != null && !loaded.isDirectory())
				{
					setCursor(Cursor.getPredefinedCursor(Cursor.WAIT_CURSOR));
					loadEverything(loaded);
					int count = tree.getRs().size();
					for(int i = 0; i < tree.getRs().size(); i++)
						count+= tree.getRs().get(i).getRoot().countChildren();
					tree.setPreferredSize(new Dimension(195,count*15));
					System.err.println("Count was "+count);
					invalidate();
					tree.updateUI();
					plot.clearPlots();
					plot.removeAll();
					plot.updateUI();
					setCursor(Cursor.getDefaultCursor());
				}
				
			}
		});
		saveBtn = new JButton("Save Workspace");
		saveBtn.setToolTipText("Save the current data sets, gates, and settings as a workspace file (*.wrk)"); 
		saveBtn.setPreferredSize(new Dimension(198,20));
		saveBtn.addActionListener(new ActionListener(){
			public void actionPerformed(ActionEvent arg0) {
				File savefile = Dialogs.getSaveFile("wrk","Save the current data sets, gates, and settings as a workspace file (*.wrk)");
				saveEverything(savefile,tree);
				tree.updateUI();
			}
		});
		scroller = new JScrollPane(tree);
		scroller.setHorizontalScrollBarPolicy(JScrollPane.HORIZONTAL_SCROLLBAR_NEVER);
		scroller.setVerticalScrollBarPolicy(JScrollPane.VERTICAL_SCROLLBAR_ALWAYS);
		scroller.setPreferredSize(new Dimension(240,600));
		plot = new FMGallery(this,tree);
		JPanel left = new JPanel(new FlowLayout(FlowLayout.LEFT,1,1) );
		this.setFocusable(true);
		this.addKeyListener(new KeyListener(){
			public void keyPressed(KeyEvent arg0) {			
				String modstring =KeyEvent.getKeyModifiersText(arg0.getModifiers());
				if(modstring.compareTo("Ctrl")==0&& arg0.getKeyCode() == KeyEvent.VK_C)
				{
					//lets copy the entire tree.
					if(tree.getSelected() != null && tree.getSelected().getParent() != null)
					{
						tree.setCopy(tree.getSelected().clone());
						System.err.println("copied!");
					}
				}
				else if(modstring.compareTo("Ctrl")==0&& arg0.getKeyCode() == KeyEvent.VK_V)
				{
					//lets paste the entire tree to the selected node.
					System.err.println("pasted");
					if(tree.getCopy() != null && tree.getSelected() != null)
					{
						FMTreeNode newcopy = ((FMTreeNode)tree.getCopy()).clone();
						FMFCS oldparent = newcopy.getSourceFCS();
						newcopy.propagateNewParent(tree.getSelected());
						tree.getSelected().getLeaves().add(newcopy);
						newcopy.getSourceFCS().setInverse(oldparent.getInverse().copy());
						tree.setCopy(newcopy);
						tree.updateUI();
						boolean ready = newcopy.getSourceFCS().checkIfReady();
						if(ready)
							tree.setState(newcopy,Color.green.getRGB());
						newcopy.getSourceFCS().propagateColors();
						plot.updateUI();
						tree.paintImmediately(tree.getVisibleRect());
					}
				}
				else if(arg0.getKeyCode() == KeyEvent.VK_DELETE)
				{
					if(tree.getSelected() != null) 
					{
						if(tree.getSelected().getParent() != null&& tree.getSelected().getGate() != null)
						{
							FMFCS fcs = ((FMTreeNode)tree.getSelected()).getSourceFCS();
							tree.getSelected().getParent().getLeaves().remove(tree.getSelected());
							if(!fcs.checkIfReady())	
							{
								tree.setState(fcs.getRoot(),Color.LIGHT_GRAY.getRGB());
							}
						}
						else
						{
							runs.remove(tree.getSelected().getSourceFCS());
						}
						plot.removeGraph(tree.getSelected(),true);
						tree.setSelected(null);
//						plot.current_plot.node = null;
						plot.updateUI();
						tree.updateUI();
						System.err.println("Deleted");
					}
				}
				else if(arg0.getKeyCode() == KeyEvent.VK_ESCAPE)
				{
					plot.getCurrent_plot().mode = 0;
					if(plot.getCurrent_plot() != null && plot.getCurrent_plot().getNode() != null)
					{
						plot.getCurrent_plot().getNode().zoomx1 = -1;
						plot.getCurrent_plot().getNode().zoomx2 = -1;
						plot.getCurrent_plot().getNode().zoomy1 = -1;
						plot.getCurrent_plot().getNode().zoomy2 = -1;
					}
					plot.getCurrent_plot().getGrf().setCurrentGate(null);
					plot.updateUI();
					plot.getCurrent_plot().enabledAllButtons();

					System.err.println("Cancelled");
				}
				else
					System.err.println("CONTROL: "+arg0.isControlDown()+" key code "+arg0.getKeyCode()+" char "+arg0.getKeyChar()+" modifiers "+modstring);
				
			}

			public void keyReleased(KeyEvent arg0) {
				// TODO Auto-generated method stub
				
			}

			public void keyTyped(KeyEvent arg0) {
				// TODO Auto-generated method stub
				
			}
			
		});

		JScrollPane galleryBars = new JScrollPane(plot,JScrollPane.VERTICAL_SCROLLBAR_ALWAYS,JScrollPane.HORIZONTAL_SCROLLBAR_NEVER);
		mainPanel.add(galleryBars,BorderLayout.CENTER);	
		left.setPreferredSize(new Dimension(200,640));
		left.add(loadFiles);
		left.add(loadBtn);
		left.add(saveBtn);
		left.add(scroller);
		mainPanel.add(left, BorderLayout.WEST);
		tabber = new JTabbedPane();
		tabber.addTab("Data",mainPanel);
		this.setFocusable(true);
		proliferator = new JPanel();
//		proliferator.setSize(new Dimension(770,620));
		profDisp = new ProliferationDisplay();
		profDisp.addMouseListener(new MouseListener(){

			public void mouseClicked(MouseEvent arg0) {
				// TODO Auto-generated method stub
				
			}

			public void mouseEntered(MouseEvent arg0) {
				// TODO Auto-generated method stub
				
			}

			public void mouseExited(MouseEvent arg0) {
				// TODO Auto-generated method stub
				
			}

			public void mousePressed(MouseEvent arg0) {
				// TODO Auto-generated method stub
				
			}

			public void mouseReleased(MouseEvent arg0) {
				// TODO Auto-generated method stub
				//lets update the information about this experiment
				try{
					TimePoint tp = profDisp.getExp().timePoints.get(arg0.getX()/(profDisp.getHSize()));
					Run r = tp.getReplicates().get(arg0.getY()/(profDisp.getHSize()));
					String out = profDisp.getExp().name+"\nTime\t"+tp.getTime()+" h.\nReplicate\t"+(1+arg0.getY()/profDisp.getHSize());
					out+="\nTotal Exp\t"+r.data.raw.size();
					out+="\nTotal Cyton\t"+r.totalCyton;
					out+="\nUndiv. Peak\t"+String.format("%4.3f",Math.pow(10,r.undivided+r.shift));
					out+="\nRatio\t"+String.format("%4.4f",tp.ratio);
					out+="\nBackground\t"+String.format("%4.4f",tp.background);
					out+="\nCV\t"+String.format("%4.4f",tp.CV);
					out+="\nScore\t"+r.lastScore;
					double[] counts = r.getConcentrations();
					for(int i = 0; i < counts.length; i++)
					{
						out+="\nDiv "+i+"\t"+counts[i]+" cells/ml";
					}
					stats.setText(out);
				}
				catch(Exception e)
				{
					//do nothing
				}
			}
			
		});
//		profDisp.setPreferredSize(new Dimension(600,600));
		proliferator.setLayout(new BorderLayout());
		Pright= new JPanel();
		Pright.setPreferredSize(new Dimension(150,getHeight()-20));
		Pright.setLayout(new FlowLayout(FlowLayout.LEFT,0,0));
		Font fon = new Font("Calibri",Font.PLAIN,11);
		JPanel upperright = new JPanel();
		upperright.setLayout(new GridLayout(2,2));
		upperright.setPreferredSize(new Dimension(150,35));
		exptCombo = new JComboBox();
		exptCombo.setPreferredSize(new Dimension(75,20));
		exptCombo.setToolTipText("Pick an experimental condition to phenotype");
		JLabel temp = new JLabel("Condition:");
		temp.setFont(fon);
		upperright.add(temp);
		upperright.add(exptCombo);
		genSpinner = new JSpinner(new SpinnerNumberModel(8,1,20,1));
		genSpinner.setPreferredSize(new Dimension(75,20));
		genSpinner.setToolTipText("The number of division classes that should be modeled");
		JLabel temp2 = new JLabel("Peaks:");
		temp2.setFont(fon);
		sizeSpinner = new JSpinner(new SpinnerNumberModel(200,30,700,5));
		sizeSpinner.setPreferredSize(new Dimension(75,20));
		sizeSpinner.setToolTipText("The size of each histogram at right in pixels. You can also use the scrollwheel to adjust the size.");
		JLabel temp3 = new JLabel("Disp Size:");
		temp3.setFont(fon);
		countsSpinner = new JSpinner(new SpinnerNumberModel(0,0,100000,1));
		countsSpinner.setPreferredSize(new Dimension(75,20));
		countsSpinner.setToolTipText("The maximum histogram cell count (in cells/mL). Set to 0 for autonormalization.");
		JLabel temp5 = new JLabel("Histogram size:");
		temp5.setFont(fon);
		JLabel temp4 = new JLabel("Solutions:");
		temp4.setFont(fon);
		upperright.add(temp4);
		repeatsSpinner = new JSpinner(new SpinnerNumberModel(100,0,10000,1));
		repeatsSpinner.setToolTipText("How many times should the data set be fitted? Solutions will be clustered to estimate degeneracy. Too many clusters means not enough constraints!");
		repeatsSpinner.addChangeListener(new ChangeListener(){

			public void stateChanged(ChangeEvent e) {
				// TODO Auto-generated method stub
				if(profDisp.getExp() != null)
				{
					profDisp.getExp().repeats = (Integer)repeatsSpinner.getValue();
					System.err.println("Trials = "+profDisp.getExp().repeats);
				}
			}
			
		});
		
		upperright.add(repeatsSpinner);		
		sizeSpinner.addChangeListener(new ChangeListener(){

			public void stateChanged(ChangeEvent arg0) {
				// TODO Auto-generated method stub
				profDisp.setHSize((Integer)sizeSpinner.getValue());
				profDisp.setRedraw_overlay(true);
				profDisp.setRedraw_needed(true);
				System.err.println("new size "+profDisp.getHSize());
				profDisp.draw();
			}
			
		});
		countsSpinner.addChangeListener(new ChangeListener(){

			public void stateChanged(ChangeEvent arg0) {
				// TODO Auto-generated method stub
				profDisp.setMaxHistogramHeight((Integer)countsSpinner.getValue());
				profDisp.setRedraw_overlay(true);
				profDisp.setRedraw_needed(true);
				//System.err.println("new hist size "+profDisp.getHSize());
				profDisp.draw();
			}
			
		});
		optFrame = new JFrame();

		drawBox = new JCheckBox("Visual?");
		drawBox.setToolTipText("Do you want to see the fitting in real time?");
		drawBox.setFont(fon);
		drawBox.setSelected(true);
		colorBox = new JCheckBox("Colors?");
		colorBox.setToolTipText("Draw the colored backgrounds? Colors reflect the current error of the fit.");
		colorBox.setFont(fon);
		colorBox.setSelected(true);
		colorBox.addActionListener(new ActionListener(){

			@Override
			public void actionPerformed(ActionEvent e) {
				// TODO Auto-generated method stub
				profDisp.setColor(colorBox.isSelected());
			}
			
		});
		PhenfcytonBtn = new JButton("Phenotype FCyton");
		PhenfcytonBtn.setToolTipText("Start phenotyping the data using the fcyton model and current fitting settings.");
		PhenfcytonBtn.addActionListener(new ActionListener(){

			public void actionPerformed(ActionEvent arg0) {
				// TODO Auto-generated method stub
				cr =  new ClickResponder(0);
				cr.start();
			}
			
		});
		PhenfcytonBtn.setFont(fon);
		PhenfcytonBtn.setPreferredSize(new Dimension(150,20));
		advancedPanel = new JPanel();
		advancedPanel.setPreferredSize(new Dimension(150,430));
		advancedPanel.setLayout(new FlowLayout(FlowLayout.LEFT,0,0));
		advancedMode = new JCheckBox("Advanced Options?",false);
		advancedMode.setFont(fon);
		advancedMode.addItemListener(new ItemListener(){

			public void itemStateChanged(ItemEvent arg0) {
				if(!advancedMode.isSelected())
				{
					advancedPanel.setVisible(false);
				}
				else
					advancedPanel.setVisible(true);
			}
			
		});
		PhenCytonBtn = new JButton("Phenotype Cyton");
		
		PhenCytonBtn.setToolTipText("Start phenotyping the data using the cyton model (see Hawkins et al 2007 PNAS) and current fitting settings.");
		PhenCytonBtn.addActionListener(new ActionListener(){

			public void actionPerformed(ActionEvent arg0) {
				// TODO Auto-generated method stub
				cr =  new ClickResponder(1);
				cr.start();
			}
			
		});
		PhenCytonBtn.setFont(fon);
		PhenCytonBtn.setPreferredSize(new Dimension(150,20));
		fitFluorescenceBtn = new JButton("Fluorescence Fit");
		
		fitFluorescenceBtn.setToolTipText("Fit fluorescence parameters and determine best-fit generational cell counts.");
		fitFluorescenceBtn.addActionListener(new ActionListener(){

			public void actionPerformed(ActionEvent arg0) {
				// TODO Auto-generated method stub
				cr =  new ClickResponder(1);
				cr.start();
			}
			
		});
		PhenCytonBtn.setFont(fon);
		PhenCytonBtn.setPreferredSize(new Dimension(150,20));
		CancelBtn = new JButton("Cancel Phenotyping");
		CancelBtn.setVisible(false);
		CancelBtn.addActionListener(new ActionListener(){

			public void actionPerformed(ActionEvent arg0) {
				// TODO Auto-generated method stub
				profDisp.getExp().abort_flag= true;
				PhenCytonBtn.setEnabled(true);
				PhenfcytonBtn.setEnabled(true);
			}
			
		});
		CancelBtn.setFont(fon);
		CancelBtn.setPreferredSize(new Dimension(150,20));	
		Pright.add(upperright);
		Pright.add(PhenfcytonBtn);
		Pright.add(advancedMode);
		JPanel tpanel = new JPanel(new GridLayout(4,2));
		tpanel.setPreferredSize(new Dimension(150,60));
		tpanel.add(temp2);
		tpanel.add(genSpinner);
		tpanel.add(temp3);
		tpanel.add(sizeSpinner);
		tpanel.add(temp5);
		tpanel.add(countsSpinner);
		tpanel.add(drawBox);
		tpanel.add(colorBox);
		advancedPanel.add(tpanel);
		advancedPanel.add(PhenCytonBtn);
		advancedPanel.add(CancelBtn);
		JPanel middleRight = new JPanel();
		fluorescenceSpinners = new JSpinner[8];
		middleRight.setLayout(new GridLayout(4,3));
		
		Font fnt = new Font("Calibri",Font.PLAIN,11);

		fluorescenceSpinners[0] = new JSpinner(new SpinnerNumberModel(0.01,0.005,0.03,0.001));
		fluorescenceSpinners[1] = new JSpinner(new SpinnerNumberModel(0.03,0.005,0.03,0.001));
		fluorescenceSpinners[2] = new JSpinner(new SpinnerNumberModel(0.45,0.4,0.6,0.01));
		fluorescenceSpinners[3] = new JSpinner(new SpinnerNumberModel(0.5,0.4,0.6,0.01));
		fluorescenceSpinners[4] = new JSpinner(new SpinnerNumberModel(0,0,1000000.0,100));
		fluorescenceSpinners[5] = new JSpinner(new SpinnerNumberModel(100000.0,0,1000000.0,10));
		fluorescenceSpinners[6] = new JSpinner(new SpinnerNumberModel(-0.002,-0.03,0.03,0.001));
		fluorescenceSpinners[7] = new JSpinner(new SpinnerNumberModel(0.002,-0.03,0.03,0.001));
		JLabel lbl = new JLabel("Width");
		lbl.setToolTipText("The width of all peaks for any histogram (CV)");

		lbl.setFont(fnt);
		middleRight.add(lbl);
		middleRight.add(fluorescenceSpinners[0]);
		middleRight.add(fluorescenceSpinners[1]);
		JLabel lbl2 = new JLabel("Ratio");
		lbl2.setToolTipText("The ratio is the fraction of fluorescence left after a division event (should be around 0.5)");
		lbl2.setFont(fnt);
		middleRight.add(lbl2);
		middleRight.add(fluorescenceSpinners[2]);
		middleRight.add(fluorescenceSpinners[3]);
		JLabel lbl3 = new JLabel("Bckgrnd.");
		lbl3.setToolTipText("The autofluorescence (background fluorescence) mostly affects the left-most part of the histogram (divided peaks).");
		lbl3.setFont(fnt);
		middleRight.add(lbl3);
		middleRight.add(fluorescenceSpinners[4]);
		middleRight.add(fluorescenceSpinners[5]);
		JLabel lbl4 = new JLabel("Shift");
		lbl4.setToolTipText("The shift is used to compensate for errors during generation 0 peak selection by the user.");
		lbl4.setFont(fnt);
		middleRight.add(lbl4);
		middleRight.add(fluorescenceSpinners[6]);
		middleRight.add(fluorescenceSpinners[7]);
		for(int i = 0; i < fluorescenceSpinners.length; i++)
			fluorescenceSpinners[i].setFont(fnt);
		middleRight.setPreferredSize(new Dimension(150,70));
		JPanel loadSaveParamPanel = new JPanel(new GridLayout(1,2));
		SaveParams = new JButton("Save Ps.");
		SaveParams.setToolTipText("Save current parameters from file");
		SaveParams.addActionListener(new ActionListener(){

			@Override
			public void actionPerformed(ActionEvent e) {
				File saveFile = Dialogs.getSaveFile("prm", "Save Parameters:");
				try {
					PrintWriter pw = new PrintWriter(saveFile);
					for(JSpinner spinner: fluorescenceSpinners)
					{
						pw.println((Double)spinner.getValue());
					}
					for(JSpinner spinner: cytonSpinners)
					{
						pw.println((Double)spinner.getValue());
					}
					pw.close();
				} catch (FileNotFoundException e1) {
					// TODO Auto-generated catch block
					e1.printStackTrace();
				}
				
			}
			
		});
		LoadParams = new JButton("Load Ps.");
		LoadParams.setToolTipText("Load parameters from file");
		LoadParams.addActionListener(new ActionListener(){

			@Override
			public void actionPerformed(ActionEvent e) {
				File loadFile = Dialogs.getOpenFile("prm", "Load parameters:");
				try {
					Scanner s = new Scanner(loadFile);
					int spinIndex = 0;
					while(s.hasNextDouble())
					{
						if(spinIndex < fluorescenceSpinners.length)
						{
							fluorescenceSpinners[spinIndex].setValue(s.nextDouble());
						}
						else
						{
							cytonSpinners[spinIndex-fluorescenceSpinners.length].setValue(s.nextDouble());
						}
						spinIndex++;
					}
					s.close();
				} catch (FileNotFoundException e1) {
					// TODO Auto-generated catch block
					e1.printStackTrace();
				}
				
			}
			
		});
		SaveParams.setFont(fnt);
		LoadParams.setFont(fnt);
		SaveParams.setPreferredSize(new Dimension(75,15));
		LoadParams.setPreferredSize(new Dimension(75,15));
		loadSaveParamPanel.add(SaveParams);
		loadSaveParamPanel.add(LoadParams);
		JLabel temp5b = new JLabel("Fluorescence Parameters:");
		temp5b.setFont(new Font("Calibri",Font.BOLD,12));
		advancedPanel.add(loadSaveParamPanel);
		advancedPanel.add(temp5b);
		advancedPanel.add(middleRight);
		JLabel temp6 = new JLabel("Cellular Parameters:");
		temp6.setFont(new Font("Calibri",Font.BOLD,12));
		advancedPanel.add(temp6);
		JPanel cs = new JPanel(new GridLayout(nparams,3));
		cs.setPreferredSize(new Dimension(150,nparams*17));
		String[] pnames = {"E[Tdiv0]","sd[Tdiv0]","E[Tdiv1+]","sd[Tdiv1+]","E[Tdie0]","sd[Tdie0]","E[Tdie1+]","sd[Tdie1+]","D"+"\u03BC","D"+"\u03C3","F0","N","kMech","pMech"};
		for(int i = 0; i < nparams; i++)
		{
			cytonSpinners[i*2]=new JSpinner(new SpinnerNumberModel(cspinners[i][1],cspinners[i][0],cspinners[i][3],1));
			cytonSpinners[i*2+1]=new JSpinner(new SpinnerNumberModel(cspinners[i][2],cspinners[i][0],cspinners[i][3],1));
		}

		for(int i = 0; i < nparams; i++)
		{
			JLabel newlbl = new JLabel(pnames[i]);
			newlbl.setFont(new Font("Calibri",Font.PLAIN,11));
			if(i%2==0)newlbl.setBackground(Color.YELLOW);
			cs.add(newlbl);
			cs.add(cytonSpinners[i*2]);
			cs.add(cytonSpinners[i*2+1]);
		}
		stats = new JTextArea("Fitting Statistics:");
		stats.setToolTipText("The current fitting statistics are shown in this text box.");
		stats.setEditable(false);
		stats.setFont(new Font("Calibri",Font.PLAIN,11));
		JScrollPane statScroller = new JScrollPane(stats);
		statScroller.setPreferredSize(new Dimension(150,220));
		advancedPanel.add(cs);
		Pright.add(advancedPanel);
		Pright.add(statScroller);
		advancedPanel.setVisible(false);
		proliferator.add(Pright,BorderLayout.WEST);
		scroller2 = new JScrollPane(profDisp);
		proliferator.add(scroller2,BorderLayout.CENTER);
		this.setTitle("FlowMax V "+Constants.FlowMaxVersion+" (C) Signaling Systems Laboratory, UCSD, 2011-2014");
		this.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
		this.setSize(1150,775);

		tabber.addTab("Phenotyping",proliferator);
		//analysis tab
		JPanel aPane = new JPanel(new BorderLayout(2,2));
		JPanel aleft = new JPanel(new BorderLayout(2,2));
		slist = new ClickList(new DefaultListModel());
		slist.addKeyListener(new KeyListener(){

			public void keyPressed(KeyEvent arg0) {
				// TODO Auto-generated method stub
				
			}

			public void keyReleased(KeyEvent arg0) {
				if(arg0.getKeyCode() == KeyEvent.VK_DELETE)
				{
					//deleting the selected!
					DefaultListModel lm = (DefaultListModel) slist.getModel();
					DefaultListModel lm2 = new DefaultListModel();
					for(int i = 0; i < lm.getSize();i++)
					{
						lm2.addElement(lm.getElementAt(i));
					}
					int[] selected = slist.getSelectedIndices();
					for(int i : selected)
					{
						lm2.remove(i);
						answerScores.remove(i);
						answers.remove(i);
						answerTimes.remove(i);
						answerCounts.remove(i);
						answerCandidates.remove(i);
						answerNames.remove(i);
					}
					slist.setModel(lm2);
				}
				
			}

			public void keyTyped(KeyEvent arg0) {
				// TODO Auto-generated method stub
				
			}
			
		});
		JScrollPane listScroller = new JScrollPane(slist);
		//top.setPreferredSize(new Dimension(800,200));
		aleft.add(listScroller, BorderLayout.CENTER);
		//grid now only contains buttons
		lButton = new JButton("Load Phenotype");
		lButton.setToolTipText("Load a phenotype from a saved .csv file");
		lButton.addActionListener(new ActionListener(){

			public void actionPerformed(ActionEvent arg0) {
				JFileChooser fc = new JFileChooser();
				fc.setApproveButtonText("Load");
				fc.setCurrentDirectory(new File("./Phenotypes"));
				fc.setFileSelectionMode(JFileChooser.FILES_ONLY);
				fc.setFileFilter(new FileFilter(){

					@Override
					public boolean accept(File f) {
						// TODO Auto-generated method stub
						if(f.getName().endsWith(".csv")) return true;
						if(f.isDirectory()) return true;
						return false;
					}

					@Override
					public String getDescription() {
						// TODO Auto-generated method stub
						return "Previously saved CSV file from phenotyping";
					}
					
				});
				fc.setMultiSelectionEnabled(true);
				int val = fc.showOpenDialog(lButton);
				if(val == 0)
				{
					File[] files = fc.getSelectedFiles();
					for(int i = 0; i < files.length; i++)
					{
						if(!files[i].isDirectory())
						{
							processPhenotypeCSV(files[i]);
							
						}
					}
				}
				
			}


			
		});
		//button
		vButton = new JButton("Draw");
		vButton.addActionListener(new ActionListener(){

			public void actionPerformed(ActionEvent arg0) {
				// TODO Auto-generated method stub
				//lets process the text
				cr = new ClickResponder(7);
				cr.start();
			}
			
		});
		//counts
		countsButton = new JButton("Draw");
		countsButton.addActionListener(new ActionListener(){

			public void actionPerformed(ActionEvent arg0) {
				// TODO Auto-generated method stub
				//lets process the text
				cr = new ClickResponder(9);
				cr.start();
			}
			
		});
		cButton = new JButton("Compare");
		cButton.setToolTipText("Calculate the probability that phenotypes are the same in the context of experimental data.");
		cButton.addActionListener(new ActionListener(){

			public void actionPerformed(ActionEvent arg0) {
				// TODO Auto-generated method stub
				//lets process the text
				cr = new ClickResponder(8);
				cr.start();
			}
			
		});
		mButton = new JButton("Generated intermediate phenotype");
		mButton.setToolTipText("Test how parameters of one phenotype will change another phenotype by generating intermediates");
		mButton.addActionListener(new ActionListener(){

			public void actionPerformed(ActionEvent arg0) {
				// TODO Auto-generated method stub
				//lets process the text
				cr = new ClickResponder(10);
				cr.start();
			}
			
		});
		vButton.setToolTipText("Click to render the best fit parameter values for each of the selected clusters of parameters. This may take a few moments.");
		countsButton.setToolTipText("Click to render the best fit solutions for the selected cluster of parameters. This may take a few moments.");
		colButton = new JButton("Dot color");
		colButton.setToolTipText("Click to select the experimental count dot color.");
		colButton.addActionListener(new ActionListener(){

			public void actionPerformed(ActionEvent arg0) {
				// TODO Auto-generated method stub
				//lets process the text
				JFrame dcframe = new JFrame("Dot color");
				DotColorer dc = new DotColorer(counter,dcframe, colButton);
				dcframe.setSize(new Dimension(200,200));
				dcframe.getContentPane().setLayout(new GridLayout(1,1));
				dcframe.getContentPane().add(dc);
				dcframe.setDefaultCloseOperation(JFrame.DO_NOTHING_ON_CLOSE);
				dcframe.setVisible(true);
				System.err.println("Visible");
			}
			
		});		
		
		
		
		
		dmenu = new JPopupMenu();
		JMenuItem copyMI = new JMenuItem("Copy to Clipboard");
		Font calibri = new Font("Calibri",Font.PLAIN,11);
		copyMI.setFont(calibri);
		dmenu.add(copyMI);
		copyMI.addActionListener(new ActionListener() {
			
			public void actionPerformed(ActionEvent e) {
				// TODO Auto-generated method stub
					ImageSelection imgSel = new ImageSelection(draw.img);
					Toolkit.getDefaultToolkit().getSystemClipboard().setContents(imgSel,null);
			}
		});
		cmenu = new JPopupMenu();
		JMenuItem copyMIc = new JMenuItem("Copy to Clipboard");
		copyMI.setFont(calibri);
		cmenu.add(copyMIc);
		copyMIc.addActionListener(new ActionListener() {
			
			public void actionPerformed(ActionEvent e) {
				// TODO Auto-generated method stub
					ImageSelection imgSel = new ImageSelection(counter.img);
					Toolkit.getDefaultToolkit().getSystemClipboard().setContents(imgSel,null);
			}
		});
		
		aleft.add(lButton, BorderLayout.NORTH);
		aPane.add(aleft,BorderLayout.WEST);

		drawTabs = new JTabbedPane();
		draw = new Drawer(this);
		//draw.setPreferredSize(new Dimension(500,100));
		draw.addMouseListener(new MouseListener(){

			public void mouseClicked(MouseEvent arg0) {
				// TODO Auto-generated method stub
				
			}

			public void mouseEntered(MouseEvent arg0) {
				// TODO Auto-generated method stub
				
			}

			public void mouseExited(MouseEvent arg0) {
				// TODO Auto-generated method stub
				
			}

			public void mousePressed(MouseEvent arg0) {
				// TODO Auto-generated method stub
				
			}

			public void mouseReleased(MouseEvent arg0) {
				if(arg0.getButton()==MouseEvent.BUTTON3)
				{
					dmenu.show(draw,arg0.getX(),arg0.getY());
				}
				
			}
			
		});
		counter = new Counter(this);
		counter.addMouseListener(new MouseListener(){

			public void mouseClicked(MouseEvent arg0) {
				// TODO Auto-generated method stub
				
			}

			public void mouseEntered(MouseEvent arg0) {
				// TODO Auto-generated method stub
				
			}

			public void mouseExited(MouseEvent arg0) {
				// TODO Auto-generated method stub
				
			}

			public void mousePressed(MouseEvent arg0) {
				// TODO Auto-generated method stub
				
			}

			public void mouseReleased(MouseEvent arg0) {
				if(arg0.getButton()==MouseEvent.BUTTON3)
				{
					cmenu.show(counter,arg0.getX(),arg0.getY());
				}
				
			}
			
		});

		counter.setPreferredSize(new Dimension(200,100));
		JScrollPane cscroller = new JScrollPane(counter);
		JScrollPane dscroller = new JScrollPane(draw);
		
		JPanel pDrawPanel = new JPanel(new BorderLayout(2,2));
		JPanel pgrid = new JPanel(new GridLayout(32,1,2,2));
		pgrid.add(vButton);
		pgrid.add(new JLabel("Tmax (h):"));
		timeSpinnerP = new JSpinner(new SpinnerNumberModel(120,10,500,10));
		timeSpinnerP.setToolTipText("The maximum plotted time (h)");
		pgrid.add(timeSpinnerP);
		pgrid.add(new JLabel("Max f %:"));
		fsmaxSpinner = new JSpinner(new SpinnerNumberModel(100,1,100,1));
		fsmaxSpinner.setToolTipText("Maximum responding fraction percent plotted.");
		pgrid.add(fsmaxSpinner);
		pgrid.add(new JLabel("Gens:"));
		gensSpinnerP = new JSpinner(new SpinnerNumberModel(8,2,20,1));
		gensSpinnerP.setToolTipText("The number of division-classes to plot");
		pgrid.add(gensSpinnerP);
		pgrid.add(new JLabel("Curves:"));
		vSpinnerP = new JSpinner(new SpinnerNumberModel(250,1,1000,1));
		vSpinnerP.setToolTipText("Number of times the solution will be sampled and visualized. Pick a number that is high enough to showcase the variability in the solution(s).");
		pgrid.add(vSpinnerP);
		pgrid.add(new JLabel("Plot height:"));
		drawsizeSpinnerP = new JSpinner(new SpinnerNumberModel(100,80,500,10));
		drawsizeSpinnerP.setToolTipText("The height of each plot.");
		pgrid.add(drawsizeSpinnerP);
		tboxP = new JCheckBox("Transparent?");
		tboxP.setFont(new Font("Calibri",Font.PLAIN,12));
		tboxP.setSelected(true);
		tboxP.setToolTipText("Render semi-transparent curves?");
		pgrid.add(tboxP);
		overlayBoxP = new JCheckBox("Overlay?");
		overlayBoxP.setFont(new Font("Calibri",Font.PLAIN,12));
		overlayBoxP.setSelected(false);
		overlayBoxP.setToolTipText("Overlay selected solutions?");
		pgrid.add(overlayBoxP);
		maxBoxP = new JCheckBox("Norm to Max?");
		maxBoxP.setFont(new Font("Calibri",Font.PLAIN,12));
		maxBoxP.setSelected(true);
		maxBoxP.setToolTipText("All lognormal distributions will be drawn normalized to their maximum if checked.");
		pgrid.add(maxBoxP);
		nameboxP = new JCheckBox("Names?");
		nameboxP.setFont(new Font("Calibri",Font.PLAIN,12));
		nameboxP.setToolTipText("Do you want to draw the names as they appear in the list at left?");
		pgrid.add(nameboxP);
		lblboxP = new JCheckBox("Labels?");
		lblboxP.setFont(new Font("Calibri",Font.PLAIN,12));
		lblboxP.setToolTipText("Do you want to draw the axes and numbers?");
		lblboxP.setSelected(true);
		pgrid.add(lblboxP);
		actualSP = new JCheckBox("Plot clusters?");
		actualSP.setFont(new Font("Calibri",Font.PLAIN,12));
		actualSP.setToolTipText("Choose to sample the average solution space instead of showing the best-fit solutions?");
		actualSP.setSelected(true);
		pgrid.add(actualSP);
		fcytonboxP = new JCheckBox("fCyton?");
		fcytonboxP.setFont(new Font("Calibri",Font.PLAIN,12));
		fcytonboxP.setToolTipText("Solve assuming fcyton model is to be used?");
		fcytonboxP.setSelected(true);
		pgrid.add(fcytonboxP);	
		String[] drawTypes = {"Distributions","Sampled","Mean/Sd."};
		drawBoxP = new JComboBox(drawTypes);
		drawBoxP.setFont(new Font("Calibri",Font.PLAIN,12));
		drawBoxP.setToolTipText("Choose how to draw the distributions");
		drawBoxP.setSelectedIndex(0);
		pgrid.add(drawBoxP);	
		
		Color[] overlayColors = {Color.black,Color.red,Color.green,Color.blue,Color.orange,Color.MAGENTA,Color.yellow,Color.gray,Color.CYAN};
		
		overlayColorButtons = new JButton[overlayColors.length];
		for(int i = 1; i <= overlayColors.length; i++)
		{

			overlayColorButtons[i-1] = new JButton("Overlay "+i);
			overlayColorButtons[i-1].setForeground(overlayColors[i-1]);
			overlayColorButtons[i-1].setToolTipText("Click to select the overlay color for dataset: "+i);
			overlayColorButtons[i-1].addActionListener(new ActionListener(){

				public void actionPerformed(ActionEvent arg0) {
					// TODO Auto-generated method stub
					//lets process the text
					JFrame dcframe = new JFrame("Overlay color");
					overlayColorer oc = new overlayColorer((JButton)arg0.getSource(),dcframe);
					dcframe.setSize(new Dimension(200,200));
					dcframe.getContentPane().setLayout(new GridLayout(1,1));
					dcframe.getContentPane().add(oc);
					dcframe.setDefaultCloseOperation(JFrame.DISPOSE_ON_CLOSE);
					dcframe.setVisible(true);
					System.err.println("Visible");
				}
				
			});	
			pgrid.add(overlayColorButtons[i-1]);
		}
		
		pgrid.setPreferredSize(new Dimension(100,1000));
		pDrawPanel.add(pgrid, BorderLayout.EAST);
		pDrawPanel.add(dscroller,BorderLayout.CENTER);
		drawTabs.addTab("Visualize Parameters", pDrawPanel);
		
		JPanel cDrawPanel = new JPanel(new BorderLayout(2,2));
		JPanel cgrid = new JPanel(new GridLayout(30,1,2,2));
		cgrid.add(countsButton);
		cgrid.add(colButton);
		cgrid.add(new JLabel("Tmax (h):"));
		timeSpinnerC = new JSpinner(new SpinnerNumberModel(144,10,500,10));
		timeSpinnerC.setToolTipText("The maximum plotted time (h)");
		cgrid.add(timeSpinnerC);
		cgrid.add(new JLabel("Cells x10^4:"));
		maxHeightSpinnerC = new JSpinner(new SpinnerNumberModel(-1,-1,1000,5));
		maxHeightSpinnerC.setToolTipText("The maximum plotted cell count. Set to -1 to calculate automatically.");
		cgrid.add(maxHeightSpinnerC);
		cgrid.add(new JLabel("Gens:"));
		gensSpinnerC = new JSpinner(new SpinnerNumberModel(8,2,20,1));
		gensSpinnerC.setToolTipText("The number of division-classes to simulate.");
		cgrid.add(gensSpinnerC);
		cgrid.add(new JLabel("Curves:"));
		vSpinnerC = new JSpinner(new SpinnerNumberModel(250,1,1000,1));
		vSpinnerC.setToolTipText("Number of times the solution will be sampled and visualized. Pick a number that is high enough to showcase the variability in the solution(s).");
		cgrid.add(vSpinnerC);
		cgrid.add(new JLabel("Plot height:"));
		drawsizeSpinnerC = new JSpinner(new SpinnerNumberModel(250,50,1000,10));
		drawsizeSpinnerC.setToolTipText("The height of each plot.");
		cgrid.add(drawsizeSpinnerC);
		
		tboxC = new JCheckBox("Transparent?");
		tboxC.setFont(new Font("Calibri",Font.PLAIN,12));
		tboxC.setSelected(true);
		tboxC.setToolTipText("Render semi-transparent curves?");
		cgrid.add(tboxC);
		nameboxC = new JCheckBox("Names?");
		nameboxC.setFont(new Font("Calibri",Font.PLAIN,12));
		nameboxC.setToolTipText("Do you want to draw the names as they appear in the list at left?");
		cgrid.add(nameboxC);
		lblboxC = new JCheckBox("Labels?");
		lblboxC.setFont(new Font("Calibri",Font.PLAIN,12));
		lblboxC.setToolTipText("Do you want to draw the axes and numbers?");
		lblboxC.setSelected(true);
		cgrid.add(lblboxC);
		totalsBox = new JCheckBox("Totals only?");
		totalsBox.setFont(new Font("Calibri",Font.PLAIN,12));
		totalsBox.setToolTipText("Visualize only the total cell counts?");
		totalsBox.setSelected(false);
		cgrid.add(totalsBox);
		fcytonboxC = new JCheckBox("fcyton?");
		fcytonboxC.setFont(new Font("Calibri",Font.PLAIN,12));
		fcytonboxC.setToolTipText("Use the fcyton model instead of the cyton model");
		fcytonboxC.setSelected(true);
		cgrid.add(fcytonboxC);	
		actualSC = new JCheckBox("Plot clusters?");
		actualSC.setFont(new Font("Calibri",Font.PLAIN,12));
		actualSC.setToolTipText("Choose to sample the average solution space instead of showing the best-fit solutions?");
		actualSC.setSelected(true);
		cgrid.add(actualSC);

		cDrawPanel.setPreferredSize(new Dimension(100,1000));
		cDrawPanel.add(cgrid, BorderLayout.EAST);
		cDrawPanel.add(cscroller,BorderLayout.CENTER);
		drawTabs.addTab("Visualize Counts", cDrawPanel);
		
		JPanel compPanel = new JPanel(new FlowLayout(FlowLayout.LEFT,2,2));
		compPanel.add(cButton);
		compPanel.add(new JLabel("FlowMax will perform a statistical analysis on two selected phenotypes,")); 
		JSeparator sep = new JSeparator(SwingConstants.HORIZONTAL);
		sep.setPreferredSize(new Dimension(800,0));
		compPanel.add(sep);
		
		compPanel.add(new JLabel("in the context of one or more experimental data sets. Generations:"));
		gensSpinnerComp = new JSpinner(new SpinnerNumberModel(8,2,20,1));
		compPanel.add(gensSpinnerComp);
		
		JSeparator sep2 = new JSeparator(SwingConstants.HORIZONTAL);
		sep2.setPreferredSize(new Dimension(800,2));
		compPanel.add(sep2);
		
		compPanel.add(mButton);
		compPanel.add(new JLabel("FlowMax will generate variants of the first phenotype containing parameters from the second phenotype."));
		
		
		JSeparator sep3 = new JSeparator(SwingConstants.HORIZONTAL);
		sep3.setPreferredSize(new Dimension(800,0));
		compPanel.add(sep3);
		
		JPanel paramPanel = new JPanel(new GridLayout(6,2));
		morphParams = new JCheckBox[12];
		for(int i = 0; i < 12; i++)
		{
			morphParams[i]=new JCheckBox(Constants.paramNames[i]);
			paramPanel.add(morphParams[i]);
		}
		compPanel.add(paramPanel);
		
		JSeparator sep4 = new JSeparator(SwingConstants.HORIZONTAL);
		sep4.setPreferredSize(new Dimension(1920,0));
		compPanel.add(sep4);


		
		drawTabs.addTab("Compare Phenotypes",compPanel);
		
		aPane.add(drawTabs,BorderLayout.CENTER);
		
		tabber.addTab("Solution Analysis", aPane);
		tabber.addChangeListener(new ChangeListener(){

			public void stateChanged(ChangeEvent arg0) {
				// TODO Auto-generated method stub
				updateFCSMapping();
			}

		});
		cp.setLayout(new BorderLayout());
		cp.add(tabber,BorderLayout.CENTER);
		cp.invalidate();
		setIconImage(Methods.getImageFromResource(("FM.png")));

		this.setVisible(true);
		loadDefaultParams();
	}

	private void loadDefaultParams() {
		//try to load parameters from the default set.
		File loadFile = new File("default.prm");
		try {
			Scanner s = new Scanner(loadFile);
			int spinIndex = 0;
			while(s.hasNextDouble())
			{
				if(spinIndex < fluorescenceSpinners.length)
				{
					fluorescenceSpinners[spinIndex].setValue(s.nextDouble());
				}
				else
				{
					cytonSpinners[spinIndex-fluorescenceSpinners.length].setValue(s.nextDouble());
				}
				spinIndex++;
			}
			s.close();
		}catch(Exception e){}
	}

	private void updateFCSMapping() {
		// TODO Auto-generated method stub
		if(tabber.getSelectedIndex()> 0)
		{
			map= new HashMap<String,ArrayList<FMFCS>>();
			exptCombo.removeAllItems();

			for(int i = 0; i < runs.size(); i++)
			{
				//lets figure out which ones are ready and sort according to condition and timepoints
				if(runs.get(i).getReady() != null)
				{
					if(map.get(runs.get(i).getCondition()) == null)
					{
						map.put(runs.get(i).getCondition(), new ArrayList<FMFCS>());
						exptCombo.addItem(runs.get(i).getCondition());
					}
					int index = 0;
					while(index < map.get(runs.get(i).getCondition()).size() &&map.get(runs.get(i).getCondition()).get(index).getTime() < runs.get(i).getTime() )
						index++;
					map.get(runs.get(i).getCondition()).add(index,runs.get(i));
				}
			}
			if(map.size() == 0)
			{
				PhenfcytonBtn.setEnabled(false);
				PhenCytonBtn.setEnabled(false);
			}else
			{
				PhenfcytonBtn.setEnabled(true);
				PhenCytonBtn.setEnabled(true);
			}
			
		}
	}
	
	private void processPhenotypeCSV(File selectedFile) {
		//lets read in the file line by line and generate appropraite phenotype clusters in the list
		Scanner s;
		try {
			s = new Scanner(selectedFile);
			double[] params = new double[nparams*3];
			int current_cluster = 0;
			double score = -1;
			double[][][] counts = null;
			double[] times = null;
			ArrayList<double[]> candidatesList = new ArrayList<double[]>();
			boolean in_cluster_candidates = false;
			while(s.hasNextLine())
			{
				String[] line = s.nextLine().split(",");
				if(line[0].startsWith("Generational"))
				{
					int timepoints = Integer.parseInt(line[5]);
					int replicates = Integer.parseInt(line[7]);
					int num_gens = Integer.parseInt(line[9]);
					counts = new double[timepoints][replicates][num_gens];
					times = new double[timepoints];
				}
				else if(line[0].startsWith("Time"))
				{
					for(int i = 0; i < counts.length; i++)
					{
						String[] split = s.nextLine().split(",");
						double time = Double.parseDouble(split[0]); //hours
						times[i] = time;
						String[][] temp = new String[counts[i].length][counts[i][0].length];
						for(int j = 0; j < temp.length; j++)
						{
							for(int k = 0; k < temp[j].length; k++)
							{
								temp[j][k] = split[1+(j*temp[j].length+k)];
							}
						}
						for(int j = 0; j < counts[i].length; j++)
						{
							for(int k = 0; k < counts[i][j].length; k++)
							{
								counts[i][j][k] = Double.parseDouble(temp[j][k]);
							}
						}
					}
				}
				else if(line[0].startsWith("Cluster "))
				{
					//we have found the parameters and their ranges
					current_cluster = Integer.parseInt(line[0].split(" ")[1]);
					s.nextLine();
					in_cluster_candidates = true;
				}
				else if(line[0].startsWith("Param"))
				{
					for(int i = 0; i < nparams; i++)
					{
						String[] s2 = s.nextLine().split(",");
						params[i] = Double.parseDouble(s2[4]);
						params[i+nparams] = Double.parseDouble(s2[5]);
						params[i+2*nparams] = Double.parseDouble(s2[6]);
					}
					answers.add(Methods.duplicate(params));
					answerNames.add(selectedFile.getName()+"_"+current_cluster);
					conditionNames.add(selectedFile.getName().split("\\.")[0]);
					answerScores.add(score);
					((DefaultListModel)slist.getModel()).addElement(selectedFile.getName()+"_"+current_cluster);
					answerTimes.add(times);
					answerCounts.add(counts);
					answerCandidates.add(candidatesList);
					candidatesList = new ArrayList<double[]>();
					in_cluster_candidates = false;
					//save everything
				//	draw.addParameters(Methods.duplicate(params), selectedFile.getName()+"_"+current_cluster, score);
				}
				else if(in_cluster_candidates)
				{
					double[] canParams = new double[Constants.nparams*3];
					for(int i = 2; i < canParams.length+2; i++)
					{
						canParams[i-2] = Double.parseDouble(line[i]);
					}
					candidatesList.add(canParams);
				}
				
			}
		} catch (FileNotFoundException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}

		
	}
	




	private void saveImage(String filename, Experiment exp,int mode)
	{
		if(mode==1)
		{
			exp.getCytonScore2();
			
			ArrayList<TimePoint> timePoints = exp.timePoints;
			for(int t = 0; t < timePoints.size(); t++)
			{
				for(int i = 0; i < timePoints.get(t).getReplicates().size(); i++)
				{
					Run r = timePoints.get(t).getReplicates().get(i);
					if( r.lastCorrelation< 10)
						r.updateDraw("Scr:", "%Diff:", new Color(200,255,200,50));
					else if( r.lastCorrelation< 20)
						r.updateDraw("Scr:", "%Diff:", new Color(255,255,200,50));
					else
						r.updateDraw("Scr:", "%Diff:", new Color(255,200,200,50));
				}
			}
		}
		else if (mode==0)
		{
			for(int i = 0; i < exp.timePoints.size(); i++)
				exp.timePoints.get(i).getFluorescenceScore2();
			ArrayList<TimePoint> timePoints = exp.timePoints;
			for(int t = 0; t < timePoints.size(); t++)
			{
				for(int i = 0; i < timePoints.get(t).getReplicates().size(); i++)
				{
					Run r = timePoints.get(t).getReplicates().get(i);
					if( r.lastScore< 10)
						r.updateDraw("Scr:", "%Diff:", new Color(200,255,200,50));
					else if( r.lastScore< 20)
						r.updateDraw("Scr:", "%Diff:", new Color(255,255,200,50));
					else
						r.updateDraw("Scr:", "%Diff:", new Color(255,200,200,50));
				}
			}
		}
		if(filename != null)
		{
			//we need to save what we just drew to the file
			BufferedImage saveImage = ProliferationDisplay.drawOnGraphics(exp,  (Integer)sizeSpinner.getValue(),colorBox.isSelected(),0.001*((Integer)(countsSpinner.getValue())));
			try {
				File f = new File(filename);
				int fnum = 2;
				while(f.exists())
				{
					f = new File(filename.substring(filename.length()-4)+fnum+".png");
					fnum++;
				}					
				ImageIO.write(saveImage, "png", new File(filename));
			} catch (IOException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}
			filename = null;
		}

	}


	
/**
 *  Does the phenotyping
 * @param gens
 * @param undivided
 * @param time
 * @param volume
 * @param data
 * @param mode
 */
	private void proliferationAnalysis(int gens, ArrayList<Double> undivided, ArrayList<Double> time, ArrayList<Integer> volume, ArrayList<ArrayList<Double>> data, int mode, String expname)
	{
		int total_reps_already = 0;
		if(profDisp.getExp() != null)
		{
			for(int i = 0; i < profDisp.getExp().timePoints.size(); i++)
			{
				total_reps_already+=profDisp.getExp().timePoints.get(i).getReplicates().size();
			}
		}

		if(mode < 3|| profDisp.getExp() == null || gens != profDisp.getExp().gens ||  profDisp.getExp().name.compareTo((String)exptCombo.getSelectedItem())!=0||total_reps_already !=  data.size())
		{
			//lets figure out replicates for each time point
			System.err.println("Started proliferation analysis");
			ArrayList<Double> times = new ArrayList<Double>();
			ArrayList<ArrayList<ArrayList<Double>>> sorted = new ArrayList<ArrayList<ArrayList<Double>>>();
			ArrayList<ArrayList<Double>> undived = new ArrayList<ArrayList<Double>>();
			ArrayList<ArrayList<Integer>> volumes = new ArrayList<ArrayList<Integer>>();
			//sorting data according to time point (assume all data sets from the same time point are replicates for the same condition)
			int height = 1;
			double maxT = 0;
			for(int i = 0; i < time.size(); i++)
			{
				boolean found = false;
				for(int j = 0; j < times.size(); j++)
				{
					if(times.get(j).compareTo(time.get(i))==0)
					{
						found = true;
						undived.get(j).add(undivided.get(i));
						volumes.get(j).add(volume.get(i));
						sorted.get(j).add(data.get(i));
						height = Math.max(height,sorted.get(j).size());
					}
				}
				if(!found)
				{
					times.add(time.get(i));
					if(time.get(i) > maxT) maxT = time.get(i);
					ArrayList<ArrayList<Double>> list = new ArrayList<ArrayList<Double>>();
					ArrayList<Double> current_peaks = new ArrayList<Double>();
					ArrayList<Integer> current_volumes = new ArrayList<Integer>();
					list.add(data.get(i));
					current_peaks.add(undivided.get(i));
					current_volumes.add(volume.get(i));
					sorted.add(list);
					undived.add(current_peaks);
					volumes.add(current_volumes);
				}
			}
			proliferator.remove(scroller2);
			profDisp.setPreferredSize(new Dimension(times.size()*300,height*300));		
			scroller2 = new JScrollPane(profDisp);
			scroller2.setPreferredSize(new Dimension(630,630));
			proliferator.add(scroller2);
			proliferator.invalidate();
			proliferator.updateUI();
			profDisp.draw();
			System.err.println("Created sorted list of data");
			Experiment e = new Experiment(sorted,undived,gens,volumes,100,times,expname);
			e.registerFitListener(this);
			e.registerFluorescenceFitListener(this);
			profDisp.setExperiment(e);
			profDisp.updateUI();
		}

		
		/**********************************************************************************************************************************
		 *   Fit both the fluorescence and fcyton/Cyton models (new model)
		 *  
		 ***********************************************************************************************************************************/
		
		if(mode< 2)
		{
			long start_time = System.currentTimeMillis();
			profDisp.updateUI();
			
			String phenotype_name = "";
			if(mode==0)phenotype_name = JOptionPane.showInputDialog(this,"Output for fcyton phenotyping: "+profDisp.getExp().name+"?",profDisp.getExp().name);
			else phenotype_name = JOptionPane.showInputDialog(this,"Output for Cyton phenotyping: "+profDisp.getExp().name+"?",profDisp.getExp().name);
			String outf = "Phenotypes"+File.separator+phenotype_name;
			File tout = new File(outf);
			tout.mkdirs();
			double[][] flims = new double[4][2];
			for(int i = 0; i < fluorescenceSpinners.length ; i++)
			{
				flims[i/2][i%2]=((Double) fluorescenceSpinners[i].getValue()).doubleValue();
				if(i%2==1)System.err.println("lims are: "+flims[i/2][0]+","+flims[i/2][1]);
			}
			boolean finished = profDisp.getExp().fitFluorescence(1000,flims);
			if(!finished) return;
			saveImage(new String(outf+File.separator+phenotype_name+"_fl.png"),profDisp.getExp(),0);
			profDisp.draw();
			gens = (Integer)genSpinner.getValue();
			profDisp.getExp().gens = gens;
			profDisp.getExp().repeats = (Integer) repeatsSpinner.getValue();
			double[][] lims = new double[nparams][2];
			for(int i = 0; i < nparams; i++)
			{
				if(i <8)
				{
					double e1 = (Double)cytonSpinners[i*2].getValue();
					double e2 = (Double)cytonSpinners[i*2+1].getValue();
					double sd1  = (Double)cytonSpinners[i*2+2].getValue();
					double sd2 = (Double)cytonSpinners[i*2+3].getValue();
					//we also want to ensure that the sds are always less than or equal to the expected values to avoid degeneracy in distribution shape
					sd1= Math.min(sd1, e1);
					sd2 = Math.min(sd2, e2);
					double s1 = Math.log(1+(Math.pow(sd1,2))/Math.pow(e1,2));
					double u1 = Math.log(e1)-0.5*s1;
					s1 = Math.sqrt(s1);
					double s2 = Math.log(1+(Math.pow(sd2,2))/Math.pow(e2,2));
					double u2 =Math.log(e2)-0.5*s2;
					s2 = Math.sqrt(s2);
					lims[i][0] = u1;
					lims[i][1] = u2;
					lims[i+1][0] = s1;
					lims[i+1][1] = s2;
					System.err.println(i+" "+limits[i][0]+" "+limits[i][1]);
					i++;
					System.err.println(i+" "+limits[i][0]+" "+limits[i][1]);
				}
				else
				{
					lims[i][0] = (Double)cytonSpinners[i*2].getValue();
					lims[i][1] = (Double)cytonSpinners[i*2+1].getValue();
					System.err.println(i+" "+limits[i][0]+" "+limits[i][1]);
				}
			}
			ArrayList<double[] > solutions = null;
			try{
				
				if(outf==null) return;
				File outFile = new File(tout.getAbsolutePath()+File.separator+phenotype_name);
				BufferedWriter out = new BufferedWriter(new FileWriter(outFile.getAbsolutePath()+".csv", false));
				if(mode==0)
				{
					out.write("Phenotype:,"+phenotype_name+",model used:,fcyton Model,Started at:,"+Methods.now());out.newLine();
				}
				else
				{
					out.write("Phenotype:,"+phenotype_name+",model used:,Cyton Model,Started at:,"+Methods.now());out.newLine();
				}
				//we can now also save the counts!!!!!!!!
				saveCounts(out,profDisp.getExp());
				if(profDisp.getExp().repeats > 0)
				{	
			
				//Fitting as described in the text
				if(mode==0)solutions =profDisp.getExp().fitCyton(gens,lims,dt,true,(Integer) repeatsSpinner.getValue());
				else solutions =profDisp.getExp().fitCyton(gens,lims,dt,false,(Integer) repeatsSpinner.getValue());

				
				if(solutions.size() == 0) return;
				int solution_count= solutions.size();
				double bestScore = solutions.get(0)[1];
				String os = "fitScore,Weighted %Diff,Tdiv0 u,Tdiv0 s,Tdiv1 u,Tdiv1 s,Tdie0 u,Tdie0 s,Tdie1 u,Tdie1 s,D u,D s, F0,Start Count,pmech,kmech,Tdiv0 u<,Tdiv0 s<,Tdiv1 u<,Tdiv1 s<,Tdie0 u<,Tdie0 s<,Tdie1 u<,Tdie1 s<,D u<,D s<, F0<,Start Count<,pmech<,kmech<,Tdiv0 u>,Tdiv0 s>,Tdiv1 u>,Tdiv1 s>,Tdie0 u>,Tdie0 s>,Tdie1 u>,Tdie1 s>,D u>,D s>, F0>,Start Count>,pmech>,kmech>";

				 out.write(os);out.newLine();
				 for(int j = 0; j < solutions.size(); j++)
				 {
				 	System.err.print(" "+j);
				 	String line = ""+solutions.get(j)[0];
					for(int k = 1; k < solutions.get(j).length; k++)
					{
						if(k==1)
							line+=","+Math.sqrt(solutions.get(j)[k]);
						else
							line+=","+solutions.get(j)[k];
					}
					out.write(line); out.newLine();
				}
				 double[] second_best = null;
				 if(solutions.size() > 1) second_best= solutions.get(1);
				 //trim the data according to 1% cutoff along the score
				for(int i = 0; i < solutions.size(); i++)
				{
					if(solutions.get(i)[1] >  bestScore+0.1) //within 0.1%
					{
						solutions.remove(i);
						i--;
					}
				}
				if(second_best != null && solutions.size() ==1)
					solutions.add(second_best);
				//now lets run the clustering by sensitivity maximization agglomeration
				boolean[] mask = new boolean[nparams*3+2];
				for(int j = 2;j <2+nparams; j++)
					mask[j]=true;
	
				if(solutions.size()==0) return;
				System.err.println("Starting clustering with "+solutions.size()+" solutions.");
				Clustering c = heirarchalClusterWithArms(solutions);
				//now lets output the clusters and the average solutions with average sensititivities
				//now lets go through and sample from each cluster solution and order them by their scores
				Clustering sorted_c = new Clustering(new ArrayList<ArrayList<double[]>>(),new ArrayList<double[]>());
				ArrayList<Double> sorted_scores = new ArrayList<Double>();
				for(int i = 0; i < c.averages.size();i++)
				{
					double[] average = c.averages.get(i);
					double[] scores = new double[1000];
					for(int sample = 0; sample < 1000; sample++)
					{
						double[] params = new double[Constants.nparams];
						for(int j = 0; j < Constants.nparams; j++)
						{
							params[j] = average[2+j]-average[2+j+Constants.nparams]+Math.random()*(average[2+j+Constants.nparams]+average[2+j+Constants.nparams*2]);
						}
						profDisp.getExp().cytonToFluorescence(params, mode==0);
						scores[sample]=profDisp.getExp().getAreaDiff();
					}
					Arrays.sort(scores);
					double median_score = scores[499];
					c.averages.get(i)[0] = median_score;
					System.err.println("Cluster "+(i+1)+" has a median fit score of "+median_score);
					//insert sort
					int j = 0;
					for(j=0; j < sorted_scores.size(); j++)
					{
						if(sorted_scores.get(j) > median_score)
							break;
					}
					sorted_scores.add(j,median_score);
					sorted_c.averages.add(j,average);
					sorted_c.clusters.add(j,c.clusters.get(i));
				}
				double bestClusterScore = sorted_scores.get(0);
				for(int i = 0; i < sorted_scores.size(); i++)
				{
					System.err.println("Cluster with median score: "+sorted_scores.get(i)+" acceptable = "+(bestClusterScore+1));
					if(sorted_scores.get(i) > bestClusterScore+1)
					{
						sorted_scores.remove(i);
						sorted_c.averages.remove(i);
						sorted_c.clusters.remove(i);
						i--;
					}
				}
				out.write("Total solutions,"+solution_count+",After filtering by score,"+solutions.size()+",Number of non-overlapping solution clusters,"+c.clusters.size()+",detrmined using sensitivity agglomeration clustering,After filtering by average cluster sample score there were,"+sorted_c.clusters.size()+",clusters remaining");out.newLine();
				for(int i = 0; i < sorted_c.clusters.size(); i++)
				{
					System.err.println("Processing cluster "+(i+1));
					out.write("Cluster "+(i+1)+" with "+sorted_c.clusters.get(i).size()+" solutions total,Average sampled cluster score was,"+sorted_scores.get(i));
					out.newLine();
					ArrayList<double[]> cluster;
					cluster = new ArrayList<double[]>();
					double[][]vals = new double[nparams][sorted_c.clusters.get(i).size()];
					for(int j = 0; j < sorted_c.clusters.get(i).size(); j++)
					{
						int k = 0;
						for(k = 0; k < cluster.size(); k++)
						{
							if(cluster.get(k)[0] > sorted_c.clusters.get(i).get(j)[0])
								break;
						}
						cluster.add(k,Methods.duplicate(sorted_c.clusters.get(i).get(j)));
					}
					System.err.println("Finished ordering and filling out vals");
					String[] names = {"Tdiv0 u","Tdiv0 s","Tdiv1 u","Tdiv1 s","Tdie0 u","Tdie0 s","Tdie1 u","Tdie1 s","D u","D s","F0","Start Count","pmech","kmech"};
					out.write(os); out.newLine();
					for(int j = 0; j < cluster.size(); j++)
					{
						System.err.print(" "+j);
						String line = ""+cluster.get(j)[0];
						for(int k = 1; k < cluster.get(j).length; k++)
						{
							if(k==1)
								line+=","+Math.sqrt(cluster.get(j)[k]);
							else
								line+=","+cluster.get(j)[k];
							if(k < nparams+2 && k > 1)
							vals[k-2][j] = cluster.get(j)[k];
						}
						out.write(line); out.newLine();
	
					}
	
					double[] average_params = new double[nparams];			
					for(int k = 0; k < 8; k+=2)
					{
						double second = Math.log(1+Math.pow(sorted_c.averages.get(i)[k+3]/sorted_c.averages.get(i)[k+2],2));
						double u = Math.log(sorted_c.averages.get(i)[k+2])-0.5*second;
						double s = Math.sqrt(second);
						average_params[k]=u;
						average_params[k+1]=s;
					}
					for(int k = 8; k < nparams; k++)
						average_params[k]=sorted_c.averages.get(i)[k+2];
					out.write("Param,Ar. Average,Ar. s.d.,Ar. SEM,Overlap minimum,Lower,Upper"); out.newLine();
					for(int k = 0; k < nparams; k++)
					{
						out.write(names[k]+","+Methods.average(vals[k])+","+Methods.std(vals[k])+","+(Methods.std(vals[k])/Math.sqrt(cluster.size()))+","+sorted_c.averages.get(i)[k+2]+","+sorted_c.averages.get(i)[k+nparams+2]+","+sorted_c.averages.get(i)[k+nparams*2+2]);out.newLine();
					}
					profDisp.getExp().cytonToFluorescence(Methods.duplicate(average_params),mode==0);
					answers.add(Methods.suffix(sorted_c.averages.get(i), 2));
					conditionNames.add(phenotype_name);
					answerNames.add(phenotype_name+"_C"+(i+1));
					answerScores.add(sorted_scores.get(i));
					ArrayList<double[]> clusterCandidates = new ArrayList<double[]>();
					for(double[] a : sorted_c.clusters.get(i))
					{
						double[] trimmed = Methods.suffix(a, 2);
						clusterCandidates.add(trimmed);
					}
					answerCandidates.add(clusterCandidates);
					double[] times = new double[profDisp.getExp().timePoints.size()];
					int max_rep = 0;
					for(int j = 0; j < profDisp.getExp().timePoints.size(); j++)
					{
						times[j] = profDisp.getExp().timePoints.get(j).getTime();
						max_rep = Math.max(max_rep,profDisp.getExp().timePoints.get(j).getReplicates().size());
					}
					answerTimes.add(times);
					double[][][] counts = new double[times.length][max_rep][];
					for(int j = 0; j < times.length; j++)
					{
						for(int k = 0; k < profDisp.getExp().timePoints.get(j).getReplicates().size(); k++)
						{
							for(int l = 0; l < profDisp.getExp().gens; l++)
							{
								counts[j][k] = profDisp.getExp().timePoints.get(j).getReplicates().get(k).getConcentrations();
							}
						}
					}
					answerCounts.add(counts);
					((DefaultListModel)slist.getModel()).addElement(phenotype_name+"_C"+(i+1));
					saveImage(new String(outFile.getAbsolutePath()+"_C"+(i+1)+".png"),profDisp.getExp(),1);
				}
				String model = "fcyton";
				if(mode==1) model = "Cyton";
				JOptionPane.showMessageDialog(this, "Finished phenotyping "+profDisp.getExp().name+". Total number of "+model+" fits:"+solution_count+" (Filtered all but "+solutions.size()+"). Total number of non-overlapping good solutions: "+sorted_c.clusters.size());
				}
				out.close();
				System.err.println("Finished everything using output file: "+outFile.getName()+" The fitting took: "+((System.currentTimeMillis()-start_time)/60000)+" minutes.");
			}catch(Exception e)
			{
				e.printStackTrace();
			}

		}

		/**********************************************************************************************************************************
		 *   Just do the basic fluorescence fitting to help with parameter visualization... 
		 *  
		 ***********************************************************************************************************************************/
		
		else if(mode==2)
		{
			profDisp.updateUI();
			
			String phenotype_name = JOptionPane.showInputDialog(this,"Output for Fluorescence Fitting: "+profDisp.getExp().name+"?",profDisp.getExp().name);
			String outf = "Phenotypes"+File.separator+phenotype_name;
			File tout = new File(outf);
			tout.mkdirs();
			double[][] flims = new double[4][2];
			for(int i = 0; i < fluorescenceSpinners.length ; i++)
			{
				flims[i/2][i%2]=((Double) fluorescenceSpinners[i].getValue()).doubleValue();
				if(i%2==1)System.err.println("lims are: "+flims[i/2][0]+","+flims[i/2][1]);
			}
			boolean finished = profDisp.getExp().fitFluorescence(1000,flims);
			if(!finished) return;
			saveImage(new String(outf+File.separator+phenotype_name+"_fl.png"),profDisp.getExp(),0);
			profDisp.draw();
		
		}	
		System.err.println("Done with analysis");

	}
		

	public static Clustering heirarchalClusterWithArms(
			ArrayList<double[]> solutions) {
		//agglomerate until k
		ArrayList<ArrayList<double[]>> clustering = new ArrayList<ArrayList<double[]>>();
		ArrayList<double[]> averages = new ArrayList<double[]>();  //average values for each cluster
		for(int j = 0; j < solutions.size(); j++)  //one solution per cluster to start
		{
			clustering.add(new ArrayList<double[]>());
			clustering.get(clustering.size()-1).add(solutions.get(j));
			averages.add(Methods.duplicate(solutions.get(j)));  
		}

		for(int i = 0; i < solutions.size()-1; i++)
		{
			System.err.println(""+((i*100)/(solutions.size())+"% done."));

			//find the two closest clusters using an appropriate distance measure
			int best_a, best_b;
			best_a=best_b = 0;
			double bestDistance = Double.NEGATIVE_INFINITY;
			for(int j = 0; j < clustering.size();j ++)	//cluster j
			{
				for(int k = j+1; k < clustering.size(); k++ )  //cluster k
				{
					double dist  = 0;
					for(int l = 2; l < nparams+2; l++)  //parameters
					{
						double d = (Math.min(averages.get(k)[l]+averages.get(k)[l+nparams*2],averages.get(j)[l]+averages.get(j)[l+nparams*2])-Math.max(averages.get(k)[l]-averages.get(k)[l+nparams],averages.get(j)[l]-averages.get(j)[l+nparams]))/(Math.abs(averages.get(k)[l]+averages.get(k)[l+nparams*2])+Math.abs(averages.get(j)[l]+averages.get(j)[l+nparams*2]));
						if(Double.isNaN(d))d=0;
						if(d < 0 )d = Double.NEGATIVE_INFINITY;			
						dist+=d;  //distance gets the distance between these two clusterings along the Lth parameter 
					}
					if(dist > bestDistance)  //we wanted to overlap
					{
						best_a=j;  //should combine clustering j and clustering k
						best_b=k;
						bestDistance=dist;
					}
				}
			}
			if(bestDistance >0)
			{
				//combine best_a and best_b
				ArrayList<double[]> nc = new ArrayList<double[]>();  //new cluster
				nc.addAll(clustering.get(best_a));	//add all from besta
				nc.addAll(clustering.get(best_b));   //add all from bestb
				double[] newA = new double[2+nparams*3];  //update average (should reflect
				for(int j = 2; j < nparams+2;j++)
				{
					double[] A = averages.get(best_a);
					double[] B = averages.get(best_b);
					if(A[j] < B[j])
					{
						double x = B[j+nparams];
						double y = A[j+nparams*2];
						double z = B[j]-A[j];
						double xint = (y*z)/(y+x);
						if(Double.isNaN(xint))xint = 0;
						newA[j]= A[j]+xint;
						newA[j+nparams] = newA[j]-Math.max(A[j]-A[j+nparams],B[j]-B[j+nparams]);
						newA[j+nparams*2] = Math.min(A[j]+A[j+nparams*2], B[j]+B[j+nparams*2])-newA[j];

					}
					else
					{
						double x = A[j+nparams];
						double y = B[j+nparams*2];
						double z = A[j]-B[j];
						double xint = (y*z)/(y+x);
						if(Double.isNaN(xint))xint = 0;
						newA[j]= B[j]+xint;
						newA[j+nparams] = newA[j]-Math.max(A[j]-A[j+nparams],B[j]-B[j+nparams]);
						newA[j+nparams*2] = Math.min(A[j]+A[j+nparams*2], B[j]+B[j+nparams*2])-newA[j];
					}
				}
				clustering.remove(best_b);
				clustering.remove(best_a);
				averages.remove(best_b);
				averages.remove(best_a);
				clustering.add(nc);
				averages.add(newA);
			}
		}//main loop
		System.err.println("Finished clustering...");
		return new Clustering(clustering,averages);
	}

	
	private void saveCounts(BufferedWriter pw, Experiment exp) {
		
		try {
			int max_reps = 0;
			for(int j = 0; j < exp.timePoints.size(); j++)
				max_reps = Math.max(max_reps,exp.timePoints.get(j).getReplicates().size());
		
			pw.write("Generational counts derived from Cell Fluorescence Model fitting for,"+exp.name+",performed on,"+Methods.now()+",timepoints,"+exp.timePoints.size()+",replicates,"+max_reps+",generatons,"+exp.gens);
			pw.newLine();
			pw.write("Time");

			for(int i = 0; i < exp.gens*(max_reps+2); i++)
				pw.write(",Gen "+(i%exp.gens));
			for(int i = 0; i < (max_reps); i++)
				pw.write(",Sum "+(i+1));
			pw.write(",Time,Avg Sum,S.d. Sum");
			pw.newLine();
			for (int i = 0; i < exp.timePoints.size(); i++)
			{
				double[][] vals = new double[exp.gens][exp.timePoints.get(i).getReplicates().size()];
				double sums[] = new double[exp.timePoints.get(i).getReplicates().size()];
				pw.write(""+exp.timePoints.get(i).getTime());
				for(int j = 0; j < exp.timePoints.get(i).getReplicates().size(); j++)
				{
					Run r = exp.timePoints.get(i).getReplicates().get(j);
					double[] counts = r.getConcentrations();
					for(int k =0; k < counts.length; k++)
					{
						vals[k][j] = counts[k];
						sums[j] += counts[k];
						pw.write(","+counts[k]);
					}
				}
				for(int j = 0; j < vals.length; j++)
				{
					pw.write(","+Methods.average(vals[j]));
				}
				for(int j = 0; j < vals.length; j++)
				{
					pw.write(","+Methods.std(vals[j]));
				}
				for(int j = 0; j < sums.length; j++)
					pw.write(","+sums[j]);
				pw.write(","+exp.timePoints.get(i).getTime()+","+Methods.average(sums)+","+Methods.std(sums));
				pw.newLine();
			}
			System.err.println("Finished outputing the counts for experiment "+exp.name);
		} catch (FileNotFoundException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
	
	}


	private int unsigned(byte b)
	{
		if(b < 0)
			return 256+b;
		return b;
	}
	
	private String getNextString(int i, FileInputStream s) {
		// TODO Auto-generated method stub
		try
		{
			String result = "";
			while(i > 0 && s.available()>=i)
			{
				result+=(char)s.read();
				i--;
			}
			return result;
		}
		catch(Exception e)
		{
			e.printStackTrace();
			return null;
		}
	}
	
	public Integer getNextInteger(int r, FileInputStream s)
	{
		String nextStr = getNextString(r,s).trim();
		return Integer.parseInt(nextStr);
	}
	/**
	 * The main function called by the JVM ... this creates a new FlowMax user interface with the system look and feel
	 * @param args
	 */
	public static void main(String[] args) {
		    try {
			    // Set System L&F
		        UIManager.setLookAndFeel(
		            UIManager.getSystemLookAndFeelClassName());
		    } 
		    catch (UnsupportedLookAndFeelException e) {
		       // handle exception
		    }
		    catch (ClassNotFoundException e) {
		       // handle exception
		    }
		    catch (InstantiationException e) {
		       // handle exception
		    }
		    catch (IllegalAccessException e) {
		       // handle exception
		    }
	
			new FlowMax();  

	}





	@Override
	public void plotterUpdated(Plotter p) {
		// TODO Auto-generated method stub
		tree.updateUI();
		plot.updateAll();
	}



	public void notifyColorChange(Color c) {
		// TODO Auto-generated method stub
		
	}

	public int getNParams() {
		// TODO Auto-generated method stub
		return nparams;
	}



	public void processCytonUpdate(double[] params, int repeat, int trial, double score, Experiment sender) {
		if(drawBox.isSelected())
		{
			profDisp.setRedraw_overlay(true);
			profDisp.draw();
			String tostats = "Fit: "+(repeat+1)+"/"+sender.repeats+" Step: "+trial+separator+String.format("Score %%%4.3f %%Area %3.1f", (100*score),sender.getAreaDiff());
			tostats+=separator+String.format("N: %,d",(int)params[11]);
			double mean = Math.exp(0.5*Math.pow(params[1],2)+params[0]);
			double std = mean*Math.sqrt(Math.exp(Math.pow(params[1],2))-1);
			tostats+=separator+String.format("Tdiv0: %4.3f  (%4.3f)",mean,std);
			mean = Math.exp(0.5*Math.pow(params[3],2)+params[2]);
			std = mean*Math.sqrt(Math.exp(Math.pow(params[3],2))-1);
			tostats+=separator+String.format("Tdiv1+: %4.3f  (%4.3f)",mean,std);
			mean = Math.exp(0.5*Math.pow(params[5],2)+params[4]);
			std = mean*Math.sqrt(Math.exp(Math.pow(params[5],2))-1);
			tostats+=separator+String.format("Tdie0: %4.3f  (%4.3f)",mean,std);
			mean = Math.exp(0.5*Math.pow(params[7],2)+params[6]);
			std = mean*Math.sqrt(Math.exp(Math.pow(params[7],2))-1);
			tostats+=separator+String.format("Tdie1+: %4.3f  (%4.3f)",mean,std);
			tostats+=separator+String.format("Prol. Cap.: %4.3f (%4.3f)",params[8],params[9]);
			tostats+=separator+String.format("F0: %4.4f",params[10]);
			stats.setText(tostats);
			for(int t = 0; t < sender.timePoints.size(); t++)
			{
				for(int i = 0; i < sender.timePoints.get(t).getReplicates().size(); i++)
				{
					Run r = sender.timePoints.get(t).getReplicates().get(i);
					if( r.lastCorrelation< 10)
						r.updateDraw("Scr:", "%Diff:", new Color(200,255,200,50));
					else if( r.lastCorrelation< 20)
						r.updateDraw("Scr:", "%Diff:", new Color(255,255,200,50));
					else
						r.updateDraw("Scr:", "%Diff:", new Color(255,200,200,50));
				}
			}
		}
		
	}

	public void processFluorescenceUpdate(TimePoint sender, double bratio, double bbackground, double bCV, double score,
			int counter, double temperature) {
		String tostats = "Optimization cycle: "+counter+Constants.separator+String.format("Score:%4.2f", score)+Constants.separator+String.format("Fit temp:%3.3f", temperature);
		//last_best_time = System.currentTimeMillis();
		if(drawBox.isSelected())
		{
			profDisp.setRedraw_overlay(true);
			profDisp.draw();
		}
		for(int i = 0; i < sender.getReplicates().size(); i++)
		{
			Run r = sender.getReplicates().get(i);
			if( r.lastScore< 15)
				r.updateDraw("%Dif:", "Cor:", new Color(200,255,200,50));
			else if(r.lastScore< 30)
				r.updateDraw("%Dif:", "Cor:", new Color(255,255,200,50));
			else
				r.updateDraw("%Dif:", "Cor:", new Color(255,200,200,50));
		}
		if(drawBox.isSelected())
		{
			tostats+=separator+String.format("Ratio: %3.3f",bratio)+separator+String.format("Bgrnd Fl: %4.1f",bbackground)+separator+String.format("CV: %3.3f",bCV)+separator+"-=-=-=-WEIGHTS-=-=-=-=-";
			for(int i = 0; i < sender.getReplicates().size(); i++)
			{
				tostats+=separator+"Rep "+(i+1);
				for(int j = 0; j < sender.getReplicates().get(i).weights.length; j++)
				{
					tostats+=String.format(separator+"Gen %d:%6.1f ",j,sender.getReplicates().get(i).weights[j]*sender.getReplicates().get(i).data.raw.size());
				}
			}
			stats.setText(tostats);
		}

		
	}


}
