/*
Copyright (c)  2013   The Regents of the University of California
All Rights Reserved
 
Permission to copy, modify and distribute any part of this FlowMax for educational, research and non-profit purposes, without fee, and without a written agreement is hereby granted, provided that the above copyright notice, this paragraph and the following three paragraphs appear in all copies.
Those desiring to incorporate this FlowMax into commercial products or use for commercial purposes should contact the Technology Transfer Office, University of California, San Diego, 9500 Gilman Drive, Mail Code 0910, La Jolla, CA 92093-0910, Ph: (858) 534-5815, FAX: (858) 534-7345, E-MAIL:invent@ucsd.edu.
IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES, INCLUDING LOST PROFITS, ARISING OUT OF THE USE OF THIS FlowMax, EVEN IF THE UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
THE FlowMax PROVIDED HEREIN IS ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.  THE UNIVERSITY OF CALIFORNIA MAKES NO REPRESENTATIONS AND EXTENDS NO WARRANTIES OF ANY KIND, EITHER IMPLIED OR EXPRESS, 
INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE, OR THAT THE USE OF THE FlowMax WILL NOT INFRINGE ANY PATENT, TRADEMARK OR OTHER RIGHTS.
*/
package FlowMax;

import java.awt.BasicStroke;
import java.awt.Color;
import java.awt.Font;
import java.awt.Graphics2D;
import java.awt.image.BufferedImage;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileWriter;
import java.io.IOException;
import java.io.PrintWriter;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Random;

import javax.imageio.ImageIO;
import javax.swing.DefaultListModel;
import javax.swing.JOptionPane;

import common.Methods;


/**
 * This code is used to generate and fit realistic CFSE curves with varying sources of experimental noise. Integrated FlowMax fitting is performed and results are subject to  sensitivity analysis and clustering.
 * @author Max
 *
 */
public class GeneratedFitting{


	private int steps= 400;  //time steps
	private int maxTime= 200;  //time total
	private static final int nparams = 14;
	private static final int gens = 8;
	private static final double dt = 0.5;
	//fitting limits
	double[][] limits = {{2.83148024007,3.93009221},{0.01667,0.832555},{1.44518587895,3.17805382948},{0.01667,0.832555},{1.79162052,4.91092146},{0.01667,0.832555},{1.79162052,4.21777428},{0.01667,0.832555},{-3,8},{0.0,10},{0.001,1},{100000,400000},{0,0},{0,0}};
	double[][] hlimits = {{24,72},{0.001,72},{6,24},{0.001,24},{6,192},{0.001,192},{6,96},{0.001,96},{-3,8},{0.001,10},{0.001,1},{100000,400000},{0,0},{0,0}};
	double[][] flimits = {{0.016,0.026},{0.45,0.50},{0,1000},{-0.001,0.001}}; //ratio, CV, Background FL, shift
	//generated values:
	double[][] glimits = {{3.637774007,3.637774007},{0.221117736,0.221117736},{2.051730329,2.051730329},{0.257146218,0.257146218},{4.252936663,4.252936663},{0.824592758,0.824592758},{3.953595517,3.953595517},{0.606092072,0.606092072},{1.457149469,1.457149469},{1.84319388,1.84319388},{0.26555396,0.26555396},{204580.9381,204580.9381},{0,0},{0,0}};
	double[][] ghlimits = {{38.94773168,38.94773168},{8.718381574,8.718381574},{8.04292174	,8.04292174},{2.102872228,2.102872228},{98.78178956,98.78178956},{97.47842795,97.47842795},{62.63156079,62.63156079},{41.72877883,41.72877883},{1.457149469,1.457149469},{1.84319388,1.84319388},{0.26555396,0.26555396},{204580.9381,204580.9381},{0,0},{0,0}};
	double[][] gflimits = {{0.018,0.018},{0.5,0.50},{0,100},{-0.001,0.001}}; //ratio, CV, Background FL, shift
	
	
	public static final int maximum = 16777216;
	public static final int minimum = 1;
	String separator = System.getProperty("line.separator");
	int tpoints = 3;
	double[] timePoints  = {24,48,72};//{24,32,48,60,80,100,144,192};//{24,60,96,144};//
	double gnoise = 0.0;  // sd of gaussian for counts in each gen.
	double snoise = 0.0;  // sd of gaussian for total counts in each tp
	int tries = 100;
	
	Random gen = new Random();
	
	public GeneratedFitting(String outf) {

		maxTime = (int)timePoints[timePoints.length-1]+1;
		steps=(int)(maxTime/dt);

			System.err.println("-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-");
			System.err.println("STARTED GENERATED INTEGRATED FITTING at "+Methods.now());
			System.err.println("-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-");
			
			//lets generate the params
			boolean generatedReasonable = false;
			double[] params = new double[nparams];
			double[][] fparams = new double[timePoints.length][4];
			ArrayList<ArrayList<ArrayList<Double>>> data = new ArrayList<ArrayList<ArrayList<Double>>>();
			ArrayList<ArrayList<Double>> undivided= new ArrayList<ArrayList<Double>>();
			ArrayList<ArrayList<Integer>> volumes= new ArrayList<ArrayList<Integer>>();
			double[][] counts = null; //Live Matrix
			ArrayList<Double> totalSolved = new ArrayList<Double>();
			
			double[][] gensolvedcounts = new double[timePoints.length][gens];
			while(!generatedReasonable)
			{
				for(int i = 0; i < params.length; i++)
				{
						params[i] = gen.nextDouble()*(ghlimits[i][1]-ghlimits[i][0])+ghlimits[i][0]; //pick random params from within limits
				}
				for(int i = 0; i <8; i+=2)  //convert to log scale params (u and s)
				{
					double log = Math.log(1+(params[i+1]*params[i+1])/(params[i]*params[i]));
					double sigma = Math.sqrt(log);
					double mean = Math.log(params[i])-0.5*log;
					params[i] = mean;
					params[i+1] = sigma;
				}
				maxTime = ((int)Methods.maximum(timePoints))+1;
				steps = (int)(maxTime/dt);
				counts = Methods.computeCytons(params, steps, gens, maxTime)[0]; //solve model
				totalSolved.clear();
				for(int t = 0; t < timePoints.length; t++)
				{
					double tot = 0;
					for(int g = 0; g < gens; g++)
					{
						gensolvedcounts[t][g]=counts[g][(int) (timePoints[t]/dt)];
						tot+=gensolvedcounts[t][g];
					}
					totalSolved.add(tot);
				}
				double minCount = Double.POSITIVE_INFINITY;
				for(int i = 0; i < timePoints.length; i++)
				{
					double m1 = 0;
					for(int g = 0; g < gens; g++)
					{
						m1+=counts[g][(int) (timePoints[i]/dt)];
					}
					if(m1 < minCount)minCount=m1;
				}
				generatedReasonable = true;
			}
			//the fluorescence params are now generated based on the flimits
			double CV = gflimits[0][0]+gen.nextDouble()*(gflimits[0][1]-gflimits[0][0]);
			double ratio= gflimits[1][0]+gen.nextDouble()*(gflimits[1][1]-gflimits[1][0]);
			double background = gflimits[2][0]+gen.nextDouble()*(gflimits[2][1]-gflimits[2][0]);
			double shift = gflimits[3][0]+gen.nextDouble()*(gflimits[3][1]-gflimits[3][0]);
			double[] gcounts = new double[gens];  //generational counts from the LiveMatrix
			for(int t=0; t < timePoints.length; t++)
			{
				fparams[t][0] = ratio;
				fparams[t][1] = CV;
				fparams[t][2] = background;
				fparams[t][3] = shift;
				//lets generate the time point
				ArrayList<ArrayList<Double>> replist = new ArrayList<ArrayList<Double>>();
				ArrayList<Double> und = new ArrayList<Double>();
				ArrayList<Integer> vol = new ArrayList<Integer>();
				data.add(replist); 
				undivided.add(und);
				volumes.add(vol);

				double scale = Math.max(0,1.0+gen.nextGaussian()*snoise);
				for(int g = 0; g < gens; g++)
				{
					gcounts[g]=counts[g][(int)(steps*(timePoints[t]/maxTime))];  //extract counts from cyton solutions
					gcounts[g]=Math.max(0, gcounts[g]*(1.0+gen.nextGaussian()*gnoise))*scale;  //randomize the count
				}
				
				for(int i = 0; i < 1; i++) //replicates
				{				
					ArrayList<Double> raw = new ArrayList<Double>();
					replist.add(raw);
					und.add(6.0); //assume six decades of resolution
					vol.add(1000000); //assume all volumes are 1 ml
					//lets populate the data

					//double total = Methods.sum(gcounts);
					double theSigma = 1;
					for(int g = 0; g < gens; g++)
					{
						double mu;
						mu= Math.log10((Math.pow(10,und.get(i))*Math.pow(ratio,g)+background))+shift;
						double sigma;
						if(g == 0)
						{
							sigma= Math.max(mu*CV,0.00001);
							theSigma = sigma;
						}
						else
							sigma=theSigma;
						double[] generated =  Methods.generateGaussians(mu, sigma, (int)gcounts[g]);    //generate fluorescence values from the gaussian distros
						for(int j = 0; j < generated.length; j++)
						{
							raw.add(Math.max(Math.log10(minimum),Math.min(Math.log10(maximum),generated[j])));
						}
						generated = null;
						System.gc();
					}
				}
			}
			ArrayList<Double> times = new ArrayList<Double>();
			for(int i = 0; i < timePoints.length; i++)
				times.add(timePoints[i]);
			///////////////////////////////////////////////////////////// DONE WITH GENERATION OF HISTOGRAMS////////////////////
			Experiment exp = new Experiment(data,undivided,gens,volumes,100,times,"Generated");				//for counts PC
			System.err.println("Fitting Fluorescence");
			exp.fitFluorescence(1000,flimits);				//fit Fluorescence for PC		
			File tout = new File(outf);
			File outFile = new File(tout.getAbsolutePath()+File.separator+outf);
			tout.mkdirs();
			saveImage(new String(outFile.getAbsolutePath()+"_fl.png"),exp,0);
			//BEST SOLUTION USING COMBINED METHOD
			System.err.println("Fitting PCyton to Histograms: ");
			//NOW FIT USING COMBINED METHOD
		//	exp.fitCyton(gens, limits, dt, true, 100);
			ArrayList<double[] > solutions = null;
			try{
				
				if(outf==null) return;



				BufferedWriter out = new BufferedWriter(new FileWriter(outFile.getAbsolutePath()+".csv", false));
				out.write(outFile+",model used:,fcyton Model,Finished at:,"+Methods.now());out.newLine();
				//we can now also save the counts!!!!!!!!
				saveCounts(out,exp);
				solutions =exp.fitCyton(gens,limits,dt,true,tries);
				if(solutions.size() == 0) return;
				int solution_count= solutions.size();
				double bestScore = solutions.get(0)[1];
				String os = "fitScore,Weighted %Diff,Tdiv0 u,Tdiv0 s,Tdiv1 u,Tdiv1 s,Tdie0 u,Tdie0 s,Tdie1 u,Tdie1 s,D u,D s, F0,Start Count,pmech,kmech,Tdiv0 u<,Tdiv0 s<,Tdiv1 u<,Tdiv1 s<,Tdie0 u<,Tdie0 s<,Tdie1 u<,Tdie1 s<,D u<,D s<, F0<,Start Count<,pmech<,kmech<,Tdiv0 u>,Tdiv0 s>,Tdiv1 u>,Tdiv1 s>,Tdie0 u>,Tdie0 s>,Tdie1 u>,Tdie1 s>,D u>,D s>, F0>,Start Count>,pmech>,kmech>";

				 out.write(os);out.newLine();
				 for(int j = 0; j < solutions.size(); j++)
				 {
				 	System.err.print(" "+j);
				 	String line = ""+solutions.get(j)[0];
					for(int k = 1; k < solutions.get(j).length; k++)
					{
						if(k==1)
							line+=","+Math.sqrt(solutions.get(j)[k]);
						else
							line+=","+solutions.get(j)[k];
					}
					out.write(line); out.newLine();
				}
				 double[] second_best = null;
				 if(solutions.size() > 1) second_best= solutions.get(1);
				 //trim the data according to 1% cutoff along the score
				for(int i = 0; i < solutions.size(); i++)
				{
					if(solutions.get(i)[1] >  bestScore+0.1) //within 0.1%
					{
						solutions.remove(i);
						i--;
					}
				}
				if(second_best != null && solutions.size() ==1)
					solutions.add(second_best);
				//now lets run the clustering by sensitivity maximization agglomeration
				boolean[] mask = new boolean[nparams*3+2];
				for(int j = 2;j <2+nparams; j++)
					mask[j]=true;
	
				if(solutions.size()==0) return;
				System.err.println("Starting clustering with "+solutions.size()+" solutions.");
				Clustering c = heirarchalClusterWithArms(solutions);
				//now lets output the clusters and the average solutions with average sensititivities
				//now lets go through and sample from each cluster solution and order them by their scores
				Clustering sorted_c = new Clustering(new ArrayList<ArrayList<double[]>>(),new ArrayList<double[]>());
				ArrayList<Double> sorted_scores = new ArrayList<Double>();
				for(int i = 0; i < c.averages.size();i++)
				{
					double[] average = c.averages.get(i);
					double[] scores = new double[1000];
					for(int sample = 0; sample < 1000; sample++)
					{
						double[] ps = new double[Constants.nparams];
						for(int j = 0; j < Constants.nparams; j++)
						{
							ps[j] = average[2+j]-average[2+j+Constants.nparams]+Math.random()*(average[2+j+Constants.nparams]+average[2+j+Constants.nparams*2]);
						}
						exp.cytonToFluorescence(params, true);
						scores[sample]=exp.getAreaDiff();
					}
					Arrays.sort(scores);
					double median_score = scores[499];
					c.averages.get(i)[0] = median_score;
					System.err.println("Cluster "+(i+1)+" has a median fit score of "+median_score);
					//insert sort
					int j = 0;
					for(j=0; j < sorted_scores.size(); j++)
					{
						if(sorted_scores.get(j) > median_score)
							break;
					}
					sorted_scores.add(j,median_score);
					sorted_c.averages.add(j,average);
					sorted_c.clusters.add(j,c.clusters.get(i));
				}
				double bestClusterScore = sorted_scores.get(0);
				for(int i = 0; i < sorted_scores.size(); i++)
				{
					System.err.println("Cluster with median score: "+sorted_scores.get(i)+" acceptable = "+(bestClusterScore+1));
					if(sorted_scores.get(i) > bestClusterScore+1)
					{
						sorted_scores.remove(i);
						sorted_c.averages.remove(i);
						sorted_c.clusters.remove(i);
						i--;
					}
				}
				out.write("Total solutions,"+solution_count+",After filtering by score,"+solutions.size()+",Number of non-overlapping solution clusters,"+c.clusters.size()+",detrmined using sensitivity agglomeration clustering,After filtering by average cluster sample score there were,"+sorted_c.clusters.size()+",clusters remaining");out.newLine();
				for(int i = 0; i < sorted_c.clusters.size(); i++)
				{
					System.err.println("Processing cluster "+(i+1));
					out.write("Cluster "+(i+1)+" with "+sorted_c.clusters.get(i).size()+" solutions total,Average sampled cluster score was,"+sorted_scores.get(i));
					out.newLine();
					ArrayList<double[]> cluster;
					cluster = new ArrayList<double[]>();
					double[][]vals = new double[nparams][sorted_c.clusters.get(i).size()];
					for(int j = 0; j < sorted_c.clusters.get(i).size(); j++)
					{
						int k = 0;
						for(k = 0; k < cluster.size(); k++)
						{
							if(cluster.get(k)[0] > sorted_c.clusters.get(i).get(j)[0])
								break;
						}
						cluster.add(k,Methods.duplicate(sorted_c.clusters.get(i).get(j)));
					}
					System.err.println("Finished ordering and filling out vals");
					String[] names = {"Tdiv0 u","Tdiv0 s","Tdiv1 u","Tdiv1 s","Tdie0 u","Tdie0 s","Tdie1 u","Tdie1 s","D u","D s","F0","Start Count","pmech","kmech"};
					out.write(os); out.newLine();
					for(int j = 0; j < cluster.size(); j++)
					{
						System.err.print(" "+j);
						String line = ""+cluster.get(j)[0];
						for(int k = 1; k < cluster.get(j).length; k++)
						{
							if(k==1)
								line+=","+Math.sqrt(cluster.get(j)[k]);
							else
								line+=","+cluster.get(j)[k];
							if(k < nparams+2 && k > 1)
							vals[k-2][j] = cluster.get(j)[k];
						}
						out.write(line); out.newLine();
	
					}
	
					double[] average_params = new double[nparams];			
					for(int k = 0; k < 8; k+=2)
					{
						double second = Math.log(1+Math.pow(sorted_c.averages.get(i)[k+3]/sorted_c.averages.get(i)[k+2],2));
						double u = Math.log(sorted_c.averages.get(i)[k+2])-0.5*second;
						double s = Math.sqrt(second);
						average_params[k]=u;
						average_params[k+1]=s;
					}
					for(int k = 8; k < nparams; k++)
						average_params[k]=sorted_c.averages.get(i)[k+2];
					out.write("Param,Ar. Average,Ar. s.d.,Ar. SEM,Overlap minimum,Lower,Upper"); out.newLine();
					exp.cytonToFluorescence(average_params, true);
					for(int k = 0; k < nparams; k++)
					{
						out.write(names[k]+","+Methods.average(vals[k])+","+Methods.std(vals[k])+","+(Methods.std(vals[k])/Math.sqrt(cluster.size()))+","+sorted_c.averages.get(i)[k+2]+","+sorted_c.averages.get(i)[k+nparams+2]+","+sorted_c.averages.get(i)[k+nparams*2+2]);out.newLine();
					}
					saveImage(new String(outFile.getAbsolutePath()+"_C"+(i+1)+".png"),exp,1);
				}
				String model = "fcyton";
				JOptionPane.showMessageDialog(null, "Finished phenotyping "+exp.name+". Total number of "+model+" fits:"+solution_count+" (Filtered all but "+solutions.size()+"). Total number of non-overlapping good solutions: "+sorted_c.clusters.size());
			
				out.close();
				System.err.println("Finished everything using output file: "+outFile.getName());
			}catch(Exception e)
			{
				e.printStackTrace();
			}
	
	}
	private void saveCounts(BufferedWriter pw, Experiment exp) {
		
		try {
			int max_reps = 0;
			for(int j = 0; j < exp.timePoints.size(); j++)
				max_reps = Math.max(max_reps,exp.timePoints.get(j).getReplicates().size());
		
			pw.write("Generational counts derived from Cell Fluorescence Model fitting for,"+exp.name+",performed on,"+Methods.now()+",timepoints,"+exp.timePoints.size()+",replicates,"+max_reps+",generatons,"+exp.gens);
			pw.newLine();
			pw.write("Time");

			for(int i = 0; i < exp.gens*(max_reps+2); i++)
				pw.write(",Gen "+(i%exp.gens));
			for(int i = 0; i < (max_reps); i++)
				pw.write(",Sum "+(i+1));
			pw.write(",Time,Avg Sum,S.d. Sum");
			pw.newLine();
			for (int i = 0; i < exp.timePoints.size(); i++)
			{
				double[][] vals = new double[exp.gens][exp.timePoints.get(i).getReplicates().size()];
				double sums[] = new double[exp.timePoints.get(i).getReplicates().size()];
				pw.write(""+exp.timePoints.get(i).getTime());
				for(int j = 0; j < exp.timePoints.get(i).getReplicates().size(); j++)
				{
					Run r = exp.timePoints.get(i).getReplicates().get(j);
					double[] counts = r.getConcentrations();
					for(int k =0; k < counts.length; k++)
					{
						vals[k][j] = counts[k];
						sums[j] += counts[k];
						pw.write(","+counts[k]);
					}
				}
				for(int j = 0; j < vals.length; j++)
				{
					pw.write(","+Methods.average(vals[j]));
				}
				for(int j = 0; j < vals.length; j++)
				{
					pw.write(","+Methods.std(vals[j]));
				}
				for(int j = 0; j < sums.length; j++)
					pw.write(","+sums[j]);
				pw.write(","+exp.timePoints.get(i).getTime()+","+Methods.average(sums)+","+Methods.std(sums));
				pw.newLine();
			}
			System.err.println("Finished outputing the counts for experiment "+exp.name);
		} catch (FileNotFoundException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
	
	}


	private void saveImage(String filename, Experiment exp,int mode)
	{
		if(mode==1)
		{
			exp.getCytonScore2();
			
			ArrayList<TimePoint> timePoints = exp.timePoints;
			for(int t = 0; t < timePoints.size(); t++)
			{
				for(int i = 0; i < timePoints.get(t).getReplicates().size(); i++)
				{
					Run r = timePoints.get(t).getReplicates().get(i);
					if( r.lastCorrelation< 10)
						r.updateDraw("Scr:", "%Diff:", new Color(200,255,200,50));
					else if( r.lastCorrelation< 20)
						r.updateDraw("Scr:", "%Diff:", new Color(255,255,200,50));
					else
						r.updateDraw("Scr:", "%Diff:", new Color(255,200,200,50));
				}
			}
		}
		else if (mode==0)
		{
			for(int i = 0; i < exp.timePoints.size(); i++)
				exp.timePoints.get(i).getFluorescenceScore2();
			ArrayList<TimePoint> timePoints = exp.timePoints;
			for(int t = 0; t < timePoints.size(); t++)
			{
				for(int i = 0; i < timePoints.get(t).getReplicates().size(); i++)
				{
					Run r = timePoints.get(t).getReplicates().get(i);
					if( r.lastScore< 10)
						r.updateDraw("Scr:", "%Diff:", new Color(200,255,200,50));
					else if( r.lastScore< 20)
						r.updateDraw("Scr:", "%Diff:", new Color(255,255,200,50));
					else
						r.updateDraw("Scr:", "%Diff:", new Color(255,200,200,50));
				}
			}
		}
		BufferedImage img;
		BufferedImage overlay;
		int max_rep = 1;
		int size =  38;
		int width = (size*3)/2;
		int height = size;
		double maxh = 0;
		double minf = 0;
		double maxf = 0;
		for(int i = 0; i < exp.timePoints.size(); i++)
			if(exp.timePoints.get(i).getReplicates().size() > max_rep) max_rep = exp.timePoints.get(i).getReplicates().size();
		ArrayList<ArrayList<double[]>> hs = new ArrayList<ArrayList<double[]>>();
		maxh = 0;
		minf = 3;
		maxf = 0;
		for(int i = 0; i < exp.timePoints.size(); i++)
		{
			if(exp.timePoints.get(i).getReplicates().size() > max_rep) max_rep = exp.timePoints.get(i).getReplicates().size();
			for(int j = 0; j < exp.timePoints.get(i).getReplicates().size(); j++)
			{	
				if(exp.timePoints.get(i).getReplicates().get(j).undivided+0.5 > maxf)
					maxf = exp.timePoints.get(i).getReplicates().get(j).undivided+0.5;
			}
		}			
		img = new BufferedImage(exp.timePoints.size()*width,max_rep*height,BufferedImage.TYPE_INT_ARGB);
		Graphics2D ig = (Graphics2D) img.getGraphics();
		ig.setColor(Color.BLACK);
		for(int i = 0; i < exp.timePoints.size(); i++)
		{
			//time i
			ArrayList<double[]> rhs = new ArrayList<double[]>();
			for(int j = 0; j < exp.timePoints.get(i).getReplicates().size(); j++)
			{
				//replicate j
				double[] h = new double[width];
				for(int k = 0; k < exp.timePoints.get(i).getReplicates().get(j).data.raw.size(); k++)
				{
					double val = exp.timePoints.get(i).getReplicates().get(j).data.raw.get(k);
					int hindex = (int)(((val-minf)/(maxf-minf))*h.length);
					if(hindex< 0) hindex = 0;
					if(hindex>=h.length) hindex = h.length-1;
					h[hindex]+=(1.0/exp.timePoints.get(i).getReplicates().get(j).volume);
					if(h[hindex]> maxh) maxh = h[hindex];
				}
				rhs.add(h);
			}
			hs.add(rhs);
		}
		for(int i = 0 ; i < exp.timePoints.size(); i++)
		{
			for(int j = 0; j < exp.timePoints.get(i).getReplicates().size(); j++)
			{					
				double[] h = hs.get(i).get(j);
				int top = j*height;
				int left = i*width;
				ig.drawRect(left, top, width, height);
				for(int k = 0; k < h.length; k++)
				{
					int length = (int)(h[k]*height/maxh);
					ig.drawLine(left+k,top+height,left+k,top+height-length);
//					stats[2]+=length;
				}
				//lets draw the start peak
				ig.setColor(Color.RED);
				ig.drawLine(left+(int)(((exp.timePoints.get(i).getReplicates().get(j).undivided-minf)/(maxf-minf))*width),top+height,left+(int)(((exp.timePoints.get(i).getReplicates().get(j).undivided-minf)/(maxf-minf))*width),top);
				//lets label this graph
				ig.setColor(Color.BLACK);
			}
		}
		overlay = new BufferedImage(exp.timePoints.size()*width,max_rep*height,BufferedImage.TYPE_INT_ARGB);
		Graphics2D g3 = (Graphics2D) overlay.getGraphics();
		g3.setColor(new Color(255,255,0,255));
		
		for(int i = 0; i < exp.timePoints.size(); i++)
		{
			//time i
			for(int j = 0; j < exp.timePoints.get(i).getReplicates().size(); j++)
			{
				//replicate j
				//lets draw the overlay
				int top = j*height;
				int left = (int) (i*width);
				try{
					//color by quality of fit
					g3.setColor(exp.timePoints.get(i).getReplicates().get(j).bcolor);
					g3.fillRect(left, top, width, height);
					
					g3.setColor(Color.black);
					double[] h = new double[width];
					double[][] gaussians = new double[exp.timePoints.get(i).getReplicates().get(j).dweights.length][width];
					double[] weights = new double[exp.timePoints.get(i).getReplicates().get(j).dweights.length];
					for(int w = 0; w < weights.length; w++)
						weights[w]=exp.timePoints.get(i).getReplicates().get(j).dweights[w];
					//System.err.println("undivided peak at "+exp.undivided.get(i).get(j)+" min/max = "+min+"/"+max);
					//lets draw the labels+scores
					if(size >= 100)
					{
						g3.setColor(Color.BLACK);
						g3.setFont(new Font("Calibri",Font.PLAIN,12));
						g3.drawString((exp.name)+" rep:"+(j+1)+" T: "+exp.timePoints.get(i).getTime()+" hrs.",left+1, top+8);
					}
					else
					{
						g3.setColor(Color.BLACK);
						g3.setFont(new Font("Calibri",Font.PLAIN,10));
						if(exp.timePoints.get(i).getReplicates().size() > 1)
							g3.drawString(exp.timePoints.get(i).getTime()+" hrs "+(j+1),left+1, top+8);
						else
							g3.drawString(exp.timePoints.get(i).getTime()+" hrs",left+1, top+8);
					}
					if(size >= 200)
					{
						g3.setColor(Color.BLUE);
						g3.setFont(new Font("Calibri",Font.PLAIN,12));
						g3.drawString(String.format("%s %4.3f %s %4.3f",exp.timePoints.get(i).getReplicates().get(j).scoreLbl,exp.timePoints.get(i).getReplicates().get(j).drawScore,exp.timePoints.get(i).getReplicates().get(j).corLbl,exp.timePoints.get(i).getReplicates().get(j).drawCor),left+1, top+20);
					}
					//draw the colored background to indicate goodness of fit
					double total_model = 0;
					double total = 0;
					double[] eh = hs.get(i).get(j);
					for(int step = 0; step < h.length; step++)
					{
						double theSigma = 1;
						total+=((double)eh[step]*height)/maxh;
						for(int gen = 0; gen <weights.length; gen++)
						{
							//for each generation
							double mu;
							mu= Math.log10((Math.pow(10,exp.timePoints.get(i).getReplicates().get(j).undivided)*Math.pow(exp.timePoints.get(i).ratio,gen)+exp.timePoints.get(i).background))+exp.timePoints.get(i).getReplicates().get(j).shift;
							double sigma;
							if(gen == 0)
							{
								sigma= Math.max(exp.timePoints.get(i).CV*mu,0.000001);
								theSigma = sigma;
							}
							else
							{
								sigma = theSigma;
							}
							double agauss = weights[gen]*(Methods.getGaussianCDF(mu, sigma, ((step+1)*(maxf-minf))/((double)width)+minf)-Methods.getGaussianCDF(mu, sigma, ((step)*(maxf-minf))/((double)width)+minf));  //area under the gaussian in this interval adjusted for weights
							gaussians[gen][step]= agauss;  
							h[step]+= gaussians[gen][step];
						}
						total_model+=h[step]*height;
					}
					int oldval = height;
					int[] oldvals = new int[gaussians.length];
					for(int var = 0; var < oldvals.length; var++)
						oldvals[var] = height;
					int sumSteps = 0;
					Color lightBlue = new Color(100,100,255,200);

					for(int step = 0; step < h.length; step++)
					{
						//draw
						g3.setStroke(new BasicStroke());
						if(size >= 30) //draw the blues
						{
							g3.setColor(lightBlue);
							for(int gen = 0; gen < gaussians.length; gen++)
							{
								int val2 = (int)(Math.floor((gaussians[gen][step]*total*height*exp.timePoints.get(i).getReplicates().get(j).getdTotalsRatio()) / (total_model) +0.5));  //draw each gaussian as well adjusted as for the total (see above)
								val2 = Math.min(val2,height);
								//System.err.println(h[step]);
								if(oldvals[gen] <= height && val2 <= height && val2 > 0 )
									g3.drawLine(left+step-1, top+height-oldvals[gen], left+step, top+height-val2);
								oldvals[gen] = val2;
							}
						}
						int val = (int)(Math.floor((h[step]*total*height*exp.timePoints.get(i).getReplicates().get(j).getdTotalsRatio()) / (total_model) +0.5));  //value from model adjuted for total experimental count, display size, total_model value and the maximum height of the histogram
						//System.err.println(h[step]);
						
						sumSteps+=val;
						if(size > 80)g3.setStroke(new BasicStroke(1.5f, BasicStroke.CAP_ROUND,BasicStroke.JOIN_ROUND));
						g3.setColor(new Color(100,255,100,200));
						val = Math.min(val,height);
						if(val <height && val > 0)
							g3.drawLine(left+step-1, top+height-oldval, left+step, top+height-val);
						oldval = val;

					}
					if(size >30)g3.setStroke(new BasicStroke(1f, BasicStroke.CAP_ROUND,BasicStroke.JOIN_ROUND));

					g3.setColor(Color.BLACK);
				}
				catch(Exception e)
				{
					
				}
			}
		}				
		if(filename != null)
		{
			//we need to save what we just drew to the file
			BufferedImage saveImage = new BufferedImage(img.getWidth(),img.getHeight(),BufferedImage.TYPE_INT_ARGB);
			saveImage.getGraphics().drawImage(img,0,0,null);
			saveImage.getGraphics().drawImage(overlay,0,0,null);
			try {
				File f = new File(filename);
				int fnum = 2;
				while(f.exists())
				{
					f = new File(filename.substring(filename.length()-4)+fnum+".png");
					fnum++;
				}
					
				ImageIO.write(saveImage, "png", new File(filename));
			} catch (IOException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}
			filename = null;
		}

	}

	public static Clustering heirarchalClusterWithArms(
			ArrayList<double[]> solutions) {
		//agglomerate until k
		ArrayList<ArrayList<double[]>> clustering = new ArrayList<ArrayList<double[]>>();
		ArrayList<double[]> averages = new ArrayList<double[]>();  //average values for each cluster
		for(int j = 0; j < solutions.size(); j++)  //one solution per cluster to start
		{
			clustering.add(new ArrayList<double[]>());
			clustering.get(clustering.size()-1).add(solutions.get(j));
			averages.add(Methods.duplicate(solutions.get(j)));  
		}

		for(int i = 0; i < solutions.size()-1; i++)
		{
			System.err.println(""+((i*100)/(solutions.size())+"% done."));

			//find the two closest clusters using an appropriate distance measure
			int best_a, best_b;
			best_a=best_b = 0;
			double bestDistance = Double.NEGATIVE_INFINITY;
			for(int j = 0; j < clustering.size();j ++)	//cluster j
			{
				for(int k = j+1; k < clustering.size(); k++ )  //cluster k
				{
					double dist  = 0;
					for(int l = 2; l < nparams+2; l++)  //parameters
					{
						double d = (Math.min(averages.get(k)[l]+averages.get(k)[l+nparams*2],averages.get(j)[l]+averages.get(j)[l+nparams*2])-Math.max(averages.get(k)[l]-averages.get(k)[l+nparams],averages.get(j)[l]-averages.get(j)[l+nparams]))/(Math.abs(averages.get(k)[l]+averages.get(k)[l+nparams*2])+Math.abs(averages.get(j)[l]+averages.get(j)[l+nparams*2]));
						if(Double.isNaN(d))d=0;
						if(d < 0 )d = Double.NEGATIVE_INFINITY;			
						dist+=d;  //distance gets the distance between these two clusterings along the Lth parameter 
					}
					if(dist > bestDistance)  //we wanted to overlap
					{
						best_a=j;  //should combine clustering j and clustering k
						best_b=k;
						bestDistance=dist;
					}
				}
			}
			if(bestDistance >0)
			{
				//combine best_a and best_b
				ArrayList<double[]> nc = new ArrayList<double[]>();  //new cluster
				nc.addAll(clustering.get(best_a));	//add all from besta
				nc.addAll(clustering.get(best_b));   //add all from bestb
				double[] newA = new double[2+nparams*3];  //update average (should reflect
				for(int j = 2; j < nparams+2;j++)
				{
					double[] A = averages.get(best_a);
					double[] B = averages.get(best_b);
					if(A[j] < B[j])
					{
						double x = B[j+nparams];
						double y = A[j+nparams*2];
						double z = B[j]-A[j];
						double xint = (y*z)/(y+x);
						if(Double.isNaN(xint))xint = 0;
						newA[j]= A[j]+xint;
						newA[j+nparams] = newA[j]-Math.max(A[j]-A[j+nparams],B[j]-B[j+nparams]);
						newA[j+nparams*2] = Math.min(A[j]+A[j+nparams*2], B[j]+B[j+nparams*2])-newA[j];

					}
					else
					{
						double x = A[j+nparams];
						double y = B[j+nparams*2];
						double z = A[j]-B[j];
						double xint = (y*z)/(y+x);
						if(Double.isNaN(xint))xint = 0;
						newA[j]= B[j]+xint;
						newA[j+nparams] = newA[j]-Math.max(A[j]-A[j+nparams],B[j]-B[j+nparams]);
						newA[j+nparams*2] = Math.min(A[j]+A[j+nparams*2], B[j]+B[j+nparams*2])-newA[j];
					}
				}
				clustering.remove(best_b);
				clustering.remove(best_a);
				averages.remove(best_b);
				averages.remove(best_a);
				clustering.add(nc);
				averages.add(newA);
			}
		}//main loop
		System.err.println("Finished clustering...");
		return new Clustering(clustering,averages);
	}

	
	public static void main(String[] args)
	{
		new GeneratedFitting(args[0]);
	}
}

