/*
Copyright (c)  2013   The Regents of the University of California
All Rights Reserved
 
Permission to copy, modify and distribute any part of this FlowMax for educational, research and non-profit purposes, without fee, and without a written agreement is hereby granted, provided that the above copyright notice, this paragraph and the following three paragraphs appear in all copies.
Those desiring to incorporate this FlowMax into commercial products or use for commercial purposes should contact the Technology Transfer Office, University of California, San Diego, 9500 Gilman Drive, Mail Code 0910, La Jolla, CA 92093-0910, Ph: (858) 534-5815, FAX: (858) 534-7345, E-MAIL:invent@ucsd.edu.
IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES, INCLUDING LOST PROFITS, ARISING OUT OF THE USE OF THIS FlowMax, EVEN IF THE UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
THE FlowMax PROVIDED HEREIN IS ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.  THE UNIVERSITY OF CALIFORNIA MAKES NO REPRESENTATIONS AND EXTENDS NO WARRANTIES OF ANY KIND, EITHER IMPLIED OR EXPRESS, 
INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE, OR THAT THE USE OF THE FlowMax WILL NOT INFRINGE ANY PATENT, TRADEMARK OR OTHER RIGHTS.
*/
package FlowMax;

import java.awt.BasicStroke;
import java.awt.Color;
import java.awt.Font;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.event.MouseWheelEvent;
import java.awt.event.MouseWheelListener;
import java.awt.image.BufferedImage;
import java.io.File;
import java.io.IOException;
import java.util.ArrayList;

import javax.imageio.ImageIO;
import javax.swing.JPanel;



import common.Methods;

public class ProliferationDisplay extends JPanel{

	private static final long serialVersionUID = 1L;
	private Experiment exp = null;
	private BufferedImage img;
	private BufferedImage overlay;
	private int max_rep = 1;
	private int hsize = 200;
	private boolean redraw_needed = true;
	private boolean redraw_overlay = true;
	private double maxh = 0;
	private double minf = 0;
	private double maxf = 0;
	private boolean drew = true;
	private boolean color = true;
	private BufferedImage pic;
	private double mh = 0;  //max height of histograms... normalized to experiment max, or set directly by the user.
	
	private ArrayList<ArrayList<double[]>> hs = new ArrayList<ArrayList<double[]>>();
//	ArrayList<ArrayList<double[]>> statistics = new ArrayList<ArrayList<double[]>>();
	
	public void setExperiment(Experiment e)
	{
		exp = e;
		redraw_needed = true;
	}
	/**
	 * In cell counts
	 * @param mh
	 */
	public void setMaxHistogramHeight(double mh)
	{
		this.mh = mh*0.001;
	}
	
	
	public int getHSize() {
		return hsize;
	}



	public void setSize(int size) {
		this.hsize = size;
	}



	public boolean isRedraw_needed() {
		return redraw_needed;
	}



	public void setRedraw_needed(boolean redraw_needed) {
		this.redraw_needed = redraw_needed;
	}



	public boolean isRedraw_overlay() {
		return redraw_overlay;
	}



	public void setRedraw_overlay(boolean redraw_overlay) {
		this.redraw_overlay = redraw_overlay;
	}



	public boolean isDrew() {
		return drew;
	}



	public void setDrew(boolean drew) {
		this.drew = drew;
	}



	public boolean isColor() {
		return color;
	}



	public void setColor(boolean color) {
		this.color = color;
	}



	public Experiment getExp() {
		return exp;
	}



	public ProliferationDisplay()
	{
		addMouseWheelListener(new MouseWheelListener() {
			
			public void mouseWheelMoved(MouseWheelEvent e) {
				// TODO Auto-generated method stub
				hsize+=(e.getWheelRotation()*10);
				if(hsize > 500) hsize=500;
				if(hsize < 50) hsize = 50;
				redraw_needed = true;
				redraw_overlay = true;
				updateUI();
			}
		});
	
		pic = Methods.getImageFromResource("PhenotypingAnalysis.png");
	}
	
	public void draw()
	{
		if(!drew) return;
		else
			updateUI();
	}
	
	public static BufferedImage drawOnGraphics(Experiment exp, int size,boolean color_background, double maxHeight)
	{
		BufferedImage img;
		BufferedImage overlay;
		int max_rep = 1;
		int width = (size*3)/2;
		int height = size;
		double maxh = 0;
		double minf = 0;
		double maxf = 0;
		for(int i = 0; i < exp.timePoints.size(); i++)
			if(exp.timePoints.get(i).getReplicates().size() > max_rep) max_rep = exp.timePoints.get(i).getReplicates().size();
		ArrayList<ArrayList<double[]>> hs = new ArrayList<ArrayList<double[]>>();
		BufferedImage result =  new BufferedImage(exp.timePoints.size()*width,max_rep*height,BufferedImage.TYPE_INT_ARGB);
		Graphics2D g2 = result.createGraphics();
		maxh = 0;
		minf = 2;
		maxf = 0;
		for(int i = 0; i < exp.timePoints.size(); i++)
		{
			if(exp.timePoints.get(i).getReplicates().size() > max_rep) max_rep = exp.timePoints.get(i).getReplicates().size();
			for(int j = 0; j < exp.timePoints.get(i).getReplicates().size(); j++)
			{	
				if(exp.timePoints.get(i).getReplicates().get(j).undivided+0.5 > maxf)
					maxf = exp.timePoints.get(i).getReplicates().get(j).undivided+0.5;
			}
		}			
		img = new BufferedImage(exp.timePoints.size()*width,max_rep*height,BufferedImage.TYPE_INT_ARGB);
		Graphics2D ig = (Graphics2D) img.getGraphics();
		ig.setColor(Color.BLACK);
		for(int i = 0; i < exp.timePoints.size(); i++)
		{
			//time i
			ArrayList<double[]> rhs = new ArrayList<double[]>();
			for(int j = 0; j < exp.timePoints.get(i).getReplicates().size(); j++)
			{
				//replicate j
				double[] h = new double[width];
				for(int k = 0; k < exp.timePoints.get(i).getReplicates().get(j).data.raw.size(); k++)
				{
					double val = exp.timePoints.get(i).getReplicates().get(j).data.raw.get(k);
					int hindex = (int)(((val-minf)/(maxf-minf))*h.length);
					if(hindex< 0) hindex = 0;
					if(hindex>=h.length) hindex = h.length-1;
					h[hindex]++;
				}
				for(int k = 0; k < h.length; k++)
				{
					h[k]*=(1.0/exp.timePoints.get(i).getReplicates().get(j).volume);
					if(h[k]> maxh) maxh = h[k];
				}
				rhs.add(h);
			}
			hs.add(rhs);
		}
		if(maxHeight > 0)
			maxh = maxHeight;
		for(int i = 0 ; i < exp.timePoints.size(); i++)
		{
			for(int j = 0; j < exp.timePoints.get(i).getReplicates().size(); j++)
			{					
				double[] h = hs.get(i).get(j);
				int top = j*height;
				int left = i*width;
				ig.drawRect(left, top, width, height);
				for(int k = 0; k < h.length; k++)
				{
					int length = (int)(h[k]*height/maxh);
					ig.drawLine(left+k,top+height,left+k,top+height-length);
//					stats[2]+=length;
				}
				//lets draw the start peak
				ig.setColor(Color.RED);
				ig.drawLine(left+(int)(((exp.timePoints.get(i).getReplicates().get(j).undivided-minf)/(maxf-minf))*width),top+height,left+(int)(((exp.timePoints.get(i).getReplicates().get(j).undivided-minf)/(maxf-minf))*width),top);
				//lets label this graph
				ig.setColor(Color.BLACK);
			}
		}
		overlay = new BufferedImage(exp.timePoints.size()*width,max_rep*height,BufferedImage.TYPE_INT_ARGB);
		Graphics2D g3 = (Graphics2D) overlay.getGraphics();
		g3.setColor(new Color(255,255,0,255));
		
		for(int i = 0; i < exp.timePoints.size(); i++)
		{
			//time i
			for(int j = 0; j < exp.timePoints.get(i).getReplicates().size(); j++)
			{
				//replicate j
				//lets draw the overlay
				int top = j*height;
				int left = (int) (i*width);
				try{
					if(color_background)
					{
						//color by quality of fit
						g3.setColor(exp.timePoints.get(i).getReplicates().get(j).bcolor);
						g3.fillRect(left, top, width, height);
					}
					g3.setColor(Color.black);
					double[] h = new double[width];
					double[][] gaussians = new double[exp.timePoints.get(i).getReplicates().get(j).dweights.length][width];
					double[] weights = new double[exp.timePoints.get(i).getReplicates().get(j).dweights.length];
					for(int w = 0; w < weights.length; w++)
						weights[w]=exp.timePoints.get(i).getReplicates().get(j).dweights[w];
					//System.err.println("undivided peak at "+exp.undivided.get(i).get(j)+" min/max = "+min+"/"+max);
					//lets draw the labels+scores
					if(size >= 120)
					{
						g3.setColor(Color.BLACK);
						g3.setFont(new Font("Calibri",Font.PLAIN,12));
						g3.drawString((exp.name)+" rep:"+(j+1)+" T: "+exp.timePoints.get(i).getTime()+" hrs.",left+2, top+10);
					}
					else
					{
						g3.setColor(Color.BLACK);
						g3.setFont(new Font("Calibri",Font.PLAIN,10));
						if(exp.timePoints.get(i).getReplicates().size() > 1)
							g3.drawString(exp.timePoints.get(i).getTime()+" h "+(j+1),left+2, top+10);
						else
							g3.drawString(exp.timePoints.get(i).getTime()+" h",left+2, top+10);
					}
					if(size >= 200)
					{
						g3.setColor(Color.BLUE);
						g3.setFont(new Font("Calibri",Font.PLAIN,12));
						g3.drawString(String.format("%s %4.3f %s %4.3f",exp.timePoints.get(i).getReplicates().get(j).scoreLbl,exp.timePoints.get(i).getReplicates().get(j).drawScore,exp.timePoints.get(i).getReplicates().get(j).corLbl,exp.timePoints.get(i).getReplicates().get(j).drawCor),left+2, top+22);
					}
					//draw the colored background to indicate goodness of fit
					double total_model = 0;
					double total = 0;
					double[] eh = hs.get(i).get(j);
					for(int step = 0; step < h.length; step++)
					{
						double theSigma = 1;
						total+=((double)eh[step]*height)/maxh;
						for(int gen = 0; gen <weights.length; gen++)
						{
							//for each generation
							double mu;
							mu= Math.log10((Math.pow(10,exp.timePoints.get(i).getReplicates().get(j).undivided)*Math.pow(exp.timePoints.get(i).ratio,gen)+exp.timePoints.get(i).background))+exp.timePoints.get(i).getReplicates().get(j).shift;
							double sigma;
							if(gen == 0)
							{
								sigma= Math.max(exp.timePoints.get(i).CV*mu,0.000001);
								theSigma = sigma;
							}
							else
							{
								sigma = theSigma;
							}
							double agauss = weights[gen]*(Methods.getGaussianCDF(mu, sigma, ((step+1)*(maxf-minf))/((double)width)+minf)-Methods.getGaussianCDF(mu, sigma, ((step)*(maxf-minf))/((double)width)+minf));  //area under the gaussian in this interval adjusted for weights
							gaussians[gen][step]= agauss;  
							h[step]+= gaussians[gen][step];
						}
						total_model+=h[step]*height;
					}
					int oldval = height;
					int[] oldvals = new int[gaussians.length];
					for(int var = 0; var < oldvals.length; var++)
						oldvals[var] = height;
					int sumSteps = 0;
					Color lightBlue = new Color(100,100,255,200);

					for(int step = 0; step < h.length; step++)
					{
						//draw
						g3.setStroke(new BasicStroke());
						if(size >= 30) //draw the blues
						{
							g3.setColor(lightBlue);
							for(int gen = 0; gen < gaussians.length; gen++)
							{
								int val2 = (int)(Math.floor((gaussians[gen][step]*total*height*exp.timePoints.get(i).getReplicates().get(j).getdTotalsRatio()) / (total_model) +0.5));  //draw each gaussian as well adjusted as for the total (see above)
								val2 = Math.min(val2,height);
								//System.err.println(h[step]);
								if(oldvals[gen] <= height && val2 <= height && val2 > 0 )
									g3.drawLine(left+step-1, top+height-oldvals[gen], left+step, top+height-val2);
								oldvals[gen] = val2;
							}
						}
						int val = (int)(Math.floor((h[step]*total*height*exp.timePoints.get(i).getReplicates().get(j).getdTotalsRatio()) / (total_model) +0.5));  //value from model adjuted for total experimental count, display size, total_model value and the maximum height of the histogram
						//System.err.println(h[step]);
						
						sumSteps+=val;
						if(size > 80)g3.setStroke(new BasicStroke(1.5f, BasicStroke.CAP_ROUND,BasicStroke.JOIN_ROUND));
						g3.setColor(new Color(100,255,100,200));
						val = Math.min(val,height);
						if(val <height && val > 0)
							g3.drawLine(left+step-1, top+height-oldval, left+step, top+height-val);
						oldval = val;

					}
					if(size >30)g3.setStroke(new BasicStroke(1f, BasicStroke.CAP_ROUND,BasicStroke.JOIN_ROUND));

					g3.setColor(Color.BLACK);
				}
				catch(Exception e)
				{
					
				}
			}
		}
		g2.drawImage(img,0,0,null);
		g2.drawImage(overlay,0,0,null);
		return result;
	}
	
	public void paintComponent(Graphics g)
	{
		super.paintComponent(g);
		drew = false;
		Graphics2D g2 = (Graphics2D)g;
		//lets draw the experiment
		if(exp == null)
		{
			g2.drawImage(pic, 0,0,this);
			return;
		}

		if(img == null|| redraw_needed)
		{
			BufferedImage result = drawOnGraphics(exp,hsize,color,mh);
			g2.drawImage(result,0,0,this);
		}	
		drew = true;
	}



	public void setHSize(int value) {
		hsize=value;
		
	}
	
	private void text(Graphics2D g2, String text, int x, int y)
	{
		Color prev = g2.getColor();
		Color shadow = new Color(prev.getRed(),prev.getGreen(),prev.getBlue(),50);
		g2.setColor(shadow);
		g2.drawString(text, x+1, y+1);
		g2.setColor(prev);
		g2.drawString(text, x, y);
	}
}

