/*
Copyright (c)  2013   The Regents of the University of California
All Rights Reserved
 
Permission to copy, modify and distribute any part of this FlowMax for educational, research and non-profit purposes, without fee, and without a written agreement is hereby granted, provided that the above copyright notice, this paragraph and the following three paragraphs appear in all copies.
Those desiring to incorporate this FlowMax into commercial products or use for commercial purposes should contact the Technology Transfer Office, University of California, San Diego, 9500 Gilman Drive, Mail Code 0910, La Jolla, CA 92093-0910, Ph: (858) 534-5815, FAX: (858) 534-7345, E-MAIL:invent@ucsd.edu.
IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES, INCLUDING LOST PROFITS, ARISING OUT OF THE USE OF THIS FlowMax, EVEN IF THE UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
THE FlowMax PROVIDED HEREIN IS ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.  THE UNIVERSITY OF CALIFORNIA MAKES NO REPRESENTATIONS AND EXTENDS NO WARRANTIES OF ANY KIND, EITHER IMPLIED OR EXPRESS, 
INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE, OR THAT THE USE OF THE FlowMax WILL NOT INFRINGE ANY PATENT, TRADEMARK OR OTHER RIGHTS.
*/
package FlowMax;

import java.awt.Color;
import java.util.ArrayList;


import common.Histogram;
import common.Matrix;
import common.Methods; 

public class Run 
{
	TimePoint parent;
	double time; 
	double[] weights;
	double[] dweights;
	double[] bweights;
	double undivided;
	Histogram data;
	double volume = 0;
	double totalCyton = -1;
	double btotalCyton = -1;
	double dtotalCyton = -1;
	double lastScore = 0;
	double lastCorrelation = 0;
	double drawScore = -1;
	double drawCor = -1;
	String scoreLbl = "Sc";
	String corLbl = "Sc";
	Color bcolor;
	double[][] gaussians = null;
	double shift = 0;
	
	
	private double[][] getGaussians()
	{
		if(gaussians== null || gaussians[0].length != data.prob.length)
		{
			gaussians = new double[weights.length][data.prob.length];
			Histogram h = data;
			double min = h.min;
			double max = h.max;
			double[] prob = h.prob;
			double step_size = (max-min)/prob.length;
			//System.err.println("shift of parent is "+parent.shift);
			for(int s = 0; s < prob.length; s++)
			{
				double theSigma = 1;
				for(int g  = 0; g < weights.length; g++)
				{
					double mu;
					mu= Math.log10((Math.pow(10,undivided)*Math.pow(parent.ratio,g)+parent.background))+shift;
					double sigma;
					if(g == 0)
					{
						sigma= Math.max(mu*parent.CV,0.00001);
						theSigma = sigma;
					}
					else
						sigma=theSigma;
					gaussians[g][s]=  Methods.getGaussianCDF(mu,sigma,min+(s)*step_size,min+(s+1)*step_size);//NormalDistQuick.cdf(mu, sigma, min+(s+1)*step_size)-NormalDistQuick.cdf(mu, sigma, min+(s)*step_size);
				}
			}
		}
		return gaussians;
	}
	
	
	
	

	public void backUpWeights()
	{
		bweights= Methods.duplicate(weights);
		btotalCyton = totalCyton;
	}
	
	public void updateDraw(String sStr,String cStr,Color back)
	{
		dweights= Methods.duplicate(weights);
		dtotalCyton = totalCyton;
		drawScore = lastScore;
		drawCor = lastCorrelation;
		scoreLbl = sStr;
		corLbl = cStr;
		bcolor = back;
	}
	
	public void restoreWeights()
	{
		weights = bweights;
		totalCyton = btotalCyton;
	}
	
	public double getCellsPerMl()
	{
		if(volume==0) return data.raw.size();
		return Math.max(1,data.raw.size()/(volume/1000000.0));
	}
	
	
	public double getTotalsRatio() {
		if(totalCyton == -1) return 1; // we haven't set this yet from the Cyton Calculations
		double ecountperml = getCellsPerMl();
		double mcountperml = totalCyton;
		//return 1;
		return mcountperml/ecountperml;
	}

	public double getdTotalsRatio() {
		if(totalCyton == -1) return 1; // we haven't set this yet from the Cyton Calculations
		double ecountperml = getCellsPerMl();
		double mcountperml = dtotalCyton;
		//return 1;
		return mcountperml/ecountperml;
	}

	public Run clone()
	{
		
		double[] ws = new double[weights.length];
		for(int i = 0; i < ws.length; i++)
			ws[i] = weights[i];
		Run clone = new Run(parent,time,undivided,ws,data);
		clone.volume = volume;
		return clone;
		
	}
	
	public Run(TimePoint parent, double time, double undivided, int volume, double[] weights, ArrayList<Double> raw)
	{
		this.parent = parent; this.time = time; this.undivided = undivided;this.weights = weights;this.volume = volume;
		data = new Histogram(raw,parent.getParent().bincount);
	}
	
	public Run(TimePoint parent, double time, double undivided, double[] weights, Histogram histo)
	{
		this.parent = parent; this.time = time; this.undivided = undivided;this.weights = weights;
		data = histo;
	}
	
	

	
	public void computeWeights()
	{
		int gens = parent.getParent().gens;
		Histogram h = data;
		double min = h.min;
		double max = h.max;
		double[] prob = h.prob;
		//double[] model = new double[prob.length];
		double step_size = (max-min)/prob.length;
		double[][] A = new double[gens][gens];
		double[][] B = new double[gens][1];
		double sigma =  Math.max(parent.CV*(Math.log10(Math.max((Math.pow(10,undivided)+parent.background),0.00000000000000000001))+shift),0.00001);
		int nonzero = gens-1;
		for(int n = 0; n < gens; n++)
		{
			double mun= Math.log10(Math.max(0.0000000000000000000001,(Math.pow(10,undivided)+parent.background)))+shift+n*Math.log10(parent.ratio);
			for(int m = 1; m < prob.length; m++)
			{
				double integraln =(Methods.getGaussianCDF(mun, sigma, min+step_size*(m-1), min+step_size*(m)));
				for(int k = 0; k < gens; k++)
				{
					double muk= Math.log10(Math.max(0.00000000000000000001,(Math.pow(10,undivided)+parent.background)))+shift+k*Math.log10(parent.ratio);
					double integralk = (Methods.getGaussianCDF(muk, sigma, min+step_size*(m-1),min+step_size*(m)));
					A[n][k]+=integraln*integralk;
				}
				B[n][0]+=prob[m]*integraln;
			}
			int mun_index = (int) (prob.length*((mun-min)/(max-min)));
			if(prob[mun_index] > 0.01/prob.length)
				nonzero = n;
		}
		if(nonzero==0)nonzero=1;
		double[][] A2 = new double[nonzero+1][nonzero+1];
		double[][] B2 = new double[nonzero+1][1];
		for(int i = 0; i < A2.length; i++)
		{
			for(int j = 0; j < A2[i].length; j++)
			{
				A2[i][j] = A[i][j];
			}
			B2[i][0] = B[i][0];
		}
		//lets turn into matrices and find the inverse
		Matrix Ma = new Matrix(A2);
		Matrix Bn = new Matrix(B2);
		Matrix C  = null;
		try
		{
			C = Ma.inverse().times(Bn);
		}
		catch(IllegalArgumentException e)
		{
			return;
		}
		double total = 0;
		weights = new double[gens];
		for(int i = 0; i < weights.length; i++)
		{
			if(i < nonzero+1)
				weights[i] = Math.max(C.getElement(i, 0),0);
			else
				weights[i] = 0;
			total+=weights[i];
		}
		if(total==0)
			total=1;
		//correct for negative values etc.			
		for(int i = 0; i < weights.length; i++)
		{
			weights[i]= weights[i]/total;  
		}
		//lets adjust the weights such that we are not modeling 0s.
		
	}
	
	public double getFluorescenceScore()
	{
		if(Math.log10(parent.background) > undivided-1)
			return 99999;
		double[][] gaussians = getGaussians();
		double[] model =new double[gaussians[0].length];
		double counts_ratio = 1;
		double[] prob  = data.prob;
		double score = 0;
		double d = 0;
		for(int i = 0; i < model.length; i++)
		{
			for(int j = 0; j < gaussians.length; j++)
			{
				model[i]+= weights[j]*gaussians[j][i];
			}
			model[i]=model[i]*counts_ratio;
			score+=Math.abs(model[i]-prob[i]);
			d+=Math.abs(model[i]-prob[i]);
		}
		if(Double.isInfinite(score))
		{
			double sum = 0;
			for(int g = 0; g < weights.length; g++)
				sum+=weights[g];
			System.err.println("Infinite score "+sum);
		}
		lastScore = score*100;//(score+Math.abs(getCellsPerMl()-totalCyton)/(getCellsPerMl()+totalCyton))/2;
		lastCorrelation = d*100;
		return score*100;
	}
	
	
	public double getFluorescenceScore2()
	{
		if(Math.log10(parent.background) > undivided-1)
			return 99999;
		gaussians = null;
		double[][] gaussians = getGaussians();
		double[] model =new double[gaussians[0].length];
		double[] prob  = data.prob;
		double score = 0;
		double diff = 0;
		for(int i = 0; i < model.length; i++)
		{
			for(int j = 0; j < gaussians.length; j++)
			{
				model[i]+= weights[j]*gaussians[j][i];
			}
			score+= Math.pow((prob[i]-model[i])*data.raw.size(),2);
			diff+=Math.abs(prob[i]-model[i]);
		}
		if(Double.isInfinite(score))
		{
			double sum = 0;
			for(int g = 0; g < weights.length; g++)
				sum+=weights[g];
			System.err.println("Infinite score "+sum);
		}
		lastScore = 100*diff;//(score+Math.abs(getCellsPerMl()-totalCyton)/(getCellsPerMl()+totalCyton))/2;
		lastCorrelation = Methods.getPearsonCorrelation(model, prob);
		return Math.sqrt(score/model.length)+data.raw.size()*(1-lastCorrelation);
		
	}
	
	/**
	 * Calculates the number of cells based on the distribution and fluorescences of the raw cells. I.e. classify each cell by division given the likelihood from the distributions
	 * @return
	 */
	public double[] getConcentrations() 
	{
		int[] counts = getCounts();
		double[] concentrations = new double[counts.length];
		for(int j = 0; j < counts.length; j++)
		{
			concentrations[j] = (1000000.0*counts[j])/volume;
		}
		return concentrations;
	}
	
	/**
	 * Calculates the number of cells based on the distribution and fluorescences of the raw cells. I.e. classify each cell by division given the likelihood from the distributions
	 * @return
	 */
	
	public int[] getCounts() 
	{

		double[] means = new double[weights.length];
		double theSigma = 1;
		for(int g  = 0; g < weights.length; g++)
		{
			double mu;
			mu= Math.log10((Math.pow(10,undivided)*Math.pow(parent.ratio,g)+parent.background))+shift;
			double sigma;
			if(g == 0)
			{
				sigma= Math.max(mu*parent.CV,0.00001);
				theSigma = sigma;
			}
			else
				sigma=theSigma;
			means[g] = mu;
		}
		int[] counts = new int[weights.length];
		double total = 0;
		for(int i = 0; i < weights.length; i++)
		{
			total+=weights[i];
		}
		for(int i = 0; i < weights.length; i++)
		{
			weights[i]/=total;
			counts[i]= (int)(data.raw.size()*weights[i]);
		}
		
		return counts;
		/*

		for(int i = 0; i < data.raw.size(); i++)
		{
			double total_p = 0;
			double[] ps = new double[weights.length];
			for(int  j= 0; j < weights.length; j++)
			{
				ps[j] = weights[j]*Methods.getGaussianPDF(means[j],theSigma,data.raw.get(i));
				total_p+=ps[j];
			}
			if(total_p ==0)total_p=1;
			for(int j = 0; j < weights.length; j++)
			{
				ps[j]/=total_p;
			}
			double rand = Math.random();
			int ind = 0;
			while(rand > 0)
			{
				rand-=ps[ind];
				ind++;
				if(ind == ps.length)
					break;
			}
			counts[ind-1]++;
		}
		return counts;
		*/
	}
	
	
	/*
	public int[] getCounts() 
	{
		double[] means = new double[weights.length];
		double theSigma = 1;
		for(int g  = 0; g < weights.length; g++)
		{
			double mu;
			mu= Math.log10((Math.pow(10,undivided)*Math.pow(parent.ratio,g)+parent.background))+shift;
			double sigma;
			if(g == 0)
			{
				sigma= Math.max(mu*parent.CV,0.00001);
				theSigma = sigma;
			}
			else
				sigma=theSigma;
			means[g] = mu;
		}
		int[] counts = new int[weights.length];
		for(int i = 0; i < data.raw.size(); i++)
		{
			int maxg = 0;
			double max_p = 0;
			for(int  j= 0; j < weights.length; j++)
			{
				double p = weights[j]*Methods.getGaussianPDF(means[j],theSigma,data.raw.get(i));
				if(p > max_p)
				{
					max_p = p;
					maxg = j;
				}
			}
			counts[maxg]++;
		}
		return counts;
	}
	*/
	/**
	 * Calculates the number of cells based on the distribution and fluorescences of the raw cells. I.e. classify each cell by division given the likelihood from the distributions
	 * @return
	 */
	/*
	public int[] getCounts() 
	{
		double[] means = new double[weights.length];
		double theSigma = 1;
		for(int g  = 0; g < weights.length; g++)
		{
			double mu;
			mu= Math.log10((Math.pow(10,undivided)*Math.pow(parent.ratio,g)+parent.background))+shift;
			double sigma;
			if(g == 0)
			{
				sigma= Math.max(mu*parent.CV,0.00001);
				theSigma = sigma;
			}
			else
				sigma=theSigma;
			means[g] = mu;
		}
		int[] counts = new int[weights.length];
	
		for(int g = 0; g < weights.length; g++)
		{
			counts[g]=(int)Math.round(weights[g]*data.raw.size());
		}
		return counts;
		*/
		/*
				double[] counts = new double[weights.length];
		for(double f : data.raw)
		{
			double[] pdfs = new double[weights.length];
			double total = 0;
			for(int g = 0; g < weights.length; g++)
			{
				double prob=weights[g]*Methods.getGaussianPDF(means[g], theSigma, f);
				total+=prob;
				pdfs[g]+=prob;
			}
			for(int g = 0; g < weights.length; g++)
			{
				pdfs[g]/=total;
				counts[g]+=pdfs[g];
			}
		}
		int[] result = new int[weights.length];
		for(int g = 0; g < weights.length; g++)
			result[g]=(int)Math.round(counts[g]);
		return result;
			}
		*/

	
	/**
	 * 
	 * @return % Area Difference adjusted for start count (i.e. this can be greater than 100%)
	 */
	public double getCytonScore()
	{
		double[][] gaussians = getGaussians();
		double[] prob  = data.prob;
		double[] model =new double[prob.length];
		double counts_ratio = getTotalsRatio();
		double diff = 0;
		double d = 0;
		for(int i = 0; i < model.length; i++)
		{
		//	String gstring = "";
			for(int j = 0; j < gaussians.length; j++)
			{
				model[i]+= weights[j]*gaussians[j][i];
			//	gstring+="\t"+weights[j]*gaussians[j][i];
			}
			model[i] = model[i]*counts_ratio;
			diff+=Math.pow(prob[i]-model[i],2);
			d+=Math.abs(prob[i]-model[i]);
		//	System.err.println(i+"\t"+model[i]+"\t"+prob[i]+gstring);
		}
//		System.err.println(diff);
//		System.exit(0);
		double score =diff;//(2*diff+(1-Methods.getPearsonCorrelation(model, prob)))/3;		
		if(Double.isInfinite(score))
		{
			double sum = 0;
			for(int g = 0; g < weights.length; g++)
				sum+=weights[g];
			System.err.println("Infinite score "+sum);
		}
		else
		{

		//	score = 100*Math.pow(score,1.0/2.0)/(model.length*totE);  //percent of cells difference in terms of distance
			//double counts_score = Math.abs((getCellsPerMl()-totalCyton)/(getCellsPerMl()+totalCyton));
			//score = (score*10+counts_score)/11;
			//lastCorrelation = counts_score;
			lastScore = 100*score;
			lastCorrelation = 50*d; //divide by 2 to normalize percentages
			return 100*diff;
		}
		return 10;
	}
	public double getAreaDiff() //this should capture both the distribution and the counts
	{
		double[][] gaussians = getGaussians();
		double[] prob  = data.prob;
		double[] model =new double[prob.length];
		double counts_ratio = getTotalsRatio();
		double diff = 0;
		for(int i = 0; i < model.length; i++)
		{
			for(int j = 0; j < gaussians.length; j++)
			{
				model[i]+= weights[j]*gaussians[j][i];
			}
			model[i] = model[i]*counts_ratio;
			diff+=Math.abs(prob[i]-model[i]);
		}
		return diff*50;  //percent combined area difference since both add up to 1 we just divide by 2 to normalize the percentage
	}
	
	public double getTrueAreaDiff() //this should capture both the distribution and the counts
	{
		double[][] gaussians = getGaussians();
		double[] prob  = data.prob;
		double[] model =new double[prob.length];
		double counts_ratio = getTotalsRatio();
		double diff = 0;
		for(int i = 0; i < model.length; i++)
		{
			for(int j = 0; j < gaussians.length; j++)
			{
				model[i]+= weights[j]*gaussians[j][i];
			}
			model[i] = model[i]*counts_ratio;
			diff+=Math.abs(prob[i]-model[i]);
		}
		return diff;  //percent combined area difference since both add up to 1 we just divide by 2 to normalize the percentage
	}
	
	public double getSquaredDeviation() //this should capture both the distribution and the counts
	{
		double[][] gaussians = getGaussians();
		double[] prob  = data.prob;
		double[] model =new double[prob.length];
		double counts_ratio = getTotalsRatio();
		double diff = 0;
		for(int i = 0; i < model.length; i++)
		{
			for(int j = 0; j < gaussians.length; j++)
			{
				model[i]+= weights[j]*gaussians[j][i];
			}
			model[i] = model[i]*counts_ratio;
			diff+=Math.pow(prob[i]-model[i],2);
		}
		return diff;  //percent combined area difference since both add up to 1 we just divide by 2 to normalize the percentage
	}


	public double getCytonScore2() //this should capture both the distribution and the counts
	{
		double[][] gaussians = getGaussians();
		double[] prob  = data.prob;
		double[] model =new double[prob.length];
		double counts_ratio = getTotalsRatio();
		double diff = 0;
		double d = 0;
		for(int i = 0; i < model.length; i++)
		{
			for(int j = 0; j < gaussians.length; j++)
			{
				model[i]+= weights[j]*gaussians[j][i];
			}
			model[i] = model[i]*counts_ratio;
			diff+=Math.pow((prob[i]-model[i]),2);
			d+=Math.abs(prob[i]-model[i]);
		}
		double score =Math.sqrt(diff)/(Math.pow(Methods.getPearsonCorrelation(model, prob),2));
		if(data.raw.size() <100)
			score = diff;
		if(Double.isInfinite(score))
		{
			double sum = 0;
			for(int g = 0; g < weights.length; g++)
				sum+=weights[g];
			System.err.println("Infinite score "+sum);
			return 100000;
		}
		else
		{
			lastScore = score;
			lastCorrelation = 50*d; //normalized 
			return score;
		}
	}
	
	public double getCytonScoreWeightedRMSD() //this should capture both the distribution and the counts
	{
		double[][] gaussians = getGaussians();
		double[] prob  = data.prob;
		double[] model =new double[prob.length];
		double counts_ratio = getTotalsRatio();
		double diff = 0;
		double d = 0;
		for(int i = 0; i < model.length; i++)
		{
			for(int j = 0; j < gaussians.length; j++)
			{
				model[i]+= weights[j]*gaussians[j][i];
			}
			model[i] = model[i]*counts_ratio;
			diff+=Math.pow((prob[i]-model[i]),2);
			d+=Math.abs(prob[i]-model[i]);
		}
		double score =Math.sqrt(diff);
		if(data.raw.size() <100)
			score = diff;
		if(Double.isInfinite(score))
		{
			double sum = 0;
			for(int g = 0; g < weights.length; g++)
				sum+=weights[g];
			System.err.println("Infinite score "+sum);
			return 100000;
		}
		else
		{
			lastScore = score;
			lastCorrelation = 50*d; //normalized 
			return score;
		}
	}
	
	
	

	public double getCorrelation()
	{

		double[][] gaussians = getGaussians();
		double[] model =new double[gaussians[0].length];
		double counts_ratio = getTotalsRatio();
		double[] prob  = data.prob;
		for(int i = 0; i < model.length; i++)
		{
			for(int j = 0; j < gaussians.length; j++)
			{
				model[i]+= weights[j]*gaussians[j][i];
			}
			model[i]=model[i]*counts_ratio;
		}
		//now lets compute score
		double score = 1-Methods.getPearsonCorrelation(model, prob);
		lastCorrelation = score;
		return score;
	}
	public double getPValue()
	{
		Histogram h = data;
		double oldmin = h.min;
		double oldmax = h.max;
		int oldbins = h.prob.length;
		data.removeEmptySpace();
//		double width= (parent.CV*Math.log10((Math.pow(10,undivided)-parent.background))+parent.shift)/5;
//		h.rebin(width);

		double best2 = Double.POSITIVE_INFINITY;
		int bbsize = 1;
		int a = 1;
		int c = 2000;
		for(int step = 100; step >= 1; step=step/10)
		{
			//System.err.println("Step = "+step+" min "+a+" max "+c);
			for(int b = a; b <=c; b+=step)
			{
				if(b > 0)
					h.rebin(b);
				double min = h.min;
				double max = h.max;
				double[] prob = h.prob;
				double[] model = new double[prob.length];
				double step_size = (max-min)/prob.length;
	
				for(int s = 0; s < model.length; s++)
				{
					double theSigma = 1;
					for(int g  = 0; g < weights.length; g++)
					{
						double mu;
						mu= Math.log10((Math.pow(10,undivided)*Math.pow(parent.ratio,g)+parent.background))+shift;
						double sigma;
						if(g == 0)
						{
							sigma= Math.max(mu*parent.CV,0.00001);
							theSigma = sigma;
						}
						else
							sigma=theSigma;
						model[s]+=weights[g]*Methods.getGaussianCDF(mu, sigma, min+(s-1)*step_size, min+(s)*step_size);
					}
					//model[s]=model[s]*getTotalsRatio();
				}
				double D = 0;
				double cdf_model = 0;
				double cdf_expt = 0;
			//	double ratio = getTotalsRatio();
				for(int i = 0; i < model.length; i++)
				{
					cdf_model+=model[i];
					cdf_expt+=prob[i];
					double d = Math.abs(cdf_model-cdf_expt);
					if(d > D)
					D=d;
				}
	
				double pvalue = Methods.KolmogorovCDF(prob.length,D);

				//System.err.println("Fit D = "+D+" pvalue = "+pvalue);
				
				if(pvalue < best2)
				{
					best2 = pvalue;
					bbsize = b;
					//System.err.println("Best P value: "+best2+" bbsize "+b+" D : "+D);
				}
			}
			a = bbsize-step;
			c = a+2*step;
		}
		//System.err.println("best P value : "+best2);
		h.min = oldmin;
		h.max = oldmax;
		h.rebin(oldbins);
		return best2;
	}
	
	public double getD()
	{
		double[][] gaussians = getGaussians();
		double[] model =new double[gaussians[0].length];
		double counts_ratio = getTotalsRatio();
		double[] prob  = data.prob;
		for(int i = 0; i < model.length; i++)
		{
			for(int j = 0; j < gaussians.length; j++)
			{
				model[i]+= weights[j]*gaussians[j][i];
			}
			model[i]=model[i]*counts_ratio;
		}
		double D = 0;
		double cdf_model = 0;

		double cdf_expt = 0;
		for(int i = 0; i < model.length; i++)
		{
			cdf_model+=model[i];
			cdf_expt+=prob[i];
			double d = Math.pow(cdf_model-cdf_expt,2);
			D+=d;
		}
		lastScore=Math.sqrt(D/model.length);
		return lastScore;
	}
	

}
