/*
Copyright (c)  2013   The Regents of the University of California
All Rights Reserved
 
Permission to copy, modify and distribute any part of this FlowMax for educational, research and non-profit purposes, without fee, and without a written agreement is hereby granted, provided that the above copyright notice, this paragraph and the following three paragraphs appear in all copies.
Those desiring to incorporate this FlowMax into commercial products or use for commercial purposes should contact the Technology Transfer Office, University of California, San Diego, 9500 Gilman Drive, Mail Code 0910, La Jolla, CA 92093-0910, Ph: (858) 534-5815, FAX: (858) 534-7345, E-MAIL:invent@ucsd.edu.
IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES, INCLUDING LOST PROFITS, ARISING OUT OF THE USE OF THIS FlowMax, EVEN IF THE UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
THE FlowMax PROVIDED HEREIN IS ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.  THE UNIVERSITY OF CALIFORNIA MAKES NO REPRESENTATIONS AND EXTENDS NO WARRANTIES OF ANY KIND, EITHER IMPLIED OR EXPRESS, 
INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE, OR THAT THE USE OF THE FlowMax WILL NOT INFRINGE ANY PATENT, TRADEMARK OR OTHER RIGHTS.
*/
package FlowMax;

import java.awt.Color;
import java.util.ArrayList;
import java.util.Random;

import common.Methods;

public class TimePoint
{
	public double CV;   //coefficient of variability or the width of the peaks
	public double background;  //background autofluorescence
	public double ratio;   //distance between peaks
	private double time;   //time after start of the experiment
	private double bestFluorescencePercentage = 100;  //not used
	private Experiment parent;   
	private ArrayList<Run> replicates = new ArrayList<Run>();   //experimental replicates belonging to this time point
	private int totalCount = 0;  //total number of cells? Not sure

	private int gens = 8;
	
	
	public void setParent(Experiment parent) {
		this.parent = parent;
	}

	public double getTime() {
		return time;
	}

	public Experiment getParent() {
		return parent;
	}

	public ArrayList<Run> getReplicates() {
		return replicates;
	}

	public int getTotalCount() {
		return totalCount;
	}

	public TimePoint clone()
	{
		TimePoint clone = new TimePoint(CV,background,ratio,time,parent,null);
		ArrayList<Run> reps = new ArrayList<Run>();
		for(int i = 0; i < replicates.size(); i++)
		{
			Run r = replicates.get(i).clone();
			r.parent = clone;
			reps.add(r);
		}
		clone.replicates = reps;
		return clone;
	}
	
	public TimePoint(double time, ArrayList<Double> undivided, ArrayList<Integer> volumes,ArrayList<ArrayList<Double>> raw, Experiment parent)
	{
		this.time = time; this.parent = parent;
		for(int i = 0; i < undivided.size(); i++)
		{
			double[] weights = new double[parent.gens];
			weights[0] = 1;
			Run r = new Run(this,time,undivided.get(i), volumes.get(i),weights, raw.get(i));
			replicates.add(r);
			totalCount+=r.data.raw.size();
		}
	}
	
	public TimePoint(double CV, double background, double ratio, double time, Experiment parent,ArrayList<Run> replicates)
	{
		this.CV = CV; this.background = background; this.ratio = ratio; this.time = time; this.parent = parent; 
		this.replicates = replicates;
	}

	private void updateListeners(double bratio, double bbackground, double bCV, double score, int cycle, double temperature)
	{
		for(int i = 0; i < parent.getFlisteners().size(); i++)
		{
			parent.getFlisteners().get(i).processFluorescenceUpdate(this,bratio,bbackground,bCV, score, cycle, temperature);
		}
	}

 

	public boolean fitFluorescence2(double[][] flims, double[] starts)
	{
		
		//first lets find the probable numbe of peaks
		Random gen = new Random();

		this.CV = starts[0];
		this.ratio = starts[1];
		this.background= starts[2];
		for(int i = 0; i < replicates.size(); i++)
			replicates.get(i).shift = starts[3];
		boolean changing = true;
		for(int i = 0; i < replicates.size(); i++)
			replicates.get(i).computeWeights();
		double current = this.getFluorescenceScore2();
		double best = current;
		double bCV = CV;
		double bbackground = background;
		double bratio = ratio;
		double[] bshift = new double[replicates.size()];
		for(int i = 0; i < replicates.size(); i++)
			bshift[i] = replicates.get(i).shift;
		int rep = 0;
		//long last_best_time = System.currentTimeMillis();
		int counter = 0;
		double T = best/2;
		double Tmin = T/1000;
		double tpcells = 0;
		for(int i = 0; i < replicates.size(); i++)
		{
			tpcells+=replicates.get(i).getCellsPerMl();
		}
		tpcells = tpcells/replicates.size(); //take the average
		double cCV = bCV;
		double cratio = bratio;
		double cbackground = bbackground;
		double[] cshift = Methods.duplicate(bshift);
		while(changing)
		{
			counter++;
			rep++;
			if(gen.nextInt(100) >5)
			{
				CV = Methods.reflectWithin(cCV+gen.nextGaussian()*cCV*0.01, flims[0][0], flims[0][1]);
			}
			else
				CV = Methods.reflectWithin(cCV+gen.nextGaussian()*cCV, flims[0][0], flims[0][1]);
			if(gen.nextInt(100) > 5)
				ratio = Methods.reflectWithin(cratio+gen.nextGaussian()*cratio*0.01, flims[1][0], flims[1][1]);
			else
				ratio = Methods.reflectWithin(cratio+gen.nextGaussian()*cratio, flims[1][0], flims[1][1]);
			if(gen.nextInt(100) > 5)
				background = Methods.reflectWithin(cbackground*Math.pow(10,-0.1+gen.nextDouble()*0.2)+1, flims[2][0], flims[2][1]);
			else
				background = flims[2][0]+gen.nextDouble()*(flims[2][1]-flims[2][0]);
			if(gen.nextInt(100) > 5)
			{
				for(int i = 0; i < replicates.size(); i++)
					replicates.get(i).shift = Methods.reflectWithin(cshift[i]+gen.nextGaussian()*cshift[i]*0.01+gen.nextGaussian()*0.001, flims[3][0], flims[3][1]);
			}
			else
			{
				for(int i = 0; i < replicates.size(); i++)
					replicates.get(i).shift = Methods.reflectWithin(cshift[i]+gen.nextGaussian()*cshift[i]*0.01+gen.nextGaussian()*0.001, flims[3][0], flims[3][1]);
			}
			for(int i = 0; i < replicates.size(); i++)
				replicates.get(i).computeWeights();
			double score = this.getFluorescenceScore2();
			double diff = (score-best);
			double chance =  Math.exp(-(diff)/(T));
			if(gen.nextDouble() < chance)
			{
				if(score < best)
				{

					best = score;
					bCV = CV;
					bratio = ratio;
					bbackground = background;
					rep = 0;
					for(int i = 0; i < replicates.size(); i++)
						bshift[i] = replicates.get(i).shift;
					updateListeners(bratio, bbackground,bCV,best,counter,T);
					T=Math.min(Tmin*1000,T*1.1);
				}
				cCV = CV;
				cratio = ratio;
				cbackground = background;
				for(int i = 0; i < replicates.size(); i++)
					cshift[i] = replicates.get(i).shift;
			}
			else
			{
				//revert
				CV= cCV; background=cbackground; ratio = cratio;
				for(int i = 0; i < replicates.size(); i++)
					 replicates.get(i).shift = cshift[i];
			}
			T=T*0.99;
			current = score;
			if(T < Tmin) changing =false;
		}

		ratio = bratio;
		CV = bCV;
		background = bbackground;
		for(int i = 0; i < replicates.size(); i++)
			 replicates.get(i).shift =bshift[i];
		for(int i = 0; i < replicates.size(); i++)
			replicates.get(i).computeWeights();
		this.getFluorescenceScore2();
		bestFluorescencePercentage = best;
		updateListeners(bratio, bbackground,bCV,best,counter,T);
		return true;
	}
	
	
	
	
	public double getCorrelation()
	{
		double result = 0;
		for(int i = 0; i < replicates.size(); i++)
		{
			result+= replicates.get(i).getCorrelation();
		}
		return result/replicates.size();
	}		
	public double getFluorescenceScore()
	{
		double result = 0;
		for(int i = 0; i < replicates.size(); i++)
		{
			result+= replicates.get(i).getFluorescenceScore();
		}
		return result/replicates.size();
	}

	public double getFluorescenceScore2()
	{
		double result = 0;
		for(int i = 0; i < replicates.size(); i++)
		{
			result+= replicates.get(i).getFluorescenceScore2();
		}
		return result/replicates.size();
	}



	private double getD()
	{
		double result = 0;
		for(int i = 0; i < replicates.size(); i++)
		{
			result+= replicates.get(i).getD();
		}
		return result/replicates.size();
	}	
	private double getMaxD()  //not used currently
	{
		double result = 0;
		for(int i = 0; i < replicates.size(); i++)
		{
			double d =  replicates.get(i).getD();
			if(d > result)
				result = d;
		}
		return result;
	}
	
	private double getMaxScore()
	{
		double result = 0;
		for(int i = 0; i < replicates.size(); i++)
		{
			double d =  replicates.get(i).getCytonScore();
			if(d > result)
				result = d;
		}
		return result;
	}	
	
	public double getPValue()
	{
		double result = 0;
		for(int i = 0; i < replicates.size(); i++)
		{
			result+= replicates.get(i).getPValue();
		}
		return result/replicates.size();
	}

	
	
}
