/*
Copyright (c)  2013   The Regents of the University of California
All Rights Reserved
 
Permission to copy, modify and distribute any part of this FlowMax for educational, research and non-profit purposes, without fee, and without a written agreement is hereby granted, provided that the above copyright notice, this paragraph and the following three paragraphs appear in all copies.
Those desiring to incorporate this FlowMax into commercial products or use for commercial purposes should contact the Technology Transfer Office, University of California, San Diego, 9500 Gilman Drive, Mail Code 0910, La Jolla, CA 92093-0910, Ph: (858) 534-5815, FAX: (858) 534-7345, E-MAIL:invent@ucsd.edu.
IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES, INCLUDING LOST PROFITS, ARISING OUT OF THE USE OF THIS FlowMax, EVEN IF THE UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
THE FlowMax PROVIDED HEREIN IS ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.  THE UNIVERSITY OF CALIFORNIA MAKES NO REPRESENTATIONS AND EXTENDS NO WARRANTIES OF ANY KIND, EITHER IMPLIED OR EXPRESS, 
INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE, OR THAT THE USE OF THE FlowMax WILL NOT INFRINGE ANY PATENT, TRADEMARK OR OTHER RIGHTS.
*/
package FlowMax;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileWriter;
import java.io.PrintWriter;
import java.util.ArrayList;
import java.util.Random;

import common.Methods;


/**
 * This code was used to measure the average amount of time it takes to perform FlowMax fitting as a function of cell generations and time points on generated realistic CFSE datasets
 * @author Max
 *
 */
public class TimingAnalysis{


	private int steps= 400;  //time steps
	private int maxTime= 200;  //time total
	private static final int nparams = 14;
	private static final int gens = 6;
	private static final double dt = 0.5;
	double[][] limits = {{1.79162052,3.93009221},{0.01667,0.832555},{1.79162052,3.5246271},{0.01667,0.832555},{1.79162052,4.91092146},{0.01667,0.832555},{1.79162052,4.21777428},{0.01667,0.832555},{-3,8},{0.0,10},{0.001,1},{10000,200000},{0,0},{0,0}};
	double[][] hlimits = {{6,72},{0.001,72},{6,48},{0.001,48},{6,192},{0.001,192},{6,96},{0.001,96},{-3,8},{0.001,10},{0.001,1},{10000,200000},{0,0},{0,0}};
	double[][] flimits = {{0.015,0.021},{0.45,0.50},{0,1000},{-0.001,0.001}}; //ratio, CV, Background FL, shift
	public static final int maximum = 16777216;
	public static final int minimum = 1;
	String separator = System.getProperty("line.separator");

	Random gen = new Random();
	
	Fit[] ffl; // [repeats]
	Fit[] fmax; // [trial][repeats]
	
	double pcytonPerfectPvalue = 1;
	double cytonPerfectPvalue = 1;
	
	public TimingAnalysis(int repetitions, int num_trials ,String outFile, int run_mode,int tpoints, int delay_in_mins) {
		if(delay_in_mins > 0)
		{
			while(delay_in_mins > 0)
			{
				System.err.println("Starting in : "+delay_in_mins+" mins");
				delay_in_mins--;
				try {
					Thread.sleep(60000);
				} catch (InterruptedException e) {
					// TODO Auto-generated catch block
					e.printStackTrace();
				}
			}
		}
		double[] timePoints = new double[tpoints];
		double[][] tplist = {{96},{48,120},{48,96,144},{24,60,96,144},{24,48,72,96,144},{24,48,72,96,120,168},{24,48,60,72,96,144,180},{6,20,34,48,72,96,144,180},{12,24,36,48,60,72,96,132,192}, {6,24,36,48,60,72,96,120,144,192}};
		if(tpoints > 10)
		{
			for(int i = 1; i < tpoints; i++)
			{
				double time = i*(200.0/(tpoints));
				time = Math.round(time/dt)/2.0;
				timePoints[i] = time;
				System.err.print(time+" ");
			}
		}
		else
		{
			timePoints = Methods.duplicate(tplist[tpoints-1]);
		}
		maxTime = (int)timePoints[timePoints.length-1]+1;
		steps=(int)(maxTime/dt);
		
		//lets generate fake data and fit it using the fitting procedure at each stage
		try{

		
		ffl = new Fit[num_trials];
		fmax = new Fit[num_trials];
		

		
		for(int repetition = 0; repetition < repetitions; repetition++)
		{
			
			for(int t = 0; t < num_trials; t++)
			{
				ffl[t] = new Fit(gens,tpoints);
				fmax[t] = new Fit(gens,tpoints);
			}
			System.err.println("-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-");
			System.err.println("STARTED REPETITION "+repetition+" at "+Methods.now());
			System.err.println("-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-");
			
			//lets generate the params
			boolean generatedReasonable = false;
			double[] params = new double[nparams];
			double[][] fparams = new double[timePoints.length][4];
			ArrayList<ArrayList<ArrayList<Double>>> data = new ArrayList<ArrayList<ArrayList<Double>>>();
			ArrayList<ArrayList<Double>> undivided= new ArrayList<ArrayList<Double>>();
			ArrayList<ArrayList<Integer>> volumes= new ArrayList<ArrayList<Integer>>();
			ArrayList<ArrayList<ArrayList<Double>>> data2 = new ArrayList<ArrayList<ArrayList<Double>>>();
			double[][] counts = null; //Live Matrix
			double[][] counts2 = null;
			double[] sparams = null;
			ArrayList<Double> totalSolved = new ArrayList<Double>();
			ArrayList<Double> totalSolved2 = new ArrayList<Double>();
			
			double[][] gensolvedcounts = new double[timePoints.length][gens];
			double[][] gensolvedcounts2 = new double[timePoints.length][gens];
			while(!generatedReasonable)
			{
				for(int i = 0; i < params.length; i++)
				{
					if(i== 1 || i==3|| i==5 || i==7)
						params[i] = gen.nextDouble()*params[i-1]; //we impose the restriction that sigma must be lower that or equal to mu since this is our range for our params during fitting!!!!!!!!!!
					else
						params[i] = gen.nextDouble()*(hlimits[i][1]-hlimits[i][0])+hlimits[i][0]; //pick random params from within limits
				}
				sparams = Methods.duplicate(params);  //linear scale so we don't need to convert later
				for(int i = 0; i <8; i+=2)  //convert to log scale params (u and s)
				{
					double log = Math.log(1+(params[i+1]*params[i+1])/(params[i]*params[i]));
					double sigma = Math.sqrt(log);
					double mean = Math.log(params[i])-0.5*log;
					params[i] = mean;
					params[i+1] = sigma;
				}
				maxTime = ((int)Methods.maximum(timePoints))+1;
				steps = (int)(maxTime/dt);
				counts = computeCytons(params, steps, gens, maxTime)[0]; //solve models
				counts2 = computeCytonsOld(params,steps,gens,maxTime)[0];  //solve models
				totalSolved.clear();
				totalSolved2.clear();
				for(int t = 0; t < timePoints.length; t++)
				{
					double tot = 0;
					double tot2 = 0;
					for(int g = 0; g < gens; g++)
					{
						gensolvedcounts[t][g]=counts[g][(int) (timePoints[t]/dt)];
						gensolvedcounts2[t][g]=counts2[g][(int) (timePoints[t]/dt)];
						tot+=gensolvedcounts[t][g];
						tot2+=gensolvedcounts2[t][g];
					}
					totalSolved.add(tot);
					totalSolved2.add(tot2);
				}
				double totalCount = 0;
				double totalCount2 = 0;
				double minCount = Double.POSITIVE_INFINITY;
				double minCount2 = Double.POSITIVE_INFINITY;
				for(int i = 0; i < timePoints.length; i++)
				{
					double m1 = 0;
					double m2 = 0;
					for(int g = 0; g < gens; g++)
					{
						totalCount+=counts[g][(int) (timePoints[i]/dt)];
						totalCount2+=counts2[g][(int) (timePoints[i]/dt)];
						m1+=counts[g][(int) (timePoints[i]/dt)];
						m2+=counts2[g][(int) (timePoints[i]/dt)];
					}
					if(m1 < minCount)minCount=m1;
					if(m2 < minCount2)minCount2=m2;
				}
				if(totalCount < 4500000 && totalCount2 < 4500000 && minCount > 100 && minCount2 > 100) // check if the solutions look good across the time points
				{
					generatedReasonable = true;
				}
				else
				{
					data.clear();
					undivided.clear();
					volumes.clear();
					data2.clear();
				}
			}
			//the fluorescence params are now generated based on the flimits
			double CV = flimits[0][0]+gen.nextDouble()*(flimits[0][1]-flimits[0][0]);
			double ratio= flimits[1][0]+gen.nextDouble()*(flimits[1][1]-flimits[1][0]);
			double background = flimits[2][0]+gen.nextDouble()*(flimits[2][1]-flimits[2][0]);
			double shift = flimits[3][0]+gen.nextDouble()*(flimits[3][1]-flimits[3][0]);
			double[] gcounts = new double[gens];  //generational counts from the LiveMatrix
			double[] gcounts2 = new double[gens];
			for(int t=0; t < timePoints.length; t++)
			{
				fparams[t][0] = ratio;
				fparams[t][1] = CV;
				fparams[t][2] = background;
				fparams[t][3] = shift;
				//lets generate the time point
				ArrayList<ArrayList<Double>> replist = new ArrayList<ArrayList<Double>>();
				ArrayList<ArrayList<Double>> replist2 = new ArrayList<ArrayList<Double>>();
				ArrayList<Double> und = new ArrayList<Double>();
				ArrayList<Integer> vol = new ArrayList<Integer>();
				data.add(replist); 
				data2.add(replist2);
				undivided.add(und);
				volumes.add(vol);

				for(int g = 0; g < gens; g++)
				{
					gcounts[g]=counts[g][(int)(steps*(timePoints[t]/maxTime))];  //extract counts from cyton solutions
					gcounts2[g]=counts2[g][(int)(steps*(timePoints[t]/maxTime))];  
				}
				for(int i = 0; i < 1; i++) //replicates
				{				
					ArrayList<Double> raw = new ArrayList<Double>();
					replist.add(raw);
					ArrayList<Double> raw2 = new ArrayList<Double>();
					replist2.add(raw2);
					und.add(6.0); //assume six decades of resolution
					vol.add(1000000); //assume all volumes are 1 ml
					//lets populate the data

					//double total = Methods.sum(gcounts);
					double theSigma = 1;
					for(int g = 0; g < gens; g++)
					{
						double mu;
						mu= Math.log10((Math.pow(10,und.get(i))*Math.pow(ratio,g)+background))+shift;
						double sigma;
						if(g == 0)
						{
							sigma= Math.max(mu*CV,0.00001);
							theSigma = sigma;
						}
						else
							sigma=theSigma;
						double[] generated =  Methods.generateGaussians(mu, sigma, (int)gcounts[g]);    //generate fluorescence values from the gaussian distros
						double[] generated2 = Methods.generateGaussians(mu, sigma, (int)gcounts2[g]);
						for(int j = 0; j < generated.length; j++)
						{
							raw.add(Math.max(Math.log10(minimum),Math.min(Math.log10(maximum),generated[j])));
						}
						for(int j = 0; j < generated2.length; j++)
						{
							raw2.add(Math.max(Math.log10(minimum),Math.min(Math.log10(maximum),generated2[j])));
						}
						generated = null;
						generated2 = null;
						System.gc();
					}
				}
			}
			ArrayList<Double> times = new ArrayList<Double>();
			for(int i = 0; i < timePoints.length; i++)
				times.add(timePoints[i]);
			///////////////////////////////////////////////////////////// DONE WITH GENERATION OF HISTOGRAMS////////////////////
			double[][] fl = new double[timePoints.length][4];
			
			Experiment exp = new Experiment(data,undivided,gens,volumes,100,times,"Generated");				//for counts PC
			System.err.println("Fitting Fluorescence for PCyton Data: "+(repetition)+"/"+repetitions);
			long start_time = System.currentTimeMillis();
			exp.fitFluorescence(1000,flimits);				//fit Fluorescence for PC
			ffl[0].elapsed_time = System.currentTimeMillis()-start_time;
			Experiment best_fl = exp;
			ffl[0].pvalue = best_fl.fluorescencePValue;
			for(int t = 0; t < best_fl.timePoints.size(); t++)
			{
				ffl[0].counts[t]=best_fl.timePoints.get(t).getReplicates().get(0).getCounts();
			}
			for(int i = 1; i < num_trials; i++)
			{
				Experiment exp2 = new Experiment(data,undivided,gens,volumes,100,times,"Generated");				//for counts PC 
				if(exp2.fluorescencePValue < best_fl.fluorescencePValue)
				{
					best_fl = exp2;  //PC
					ffl[i].pvalue = best_fl.fluorescencePValue;
					for(int t = 0; t < best_fl.timePoints.size(); t++)
					{
						ffl[i].counts[t]=best_fl.timePoints.get(t).getReplicates().get(0).getCounts();
					}
				}
				else
				{
					ffl[i]= ffl[i-1].clone();
				}
			}

	
			if(run_mode != 2)
			{
				//////////////////////////////////////////////////////// DONE WITH FLUORESCENCE FITTING///////////////////////////////////
				//we only use the first fl since we only use the first one in the real workflow
				for(int i = 0; i < exp.timePoints.size(); i++)
				{
					//fl and fl2 represent the best fit fluorescence parameters from both models and only the first fit attempt
					fl[i][0] = exp.timePoints.get(i).CV;
					fl[i][1] = exp.timePoints.get(i).ratio;
					fl[i][2] = exp.timePoints.get(i).background;
					fl[i][3] = exp.timePoints.get(i).getReplicates().get(0).shift;
				}
				double[][] LiveMatrix = new double[exp.timePoints.size()][gens];
				ArrayList<Double> totalCounts = new ArrayList<Double>();
				ArrayList<Double> totalCounts2 = null;
				//EXTRACT COUNTS FROM FLUORESCENCE FIT USING FIRST FIT FLUORESCENCE PARAMETERS (ASSUME SOMETHING LIKE FLOWJO)
				for(int i = 0; i < exp.timePoints.size(); i++)
				{
					int[] Tcounts = exp.timePoints.get(i).getReplicates().get(0).getCounts(); //counts are counted by assigning each cell to the most probable gaussian
					double total = 0;
					double total2 = 0;
					for(int j = 0; j < Tcounts.length; j++)
					{
						LiveMatrix[i][j] = Tcounts[j];
						total+= Tcounts[j];
					}
					totalCounts.add(total);
					if(totalCounts2 != null)
						totalCounts2.add(total2);
				}
				//////////////////////////////////////////////////////// DONE WITH BAD COUNT FITTING///////////////////////////////////
				//BEST SOLUTION USING COMBINED METHOD
				Experiment best_exp = null;
				best_exp = exp;  //the original experiment that was already used to fit the fluorescences and contains all of the raw data
				System.err.println("Fitting PCyton to Histograms: "+(repetition)+"/"+repetitions);
				//NOW FIT USING COMBINED METHOD
				double bestFlowMaxScore = 999999;
				for(int trial = 0; trial < num_trials; trial++)
				{
					long start = System.currentTimeMillis();
					exp.fitCyton(gens,limits,dt,true,1); //PCyton (fitting is repeated x times within this method)
					if(exp.bestScore < bestFlowMaxScore)
					{
						bestFlowMaxScore = exp.bestScore;
						fmax[trial].params = Methods.duplicate(exp.abestparams);
					}
					else
						fmax[trial].params = Methods.duplicate(fmax[trial-1].params);
					fmax[trial].elapsed_time = System.currentTimeMillis()-start;
				}
			    //REMAKE THESE EXPERIMENTS FOR p-value CALCULATIONS
			    Experiment e1 =  new Experiment(data,undivided,gens,volumes,100,times,"GeneratedForCounts");          //used to determine the fluorescence histograms
				for(int trial = 0; trial < num_trials; trial++)
				{
				    //GIVE FITTED FLUORESCENCE PARAMETERS TO USE (from fitting)
					for(int i = 0; i < e1.timePoints.size(); i++)
					{
						e1.timePoints.get(i).CV = fl[i][0];
						e1.timePoints.get(i).ratio = fl[i][1];
						e1.timePoints.get(i).background = fl[i][2];
						e1.timePoints.get(i).getReplicates().get(0).shift = fl[i][3];
					}
				    BufferedWriter out = new BufferedWriter(new FileWriter("BestOf"+trial+"_"+outFile, true)); //output stream opened
				    System.err.println("-=-=-=-=-=-=- BEST OF "+(trial+1)+" -=-=-=-=-=-=-=-=-");
				    //UPDATE WEIGHTS BASED ON PARAMETERS
				    e1.cytonToFluorescence(fmax[trial].params, true);
				    //FIGURE OUT p-values!!!!
				    fmax[trial].pvalue=e1.getPValue();
				    for(int i = 0; i < e1.timePoints.size(); i++)
				    	fmax[trial].counts[i]=Methods.duplicate(e1.timePoints.get(i).getReplicates().get(0).getCounts());
				    System.err.println("Calculating p-values for FlowMax fitting: "+fmax[trial].pvalue);
				    //SET TO PERFECT FLUORESCENCE FOR THE REST:
					for(int i = 0; i < e1.timePoints.size(); i++)
					{
						e1.timePoints.get(i).CV = CV;
						e1.timePoints.get(i).ratio = ratio;
						e1.timePoints.get(i).background = background;
						e1.timePoints.get(i).getReplicates().get(0).shift = shift;
					}

					//TRANSLATE PARAMETERS
					for(int ii=0; ii <= 6; ii+=2)
					{

						//PCyton FlowMax
						double mean = Math.exp(fmax[trial].params[ii]+0.5*fmax[trial].params[ii+1]*fmax[trial].params[ii+1]);
						double sigma = mean*Math.sqrt(Math.exp(fmax[trial].params[ii+1]*fmax[trial].params[ii+1])-1); 
						fmax[trial].params[ii]=mean;
						fmax[trial].params[ii+1]=sigma;
					}
						
					String output= repetition+",PCyt FlowMax:,"+fmax[trial].pvalue+",PCyt Fl:,"+ffl[trial].pvalue+",Real:";
				    for(int i = 0; i < sparams.length; i++)
				    {
				    	output+=","+sparams[i];
				    }
				    output+=",PCyton FlowMax:";
				    for(int i = 0; i < params.length; i++)
				    {
				    		output+=","+fmax[trial].params[i];
				    }
				    output+=",CV,"+CV+",Ratio,"+ratio+",Bkgrnd,"+background+",Shft,"+shift+",PCyton F Ps";
				    for(int i = 0; i < best_exp.timePoints.size(); i++)
				    	output+=","+best_exp.timePoints.get(i).CV+","+best_exp.timePoints.get(i).ratio+","+best_exp.timePoints.get(i).background;
					output+=",Generated Counts:";
					for(int t = 0; t < timePoints.length; t++)
					{
						for(int j = 0; j < gens; j++)
							output+=","+counts[j][(int)(timePoints[t]/dt)];
					}
					output+=",PCyton Fl Counts:";
					for(int t = 0; t < timePoints.length; t++)
					{
						for(int j = 0; j < gens; j++)
							output+=","+ffl[trial].counts[t][j];
					}
					output+=",PCyton FlowMax Counts:";
				    //UPDATE WEIGHTS BASED ON PARAMETERS
					for(int t = 0; t < timePoints.length; t++)
					{
						for(int j = 0; j < gens; j++)
							output+=","+fmax[trial].counts[t][j];
					}
					output+=",Fl_Elapsed,"+ffl[trial].elapsed_time;
					output+=",FlowMax_Elapsed,"+fmax[trial].elapsed_time;					
					out.write(output+separator);
					out.close();
				}

			} //end of trial loop
			double average_time =(((double)(System.currentTimeMillis()-start_time))/(60000*(repetition+1))); 
			System.err.println("Average itteration time: "+average_time+" mins. Come back in "+(average_time*(repetitions-1-repetition))+" mins.");
		}

		}catch(Exception e)
		{
			e.printStackTrace();
		}
	}
	/**
	 * Helper class used to represent a set of parameters that fully describe one Cyton Model. 
	 * @author Max
	 *
	 */
	private class Experiment2 implements Cloneable
	{
		double[] params = new double[nparams];
		double lastScore = Double.POSITIVE_INFINITY;
		double counts_score = 0;
		double count_correlation = 0;
		double prop_score =0;
		double prop_correlation = 0;
		double area_diff = 0;

	//	ArrayList<Integer> path = new ArrayList<Integer>();
		/**
		 * Constructor
		 * @param params
		 */
		public Experiment2(double[] params)
		{
			for(int i = 0; i < params.length; i++)
				this.params[i] = params[i];
		}
		
		/**
		 * Score calculated analogously to the score from Histograms
		 * @param LiveMatrix
		 * @param times
		 * @param polyFit
		 * @param type
		 * @return
		 */
		public double calculateScoreSimple(double[][] LiveMatrix, ArrayList<Double> times,ArrayList<Double> polyFit,String type) 
		{
			double score = 0;
			double[][] results;
			int maxTime = times.get(times.size()-1).intValue() + 1;
			int steps = (int)(maxTime/dt);
			
			if(type.compareTo("Original")==0)
				results = computeCytonsOld(params, steps, gens, maxTime)[0];
			else
				results = computeCytons(params,steps,gens,maxTime)[0];
			for(int i = 0; i < times.size(); i++)
			{
				double total_m = 0;
				double total_e = 0;
				for(int j = 0; j < gens; j++)
				{
					total_m += results[j][(int)(times.get(i)/dt)];
					total_e += LiveMatrix[i][j];
				}
				for(int j = 0; j < gens; j++)
				{
					score+=Math.pow( results[j][(int)(times.get(i)/dt)] - LiveMatrix[i][j],2);
				}
			}
			return score;
		}
		
		/**
		 * computes the cytons and calculates score based on difference between the current fit and the experimental data
		 * SCORE: Sum over Times(Sum over generations(Difference in proportions^2)+Difference in total count^2)
		 * @param LiveMatrix Experimental Data in the form of proportions at particular times
		 * @param times Those times
		 * @param polyFit The total counts at each time
		 * @return
		 */
		public double calculateScoreOld(double[][] LiveMatrix, ArrayList<Double> times,ArrayList<Double> polyFit,String type) 
		{
			double zero = 0.0001;
			double score = 0;
			lastScore = Double.POSITIVE_INFINITY;
			counts_score = 0;
			count_correlation = 0;
			prop_score =0;
			prop_correlation = 0;
			area_diff = 0;
			double T = maxTime;
			int Generations = gens;
			double[][][] results = null;
			if(type.compareTo("Original")==0)
				results = computeCytonsOld(params, steps, 10, maxTime);
			else
				results = computeCytons(params,steps,10,maxTime);
			//lets just compute the global difference between the LiveMatrix and the results LiveMatrix
			double dt =(((double)T)/steps);
			int currentFitLine = 0;
			double[] x1 = new double[times.size()];
			double[] x2 = new double[times.size()];
			double last_count_e =-1;
			double model_area = 0;

			boolean integrating = false; //started looking at areas
			for(int j = 0; j < steps; j++)
			{
				if(integrating)
				{
					double val_current = 0;
					double val_before = 0;
					for(int i = 0; i < Generations; i++)
					{
						val_current+=(results[0][i][j]);
						val_before+=(results[0][i][j-1]);
					}
					model_area+= (val_before*dt)+dt*(val_current-val_before)/2;
				}
				if(currentFitLine < times.size() && dt*j <= times.get(currentFitLine) && dt*(j+1) > times.get(currentFitLine))
				{
					int lastnonzero = 0;
					double totalE = 0; //total count at this time from experiment
					for(int i = 0; i < LiveMatrix[currentFitLine].length; i++)//for each generation at the current time slice
					{
						totalE+=LiveMatrix[currentFitLine][i];  //assume that this is the count from experimental data
						if(LiveMatrix[currentFitLine][i] > 0) lastnonzero = i;
					}
					if(last_count_e != -1)
					{
						double dte = (times.get(currentFitLine)-times.get(currentFitLine-1));
						double area_expt= dte*last_count_e+dte*(polyFit.get(currentFitLine)-last_count_e)/2;
						area_diff+=Math.abs((area_expt-model_area)/(area_expt+model_area));
						model_area = 0;
						last_count_e = polyFit.get(currentFitLine);
					}
					else
					{
						integrating = true;
						last_count_e = polyFit.get(currentFitLine);
					}
					double totalCells = 0;   //this is from model
					double proportions = 0;  //this is the proportions score summation
					for(int i = 0; i < Generations; i++)
					{
						totalCells+=results[0][i][j];  //count up cells total at this time for all generations
					}			
					int effective_generations = 0;
					if(totalCells == 0)
					{
						totalCells = 1;
						for(int i = 0; i < Generations; i++)
							results[0][i][j] = 1.0/Generations;
					}
					double remaining = 0;
					effective_generations = 0;
					double[] ex = new double[Generations];
					double[] mx = new double[Generations];
					for(int i = 0; i <Generations; i++)
					{
						if(i < lastnonzero|| i < 6) 
						{
							double model =Math.max(results[0][i][j]/totalCells,zero);
							double experiment = Math.max(LiveMatrix[currentFitLine][i]/totalE,zero);
							mx[i]= model;
							ex[i]=experiment;
							double  adding = Math.pow((experiment-model),2);
							proportions+=adding ; //assume experiment's last non-zero proportion is the remainder of proportions added up
							effective_generations++;
						}
						else //just add up the rest since the last exp generation represents that generation plus all others.
						{
							effective_generations++;
							remaining = results[0][i][j];
							for(int k=i+1; k< Generations; k++)
								remaining += results[0][k][j];
							double model =Math.max(remaining/totalCells,zero);
							double experiment = Math.max(LiveMatrix[currentFitLine][i]/totalE,zero);
							mx[i]= model;
							ex[i]=experiment;
							double  adding = Math.pow((experiment-model),2);
							proportions+=adding; //assume experiment's last non-zero proportion is the remainder of proportions added up 
							effective_generations++;  //this is basically a count of all of the times we added a proportion score
							break;
						}
					}
					proportions=proportions/effective_generations;
					prop_correlation += (1-Methods.getPearsonCorrelation(ex, mx));
					prop_score +=proportions;
					//TOTAL COUNTS
					double experiment = polyFit.get(currentFitLine);
					double model = totalCells;
					counts_score+=Math.pow((experiment/(experiment+model)-model/(experiment+model)),2);
					if(model ==0)counts_score=Double.POSITIVE_INFINITY;
					x1[currentFitLine]=totalCells;
					x2[currentFitLine]=polyFit.get(currentFitLine);
					currentFitLine++;
				}
			}
			counts_score =counts_score/times.size();	//now normalized to times
			prop_score = prop_score/times.size();
			prop_correlation = prop_correlation/times.size();
			count_correlation =(1-Methods.getPearsonCorrelation(x1,x2));
			area_diff = area_diff/times.size();
			
			score = (5*prop_score+counts_score+area_diff+2*prop_correlation+count_correlation*2)/11;
			lastScore = score;

		//	lastNumDiv = results[2];
		//	lastNumDie = results[3];
			if(!Double.isNaN(score))
				return score;
			else
				return Double.POSITIVE_INFINITY;
		}	
	
		/**
		 * Complicated scoring function optimized to produce the best results using the counts only. 
		 * @param LiveMatrix
		 * @param times
		 * @param polyFit
		 * @param type
		 * @return
		 */
		public double calculateScore(double[][] LiveMatrix, ArrayList<Double> times,ArrayList<Double> polyFit,String type) 
		{
			double zero = 0.0001;
			double score = 0;
			lastScore = Double.POSITIVE_INFINITY;
			counts_score = 0;
			count_correlation = 0;
			prop_score =0;
			prop_correlation = 0;
			area_diff = 0;
			double T = maxTime;
			int Generations = gens;
			double[][][] results = null;
			if(type.compareTo("Original")==0)
				results = computeCytonsOld(params, steps,gens, maxTime);
			else
				results = computeCytons(params,steps,gens,maxTime);
			//lets just compute the global difference between the LiveMatrix and the results LiveMatrix
			double dt =(((double)T)/steps);
			int currentFitLine = 0;
			double[] x1 = new double[times.size()];
			double[] x2 = new double[times.size()];
			double last_count_e =-1;

			for(int t = 0; t < times.size(); t++)
			{
				int j = (int)(times.get(t)/dt);
				if(currentFitLine < times.size() && dt*j <= times.get(currentFitLine) && dt*(j+1) > times.get(currentFitLine))
				{
					int lastnonzero = 0;
					double totalE = 0; //total count at this time from experiment
					for(int i = 0; i < LiveMatrix[currentFitLine].length; i++)//for each generation at the current time slice
					{
						totalE+=LiveMatrix[currentFitLine][i];  //assume that this is the count from experimental data
						if(LiveMatrix[currentFitLine][i] > 0) lastnonzero = i;
					}
					if(last_count_e != -1)
					{
						last_count_e = polyFit.get(currentFitLine);
					}
					else
					{
						last_count_e = polyFit.get(currentFitLine);
					}
					double totalCells = 0;   //this is from model
					double proportions = 0;  //this is the proportions score summation
					for(int i = 0; i < Generations; i++)
					{
						totalCells+=results[0][i][j];  //count up cells total at this time for all generations
					}			
					int effective_generations = 0;
					if(totalCells == 0)
					{
						totalCells = 1;
						for(int i = 0; i < Generations; i++)
							results[0][i][j] = 1.0/Generations;
					}
					double remaining = 0;
					effective_generations = 0;
					double[] ex = new double[Generations];
					double[] mx = new double[Generations];
					for(int i = 0; i <Generations; i++)
					{
						if(i < lastnonzero|| i < 6) 
						{
							double model =Math.max(results[0][i][j]/totalCells,zero);
							double experiment = Math.max(LiveMatrix[currentFitLine][i]/totalE,zero);
							mx[i]= model;
							ex[i]=experiment;
							double  adding = Math.pow((experiment-model),2);
							proportions+=adding ; //assume experiment's last non-zero proportion is the remainder of proportions added up
							effective_generations++;
						}
						else //just add up the rest since the last exp generation represents that generation plus all others.
						{
							effective_generations++;
							remaining = results[0][i][j];
							for(int k=i+1; k< Generations; k++)
								remaining += results[0][k][j];
							double model =Math.max(remaining/totalCells,zero);
							double experiment = Math.max(LiveMatrix[currentFitLine][i]/totalE,zero);
							mx[i]= model;
							ex[i]=experiment;
							double  adding = Math.pow((experiment-model),2);
							proportions+=adding; //assume experiment's last non-zero proportion is the remainder of proportions added up 
							effective_generations++;  //this is basically a count of all of the times we added a proportion score
							break;
						}
					}
					proportions=proportions/effective_generations;
					prop_correlation += (1-Methods.getPearsonCorrelation(ex, mx));
					prop_score +=proportions;
					//TOTAL COUNTS
					double experiment = polyFit.get(currentFitLine);
					double model = totalCells;
					counts_score+=Math.pow((experiment/(experiment+model)-model/(experiment+model)),2);
					if(model ==0)counts_score=Double.POSITIVE_INFINITY;
					x1[currentFitLine]=totalCells;
					x2[currentFitLine]=polyFit.get(currentFitLine);
					currentFitLine++;
				}
			}
			counts_score =counts_score/times.size();	//now normalized to times
			prop_score = prop_score/times.size();
			prop_correlation = prop_correlation/times.size();
			count_correlation =(1-Methods.getPearsonCorrelation(x1,x2));
			
			score = (5*prop_score+counts_score+2*prop_correlation+count_correlation*2)/10;
			lastScore = score;

		//	lastNumDiv = results[2];
		//	lastNumDie = results[3];
			if(!Double.isNaN(score))
				return score;
			else
				return Double.POSITIVE_INFINITY;
		}	

	
		
		/**
		 * A guassian random distribution about the current value is used
		 * @param howmuch
		 */
		public void mutate(boolean mode)
		{

			//howmuch is a proportion of the total for each param
			//int[] sign = {1,-1};
			for(int i = 0; i < nparams; i++ )
			{
				double chance = gen.nextDouble();
				if(mode) //coarse grained fitting mode (taking larger jumps)
				{
					if(chance > 0.05)
						params[i] = Methods.reflectWithin(params[i]+gen.nextGaussian()*params[i]*0.01,limits[i][0],limits[i][1]);
					else
						params[i] = Methods.reflectWithin(params[i]+gen.nextGaussian()*params[i]*3,limits[i][0],limits[i][1]);
				}
				else  //fine grained about the minimum
				{
					if(chance > 0.05)
						params[i] = Methods.reflectWithin(params[i]+gen.nextGaussian()*params[i]*0.001,limits[i][0],limits[i][1]);
					else
						params[i] = Methods.reflectWithin(params[i]+gen.nextGaussian()*params[i]*0.1,limits[i][0],limits[i][1]);
				}
			}

		}
		
	
		public Experiment2 clone()
		{
			Experiment2 clone = null;
			try {
				clone = (Experiment2) super.clone();
				for(int i = 0; i < params.length; i++)
					clone.params[i] = this.params[i];
				//clone.lastDeath = this.lastDeath.clone();
				//clone.lastLive = this.lastLive.clone();
				clone.lastScore = this.lastScore;
				clone.prop_score = this.prop_score;
				clone.count_correlation = this.count_correlation;
				clone.counts_score = this.counts_score;
				clone.prop_correlation = this.prop_correlation;
				clone.area_diff = this.area_diff;
			} catch (CloneNotSupportedException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}
			
			return clone;
		}
		
	}




	public Experiment2 fitParameters2(double[][] LiveMatrix, ArrayList<Double> times, ArrayList<Double> totalCounts, boolean mode)
	{
		Experiment2 bestExpt = null;  //best experiment
		Experiment2 currentExpt = null;// current position
		int RandomCount=4000;
		int countme = 0;
		int pcount = nparams;

		//lets initialize T0 by measuring a completely random sampling
		double bestscore = Double.POSITIVE_INFINITY;
		while(countme < RandomCount)  //search for a good starting place
		{
			countme++;
			double[] params = new double[pcount];
			for(int i = 0; i < pcount; i++)
				params[i] = limits[i][0]+gen.nextDouble()*(limits[i][1]-limits[i][0]);
			Experiment2 newEx = new Experiment2(params);
			double score = 0;
			if(mode)
				score= newEx.calculateScoreSimple(LiveMatrix, times, totalCounts,"V3");
			else
				score = newEx.calculateScoreSimple(LiveMatrix, times, totalCounts,"Original");
			if(bestExpt==null ||score < bestscore)
			{
				bestExpt=newEx;
				bestscore = score;
			}
		}
		currentExpt = bestExpt;  
		countme=0;
		try
		{
			double T = bestExpt.lastScore/10;
			double Tstop = T/10000;
			int counter  = 0;
			int same_count = 0;
			while(T > Tstop)
			{
				double[] params = new double[pcount];
				for(int i  = 0; i <currentExpt.params.length ; i++)
					params[i]=currentExpt.params[i];
				Experiment2 candidate = new Experiment2(params);
				candidate.mutate(true);				//neighbor picked
				double score = 0;
				if(!mode) score= candidate.calculateScoreSimple(LiveMatrix, times, totalCounts,"Original");
				else score= candidate.calculateScoreSimple(LiveMatrix, times, totalCounts,"V3");
				double diff = score-bestscore; //dCk
				double chance =  Math.exp(-(diff)/(T));
				if(gen.nextDouble() < chance)
				{
					//we are good
					currentExpt = candidate;
					if(score < bestscore)
					{
						bestExpt = currentExpt.clone();
						same_count=0;
						bestscore = score;
						T=Math.min(Tstop*100000,T*1.1);
					}
				}
				T=T*0.9995;
				counter++;
				same_count++;
				
			}
			T = bestExpt.lastScore/10;
			Tstop = T/1000;
			while(T > Tstop || counter < 1000)
			{
				counter++;
				double[] params = new double[pcount];
				for(int i  = 0; i <bestExpt.params.length ; i++)
					params[i]=bestExpt.params[i];
				Experiment2 candidate = new Experiment2(params);
				candidate.mutate(false);				//neighbor picked
				double score = 0;
				if(!mode) score= candidate.calculateScoreSimple(LiveMatrix, times, totalCounts,"Original");
				else score= candidate.calculateScoreSimple(LiveMatrix, times, totalCounts,"V3");
				double diff = score-bestscore; //dCk
				double chance =  Math.exp(-(diff)/(T));
				if(gen.nextDouble() < chance)
				{
					//we are good
					currentExpt = candidate;
					if(score < bestscore)
					{
						bestExpt = currentExpt.clone();
						same_count=0;
						bestscore = score;
						T=Math.min(Tstop*1000,T*3);
					}
				}
				T=T*0.995;				
			}
		}catch(Exception e)
		{
			e.printStackTrace();
		}
		//OUTPUT to the output file after each solution is found
		System.err.println(bestscore);
		return bestExpt.clone();	//best solution
	}
	
	public Experiment2 fitParameters3(double[][] LiveMatrix, ArrayList<Double> times, ArrayList<Double> totalCounts, boolean mode)
	{
		Experiment2 bestExpt = null;  //best experiment
		Experiment2 currentExpt = null;// current position
		int RandomCount=4000;
		int countme = 0;
		int pcount = nparams;

		//lets initialize T0 by measuring a completely random sampling
		double bestscore = Double.POSITIVE_INFINITY;
		while(countme < RandomCount)  //search for a good starting place
		{
			countme++;
			double[] params = new double[pcount];
			for(int i = 0; i < pcount; i++)
				params[i] = limits[i][0]+gen.nextDouble()*(limits[i][1]-limits[i][0]);
			Experiment2 newEx = new Experiment2(params);
			double score = 0;
			if(mode)
				score= newEx.calculateScore(LiveMatrix, times, totalCounts,"V3");
			else
				score = newEx.calculateScore(LiveMatrix, times, totalCounts,"Original");
			if(bestExpt==null ||score < bestscore)
			{
				bestExpt=newEx;
				bestscore = score;
			}
		}
		currentExpt = bestExpt;  
		countme=0;
		try
		{
			double T = bestExpt.lastScore/10;
			double Tstop = T/10000;
			int counter  = 0;
			int same_count = 0;
			while(T > Tstop)
			{
				double[] params = new double[pcount];
				for(int i  = 0; i <currentExpt.params.length ; i++)
					params[i]=currentExpt.params[i];
				Experiment2 candidate = new Experiment2(params);
				candidate.mutate(true);				//neighbor picked
				double score = 0;
				if(!mode) score= candidate.calculateScore(LiveMatrix, times, totalCounts,"Original");
				else score= candidate.calculateScore(LiveMatrix, times, totalCounts,"V3");
				double diff = score-bestscore; //dCk
				double chance =  Math.exp(-(diff)/(T));
				if(gen.nextDouble() < chance)
				{
					//we are good
					currentExpt = candidate;
					if(score < bestscore)
					{
						bestExpt = currentExpt.clone();
						same_count=0;
						bestscore = score;
						T=Math.min(Tstop*10000,T*1.1);
					}
				}
				T=T*0.9995;
				counter++;
				same_count++;
				
			}
			T = bestExpt.lastScore/10;
			Tstop = T/1000;
			while(T > Tstop || counter < 1000)
			{
				double[] params = new double[pcount];
				for(int i  = 0; i <bestExpt.params.length ; i++)
					params[i]=bestExpt.params[i];
				Experiment2 candidate = new Experiment2(params);
				candidate.mutate(false);				//neighbor picked
				double score = 0;
				if(!mode) score= candidate.calculateScore(LiveMatrix, times, totalCounts,"Original");
				else score= candidate.calculateScore(LiveMatrix, times, totalCounts,"V3");
				double diff = score-bestscore; //dCk
				double chance =  Math.exp(-(diff)/(T));
				if(gen.nextDouble() < chance)
				{
					//we are good
					currentExpt = candidate;
					if(score < bestscore)
					{
						bestExpt = currentExpt.clone();
						same_count=0;
						bestscore = score;
						T=Math.min(Tstop*1000,T*3);
					}
				}
				T=T*0.995;				
			}
		}catch(Exception e)
		{
			e.printStackTrace();
		}
		//OUTPUT to the output file after each solution is found
		System.err.println(bestscore);
		return bestExpt.clone();	//best solution
	}
	private double getLogNormalCDF(double mu, double sigma, double a,
			double b) {
		if(a == 0 || b==0) return 0;
		double k = 8*(Math.PI-3)/(3*Math.PI*(4-Math.PI));
		double r2 = Math.sqrt(2);
		double A = (Math.log(a)-mu)/(r2*sigma);
		double B = (Math.log(b)-mu)/(r2*sigma);
		double A2 = A*A;
		double B2 = B*B;
		double k2 = 4.0/Math.PI;
		double left = (B/Math.abs(B))*Math.sqrt(1-Math.exp(-B2*(k2+k*B2)/(1+k*B2)));
		double right = (A/Math.abs(A))*Math.sqrt(1-Math.exp(-A2*(k2+k*A2)/(1+k*A2)));
		return 0.5*(left-right);
	}
	
	private double getLogNormalCDF(double mu, double sigma, 
			double b) {
		double k = 8*(Math.PI-3)/(3*Math.PI*(4-Math.PI));
		double r2 = Math.sqrt(2);
		double B = (Math.log(b)-mu)/(r2*sigma);
		double B2 = B*B;
		double k2 = 4.0/Math.PI;
		double left = (B/Math.abs(B))*Math.sqrt(1-Math.exp(-B2*(k2+k*B2)/(1+k*B2)));
		return 0.5*(1+left);
	}
	
	private double getGaussianCDF(double mu, double sigma, double a) {
		double k = 8*(Math.PI-3)/(3*Math.PI*(4-Math.PI));
		double r2 = Math.sqrt(2);
		double A = (a-mu)/(r2*sigma);
		double A2 = A*A;
		double k2 = 4.0/Math.PI;
		double right = (A/Math.abs(A))*Math.sqrt(1-Math.exp(-A2*(k2+k*A2)/(1+k*A2)));
		return 0.5*(1+right);
	}
	
	private double getGaussianCDF(double mu, double sigma, double a,
			double b) {
		double k = 8*(Math.PI-3)/(3*Math.PI*(4-Math.PI));
		double r2 = Math.sqrt(2);
		double A = (a-mu)/(r2*sigma);
		double B = (b-mu)/(r2*sigma);
		double A2 = A*A;
		double B2 = B*B;
		double k2 = 4.0/Math.PI;
		double left = (B/Math.abs(B))*Math.sqrt(1-Math.exp(-B2*(k2+k*B2)/(1+k*B2)));
		double right = (A/Math.abs(A))*Math.sqrt(1-Math.exp(-A2*(k2+k*A2)/(1+k*A2)));
		return 0.5*(left-right);
	}
	

	/**
	 * New and improved computeCytons with pFs affecting both division and death
	 * @param params
	 * @return
	 */
	private double[][][] computeCytons(double[] params, int steps, int Generations, int T)
	{
		//lets itterate over all steps and then output
		double[][] NumDiv = new double[Generations][steps];
		double[][] NumDie = new double[Generations][steps];
		double[] divPDF0 = new double[steps];
		double[] deathPDF0 = new double[steps];
		double[] divPDFblast = new double[steps];
		double[] deathPDFblast = new double[steps];
		double[] mechDeathPDF = new double[steps];
		double[] mechDeathCDF = new double[steps];
		double[][] LiveMatrix = new double[Generations+1][steps+1];
		double[][] DeadMatrix = new double[Generations+1][steps+1];
		double[] pF = new double[Generations+1];
		double dt =(((double)T)/steps);
		double pF0 = params[10];  //fit this as well
		try
		{
			double[] ddcurve = new double[Generations];
			ddcurve[0] = 1-pF0;
			double[]ddcdf = new double[Generations+1];
			ddcdf[0] = 1-pF0;
			for(int j = 1; j < Generations; j++)
			{
				ddcurve[j] = pF0*(getGaussianCDF(params[8],params[9],j-1,j)/(1-getGaussianCDF(params[8],params[9],0)));
				ddcdf[j] = ddcdf[j-1]+ddcurve[j];
			}
			ddcdf[Generations] = ddcdf[Generations-1]+pF0*(getGaussianCDF(params[8],params[9],Generations-1,Generations)/(1-getGaussianCDF(params[8],params[9],1)));
			pF[0] = params[10];
			for(int i = 1; i < Generations; i++)
			{
				if((1-ddcdf[i])/(1-ddcdf[i-1]) > 0)
					pF[i] = (1-ddcdf[i])/(1-ddcdf[i-1]);
				else
					pF[i] = 0;
			}
			for(int i = 1; i < steps; i ++)
			{
				double before = (i-1)*dt;
				double after = (i*dt);
				divPDF0[i]=getLogNormalCDF(params[0], params[1], before,after);
				divPDFblast[i]=getLogNormalCDF(params[2], params[3], before,after);
				deathPDF0[i]=getLogNormalCDF(params[4], params[5], before,after);
				deathPDFblast[i]=getLogNormalCDF(params[6], params[7], before,after);
			}
		}
		catch(Exception e)
		{
			e.printStackTrace();
		}
		double sumdied = 0;
		double sumdived = 0;
		for(int i = 1 ;i < steps; i++)
		{
			sumdied+=deathPDF0[i];  //up to AND including (some already die at time 1 )
			if(i > 1)  //up to but not including (none have divided yet at time 1 since you would have had to divide at time 0)
				sumdived+=divPDF0[i-1];
			NumDiv[0][i] = pF[0]*params[11]*(1-params[12]*mechDeathCDF[i])*divPDF0[i];
			NumDie[0][i] = (1-pF[0])*params[11]*((1-params[12]*mechDeathCDF[i])*deathPDF0[i]+(1-(1-pF[0])*sumdied)*params[12]*mechDeathPDF[i]);  //haven't divided, and died in one of two ways: mechanical death or normal lognormal death
		}
		for(int i = 1; i < Generations; i++)
		{		
			for(int j = 1; j < steps ; j++)
			{
				double totalDividedFromPreviousGen = 0;
				for(int jprime = 0; jprime <= j-1; jprime++)
				{
					totalDividedFromPreviousGen=NumDiv[i-1][jprime];  //the cohort from previous generation that divided at the same time
					double chanceOfCohortDividingRightNow = divPDFblast[j-jprime];  //for this cohort it is now time j-jprime
					NumDiv[i][j] += 2*pF[i]*totalDividedFromPreviousGen*chanceOfCohortDividingRightNow; //division at this time for generation i
					double chanceOfCohortDyingRightNow = (1-pF[i])*deathPDFblast[j-jprime];  //the cohort's age is j-jprime
					NumDie[i][j] += 2*totalDividedFromPreviousGen*chanceOfCohortDyingRightNow;  //death right now for this cohort added to the total pool of dead at this time for this gen
				}			
			}
		}
		//lets populate the alive and death matrices
		for(int i = 0; i < Generations; i++)
		{
			if(i==0)
				LiveMatrix[i][0] = params[11];
			else
				LiveMatrix[i][0] = 0;  //not necessary really
			double sumDiei = 0;
			sumDiei += NumDie[i][0];
			//TODO Fix dead matrix such that you don't have negative cells registering as dead
			DeadMatrix[i][0] = NumDie[i][0];  //first round of deaths
			for(int j = 1; j < steps; j++)
			{
				if(i==0) LiveMatrix[i][j] = LiveMatrix[i][j-1]-NumDiv[i][j]-NumDie[i][j];
				else LiveMatrix[i][j] = LiveMatrix[i][j-1]+2*NumDiv[i-1][j]-NumDie[i][j]-NumDiv[i][j];

				if(LiveMatrix[i][j] < 0) LiveMatrix[i][j] = 0;
				DeadMatrix[i][j] = sumDiei;
			}
		}
		double[][][] results = new double[4][Generations][steps];
		results[0] = LiveMatrix; 
		results[1] = DeadMatrix;  //not used
		results[2] = NumDiv; //not used
		results[3] = NumDie; //not used
		return results;
	}


	private double[][][] computeCytonsOld(double[] params, int steps, int Generations, int T)
	{
		//lets itterate over all steps and then output
		double[][] NumDiv = new double[Generations][steps];
		double[][] NumDie = new double[Generations][steps];
		double[] divPDF0 = new double[steps];
		double[] deathPDF0 = new double[steps];
		double[] divPDFblast = new double[steps];
		double[] deathPDFblast = new double[steps];
	//	double[] divCDF0 = new double[steps];
	//	double[] deathCDF0 = new double[steps];
		double[] divCDFblast  = new double[steps];
		double[] deathCDFblast = new double[steps];
		double[][] LiveMatrix = new double[Generations+1][steps+1];
		double[][] DeadMatrix = new double[Generations+1][steps+1];
		//double[][] NumIdle = new double[Generations][steps];
		double[] pF = new double[Generations];
		double dt =(((double)T)/steps);
		double pF0 = params[10];  //fit this as well
		try
		{
			double[] ddcurve = new double[Generations];
			ddcurve[0] = 1-pF0;
			double[]ddcdf = new double[Generations+1];
			ddcdf[0] = 1-pF0;
			for(int j = 1; j < Generations; j++)
			{
				ddcurve[j] = pF0*(getGaussianCDF(params[8],params[9],j,j+1)/(1-getGaussianCDF(params[8],params[9],1)));
				ddcdf[j] = ddcdf[j-1]+ddcurve[j];
			}
			ddcdf[Generations] = ddcdf[Generations-1]+pF0*(getGaussianCDF(params[8],params[9],Generations-1,Generations)/(1-getGaussianCDF(params[8],params[9],1)));
			pF[0] = params[10];
			for(int i = 1; i < Generations; i++)
			{
				if((1-ddcdf[i+1])/(1-ddcdf[i]) > 0)
					pF[i] = (1-ddcdf[i+1])/(1-ddcdf[i]);
				else
					pF[i] = 0;
			}
			for(int i = 1; i < steps; i ++)
			{
				double before = (i-1)*dt;
				double after = (i*dt);
				divPDF0[i]=getLogNormalCDF(params[0], params[1], before,after);
			//	divCDF0[i]=divCDF0[i-1]+divPDF0[i];
				divPDFblast[i]=getLogNormalCDF(params[2], params[3], before,after);
				divCDFblast[i]=divCDFblast[i-1]+divPDFblast[i];
				deathPDF0[i]=getLogNormalCDF(params[4], params[5], before,after);
			//	deathCDF0[i]=deathCDF0[i-1]+deathPDF0[i];
				deathPDFblast[i]=getLogNormalCDF(params[6], params[7], before,after);
				deathCDFblast[i]=deathCDFblast[i-1]+deathPDFblast[i];			
				//mechDeathPDF[i]=(1-Math.exp(-params[13]*(i)*dt))-(1-Math.exp(-params[13]*((i-1)*dt)));
				//mechDeathCDF[i] = (1-Math.exp(-params[13]*(i)*dt));
			}
		}
		catch(Exception e)
		{
			e.printStackTrace();
		}
		double sumdied = 0;
		double sumdived = 0;
		for(int i = 0 ;i < steps; i++)
		{
			sumdied+=deathPDF0[i];  //up to AND including (some already die at time 0 )
			if(i > 0)  //up to but not including (none have divided yet at time 0 since you would have had to divide at time -1)
				sumdived+=divPDF0[i-1];
			NumDiv[0][i] = pF[0]*params[11]*(1-sumdied)*divPDF0[i];
			NumDie[0][i] = params[11]*(1-pF[0]*sumdived)*deathPDF0[i];
		}
		for(int i = 1; i < Generations; i++)
		{		
			for(int j = 1; j < steps ; j++)
			{
				double totalDividedFromPreviousGen = 0;
				for(int jprime = 0; jprime <= j-1; jprime++)
				{
					totalDividedFromPreviousGen=NumDiv[i-1][jprime];  //the cohort from previous generation that divided at the same time
					double probNotDead = 1-deathCDFblast[j-1-jprime];  //probability that that cohort did not die up until this time (low for lower jprime)
					double probNotDiv = 1;//probability that that cohort did not divide up until 2 time steps ago (could have divided last time but this is already accounted for)
					if(j-jprime >= 2)  //if enough time has passed to make this possible
						probNotDiv-=pF[i]*divCDFblast[j-2-jprime];  //subtract the fraction expected to divide
					double survivingFractionOfCohort = probNotDead; //probability that this cohort survives still
					double chanceOfCohortDividingRightNow = divPDFblast[j-jprime];  //for this cohort it is now time j-jprime
					NumDiv[i][j] += 2*pF[i]*totalDividedFromPreviousGen*survivingFractionOfCohort*chanceOfCohortDividingRightNow; //division at this time for generation i
					double chanceOfCohortStillUndivided = probNotDiv; //probability that this cohort hasn't divided yet
					double chanceOfCohortDyingRightNow = deathPDFblast[j-jprime];  //the cohort's age is j-jprime
					NumDie[i][j] += 2*totalDividedFromPreviousGen*chanceOfCohortStillUndivided*chanceOfCohortDyingRightNow;  //death right now for this cohort added to the total pool of dead at this time for this gen
				}			
			}
		}
		//lets populate the alive and death matrices
		for(int i = 0; i < Generations; i++)
		{
			if(i==0)
				LiveMatrix[i][0] = params[11];
			else
				LiveMatrix[i][0] = 0;  //not necessary really
			double sumDiei = 0;
			sumDiei += NumDie[i][0];
			//TODO Fix dead matrix such that you don't have negative cells registering as dead
			DeadMatrix[i][0] = NumDie[i][0];  //first round of deaths
			for(int j = 1; j < steps; j++)
			{
				if(i==0) LiveMatrix[i][j] = LiveMatrix[i][j-1]-NumDiv[i][j]-NumDie[i][j];
				else LiveMatrix[i][j] = LiveMatrix[i][j-1]+2*NumDiv[i-1][j]-NumDie[i][j]-NumDiv[i][j];
				if(LiveMatrix[i][j] < 0) LiveMatrix[i][j] = 0;
				DeadMatrix[i][j] = sumDiei;
			}
		

		}
		double[][][] results = new double[4][Generations][steps];
		results[0] = LiveMatrix;
		results[1] = DeadMatrix;// not used
		results[2] = NumDiv; //not used
		results[3] = NumDie; //not used
		return results;
	}
	

	
	public static void main(String[] args)
	{
		if(args.length <4) 
		{
			System.err.println("Usage: FakeBlah repetitions_to_run trials_per_rep output_file_name mode(1=all 0=PCyton Only) [num tpoints=10] [min delay]");
			System.exit(1);
		}
		if(args.length < 5)
			new TimingAnalysis(Integer.parseInt(args[0]),Integer.parseInt(args[1]),args[2],Integer.parseInt(args[3]),10,0);
		else if(args.length < 6)
			new TimingAnalysis(Integer.parseInt(args[0]),Integer.parseInt(args[1]),args[2],Integer.parseInt(args[3]),Integer.parseInt(args[4]),0);
		else
			new TimingAnalysis(Integer.parseInt(args[0]),Integer.parseInt(args[1]),args[2],Integer.parseInt(args[3]),Integer.parseInt(args[4]),Integer.parseInt(args[5]));
	}
}

