/*
Copyright (c)  2013   The Regents of the University of California
All Rights Reserved
 
Permission to copy, modify and distribute any part of this FlowMax for educational, research and non-profit purposes, without fee, and without a written agreement is hereby granted, provided that the above copyright notice, this paragraph and the following three paragraphs appear in all copies.
Those desiring to incorporate this FlowMax into commercial products or use for commercial purposes should contact the Technology Transfer Office, University of California, San Diego, 9500 Gilman Drive, Mail Code 0910, La Jolla, CA 92093-0910, Ph: (858) 534-5815, FAX: (858) 534-7345, E-MAIL:invent@ucsd.edu.
IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES, INCLUDING LOST PROFITS, ARISING OUT OF THE USE OF THIS FlowMax, EVEN IF THE UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
THE FlowMax PROVIDED HEREIN IS ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.  THE UNIVERSITY OF CALIFORNIA MAKES NO REPRESENTATIONS AND EXTENDS NO WARRANTIES OF ANY KIND, EITHER IMPLIED OR EXPRESS, 
INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE, OR THAT THE USE OF THE FlowMax WILL NOT INFRINGE ANY PATENT, TRADEMARK OR OTHER RIGHTS.
*/
package FlowMax;

/**
 * TritiatedCellGenerator is a program that allows you to simulate tritiated thymidine incorporation using the framework of the pcyton model. The input is a FlowMax solution in csv format
 * and the output is another csv file which will contain the 3H incorporation over time. 
 * 
 * Author Maxim Shokhirev. UCSD Signaling Systems Laboratory. UCSD. (C) 2012 
 *  
 */

import java.io.File;
import java.io.FileNotFoundException;
import java.io.PrintWriter;
import java.util.LinkedList;
import java.util.Random;
import java.util.Scanner;

import javax.swing.DefaultListModel;

import common.Dialogs;
import common.Methods;

public class TritiatedCellGenerator {

	public TritiatedCellGenerator() throws FileNotFoundException
	{
		File open = Dialogs.getOpenFile("csv", "A solution file");
		LinkedList<double[]> answers = processPhenotypeCSV(open);
		PrintWriter pw = new PrintWriter(Dialogs.getSaveFile("csv", "Save?"));
		Random gen = new Random();
		int steps = 400;  //total steps for model
		int time = 200;  //total time (hrs)
		int gens = 10;  //number of generations to calculate
		int trials = 250;  //samples each solution this many times to generate the distribution of 3H incorporation as a function of time.
		for(int i = 0; i < answers.size(); i++)
		{
			double[][] sums = new double[steps][trials];
			for(int trial = 0; trial < trials; trial++)
			{
				double[] params1 = Methods.getLogParams(answers.get(i),Constants.nparams);
				double[] params = new double[Constants.nparams];
				for(int p = 0; p < params.length; p++)
				{
					params[p] = gen.nextDouble()*(params1[p+Constants.nparams*2]+params1[p+Constants.nparams])+params1[p]-params1[p+Constants.nparams];
				}
				double[][] solution = Methods.computeCytons(params, steps, gens, time)[0];
				for(int t = 0; t <steps; t++)
				{
					for(int g = 1; g < gens; g++)
					{
						sums[t][trial]+=solution[g][t];
					}
				}
			}
			pw.println("Solution Set "+(i+1)+" from "+open.getName());
			for(int c = 0; c < steps; c++)
				pw.print((c*0.5)+",");
			pw.println();
			for(int r = 0; r < trials; r++)
			{
				for(int c = 0; c < steps; c++)
				{
					pw.print(sums[c][r]+",");
				}
				pw.println();
			}
			pw.println();
			pw.println();
			pw.println();
			for(int c = 0; c < steps; c++)
				pw.print(Methods.average(sums[c])+",");
			pw.println();
			for(int c = 0; c < steps; c++)
				pw.print(Methods.std(sums[c])+",");
			
		}
		pw.close();
		
	}
	
	/**
	 * @param args
	 */
	public static void main(String[] args) {
		try {
			new TritiatedCellGenerator();
		} catch (FileNotFoundException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		

	}

	private LinkedList<double[]> processPhenotypeCSV(File selectedFile) {
		//lets read in the file line by line and generate appropraite phenotype clusters in the list
		Scanner s;
		try {
			s = new Scanner(selectedFile);
			double[] params = new double[Constants.nparams*3];
			LinkedList<double[]> answers = new LinkedList<double[]>();
			while(s.hasNextLine())
			{
				String[] line = s.nextLine().split(",");
				if(line[0].startsWith("Cluster "))
				{
					//we have found the parameters and their ranges
					s.nextLine();
				}
				else if(line[0].startsWith("Param"))
				{
					for(int i = 0; i < Constants.nparams; i++)
					{
						String[] s2 = s.nextLine().split(",");
						params[i] = Double.parseDouble(s2[4]);
						params[i+Constants.nparams] = Double.parseDouble(s2[5]);
						params[i+2*Constants.nparams] = Double.parseDouble(s2[6]);
					}
					answers.add(Methods.duplicate(params));
				}
			}
			//now lets calculate the
			return answers;
		} catch (FileNotFoundException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		return null;
		
	}
}
