package common;

import java.awt.Color;
import java.awt.Dimension;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.geom.AffineTransform;
import java.awt.image.BufferedImage;
import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;

import javax.imageio.ImageIO;
import javax.swing.JPanel;

public class Dendogramer extends JPanel
{
	protected ArrayList<? extends FCS> rs = null;
	protected int tab = 15;
	protected BufferedImage compen;
	
	private TreeNode copy = null;
	protected TreeNode selected;
	protected ArrayList<TreeNode> list = new ArrayList<TreeNode>();
	protected ArrayList<Integer> xoffs = new ArrayList<Integer>();
	protected HashMap<TreeNode,NodeState> nodeStates = new HashMap<TreeNode,NodeState>();
	

	
	//protected HashMap<TreeNode,BufferedImage> icons = new HashMap<TreeNode,BufferedImage>();  //should be at most the size of the tab (will be shrunk accordingly!)
	
	public void setState(TreeNode tn, boolean collapsed, int color)
	{
		if(nodeStates.get(tn) == null) nodeStates.put(tn, new NodeState(collapsed,color));
		nodeStates.get(tn).collapsed = collapsed;
		nodeStates.get(tn).color = color;
		System.err.println("Color set to "+color);
	}
	
	public void setState(TreeNode tn, int color)
	{
		
		if(nodeStates.get(tn) == null) nodeStates.put(tn, new NodeState(false,color));
		
		nodeStates.get(tn).color = color;
	}

	/*
	public void addIcon(TreeNode tn, BufferedImage img)
	{
		icons.put(tn,img);
	}
	*/
	
	public void updateSize()
	{
		Dimension size = getSize();
		if(rs != null)
			setPreferredSize(new Dimension(size.width,tab*list.size()));
	//	updateUI();
	}
	
	public void updateUI()
	{
		updateSize();
		super.updateUI();
	}
	
	public int getTab() {
		return tab;
	}

	public TreeNode getCopy() {
		return copy;
	}

	public void setCopy(TreeNode copy) {
		this.copy = copy;
	}

	public TreeNode getSelected() {
		return selected;
	}

	public void setSelected(TreeNode selected) {
		this.selected = selected;
	}

	public ArrayList<? extends FCS> getRs() {
		return rs;
	}

	public ArrayList<TreeNode> getList() {
		return list;
	}

	public void setRs(ArrayList<? extends FCS> rs) {
		this.rs = rs;
	}

	public Dendogramer()
	{
		try {
			compen =Methods.getImageFromResource("CompensateTiny.png");
		} catch (Exception e) {
			// TODO Auto-generated catch block
			
		}
	}

	public void paintComponent(Graphics g)
	{
		super.paintComponent(g);
		Graphics2D g2 = (Graphics2D)g;
		g2.setColor(new Color(150,150,150));
		g2.fillRect(0,0,getWidth(), getHeight());
		g2.setColor(new Color(220,220,220));
		g2.fillRect(1,1,getWidth()-2, getHeight()-2);
		list.clear();
		xoffs.clear();
		g2.setColor(Color.black);
		if(rs == null||rs.size()==0)
		{
			g2.drawString("No data to show yet.", 10, 10);
			return;
		}
		int current_y = 0;
		for(int i = 0; i < rs.size(); i++)
		{
			if(rs.get(i) != null &&rs.get(i).getRoot() != null)
			{
				if(rs.get(i).getRoot().getGate() == null)
				{
					g2.drawImage(compen, getWidth()-35,current_y,null);
					
					if(rs.get(i).getRoot().equals(selected))
						g2.setColor(Color.BLUE);
					else
						g2.setColor(Color.BLACK.brighter());
					if(rs.get(i).getTime() >= 0)
						drawShadowText(g2,rs.get(i).getExperiment()+" @ "+rs.get(i).getTime()+" hrs.",tab+1,current_y+tab-5);
					else
						drawShadowText(g2,rs.get(i).getExperiment(),tab+1,current_y+tab-5);
					
					
					/*
					if(icons.get(rs.get(i).getRoot())!= null)
					{
						BufferedImage bi = icons.get(rs.get(i).getRoot());
						double scalex = tab/((double)bi.getWidth());
						double scaley = tab/((double)bi.getHeight());
						AffineTransform at = AffineTransform.getTranslateInstance(0, current_y-10);
						at.scale(scalex, scaley);
						
						g2.drawImage(bi,at,this);
					}
					*/

				}
				list.add(rs.get(i).getRoot());
				xoffs.add(0);
				current_y = recursiveDraw(rs.get(i).getRoot(),0,current_y,g2)+tab;
			}
		}
	}
	
	private void drawShadowText(Graphics2D g2, String text, int x, int y)
	{
		Color prev = g2.getColor();
		Color shadow = new Color(prev.getRed(),prev.getGreen(),prev.getBlue(),50);
		g2.setColor(shadow);
		g2.drawString(text, x+1, y+1);
		g2.setColor(prev);
		g2.drawString(text, x, y);
	}
	
	private int recursiveDraw(TreeNode n, int xoff, int yoff, Graphics2D g2)
	{
		int current_y = yoff;
		Iterator<TreeNode> itr = n.leaves.iterator();
		//draw self then draw children
		if(n.getGate() != null)
		{
			if(selected != null && selected.equals(n))
				g2.setColor(Color.BLUE);
			else
				g2.setColor(Color.BLACK);
				drawShadowText(g2,n.getGate().name,xoff+tab,current_y+tab-5);
			if(n.getGate().c != null)
				g2.setColor(n.getGate().c);
			else
				g2.setColor(Color.WHITE);
			g2.fillOval(getWidth()-35, current_y, tab, tab);
			g2.setColor(Color.BLACK);
			g2.drawOval(getWidth()-35, current_y, tab, tab);

		}
		boolean collapsed = false;
		Color fill = Color.LIGHT_GRAY;
		if(nodeStates.get(n) != null)
		{
			if(nodeStates.get(n).color < 0)
				fill = new Color((nodeStates.get(n).color<<8)>>8);
			else
				fill = new Color(nodeStates.get(n).color);
			System.err.println("Color was: "+nodeStates.get(n).color);
			collapsed = nodeStates.get(n).collapsed;
		}
		if(n.getLeaves().size() > 0)
		{
			g2.setColor(fill);
			g2.fillRect(1+xoff, yoff+1, tab-1, tab-2);
			
			g2.setColor(Color.BLACK);
			g2.drawRect(1+xoff, yoff+1, tab-2, tab-2);
			g2.drawLine(3+xoff,yoff+tab/2,tab-3+xoff,yoff+tab/2);
			if(!collapsed)
			{
				g2.drawLine(tab/2+xoff,yoff+3,tab/2+xoff,yoff+tab-3);
			}
			else
			{
				while(itr.hasNext())
				{
					TreeNode n2 = itr.next();
					list.add(n2);
					xoffs.add(xoff+tab/2);
					current_y= recursiveDraw(n2,xoff+tab/2,current_y+tab,g2);
				}
			}
		}
		return current_y;
	}
	
	public void Click(int x, int y)
	{
		try{
		int xoff = xoffs.get(y/tab);
		if(x > xoff && x < xoff+tab)
		{
			TreeNode click = getClicked(y);
			if(nodeStates.get(click)==null)nodeStates.put(click,new NodeState(false,Color.LIGHT_GRAY.getRGB()));
			nodeStates.get(click).collapsed = !nodeStates.get(click).collapsed;
			updateUI();
		}
		}catch(IndexOutOfBoundsException e)
		{
			//do nothing
		}
	}
	
	public TreeNode getClicked(int x, int y)
	{
		System.err.println("Clicked: "+x+","+y);
		try{
			int xoff = xoffs.get(y/tab);
			if(x > xoff+tab)
			{
				try{
					return list.get((y)/tab);
				}catch(IndexOutOfBoundsException e)
				{
					return null;
				}
			}
			return null;
		}
		catch(IndexOutOfBoundsException e)
		{
			return null;
		}
	}
	
	private TreeNode getClicked(int y)
	{
		try{
			return list.get((y)/tab);
		}catch(IndexOutOfBoundsException e)
		{
			return null;
		}
	}
	
	public NodeState getState(TreeNode tn)
	{
		return nodeStates.get(tn);
	}

}