package common;

import java.awt.Color;
import java.io.File;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.LinkedList;

public class FCS 
{
	private File sourceFile;
	//protected static final int max_points_shown = 100000;

	private String[] names = null;
	private int total = 0;
	private String date = "";
	private int volume = 1000000;
	protected Matrix inverse;
	protected LinkedList<Point> data;
	protected LinkedList<Point> compdata;
	private String experiment = "Unnamed experiment";
	private String condition = ""; //this is what is set by the user to signify that it is part of one condition
	protected TreeNode root;
	protected int[] maxes = null;
	private FCSSettings settings;
	protected int num_par = 1;
	private double time = -1;
	private boolean compensation_applied = false;
	
	public boolean isLoaded() {
		return data.size()==total;
	}

	public int getNum_par() {
		return num_par;
	}
	
	public LinkedList<Point> getCompensatedPoints()
	{
		if(!compensation_applied)
		{
			compdata = new LinkedList<Point>();
			Iterator<Point> itr = data.iterator();
			System.err.println("FCS.getCompensatedPoints");
			while(itr.hasNext())
			{
				Point p = itr.next();
				Point cpoint = new Point(p.ps.length);
				for(int j = 0; j < inverse.getNrow(); j++)
				{
					for(int k = 0; k < inverse.getNcol(); k++)
					{
						cpoint.ps[j]+=inverse.getElement(j,k)*p.ps[k];  //apply compensation inversion matrix
					}
				}
				cpoint.color = p.color;
				compdata.add(cpoint);
			}
			System.err.println("FCS.getCompensatedPoints");
			compensation_applied = true;
		}
		
		return compdata;
	}

	private void compensatePoint(Point p)
	{
		Point cpoint = new Point(p.ps.length);
		for(int j = 0; j < inverse.getNrow(); j++)
		{
			for(int k = 0; k < inverse.getNcol(); k++)
			{
				cpoint.ps[j]+=inverse.getElement(j,k)*p.ps[k];  //apply compensation inversion matrix
			}
		}
		p.ps = cpoint.ps;
	}



	public FCS(File source)
	{
		sourceFile = source;
	}
	
	/**
	 * Shallow copy
	 * @param copy
	 */
	public FCS(FCS copy)
	{
		this.sourceFile = copy.sourceFile;
		this.names = copy.names;
		this.total = copy.total;
		this.date = copy.date;
		this.volume = copy.volume;
		this.inverse = copy.inverse;
		this.data = copy.data;
		this.experiment = copy.experiment;
		this.condition = copy.condition;
		this.root = copy.root;
		this.maxes = copy.maxes;
		this.settings = copy.settings;
		this.num_par = copy.num_par;
		this.time = copy.time;
	}
	
	/**
	 * Standard constructor when sorting!
	 * @param name
	 */
	public FCS(String name, int EVENT_PARAMETERS, String[] initial_param_names, FCSSetting s) {
		experiment = name;
		maxes = new int[EVENT_PARAMETERS];
		names = new String[EVENT_PARAMETERS];
		settings = new FCSSettings(EVENT_PARAMETERS);
		try{
		for(int i = 0; i < EVENT_PARAMETERS; i++)
		{
			maxes[i] = s.max;
			names[i] = new String(initial_param_names[i]);
			
			settings.settings[i].amp_decades = s.amp_decades;
			settings.settings[i].amp_zero = s.amp_zero;
			settings.settings[i].bits= s.bits;
			settings.settings[i].max = s.max;
			settings.settings[i].min = s.min;
			settings.settings[i].range = s.range;
			settings.settings[i].usefulBits = s.usefulBits;
		}
		}catch (Exception e)
		{
			System.err.println("Got naming exception in FCS constructor");
		}
		date = Methods.now();
		//
		initializeDataSize(EVENT_PARAMETERS);
	}


	public String getExperiment() {
		return experiment;
	}

	public void setTime(double dtime) {
		// TODO Auto-generated method stub
		time = dtime;
	}

	public void setExperiment(String experiment) {
		this.experiment = experiment;
	}

	public void initializeDataSize(int num_par)
	{
		this.num_par = num_par;
		inverse = Matrix.identityMatrix(num_par);
		data = new LinkedList<Point>();
		maxes = new int[num_par];
	}
	
	public void initializeTrueRoot()
	{
		TreeNode trueRoot = new TreeNode(getArrayData(), this);
		root = new TreeNode(new LinkedList<Point>(),this);
		trueRoot.leaves.add(root);
		root.setParent(trueRoot);
	}
	
	public LinkedList<Point> getArrayData()
	{
		if(data == null) data = new LinkedList<Point>();
		LinkedList<Point> result = new LinkedList<Point>();
		for(int i = 0; i < data.size(); i++)
			result.add(data.get(i));
		return result;
	}
	
	public String[] getNames() {
		return names;
	}
	public void setNames(String[] names) {
		this.names = names;
	}
	public int getTotal() {
		return total;
	}
	public void setTotal(int total) {
		this.total = total;
	}

	public void setVolume(int volume) {
		this.volume = volume;
	}
	public void setSpillover(String spillover) {
			inverse = Matrix.getInverse(spillover);
	}


	public void setNumPar(int num_par) {
		// TODO Auto-generated method stub
		this.num_par = num_par;
		maxes= new int[num_par];
		names = new String[num_par];
		inverse = Matrix.identityMatrix(num_par);
		for(int i = 0; i < num_par; i++)
			names[i] = "Par "+(i+1);
	}


	public void appendDataPoint(int[] next) {
		//we have to add the value to the array
		Point p = new Point(next);
		data.add(p);
		p.color = -1;
		int color = getColor(root, p);
		p.color = color;
		for(int i = 0; i < maxes.length; i++)
		{
			if(next[i] > maxes[i]) maxes[i] = next[i];
		}
		compensation_applied = false;
	}

	public void setInverse(Matrix mat) {
		inverse = mat;
		compensation_applied  = false;
		
	}
	
	/**
	 * Go through all points and assign color based on gate
	 */
	public void propagateColors()
	{
		Iterator<Point> citr = getCompensatedPoints().iterator();
		Iterator<Point> itr = data.iterator();
		while(itr.hasNext()&&citr.hasNext())
		{
			Point p = citr.next();
			p.color = -1;
			int color = getColor(root, p);
			p.color = color;
			Point uncompensated = itr.next();
			uncompensated.color = color;
		}
	}
	
	public void reloadFromFCS()
	{
		if(sourceFile != null)
		{
			FCS copy = Methods.processFCSFile(sourceFile);
			this.sourceFile = copy.sourceFile;
			this.names = copy.names;
			this.total = copy.total;
			this.date = copy.date;
			this.volume = copy.volume;
			//this.inverse = copy.inverse; the inverse comes from workspace files, not from FCS files (although some FCS also have that)
			this.data = copy.data;
			this.experiment = copy.experiment;
			//this.condition = copy.condition;
			this.maxes = copy.maxes;
			this.settings = copy.settings;
			this.num_par = copy.num_par;
			//this.time = copy.time;
			compensation_applied  = false;
		}
		else
		{
			
		}
	}
	
	private int getColor(TreeNode tn, Point p)
	{
		int red = 0;
		int green = 0;
		int blue = 0;
		int count = 0;
		Iterator<TreeNode> itr = tn.leaves.iterator();
		while(itr.hasNext())
		{
			TreeNode next = itr.next();
			if(next.getGate() != null && next.getGate().checkInside(p))
			{

				int col = getColor(next, p);
				if(col != -1)
				{
					
					red+=(col>>16)&0x000000FF;
					green+=(col>>8)&0x000000FF;
					blue+=col&0x000000FF;
					count++;
				}
			}
		}
		if(count > 0)
		{
			return ((red/count)<<16)|((green/count)<<8)|(blue/count);
		}
		else
		{
			if(tn.getGate() != null && tn.getGate().c != null)
				return tn.getGate().c.getRGB();
			else
				return 0;
		}
	}
	
	public File getSourceFile() {
		return sourceFile;
	}

	public void setSourceFile(File sourceFile) {
		this.sourceFile = sourceFile;
	}

	public String getDate() {
		return date;
	}

	public void setDate(String date) {
		this.date = date;
	}

	public String getCondition() {
		return condition;
	}

	public void setCondition(String condition) {
		this.condition = condition;
	}

	public int getVolume() {
		return volume;
	}

	public Matrix getInverse() {
		return inverse;
	}

	public LinkedList<Point> getData() {
		return data;
	}

	public TreeNode getRoot() {
		if(root==null)
		{
			initializeTrueRoot();
		}
		return root;
	}

	public int[] getMaxes() {
		return maxes;
	}

	public FCSSettings getSettings() {
		return settings;
	}

	public double getTime() {
		return time;
	}

	public void setSettings(FCSSettings fcsSettings) {
		this.settings = fcsSettings;
		
	}

	public void setData(MaxList<Point> data2) {
		data = data2;
		
	}

	public void setMaxes(int[] newmaxes) {
		maxes = newmaxes;
		
	}

	public void setRoot(TreeNode treeNode) {
		root = treeNode;
		
	}



}

