package common;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.PrintWriter;
import java.util.ArrayList;
import java.util.Scanner;

import javax.swing.JFrame;

/**
 * Wrapper class for most FCS related storage, saving, and loading routines. All GUIs should extends this and add the actual graphical components.
 * @author Max
 *
 */
public abstract class GUIGuts extends JFrame implements PlotterListener {
	protected ArrayList<FCS> runs = new ArrayList<FCS>();

	public static final double version = 2.0;
	
	public void saveEverything(File f, Dendogramer tree)
	{
		//save each FCS file followed by each TreeNode
		try {
			PrintWriter pw = new PrintWriter(f);
			char fsep = System.getProperty("file.separator").charAt(0);
			String working_dir = (new File(".")).getAbsolutePath();
			for(int i = 0; i < tree.rs.size(); i++)
			{
				
				pw.print(Methods.getRelativePath(tree.rs.get(i).getSourceFile().getAbsolutePath(),working_dir).replace(fsep, '|')+";"+tree.rs.get(i).getCondition()+";"+tree.rs.get(i).getTime()+";"+tree.rs.get(i).getVolume());
				pw.println();
				//now lets save the matrix
				pw.println("COMPENSATION");
				for(int j = 0; j < tree.rs.get(i).getInverse().getNrow(); j++)
				{
					pw.print(tree.rs.get(i).getInverse().getElement(j, 0));
					for(int k = 1; k < tree.rs.get(i).getInverse().getNcol(); k++)
					{
						pw.print(","+tree.rs.get(i).getInverse().getElement(j, k));
					}
					pw.println();
				}
			}
			pw.println("TREE"); //breadth first
			for(int i = 0; i < tree.rs.size(); i++)
			{
				tree.rs.get(i).getRoot().saveSelf(pw);
			//	pw.println("DONE");
			}
			pw.close();
		} catch (FileNotFoundException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
	}

	public abstract void plotterUpdated(Plotter p);
	

}
