package common;

import java.awt.Color;
import java.awt.Dimension;
import java.awt.FlowLayout;
import java.awt.Font;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.util.ArrayList;
import java.util.Iterator;

import javax.swing.ImageIcon;
import javax.swing.JPanel;


public abstract class Gallery extends JPanel
{
	protected ArrayList<Plotter> plots = new ArrayList<Plotter>();
	protected Plotter current_plot = null; // the current plotter accessed (clicked)
	protected GUIGuts owner;
	
	public Gallery(GUIGuts owner)
	{
		this.owner = owner;
	}

	public ArrayList<Plotter> getPlots() {
		return plots;
	}


	
	public Plotter getCurrent_plot() {
		return current_plot;
	}

	public void paintComponent(Graphics g)
	{
		super.paintComponent(g);
		Graphics2D g2 = (Graphics2D)g;
		g2.setColor(Color.WHITE);
		g2.fillRect(0,0,this.getWidth(),this.getHeight());
		if(plots.size()==0)
		{
			drawInstructions(g2);
		}
	}
	
	/**
	 * Overload this class to implement your own version of the instruction screen.
	 * @param g2
	 */
	public abstract void drawInstructions(Graphics2D g2);
	
	public void makeCurrentPlot(TreeNode tn) {
		boolean found = false;
		for(int i = 0; i < plots.size(); i++)
		{
			if(plots.get(i).node != null && plots.get(i).node.equals(tn))
			{
				current_plot = plots.get(i);
				found = true;
				break;
			}
		}
		if(!found)
		{
			Plotter p = new Plotter(this,tn);
			addNewGraph(p);
			current_plot = p;
		}
		current_plot.updateDrawParams();

	}

	public void removeGraph(TreeNode tn, boolean children)
	{
		if(children)
		{
			for(TreeNode t : tn.leaves)
			{
				removeGraph(t,children); //remove all children as well since they no longer exist
			}
		}
		for(int i = 0; i < plots.size(); i++)
		{
			if(plots.get(i).node.equals(tn)) //found the plot with the appropriate node...
			{
				
				if(current_plot.equals(plots.get(i)))
				{
					if(i == plots.size()-1)
					{
						if(i > 0)
							current_plot = plots.get(i-1);
						else
							current_plot = null;
					}
					else
						current_plot = plots.get(i+1);
				}
				plots.remove(i);
				break;
			}
		}
		
		repackGraphs();
	}
	
	public void addNewGraph(Plotter p)
	{
		if(!plots.contains(p))
		{
			plots.add(p);
			p.registerPlotterListener(owner);
			p.setPreferredSize(new Dimension(320,350));
			repackGraphs();
			current_plot = p;
		}

	}
	
	public void repackGraphs()
	{
		int graph_width  =(getVisibleRect().width/320);
		removeAll();
		setLayout(new FlowLayout(FlowLayout.LEFT,1,1));
		if(graph_width ==0) graph_width = 1;
		Dimension d = new Dimension(320*graph_width,350*(int)Math.ceil((((double)plots.size())/graph_width)));
		setPreferredSize(d);
		System.err.println("D:"+d.width+" "+d.height+" grpah_width = "+graph_width);
		
		for(int i = 0; i < plots.size(); i++)
		{
			add(plots.get(i));
		}

		invalidate();
		updateUI();
	}
	
	
	public void updateAll()
	{
		for(int i = 0; i < plots.size(); i++)
		{
			plots.get(i).grf.redraw_needed = true;
			plots.get(i).grf.updateUI();
			plots.get(i).stats.refresh();
		}
		updateUI();
	}






	public void clearPlots() {
		plots.clear();
	}



	public void setCurrent_plot(Plotter p) {
		current_plot = p;
		
	}

	public void refreshAllStats() {
		for(Plotter p : plots)
		{
			p.stats.refresh();
		}
	}
	
}
