package common;


import java.awt.Color;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.LinkedList;

/**
 * A wrapper class describing a two-dimensional polygon (see DoublePolygon) or a 1D range 
 * @author Max
 *
 */
public class Gate{
	public String name = "";
	public int x, y;
	public DoublePolygon p = new DoublePolygon();
	public Color c = Color.BLACK;  //the color that all points in this gate should have!!!
	public boolean logx;
	public boolean logy;
	public int gatetype = 0; //gateType is: 0 = polygon gate, 1 = rectangle gate, 2 = 1D gate
	
	public Gate(int param, boolean logx, double min, double yvalue, String name)
	{
		gatetype = 2;
		this.x = param;
		this.y = (int) yvalue;
		this.logx = logx;
		this.name = name;
		p.addPoint(min, min);
	}
	
	public boolean checkInside(Point pt) {
		try
		{
			if(gatetype < 2)
			{
				return inside(pt.ps[x], pt.ps[y]);
			}
			else return inside(pt.ps[x]);
		}
		catch(Exception e)
		{
			return false;
		}
	}

	public Gate clone()
	{
		Gate clone= new Gate(x,y,logx,logy,new String(name),c.getRGB());
		if(gatetype==2)
		{
			clone.gatetype=2;
		}
		for(int i = 0; i < p.xpoints.length; i++)
			clone.addPoint(p.xpoints[i], p.ypoints[i]);
		return clone;
	}
	
	
	public void update1DMax(double max)
	{
		if(gatetype == 2)p.ypoints[0] = max;
		if(p.ypoints[0] < p.xpoints[0])
		{
			double temp = p.xpoints[0];
			p.xpoints[0] = p.ypoints[0];
			p.ypoints[0] = temp;
		}
	}
	
	public Gate(int param1, int param2, boolean logx, boolean logy,String name, int col)
	{
		this.logx=logx;
		this.logy=logy;
		this.x = param1;
		this.y = param2;
		this.name = name;
		this.c = new Color(col);
	}
	
	public void setGateType(int type)
	{
		gatetype = type;
	}
	
	public LinkedList<Point> getGatedPoints(LinkedList<Point> ppoints)
	{
		LinkedList<Point> result = new LinkedList<Point>();
		if(p.getNpoints() < 3 && gatetype < 2) return new LinkedList<Point>(); //we are trying to gate on a non-polygon so return a blank set.
		Iterator<Point> itr = ppoints.iterator();
		while(itr.hasNext())
		{
			Point p = itr.next();
			if(gate(p.ps))
			{
				result.add(p);
//				if(c != null)
//					points.get(i).color = c.getRGB();
			}
		}
		return result;
	}
	
	
	
	public void addPoint(double x2, double y2)
	{
		p.addPoint(x2,y2);
	}
	

	
	private void translate1D(double dx, double dy)
	{
		p.xpoints[0]+=dx;
		p.ypoints[0]+=dx;
		y+=dy;
		System.err.println("Translating");
	}
	
	
	/**
	 * Moves all points in the DoublePolygon by x and y
	 * @param x
	 * @param y
	 */
	public void translate(double x, double y)
	{
		if(gatetype <2)
			p.translate(x, y);
		else
			translate1D(x,y);
		
	}
	
	/**
	 * Updates just one point at the specified index
	 * @param index
	 * @param x
	 * @param y
	 */
	public void adjust(int index, double x, double y)
	{
		//System.err.println("Adjust: "+index+" x "+x+" y "+y);
		if(gatetype < 2)
			p.adjust(index,x,y);
		else
		{
			p.adjust1D(index,x);
			this.y=(int) y;
		}
	}
	
	public double getDistanceToGate(double px, double py, double xmax, double ymax, double xmin, double  ymin)
	{
		if(gatetype < 2)
			return p.getMinimumDistanceToSegments(px, py,xmax, ymax, xmin, ymin);
		else
			return p.distanceToSegment(p.xpoints[0], y, p.ypoints[0], y, px, py, xmax, ymax, xmin, ymin);
	}
	
	public int getClosestVertexWithinD(double px, double py,double xmax, double ymax, double xmin, double  ymin,double d)
	{
		if(gatetype < 2)
			return p.getClosestVertexWithinD(px, py, xmax,  ymax, xmin,  ymin,d);
		else // 1D
		{
			double mind = p.distanceToVertext(p.xpoints[0], y, px, py, xmax, ymax, xmin, ymin, d);
			double maxd = p.distanceToVertext(p.ypoints[0], y, px, py, xmax, ymax, xmin, ymin, d);
			System.err.println("min d "+mind+" max d "+maxd);
			if( mind<maxd)
				return 0;  //the minimum vertex
			else
				return 1;  // the max vertex
		}
	}
	
	public boolean deletePoint(int index)
	{
		if(index < p.getNpoints() && p.getNpoints() > 3)
		{
			p.deletePoint(index);
			return true;
		}
		return false;
	}
	
	/**
	 * Only called in the case of a 1D gate so just one value
	 * @param val
	 * @return
	 */
	public boolean inside(double val)
	{
		double min = Math.min(p.xpoints[0], p.ypoints[0]);
		double max = Math.max(p.xpoints[0], p.ypoints[0]);
		return (val >= min && val <= max);
	}
	
	public boolean inside(double x2, double y2)
	{
		//lets check if a point in inside the gate!!!! Only
		return p.inside(x2, y2);
	}
	
	/**
	 * Returns true if the event lies within this gate
	 * @param event
	 * @return
	 */
	public boolean gate(int[] event)
	{
		double x2 = event[x];
		
		double y2 = event[0];
		if( gatetype<2) y2 = event[y];
		else
		{
			return inside(x2);
		}
		return inside(x2,y2);
	}

	public void updateLastPoint(double pointx, double pointy) {
		p.xpoints[p.getNpoints()-1]=pointx;
		p.ypoints[p.getNpoints()-1]=pointy;
	}

}
