package common;

import java.awt.BasicStroke;
import java.awt.Color;
import java.awt.Font;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.image.BufferedImage;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.LinkedList;

import javax.swing.JPanel;



public class Graph extends JPanel {
	
	protected int maxh = 1;
	protected TreeNode node = null;
	protected static final int max_num_points = 50000;

	protected boolean redraw_needed = true;
	protected BufferedImage img = null;	
	protected Gate currentGate = null;
	public int currentVertex = -1;
	public int events = 0;
	public String gateName = "";
	public String expName = "";
	protected Plotter parent;
	protected int ymargin = 45;
	protected int xmargin = 25;
	
	
	
	public TreeNode getNode() {
		return node;
	}

	public void setNode(TreeNode node) {
		this.node = node;
	}

	public boolean isRedraw_needed() {
		return redraw_needed;
	}

	public void setRedraw_needed(boolean redraw_needed) {
		this.redraw_needed = redraw_needed;
	}

	public Gate getCurrentGate() {
		return currentGate;
	}

	public void setCurrentGate(Gate currentGate) {
		this.currentGate = currentGate;
	}

	public Plotter getParent() {
		return parent;
	}

	public void setParent(Plotter parent) {
		this.parent = parent;
	}

	public Graph(TreeNode node, Plotter parent)
	{
		this.node = node;
		this.parent = parent;
	}
	
	protected void drawOnGraphics(Graphics2D g2, boolean drawAxes)
	{
	//	g2.setColor(Color.WHITE);
	//	g2.fillRect(0,0,this.getWidth(),this.getHeight());
		if(node == null )
		{
			g2.setColor(Color.gray);
			g2.drawString("Please pick a data set or gate to display.", 11, 11);
			g2.drawLine(2,6,9,6);
			g2.drawLine(2,6,5,3);
			g2.drawLine(2,6,5,9);
			g2.setColor(Color.black);
			g2.drawString("Please pick a data set or gate to display.", 10, 10);
			g2.drawLine(1,5,8,5);
			g2.drawLine(1,5,4,2);
			g2.drawLine(1,5,4,8);
			return;
		}
		if(node.xmax == 0 || node.ymax==0)
		{
			node.changeXIndex(node.xindex);
			node.changeYIndex(node.yindex);
		}
		//draw the treenode
		//System.err.println("Start draw");

		if(redraw_needed)
		{
			//System.err.println("Redrawing everything");
			maxh = 1;
			redraw_needed = false;

			if(node.yindex == -1)
			{
				//we are drawing a 1d histogram
				int[] h = new int[this.getWidth()-ymargin];
				int[] hr = new int[this.getWidth()-ymargin];
				int[] hg = new int[this.getWidth()-ymargin];
				int[] hb = new int[this.getWidth()-ymargin];
				LinkedList<Point> ps = node.getPoints();
				Iterator<Point> itr = ps.iterator();
				int counter = 0;
				while(itr.hasNext() && counter++ < max_num_points)
				{
					Point p = itr.next();
					int x = p.ps[node.xindex];
					int hx;
					double xmin2, xmax2;
					xmin2 = node.xmin;
					xmax2 = node.xmax;
					if(node.xlog)
					{
						xmin2 = Math.log10(node.xmin);
						xmax2 =Math.log10(node.xmax);
						double x2 = Math.log10(x);
						double f = (x2-xmin2)/(xmax2-xmin2);
						hx = (int)(f*h.length);
					}
					else
					{
						hx = (int) ((((double)(x-xmin2))/(xmax2-xmin2))*(h.length));
					}
					try{
						h[hx]++;
						if(h[hx] > maxh) 
							maxh =h[hx];
						hr[hx]+=new Color(p.color).getRed();
						hg[hx]+=new Color(p.color).getGreen();
						hb[hx]+=new Color(p.color).getBlue();
					}catch(Exception e)
					{
						
					}
				}
				//draw image
				//System.err.println("maxh = "+maxh+" total points = "+node.getPoints().length);
				img = new BufferedImage(this.getWidth()-ymargin,this.getHeight()-xmargin,BufferedImage.TYPE_INT_ARGB);
				Graphics2D g3  =img.createGraphics();
				for(int i = 0; i < h.length; i++)
				{
					int count = h[i];
					int col = (int)((double)(count*(img.getHeight()))/maxh);
					Color c = new Color(0,0,0,200);
					if(count > 0) c = new Color(hr[i]/count,hg[i]/count,hb[i]/count,200);
					g3.setColor(c);
					for(int j = 1; j < img.getHeight(); j++)
					{
						if(j <= col)
						{
							g3.drawLine(i, ((img.getHeight())-j), i, ((img.getHeight())-j));
						}
					}
				}
				//System.err.println("Max count = "+maxcount);
				//also we want to draw the model overlay if there is such a thing
			}
			else
			{
				//calculate 2d histo
				int[][] h = new int[this.getWidth()-ymargin][this.getHeight()-xmargin];
				int[][] hr = new int[this.getWidth()-ymargin][this.getHeight()-xmargin];
				int[][] hg = new int[this.getWidth()-ymargin][this.getHeight()-xmargin];
				int[][] hb = new int[this.getWidth()-ymargin][this.getHeight()-xmargin];
				LinkedList<Point> ps = node.getPoints();

				if(ps!= null)
				{
					Iterator<Point> itr = ps.iterator();
					int count = 0;
					while(itr.hasNext() && count++ < max_num_points)
					{
						Point p = itr.next();
						double x = p.ps[node.xindex];
						double y = p.ps[node.yindex];
						int hx, hy;
						double xmin2, ymin2,xmax2,ymax2;
						xmin2 = node.xmin;
						ymin2 = node.ymin;
						xmax2 = node.xmax;
						ymax2 = node.ymax;
						if(node.xlog)
						{
							xmin2 = Math.log10(node.xmin);
							xmax2 =Math.log10(node.xmax);
							double x2 = Math.log10(x);
							double f = (x2-xmin2)/(xmax2-xmin2);
							hx = (int)(f*hr.length);
						}
						else
						{
							hx = (int) (((x-xmin2)/(xmax2-xmin2))*(h.length));
						}
						if(node.ylog)
						{
							ymin2 = Math.log10(node.ymin);
							ymax2 =Math.log10(node.ymax);
							double y2 = Math.log10(y);
							double f = (y2-ymin2)/(ymax2-ymin2);
							hy = (int)(f*h[0].length);
						}
						else
						{	
							hy = (int) ((((y-ymin2))/(ymax2-ymin2))*(h[0].length));
						}
						if(hx > -1 && hx < h.length && hy > -1 && hy < h[hx].length)
						{
							h[hx][hy]++;
							hr[hx][hy]+= new Color(p.color).getRed();
							hg[hx][hy]+= new Color(p.color).getGreen();
							hb[hx][hy]+= new Color(p.color).getBlue();
							if(h[hx][hy] > maxh) maxh =h[hx][hy];
						}
					}
				}
				System.err.println("Draw image");
				//draw image
				img = new BufferedImage(this.getWidth()-ymargin,this.getHeight()-xmargin,BufferedImage.TYPE_INT_ARGB);
				for(int i = 0; i < hr.length; i++)
				{
					for(int j = 0; j < hr[i].length; j++)
					{
						int count = h[i][j];
						int countr = hr[i][j];
						int countg = hg[i][j];
						int countb = hb[i][j];
						float frac = ((float)count)/maxh;
						float fracr = ((float)countr)/(count*256);
						float fracg = ((float)countg)/(count*256);
						float fracb = ((float)countb)/(count*256);
						float rc = 0;
						float gc = 0;
						float bc = 0;
						if(ps.size() > 1000 && maxh > 3)
						{
							rc = Math.min(1f,Math.max(0,fracr));
							gc = Math.min(1f,Math.max(0,fracg));
							bc = Math.min(1f,Math.max(0,fracb));
						}
						if(count == 0)
							img.setRGB(i, img.getHeight()-j-1, Color.WHITE.getRGB());
						else
						{
							Color c = new Color(rc,gc,bc,Math.min(1,frac*0.7f+0.5f));
							int rgb = c.getRGB();
							img.setRGB(i, img.getHeight()-j-1,rgb );
						}
					}
				}

			}
			if(node != null &&node.getPoints() != null) events = node.getPoints().size();
			if(node != null && node.getGate() != null &&node.getGate().name != null) gateName = node.getGate().name;
			if(node != null && node.getSourceFCS() != null) expName = node.getSourceFCS().getExperiment();
			System.err.println("Redrawing image for graph with "+events+" events");
		}

		g2.drawImage(img, ymargin, 0, this);
		g2.setFont(new Font("Calibri",Font.PLAIN,13));
		g2.setColor(Color.RED);

//		g2.drawString(String.format("%20s %20s %10d",expName,gateName,events), 2+ymargin, 8);
		g2.setFont(new Font("Calibri",Font.PLAIN,11));
		//now lets draw the gates!
		//System.err.println("Draw gates");
		g2.setColor(Color.BLUE.darker());
		for(int i = 0; i < node.leaves.size(); i++)
		{
			Gate gate = node.leaves.get(i).getGate();
			if(gate.gatetype < 2)
			{
				if(gate.x == node.xindex && gate.y == node.yindex)
				{
					int[] xs = new int[gate.p.getNpoints()];
					int[] ys = new int[gate.p.getNpoints()];
					for(int j = 0; j < gate.p.getNpoints(); j++)
					{
						double x = gate.p.xpoints[j];
						double y = gate.p.ypoints[j];
						int hx, hy;
						double xmin2, ymin2,xmax2,ymax2;
						xmin2 = node.xmin;
						ymin2 = node.ymin;
						xmax2 =node. xmax;
						ymax2 = node.ymax;
						if(node.xlog)
						{
							xmin2 = Math.log10(node.xmin);
							xmax2 =Math.log10(node.xmax);
							double x2 = Math.log10(x);
							double f = (x2-xmin2)/(xmax2-xmin2);
							hx = (int)(f*img.getWidth());
						}
						else
						{
								hx = (int) ((((double)(x-xmin2))/(xmax2-xmin2))*(img.getWidth()));
						}
						if(node.ylog)
						{
							ymin2 = Math.log10(node.ymin);
							ymax2 =Math.log10(node.ymax);
							double y2 = Math.log10(y);
							double f = (y2-ymin2)/(ymax2-ymin2);
							hy = (int)(f*img.getHeight());
						}
						else
						{
								hy = (int) ((((double)(y-ymin2))/(ymax2-ymin2))*(img.getHeight()));
						}
						xs[j] = hx+ymargin;
						ys[j] = img.getHeight()-hy;
					}
					g2.drawPolygon(xs, ys, xs.length);
					g2.drawString(gate.name, xs[0]-20, (int)Math.max(ys[0]-15,15));
					if(node != null &&node.getPoints() != null && node.leaves.get(i) != null && node.leaves.get(i).getPoints() != null)
						g2.drawString(String.format("%1.2f",100*((double)node.leaves.get(i).getPoints().size())/node.getPoints().size())+"%", xs[0]-20, (int)Math.max(ys[0]-5,25));
				}
			}
			else //1D gate
			{
				if(node.yindex == -1 && node.xindex == gate.x)
				{
					//we draw this 1D gate
					double low = gate.p.xpoints[0];
					double high = gate.p.ypoints[0];
					int gatey = gate.y;
					int hx1,hx2;
					double xmin2, xmax2;
					xmin2 = node.xmin;
					xmax2 =node. xmax;
					
					if(node.xlog)
					{
						xmin2 = Math.log10(node.xmin);
						xmax2 =Math.log10(node.xmax);
						double x = Math.log10(low);
						double f = (x-xmin2)/(xmax2-xmin2);
						hx1 = (int)(f*img.getWidth());
						double x2 = Math.log10(high);
						double f2 = (x2-xmin2)/(xmax2-xmin2);
						hx2 = (int)(f2*img.getWidth());
					}
					else
					{
							hx1 =(int) ((((double)(low-xmin2))/(xmax2-xmin2))*(img.getWidth()));
							hx2 = (int) ((((double)(high-xmin2))/(xmax2-xmin2))*(img.getWidth()));
					}
					hx1 = hx1+ymargin;
					hx2 = hx2+ymargin;
					g2.drawLine(hx1,gatey,hx2,gatey);
					g2.drawLine(hx1,gatey-5,hx1,gatey+5);
					g2.drawLine(hx2,gatey-5,hx2,gatey+5);
					g2.drawString(gate.name, (hx1+hx2)/2-20, gatey-15);
					if(node != null &&node.getPoints() != null && node.leaves.get(i) != null && node.leaves.get(i).getPoints() != null)
						g2.drawString(String.format("%1.2f",100*((double)node.leaves.get(i).getPoints().size())/node.getPoints().size())+"%", (hx1+hx2)/2-20, gatey-25);
				}
			}
		}
		//lets draw the active gate currently
		if(currentGate != null)
		{
			g2.setColor(Color.GREEN.darker());
			Gate gate = currentGate;
			if(gate.gatetype < 2)  //polygon
			{
				int[] xs = new int[gate.p.getNpoints()];
				int[] ys = new int[gate.p.getNpoints()];
				for(int j = 0; j < gate.p.getNpoints(); j++)
				{
					double x = gate.p.xpoints[j];
					double y = gate.p.ypoints[j];
					int hx, hy;
					double xmin2, ymin2,xmax2,ymax2;
					xmin2 = node.xmin;
					ymin2 = node.ymin;
					xmax2 = node.xmax;
					ymax2 = node.ymax;
					if(node.xlog)
					{
						xmin2 = Math.log10(node.xmin);
						xmax2 =Math.log10(node.xmax);
						double x2 =Math.log10(x);
						double f = (x2-xmin2)/(xmax2-xmin2);
						hx = (int)(f*img.getWidth());
					}
					else
					{
						hx = (int) ((((double)(x-xmin2))/(xmax2-xmin2))*(img.getWidth()));
					}
					if(node.ylog)
					{
						ymin2 = Math.log10(node.ymin);
						ymax2 =Math.log10(node.ymax);
						double y2 =  Math.log10(y);
						double f = (y2-ymin2)/(ymax2-ymin2);
						hy = (int)(f*img.getHeight());
					}
					else
					{
						hy = (int) ((((double)(y-ymin2))/(ymax2-ymin2))*(img.getHeight()));
					}
					xs[j] = hx+ymargin;
					ys[j] = img.getHeight()-hy;
					//System.err.println("Drawing at "+xs[j]+","+ys[j]);
				}
				if(currentVertex != -1)
				{
					g2.setStroke(new BasicStroke(2f));
					for(int i = 0; i < xs.length; i++)
					{
						if(i == currentVertex)	g2.fillOval(xs[i]-3, ys[i]-3, 7, 7);
						else g2.fillOval(xs[i]-1, ys[i]-1, 3, 3);
					}
				}
				g2.drawPolyline(xs, ys, xs.length);
				if(currentVertex != -1)
					g2.setStroke(new BasicStroke());
			}
			else
			{
				//1D current gate so just draw the extended line
				
				g2.drawLine(node.zoomx1,node.zoomy1,node.zoomx2,node.zoomy1);
				g2.drawLine(node.zoomx1,node.zoomy1-5,node.zoomx1,node.zoomy1+5);
				g2.drawLine(node.zoomx2,node.zoomy1-5,node.zoomx2,node.zoomy1+5);
			}
		}
		//lets draw the zoomboox
		if(node.zoomx1 != -1 && node.zoomy1 != -1 && node.zoomx2 !=-1 && node.zoomy2 != -1 && (parent.mode==1||parent.mode==3))
		{
			if(parent.mode == 1)
				g2.setColor(Color.gray);
			else
				g2.setColor(Color.green);
			int zx1 = Math.min(node.zoomx1, node.zoomx2);
			int zx2 = Math.max(node.zoomx1, node.zoomx2);
			int zy1 = Math.min(node.zoomy1,node. zoomy2);
			int zy2 = Math.max(node.zoomy1, node.zoomy2);
			g2.drawRect(zx1, zy1, zx2-zx1, zy2-zy1);
		}		
		if(drawAxes)
		{
			g2.setColor(Color.BLACK);
			g2.drawLine(ymargin,getHeight()-xmargin,getWidth(),getHeight()-xmargin);
			g2.drawLine(ymargin,0,ymargin,getHeight()-xmargin);
			//lets draw the axes and label them
			g2.setColor(Color.BLACK);
			g2.setFont(new Font("Times",Font.PLAIN,10));
			//lets draw the axes labels
			String xlabel = node.getSourceFCS().getNames()[node.xindex];
			if(node.yindex > -1)
			{
				String ylabel = node.getSourceFCS().getNames()[node.yindex];
				if(ylabel != null)drawRotatedText(Math.PI/2, ylabel, 10, getHeight()/2, g2);
				if(xlabel != null)g2.drawString(xlabel, getWidth()/2 - xlabel.length()*2, getHeight());
			}
			else  //1D histogram
			{
				drawRotatedText(Math.PI/2,"# Evts.", 10,getHeight()/2 - 25,g2);
				if(xlabel != null)g2.drawString("log10("+xlabel+")", getWidth()/2 - xlabel.length()*2, getHeight()-2);
			}
	//-=-=-=-=-=-=-=-=-= AXES -=-=-=-=-=-=-=-=-=-=-=
			int num_ticks = Math.min(Math.max(2,getWidth()/50),10);
			for(int i = 0; i < num_ticks+1; i++)
			{
				//draw the x axis ticks
				double dx = (((double)i)/num_ticks)*(getWidth()-ymargin);
				g2.drawLine(ymargin+((int)dx),getHeight()-xmargin,ymargin+((int)dx),getHeight()-(xmargin)+5);
				//draw the y axis ticks
				double dy = (((double)i)/num_ticks)*(getHeight()-xmargin);
				g2.drawLine(ymargin,(int)dy,ymargin-5,(int)dy);
				//draw the x axis tick label
				if(!node.xlog)
				{
					g2.drawString(String.format("%3.1e",(node.xmin+(((double)i)/num_ticks)*(node.xmax-node.xmin))),ymargin+((int)dx),getHeight()-(xmargin)+15);
				}
				else
				{
					double xmin2 = Math.log10(node.xmin);
					double xmax2 =Math.log10(node.xmax);
					g2.drawString(String.format("%3.1f",xmin2+((xmax2-xmin2)*((((double)i)/num_ticks)))),ymargin+((int)dx),getHeight()-(xmargin)+15);
					//g2.drawString("10",ymargin+((int)dx)-3,getHeight()-(xmargin)+25);
				}
				//draw the y axis tick label
				if(i < num_ticks)
				{
					if(node.yindex> -1)
					{
						if(!node.ylog)
						{
							g2.drawString(String.format("%3.1e",(node.ymin+(((double)i)/num_ticks)*(node.ymax-node.ymin))),ymargin-25,(int) ((((double)num_ticks-i)/num_ticks)*(getHeight()-xmargin)));
						}
						else
						{
							double ymin2 = Math.log10(node.ymin);
							double ymax2 =Math.log10(node.ymax);
		
							g2.drawString(String.format("%3.1f",ymin2+((ymax2-ymin2)*((((double)i)/num_ticks)))),ymargin-22,(int) ((((double)num_ticks-i)/num_ticks)*(getHeight()-xmargin)));
							//g2.drawString("10",ymargin-30,(int) ((((double)num_ticks-i)/num_ticks)*(getHeight()-xmargin)+8));
						}
					}
					else
					{
						g2.drawString(String.format("%d",(int)((((double)i)/num_ticks)*(maxh))),ymargin-25,(int) ((((double)num_ticks-i)/num_ticks)*(getHeight()-xmargin)));
					}
				}
			}

			//lets draw the x box
			g2.setColor(new Color(200,160,160));
			g2.fillRect(2, 0, 10, 10);
			g2.setColor(Color.black);
			g2.drawRect(2, 0, 10, 10);
			g2.drawLine(5,3,9,7);
			g2.drawLine(5,7,9,3);
			g2.setColor(Color.LIGHT_GRAY);
			g2.drawLine(3, 11, 13, 11);
			g2.drawLine(13, 1, 13, 11);

		}
	}
	
	public BufferedImage getCopyImage()
	{
		BufferedImage bi = new BufferedImage(getWidth(),getHeight(),BufferedImage.TYPE_INT_ARGB);
		Graphics2D g2 = bi.createGraphics();
		drawOnGraphics(g2,true);
		return bi;
	}
	
	public BufferedImage getHistogramOnlyImage()
	{
		BufferedImage bi = new BufferedImage(getWidth(),getHeight(),BufferedImage.TYPE_INT_ARGB);
		Graphics2D g2 = bi.createGraphics();
		drawOnGraphics(g2,false);
		return bi;
	}
	
	public void paintComponent(Graphics g)
	{
		super.paintComponent(g);
		Graphics2D g2 = (Graphics2D)g;
		g2.setColor(Color.white);
		g2.fillRect(0, 0, getWidth(), getHeight());
		drawOnGraphics(g2,true);
		g2.setColor(Color.gray);
		g2.drawLine(0,0,0,getHeight());
	}
	
	private void drawRotatedText(double theta, String string, int i, int j,
			Graphics2D g2) {
		g2.translate(i,j);
		g2.rotate(theta);
		g2.drawString(string,0,0);
		g2.rotate(-theta);
		g2.translate(-i, -j);
		
	}

	public int getXMargin() {
		// TODO Auto-generated method stub
		return xmargin;
	}
	
	public int getYMargin() {
		// TODO Auto-generated method stub
		return ymargin;
	}
}
