package common;
import java.io.Serializable;


public class IntPolygon implements Serializable{
	public int[] xpoints = new int[0];
	public int[] ypoints = new int[0];
	public int maxx  = 0;
	public int maxy = 9999999;
	public int minx = 0;
	public int miny = 99999999;
	public int npoints=0;
	
	public int getNpoints() {
		return npoints;
	}
	
	public IntPolygon clone()
	{
		IntPolygon clone = new IntPolygon();
		for(int i = 0; i < xpoints.length; i++)
			clone.addPoint(xpoints[i], ypoints[i]);
		return clone;
	}
	
	public void translate(int x, int y)
	{
		for(int i  = 0; i < npoints; i++)
		{
			movePoint(i, xpoints[i]+x,ypoints[i]+y);
		}
		recalculateBoundingBox();
	}

	private void movePoint(int index, int x, int y)
	{
		if(index <npoints){
			xpoints[index] = x;
			ypoints[index] = y;
		}
		if(index == 0)
		{
			xpoints[xpoints.length-1]=x;
			ypoints[xpoints.length-1]=y;
		}
	}
	
	public void updateLastPoint(int pointx, int pointy) {
		xpoints[npoints-1]=pointx;
		ypoints[npoints-1]=pointy;
	}
	
	public void adjust(int index, int x, int y)
	{
		movePoint(index,x,y);
		recalculateBoundingBox();
	}
	
	private void recalculateBoundingBox()
	{
		maxx  = 0;
		maxy = 0;
		minx = 999999;
		miny =999999;
		for(int i = 0; i < npoints; i++)
		{
			if(xpoints[i] > maxx )maxx = xpoints[i];
			else if(xpoints[i] < minx )minx = xpoints[i];
			if(ypoints[i] > maxy )maxy = ypoints[i];
			else if(ypoints[i] < miny )miny = ypoints[i];
		}
	}
	
	public void addPoint(int x, int y)
	{
		npoints++;
		int[] newx = new int[npoints];
		int[] newy = new int[npoints];
		//String str = "";
		for(int i = 0; i < xpoints.length; i++)
		{
			newx[i] = xpoints[i];
			newy[i] = ypoints[i];
			if(newx[i] > maxx)maxx = newx[i];
			if(newx[i] < minx) minx = newx[i];
			if(newy[i] > maxy)maxy = newy[i];
			if(newy[i] < miny) miny = newy[i];
		//	str+="("+newx[i]+","+newy[i]+")";
		}
		newx[npoints-1] = x;
		newy[npoints-1] = y;
		if(newx[npoints-1] > maxx)maxx = newx[npoints-1];
		if(newx[npoints-1] < minx) minx = newx[npoints-1];
		if(newy[npoints-1] > maxy)maxy = newy[npoints-1];
		if(newy[npoints-1] < miny) miny = newy[npoints-1];
		xpoints = newx;
		ypoints = newy;
	//	str+="("+newx[npoints-1]+","+newy[npoints-1]+")";
		//System.err.println("NPOINTS = "+npoints+" "+str);
	}
	
	public void deletePoint(int index)
	{
		if(index < npoints-1)
		{
			int[] xs = new int[xpoints.length-1]; //remember that points contains all of the points plus an extra one at the end that is a copy of the first
			int[] ys = new int[ypoints.length-1];
			maxx  = 0;
			maxy = 0;
			minx = 999999;
			miny =999999;
			for(int i = 0; i < xpoints.length-1; i++)
			{
				if(i < index)
				{
					xs[i] = xpoints[i];
					ys[i] = ypoints[i];
					if(xs[i] > maxx) maxx = xs[i];
					if(xs[i] < minx) minx = xs[i];
					if(ys[i] > maxy) maxy = ys[i];
					if(ys[i] < miny) miny = ys[i];
				}
				else if( i > index)
				{
					xs[i-1] = xpoints[i];
					ys[i-1] = ypoints[i];
					if(xs[i-1] > maxx) maxx = xs[i-1];
					if(xs[i-1] < minx) minx = xs[i-1];
					if(ys[i-1] > maxy) maxy = ys[i-1];
					if(ys[i-1] < miny) miny = ys[i-1];
				}
			}
			xs[xs.length-1] = xs[0];
			ys[ys.length-1] = ys[0];
			xpoints = xs;
			ypoints = ys;
			npoints--;
		}
	}
	
	public int getMinimumDistanceToSegments(int x,int y, int xmax, int ymax, int xmin, int  ymin)
	{
		int d = 99999;
		for(int i = 0; i < npoints-1;i++)
		{
			int x1 = xpoints[i];
			int y1 = ypoints[i];
			int x2 = 0;
			int y2 = 0;
			if(i < npoints-1)
			{
				x2 = xpoints[i+1];
				y2 = ypoints[i+1];
			}
			int dist = distanceToSegment(x1,y1,x2,y2,x,y,xmax,ymax,xmin,ymin);
			if(dist >= 0 && dist < d) d = dist;
		}
		return d;
	}
	
	public int getClosestVertexWithinD(int x, int y, int xmax, int ymax, int xmin, int ymin, int max)
	{
		int bestIndex = -1;
		int d = 99999999;
		for(int i = 0; i < npoints;i++)
		{
			int dist = distanceToVertext(xpoints[i], ypoints[i], x, y, xmax, ymax, xmin, ymin, max);
			if(dist < d && dist < max)
			{
				d = dist;
				bestIndex = i;
			}
		}
		return bestIndex;
	}
	
	public int distanceToVertext(int vx, int vy,int x, int y, int xmax, int ymax, int xmin, int ymin, int max )
	{
		x = (x-xmin)/(xmax-xmin);
		y = (y-ymin)/(ymax-ymin);
		int x1 = (vx-xmin)/(xmax-xmin);
		int y1 = (vy-ymin)/(ymax-ymin);
		return (x-x1)*(x-x1)+(y-y1)*(y-y1);
	}
	
	public boolean inside(int x, int y)
	{
		if(npoints < 3) return false;

		//lets check if the point is within the bounding box
		if(x > maxx || x <= minx || y > maxy || y < miny) return false;
		int left = 0;
		int right = 0;
		for(int i = 0; i < npoints-1; i++)
		{
			int result =  intersect(xpoints[i],ypoints[i], xpoints[i+1], ypoints[i+1], x, y);
			if(result == -1) left++;
			else if(result == 1) right++;
			else if(result == 0) return true;
		}
		return (left%2==1)&&(right%2==1);
	}
	
	public int distanceToSegment(int x1, int y1, int x2, int y2, int x, int y, int xmax, int ymax, int xmin, int  ymin)
	{
		x1 -= xmin;
		x2 -= xmin;
		y1 -= ymin;
		y2 -= ymin;
		x -= xmin;
		y -= ymin;
		y1 /= (ymax-ymin);
		y2 /= (ymax-ymin);
		y /= (ymax-ymin);
		x1 /= (xmax-xmin);
		x2 /= (xmax-xmin);
		x/= (xmax-xmin);
		if(x2 != x1 && y2 != y1)
		{
			int m = (y2-y1)/(x2-x1);
			int m2 = -1/m;
			//lets solve for b
			int b = y1-x1*m;
			int b2 = y-x*m2;
			int xint = (b2-b)/(m-m2);
			int yint = m*xint+b;
			//check if xint on the line
			if (xint >= Math.min(x2,x1) && xint <= Math.max(x2,x1))
			{
				return (int)Math.sqrt((x-xint)*(x-xint)+(y-yint)*(y-yint));
			}
			else
			{
				return (int)Math.min( Math.sqrt((x-x1)*(x-x1)+(y-y1)*(y-y1)),  Math.sqrt((x-x2)*(x-x2)+(y-y2)*(y-y2)));
			}
		}
		else if(x2 == x1)
		{
			//we have an infinite slope for the segment we are looking at
			if(y <= Math.max(y1, y2) && y >= Math.min(y1, y2))
			{
				return Math.abs(x-x1); //horizontal distance to the line
			}
			else
				return (int)Math.min( Math.sqrt((x-x1)*(x-x1)+(y-y1)*(y-y1)),  Math.sqrt((x-x2)*(x-x2)+(y-y2)*(y-y2)));
		}
		else
		{
			//we have a horizontal line segment
			if(x <= Math.max(x1,x2) && x >= Math.min(x1,x2))
				return Math.abs(y-y1);  //vertical distance to the line
			else
				return (int)Math.min( Math.sqrt((x-x1)*(x-x1)+(y-y1)*(y-y1)),  Math.sqrt((x-x2)*(x-x2)+(y-y2)*(y-y2)));
		}
	}
	
	
	private int intersect(int x1, int y1, int x2, int y2, int x, int y)
	{
		//check if there is an intersection
		if(x2 != x1 && y2 != y1)
		{
			int m = (y2-y1)/(x2-x1);
			//lets solve for b
			int b = y1-m*x1;			
			int yint = y;
			int xint = (yint-b)/m;
			//check if yint on the line
			if (xint >= Math.min(x2,x1) && xint <= Math.max(x2,x1))
			{
				if(xint < x) return -1;
				else if(xint > x) return 1;
			}
			else
			{
				return -3;  //total fail!
			}
		}
		else if(x2 == x1)
		{
			//we have an infinite slope for the segment we are looking at
			if(y <= Math.max(y1, y2) && y >= Math.min(y1, y2))
			{
				if(x1 < x) return -1;
				else if(x1 > x) return 1;
				else if (x1==x) return 0;
			}
			else
				return -3;
		}
		else if(y1 == y2)
		{
			//we have a horizontal line segment
			if(x <= Math.max(x1,x2) && x >= Math.min(x1,x2))
				return -3; //don't count me
			else
				return -3;
		}
		return 0;
	}
	
	private boolean intersect(double x1, double y1, double x2, double y2, double x3, double y3, double x4, double y4)
	{
		//lets find the intersection point of these two lines.
		if(x1==x2) x2= x2+0.0000001;
		if(x4==x3) x3 = x3+0.0000001;
		double ma = (y2-y1)/(x2-x1);
		double mb = (y4-y3)/(x4-x3);
		if(ma==mb) ma+=0.0000001;
		double ba = y1-ma*x1;
		double bb = y3-mb*x3;
		double xint = (bb-ba)/(ma-mb);
		double yint = ma*xint+ba;
		return yint < Math.max(y2,y1)&&yint > Math.min(y2,y1) && xint < Math.max(x2,x1)&&xint > Math.min(x2,x1) &&
		yint < Math.max(y4,y3)&&yint > Math.min(y4,y3) && xint < Math.max(x4,x3)&&xint > Math.min(x4,x3);
	}

	public void adjust1D(int index, int x) {
		// TODO Auto-generated method stub
		//lets just adjust the one point here
		if(index==0) 
		{
			xpoints[0]=x;
		}
		else if(index==1)
		{
			ypoints[0]=x;
		}
	}
	
	public int getArea()
	{
		int count = 0;
		for(int i = minx; i <= maxx; i++)
		{
			for(int j = miny; j <= maxy; j++)
			{
				if(inside(i,j)) count++;
			}
		}
		return count;
	}
}


