package common;


public class Matrix {

	private double[][] matrix;


    private int permutationIndex[] = null;          // row permutation index
    private double rowSwapIndex = 1.0D;             // row swap index
    private double tiny = 1.0e-100;                 // small number replacing zero in LU decomposition
	
    public static Matrix identityMatrix(int size)
    {
    	Matrix m = new Matrix(size,size);
    	for(int i = 0; i < size; i++)
    		m.matrix[i][i] = 1;
    	m.permutationIndex = new int[size];
    	return m;
    }
    
    public Matrix(int r, int c)
    {
    	matrix = new double[r][c];
    	permutationIndex = new int[r];
    }
    
    public Matrix(double[][] arr)
    {
    	matrix = arr;
    	permutationIndex = new int[arr.length];
    }
    
	public static Matrix getInverse(String spillover)
	{
		//lets parse this and turn it into an inverse (should be comma-separated subtraction values for all 4 colors)
		
		String[] split = spillover.split(",");
		int mat_size = (int)Math.sqrt(split.length);
		double[][] s = new double[mat_size][mat_size];
		for(int i = 0; i < split.length; i++)
		{
			if(i%mat_size==0)System.err.println();
			if(i/mat_size == i%mat_size)
				s[i/mat_size][i%mat_size] = 1+(0.01*Double.parseDouble(split[i]));
			else
				s[i/mat_size][i%mat_size] = (0.01*Double.parseDouble(split[i]));
			System.err.print(s[i/mat_size][i%mat_size]+" ");
			
		}
		Matrix comp = new Matrix(s);
		Matrix inverse =comp.inverse();
		return inverse;
	}
    
    public Matrix times(Matrix b)
    {
    	Matrix c = new Matrix(getNrow(), b.getNcol());
    	try{
    		for(int i = 0; i < getNrow(); i++)
    			for(int j = 0; j < b.getNcol(); j++)
    				for(int k = 0; k < getNcol(); k++)
    				{
    					c.matrix[i][j]+=matrix[i][k]*b.matrix[k][j];
    				}
    		return c;
    	}
    	catch(Exception e)
    	{
    		return null;
    	}
    }
    
	// INVERSE
	// Inverse of a square matrix [instance method]
	public Matrix inverse(){
    	int n = getNrow();
    	if(n!=getNcol())throw new IllegalArgumentException("Matrix is not square");
    	Matrix invmat = new Matrix(n, n);

        if(n==1){
            double[][] hold = this.getArrayCopy();
            if(hold[0][0]==0.0)throw new IllegalArgumentException("Matrix is singular");
            hold[0][0] = 1.0/hold[0][0];
            invmat = new Matrix(hold);
        }
        else{
            if(n==2){
                double[][] hold = this.getArrayCopy();
                double det = hold[0][0]*hold[1][1] - hold[0][1]*hold[1][0];
                if(det==0.0)throw new IllegalArgumentException("Matrix is singular");
                double[][] hold2 = new double[2][2];
                hold2[0][0] = hold[1][1]/det;
                hold2[1][1] = hold[0][0]/det;
                hold2[1][0] = -hold[1][0]/det;
                hold2[0][1] = -hold[0][1]/det;
                invmat = new Matrix(hold2);
            }
            else{
    	        double[] col = new double[n];
    	        double[] xvec = new double[n];
    	        double[][] invarray = invmat.getMatrix();
    	        Matrix ludmat;

    	        ludmat = this.luDecomp();
    	        for(int j=0; j<n; j++){
        		    for(int i=0; i<n; i++)col[i]=0.0D;
        		    col[j]=1.0;
        		    xvec=ludmat.luBackSub(col);
        		    for(int i=0; i<n; i++)invarray[i][j]=xvec[i];
    	        }
    	    }
    	}
   		return invmat;
	}
    
	public double[][] getArrayCopy() {
		double[][] copy = new double[getNrow()][getNcol()];
		for(int i = 0; i < matrix.length; i++)
		{
			for(int j = 0; j < matrix[i].length; j++)
			{
				copy[i][j] = matrix[i][j];
			}
		}
		return copy;
	}

	public Matrix copy()
	{
		return new Matrix(getArrayCopy());
	}
	
	// LU DECOMPOSITION OF MATRIX A
	// For details of LU decomposition
	// See Numerical Recipes, The Art of Scientific Computing
	// by W H Press, S A Teukolsky, W T Vetterling & B P Flannery
    // Cambridge University Press,   http://www.nr.com/
    // This method has followed their approach but modified to an object oriented language
    // Matrix ludmat is the returned LU decompostion
    // int[] index is the vector of row permutations
    // rowSwapIndex returns +1.0 for even number of row interchanges
    //       returns -1.0 for odd number of row interchanges
    private Matrix luDecomp(){
    	if(getNrow()!=getNcol())throw new IllegalArgumentException("A matrix is not square");
    	int n = getNrow();
    	int imax = 0;
    	double dum = 0.0D, temp = 0.0D, big = 0.0D;
    	double[] vv = new double[n];
    	double sum = 0.0D;
    	double dumm = 0.0D;


      	Matrix ludmat = copy();
    	double[][] ludarray = ludmat.getMatrix();

		ludmat.rowSwapIndex=1.0D;
    	for (int i=0;i<n;i++) {
	    	big=0.0D;
	    	for (int j=0;j<n;j++)if  ((temp=Math.abs(ludarray[i][j])) > big) big=temp;
    		if (big == 0.0D){
     	    	for(int k=0;k<n;k++)for(int j=0;j<n;j++)ludarray[k][j]=Double.NaN;
     	    	return ludmat;
     		}
			vv[i]=1.0/big;
    	}
    	for (int j=0;j<n;j++) {
	    	for (int i=0;i<j;i++) {
		    	sum=ludarray[i][j];
		    	for (int k=0;k<i;k++) sum -= ludarray[i][k]*ludarray[k][j];
		    	ludarray[i][j]=sum;
	    	}
	    	big=0.0D;
	    	for (int i=j;i<n;i++) {
				sum=ludarray[i][j];
    			for (int k=0;k<j;k++)
			    	sum -= ludarray[i][k]*ludarray[k][j];
		    		ludarray[i][j]=sum;
 					if ((dum=vv[i]*Math.abs(sum)) >= big) {
			        	big=dum;
			        	imax=i;
		    		}
	    		}
	    		if (j != imax) {
		    	for (int k=0;k<n;k++) {
			    	dumm=ludarray[imax][k];
			    	ludarray[imax][k]=ludarray[j][k];
					ludarray[j][k]=dumm;
		    	}
		    	ludmat.rowSwapIndex = -ludmat.rowSwapIndex;
				vv[imax]=vv[j];
    		}
	    	ludmat.permutationIndex[j]=imax;

	    	if(ludarray[j][j]==0.0D){
	        	ludarray[j][j]=this.tiny;
	    	}
	    	if(j != n-1) {
		    	dumm=1.0/ludarray[j][j];
				for (int i=j+1;i<n;i++){
			    		ludarray[i][j]*=dumm;
    	    		}
    		}
    	}
    	return ludmat;
    }
    
	// Solves the set of n linear equations A.X=B using not A but its LU decomposition
	// bvec is the vector B (input)
	// xvec is the vector X (output)
	// index is the permutation vector produced by luDecomp()
	public double[] luBackSub(double[] bvec){
    	int ii = 0,ip = 0;
    	int n=bvec.length;
    	if(n!=getNcol())throw new IllegalArgumentException("vector length is not equal to matrix dimension");
    	if(getNcol()!=getNrow())throw new IllegalArgumentException("matrix is not square");
    	double sum= 0.0D;
    	double[] xvec=new double[n];
    	for(int i=0; i<n; i++){
        	xvec[i]=bvec[i];
    	}
    	for (int i=0;i<n;i++) {
	    	ip=this.permutationIndex[i];
	    	sum=xvec[ip];
	    	xvec[ip]=xvec[i];
	    	if (ii==0){
		    	for (int j=ii;j<=i-1;j++){
		        	sum -= this.matrix[i][j]*xvec[j];
		    	}
		    }
	    	else{
	        	if(sum==0.0) ii=i;
	    	}
	    	xvec[i]=sum;
    	}
    	for(int i=n-1;i>=0;i--) {
	    	sum=xvec[i];
	    	for (int j=i+1;j<n;j++){
	        	sum -= this.matrix[i][j]*xvec[j];
	    	}
	    	xvec[i]= sum/matrix[i][i];
    	}
    	return xvec;
	}


	
	public double[][] getMatrix() {
		return matrix;
	}

	public int getNumberOfColumns() {
		try{
			return matrix[0].length;
		}catch(IndexOutOfBoundsException e)
		{
			return 0;
		}
	}

	public int getNumberOfRows() {
		return matrix.length;
	}
	
	public int getNrow()
	{
		return matrix.length;
	}
	
	public int getNcol()
	{
		return getNumberOfColumns();
	}

	public double getElement(int j, int k) {
		// TODO Auto-generated method stub
		return matrix[j][k];
	}
	
	public void setElement(int j, int k, double value) {
		// TODO Auto-generated method stub
		try{
			matrix[j][k] = value;
		}catch (IndexOutOfBoundsException e)
		{
			
		}
	}


}
