package common;
import java.awt.BasicStroke;
import java.awt.Color;
import java.awt.Font;
import java.awt.Graphics2D;
import java.awt.image.BufferedImage;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Calendar;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Random;

import javax.imageio.ImageIO;
import javax.swing.ImageIcon;



public class Methods {

	
	public static final double KolmogorovCDF(int n, double d)
	{
		double z = d/Math.sqrt(n);
		double[] fj = {-2.0,-8.0,-18.0,-32.0};
		double[] r = new double[4];
		final double w = 2.50662827;
		   // c1 - -pi**2/8, c2 = 9*c1, c3 = 25*c1
		final double c1 = -1.2337005501361697;
		final double c2 = -11.103304951225528;
		final double c3 = -30.842513753404244;

		 double u = Math.abs(z);
		 double p;
		   if (u < 0.2) {
		      p = 1;
		   } else if (u < 0.755) {
		      double v = 1./(u*u);
		      p = 1 - w*(Math.exp(c1*v) + Math.exp(c2*v) + Math.exp(c3*v))/u;
		   } else if (u < 6.8116) {
		      r[1] = 0;
		      r[2] = 0;
		      r[3] = 0;
		      double v = u*u;
		      int maxj = (int)Math.max(1,Math.round(3./u));
		      for (int j=0; j<maxj;j++) {
		         r[j] = Math.exp(fj[j]*v);
		      }
		      p = 2*(r[0] - r[1] +r[2] - r[3]);
		   } else {
		      p = 0;
		   }
		   return p;
	}
	
	/**
	 * Returns either the image in the current jar/folder or a blank image with nothing but a red X
	 * @param name
	 * @return
	 */
	public static final BufferedImage getImageFromResource(String name)
	{
		try {
			BufferedImage img = ImageIO.read(Methods.class.getClassLoader().getResource(name));
			return img;
		} catch (Exception e) {
			BufferedImage img = new BufferedImage(150,150,BufferedImage.TYPE_INT_ARGB);
			Graphics2D g2 = (Graphics2D) img.getGraphics();
			g2.setColor(Color.BLACK);
			g2.setFont(new Font("Calibri",Font.PLAIN,10));
			g2.drawString(name, 2, 1);
			g2.setColor(Color.RED);
			g2.drawLine(0,0,img.getWidth(),img.getHeight());
			g2.drawLine(img.getWidth(),0,0,img.getHeight());
			
			return img;
		}
	}
	
	/**
	 * Returns the wrapped ImageIcon in the current dir/jar or a blank image if not found.
	 * @param name
	 * @return
	 */
	public static final ImageIcon getImageIcon(String name)
	{
		return new ImageIcon(getImageFromResource(name));
	}
	
	
	public static String getRelativePath(String string, String other)
	{
		byte[] a = string.getBytes();
		byte[] b = other.getBytes();
		int ind = 0;
		String result = "";
		while(ind < Math.min(a.length,b.length))
		{
			if(a[ind]!=b[ind]) break;
			ind++;
		}
		String suffix = string.substring(ind,string.length());

		byte sep = System.getProperty("file.separator").getBytes()[0];
		if(suffix.indexOf(":") != -1)
			return suffix;
		while(ind < b.length)
		{
			if(b[ind] == sep)
				result+=".."+(char)sep;
			ind++;
		}
		if(result.length() == 0) result = "."+(char)sep;
		return result+suffix;
	}
	
	
	/**
	 * Returns the non-common part of string when compared to other
	 * @param string
	 * @param other
	 * @return
	 */
	public static String getUniqueString(String string, String other)
	{
		byte[] a = string.getBytes();
		byte[] b = other.getBytes();
		int ind = 0;
		while(ind < Math.min(a.length,b.length))
		{
			if(a[ind]!=b[ind]) break;
			ind++;
		}

		return string.substring(ind,string.length());
	}
	
	
	private static Random rng = new Random(
		      Calendar.getInstance().getTimeInMillis() +
		      Thread.currentThread().getId());
	
	public static double sampleGamma(double k, double theta) {
	    boolean accept = false;
	    if (k < 1) {
	 // Weibull algorithm
	 double c = (1 / k);
	 double d = ((1 - k) * Math.pow(k, (k / (1 - k))));
	 double u, v, z, e, x;
	 do {
	  u = rng.nextDouble();
	  v = rng.nextDouble();
	  z = -Math.log(u);
	  e = -Math.log(v);
	  x = Math.pow(z, c);
	  if ((z + e) >= (d + x)) {
	   accept = true;
	  }
	 } while (!accept);
	 return (x * theta);
	    } else {
	 // Cheng's algorithm
	 double b = (k - Math.log(4));
	 double c = (k + Math.sqrt(2 * k - 1));
	 double lam = Math.sqrt(2 * k - 1);
	 double cheng = (1 + Math.log(4.5));
	 double u, v, x, y, z, r;
	 do {
	  u = rng.nextDouble();
	  v = rng.nextDouble();
	  y = ((1 / lam) * Math.log(v / (1 - v)));
	  x = (k * Math.exp(y));
	  z = (u * v * v);
	  r = (b + (c * y) - x);
	  if ((r >= ((4.5 * z) - cheng)) ||
	                    (r >= Math.log(z))) {
	   accept = true;
	  }
	 } while (!accept);
	 return (x * theta);
	    }
	  }

	public static FCS processFCSFile(File f)  {

		FCS fcs= new FCS(f);
		System.err.println("Processing: "+fcs.getExperiment());
		try
		{
			int num_par = 14;
			FileInputStream s = new FileInputStream(f);
			int start_text,end_text,start_data,end_data;
			getNextString(10,s);  //version
			start_text = getNextInteger(8,s);
			end_text = getNextInteger(8,s);
			start_data=getNextInteger(8,s);
			end_data=getNextInteger(8,s);
			System.err.println("Start text: "+start_text+" end text: "+end_text+" start data "+start_data+" end data "+end_data);
			//lets fast forward to the start of the text section
			long c = 42;
			s.skip(start_text-c);
			c=start_text;
			char del = getNextString(1,s).charAt(0);
			System.err.println("DELIMITER:\""+del+"\"");
			String keyword = "";
			String value = "";
			int mode = 0;  //0= in keyword, 1= in value
			while(c!=end_text)
			{
				char ch = (char)s.read();
				c++;
				if(ch==del && mode==0)
				{
					mode = 1;
				}
				else if(ch == del && mode==1)
				{
					mode = 0;
					//evaluate keyword/value pair
					int val = 0;
					try{
						val = Integer.parseInt(value.trim());
					}
					catch(Exception e)
					{
						//do nothing since we probably just didn't get an int and will be using the value!
					}
					System.err.println("KEYWORD:"+keyword+" VALUE: "+value);
					if(keyword.startsWith("$P")&&keyword.endsWith("S"))
					{
						int param = Integer.parseInt(keyword.substring(2,keyword.length()-1));
						fcs.getNames()[param-1]=value;
					}
					else if(keyword.startsWith("$PAR"))
					{
						num_par=val;
						fcs.setSettings(new FCSSettings(num_par));
						fcs.setNumPar(num_par);
					
					}
					else if(keyword.startsWith("$P"))
					{
						
						if(keyword.endsWith("B"))  //bits total (i.e. 32 for integer)
						{
							int index = Integer.parseInt(keyword.substring(2,keyword.indexOf('B')));
							fcs.getSettings().settings[index-1].bits = val;
						}
						else if(keyword.endsWith("E")) //amplification
						{
							int index = Integer.parseInt(keyword.substring(2,keyword.indexOf('E')));
							fcs.getSettings().settings[index-1].amp_decades = Integer.parseInt(value.split(",")[0]);
							fcs.getSettings().settings[index-1].amp_zero = Double.parseDouble(value.split(",")[1]);
						}
						else if(keyword.endsWith("S")) //name of param
						{
							int index = Integer.parseInt(keyword.substring(2,keyword.indexOf('S')));
							fcs.getSettings().settings[index-1].name = value;
						}
						else if(keyword.endsWith("R"))  //range of the parameter
						{
							int index = Integer.parseInt(keyword.substring(2,keyword.indexOf('R')));
							fcs.getSettings().settings[index-1].range = val;
							fcs.getSettings().settings[index-1].calculateUsefulBits();
						}
					}
					else if(keyword.startsWith("$TOT"))
					{
						fcs.setTotal(val);
					}
					else if(keyword.startsWith("#SAMPLE"))
					{
						fcs.setExperiment(value);
					}
					else if(keyword.startsWith("SAMPLE ID"))
					{
						fcs.setExperiment(value);
					}
					else if(keyword.startsWith("$VOL"))
					{
						fcs.setVolume(val);
					}
					keyword ="";
					value="";
				}
				else if(mode==0)
				{
					keyword+=ch;
				}
				else if(mode==1)
				{
					value+=ch;
				}
			}
			System.err.println("Start data: "+start_data+" End data: "+end_data+" numpar "+num_par+" current position "+c+" going to skip "+(start_data-c-1));
			if(start_data-c-1 > 0)
				s.skip(start_data-c-1);
			else
				s.skip(1);
			c= start_data;
			//process data
			fcs.initializeDataSize(num_par);
			//System.err.println("Initialize data to hold "+(fcs.getData().size()-1)+" events");
			int count = 0;
			int oldper=0;
			byte[] buff = new byte[end_data-start_data+1];
			int b = 0;
			s.read(buff);
			
			
			while(c < end_data)
			{				
				int per = (int) ((c*100)/end_data);
				if(per!=oldper)
				{
					if(per%10==0)
						System.err.print(""+per);
				else
						System.err.print(".");
					oldper=per;
				}
				Point newp = new Point(num_par);
				for(int w = 0; w < num_par; w++) //for each parameter load in the values
				{
					int[] chars= new int[fcs.getSettings().settings[w].bits/8];
					try{
						for(int j = 0; j < chars.length; j++)
						{
							chars[j]=unsigned(buff[b++]);
						}
					}
					catch(Exception e)
					{
						System.err.println("Couldn't read a value");
					}
					c+=chars.length;
					int val=0;
					int mult = 1;

					for(int j = chars.length-1-(fcs.getSettings().settings[w].bits-fcs.getSettings().settings[w].usefulBits)/8; j >=0; j--)
					{
						val+=mult*chars[j];
						mult=mult*256;
					}
					//val = chars[1];
					if(fcs.getSettings().settings[w].amp_decades > 0)
					{
						val = (int)Math.pow(10,((((double)val)*fcs.getSettings().settings[w].amp_decades)/fcs.getSettings().settings[w].range));
					}
					newp.ps[w] = val;
					if(fcs.getMaxes()[w] < val)
					{
						fcs.getMaxes()[w] = val;
					//	System.err.println(" maxes for "+w+" is "+val);
					}
					//out+=val+"\t";
				}
				fcs.getData().add(newp);
				//System.err.println(out);
				count++;
			}
			System.err.println("Count was "+count);
		}
		catch(Exception e)
		{
			e.printStackTrace();
			return null;
		}
		return fcs;
	}
	
	  public static FCS processFCSFileWIthNoData(File f) {

		FCS fcs= new FCS(f);
		System.err.println("Processing: "+fcs.getExperiment());
		try
		{
			int num_par = 14;
			FileInputStream s = new FileInputStream(f);
			int start_text,end_text,start_data,end_data;
			getNextString(10,s);  //version
			start_text = getNextInteger(8,s);
			end_text = getNextInteger(8,s);
			start_data=getNextInteger(8,s);
			end_data=getNextInteger(8,s);
			System.err.println("Start text: "+start_text+" end text: "+end_text+" start data "+start_data+" end data "+end_data);
			//lets fast forward to the start of the text section
			long c = 42;
			s.skip(start_text-c);
			c=start_text;
			char del = getNextString(1,s).charAt(0);
			System.err.println("DELIMITER:\""+del+"\"");
			String keyword = "";
			String value = "";
			int mode = 0;  //0= in keyword, 1= in value
			while(c!=end_text)
			{
				char ch = (char)s.read();
				c++;
				if(ch==del && mode==0)
				{
					mode = 1;
				}
				else if(ch == del && mode==1)
				{
					mode = 0;
					//evaluate keyword/value pair
					int val = 0;
					try{
						val = Integer.parseInt(value);
					}
					catch(Exception e)
					{
						//probably not meant to be an int
					}
					System.err.println("KEYWORD:"+keyword+" VALUE: "+value);
					if(keyword.startsWith("$P")&&keyword.endsWith("S"))
					{
						int param = Integer.parseInt(keyword.substring(2,keyword.length()-1));
						fcs.getNames()[param-1]=value;
					}
					else if(keyword.startsWith("$PAR"))
					{
						num_par=val;
						fcs.setSettings(new FCSSettings(num_par));
						fcs.setNumPar(num_par);
					
					}
					else if(keyword.startsWith("$P"))
					{
						
						if(keyword.endsWith("B"))  //bits total (i.e. 32 for integer)
						{
							int index = Integer.parseInt(keyword.substring(2,keyword.indexOf('B')));
							fcs.getSettings().settings[index-1].bits = val;
						}
						else if(keyword.endsWith("E")) //amplification
						{
							int index = Integer.parseInt(keyword.substring(2,keyword.indexOf('E')));
							fcs.getSettings().settings[index-1].amp_decades = Integer.parseInt(value.split(",")[0]);
							fcs.getSettings().settings[index-1].amp_zero = Double.parseDouble(value.split(",")[1]);
						}
						else if(keyword.endsWith("S")) //name of param
						{
							int index = Integer.parseInt(keyword.substring(2,keyword.indexOf('S')));
							fcs.getSettings().settings[index-1].name = value;
						}
						else if(keyword.endsWith("R"))  //range of the parameter
						{
							int index = Integer.parseInt(keyword.substring(2,keyword.indexOf('R')));
							fcs.getSettings().settings[index-1].range = val;
							fcs.getSettings().settings[index-1].calculateUsefulBits();
						}
					}
					else if(keyword.startsWith("$TOT"))
					{
						fcs.setTotal(val);
					}
					else if(keyword.startsWith("#SAMPLE"))
					{
						fcs.setExperiment(value);
					}
					else if(keyword.startsWith("SAMPLE ID"))
					{
						fcs.setExperiment(value);
					}
					else if(keyword.startsWith("#CFLOW4COLORCOMP"))
					{
					//	fcs.setSpillover(value);  not anymore because it seems that we are using raw data...
					}
					else if(keyword.startsWith("$VOL"))
					{
						fcs.setVolume(val);
					}
					keyword ="";
					value="";
				}
				else if(mode==0)
				{
					keyword+=ch;
				}
				else if(mode==1)
				{
					value+=ch;
				}
			}
			System.err.println("Start data: "+start_data+" End data: "+end_data+" numpar "+num_par+" current position "+c+" going to skip "+(start_data-c-1));
			if(start_data-c-1 > 0)
				s.skip(start_data-c-1);
			else
				s.skip(1);
			c= start_data;
			//process data
			fcs.initializeDataSize(num_par);
			//System.err.println("Initialize data to hold "+(fcs.getTotal())+" events");
		}
		catch(Exception e)
		{
			e.printStackTrace();
			return null;
		}
		
		return fcs;
	}

	private static int unsigned(byte b)
	{
		if(b < 0)
			return 256+b;
		return b;
	}
	

	private static String getNextString(int i, FileInputStream s) {
		// TODO Auto-generated method stub
		try
		{
			String result = "";
			while(i > 0 && s.available()>=i)
			{
				result+=(char)s.read();
				i--;
			}
			return result;
		}
		catch(Exception e)
		{
			e.printStackTrace();
			return null;
		}
	}
	
	public static Integer getNextInteger(int r, FileInputStream s)
	{
		String nextStr = getNextString(r,s).trim();
		try{
			return Integer.parseInt(nextStr);
		}
		catch(Exception e)
		{
			return 0;
		}
	}

	
	
	private ArrayList<ArrayList<ArrayList<Double>>> deepCopy(
			ArrayList<ArrayList<ArrayList<Double>>> list) 
	{
		ArrayList<ArrayList<ArrayList<Double>>> copy = new ArrayList<ArrayList<ArrayList<Double>>>();
		for(int i = 0; i < list.size(); i++)
		{
			copy.add(new ArrayList<ArrayList<Double>>());
			for(int j = 0; j < list.get(i).size(); j++)
			{
				copy.get(i).add(new ArrayList<Double>());
				for(int k = 0; k < list.get(i).get(j).size(); k++)
					copy.get(i).get(j).add(list.get(i).get(j).get(k));
			}
		}
		return copy;
	}
	
	/**
	 * Works for two sorted in ascending order arrays a,b
	 * Uses the Mann-Whitney test to determine if A is statistically less than B (probability of U from A )
	 * @param a
	 * @param b
	 * @return probability that some value in a is less than some value in b
	 */
	public static double getProbabilityLess(double[] a, double[]b)
	{
		int acount = 0;
		int apos = 0;
		int bpos = 0;
		while(apos != a.length && bpos != b.length )
		{
			double x = a[apos];
			double y = b[bpos];
			if(x < y)
			{
				acount+=(b.length-bpos);
				apos++;
			}
			else if(x > y)
			{
				bpos++;
			}
			else
			{
				apos++;
				bpos++;
				acount+=(b.length-bpos);
			}
		}
		
		int U1 = acount;
		double alen = a.length;
		double blen = b.length;
		//The binomial distribution approaches the normal for more than 20 values):
		double H0mean = alen*blen*0.5;
		double H0sd = Math.sqrt((alen*blen*(alen+blen+1))/12.0);
		//Probability of more extreme solution given that they are the same (one sided on the left). 
		//The probability should be high if U1 is large since U1 counts the number of ranks in b that are higher (worse scores)
		//therefore if U1 is high then the gaussian cdf should approach 1 (in other words we integrate most of the distribution)
		double p = getGaussianCDF(H0mean, H0sd, U1);  // if the sum of the ranks is high then the value will be away from the mean on the right. 
		return p;
	}
	

	
    public static double getSegmentedPearsonCorrelation(double[] scores1,double[] scores2,int div){
    	int size = scores1.length/div;
    	double result = 0;
    	int cnt = 0;
    	for(int i = 0; i < div; i++)
    	{
			double[] a = new double[size];
			double[] b = new double[size];
    		for(int j = i*size;j < (i+1)*size; j++)
    		{
    			a[j-i*size]=scores1[j];
    			b[j-i*size]=scores2[j];
    		}
    		double p = getPearsonCorrelation(a,b);
    		if(!Double.isNaN(p))
    		{
    			result *= getPearsonCorrelation(a, b);
    			cnt++;
    		}
    	}
    	if(cnt == 0) return 0;
    	return Math.pow(result,1.0/cnt);
    }
    
    public static double[] prefix(double[] d, int index)
    {
    	double[] prefix = new double[index];
    	for(int i = 0; i < Math.min(index,d.length); i++)
    	{
    		prefix[i]=d[i];
    	}
    	return prefix;
    }
    
    public static double[] suffix(double[] d, int index)
    {
    	double[] suffix = new double[d.length-index];
    	for(int i = index; i <d.length; i++)
    	{
    		suffix[i-index]=d[i];
    	}
    	return suffix;
    }
    
    public static double[] slice(double[] d, int start, int stop)
    {
    	double[] slice = new double[stop-start];
    	for(int i = start; i < stop; i++)
    	{
    		slice[i-start]=d[i];
    	}
    	return slice;
    }
    /**
     * Centered moving average
     * @param d
     * @param window
     * @return
     */
    public static double[] movingAverage(double[] d, int window)
    {
    	double[] result = new double[d.length-window];
    	for(int i = window/2; i < d.length-window/2; i++)
    	{
    		for(int j = -window/2; j < window/2; j++)
    		{
    			result[i-window/2]+=d[i+j];
    		}
    		result[i-window/2]=result[i-window/2]/window;
    	}
    	return result;
    }
    
    public static double[] segmentedAverage(double[] d, int count)
    {
    	double[] result = new double[count];
    	for(int i = 0; i < count ;i++)
    	{
    		for(int j = 0; j < (d.length/count); j++)
    		{
    			result[i]+=d[i*(d.length/count)+j];
    		}
    		result[i]=result[i]/(d.length/count);
    	}
    	return result;
    }
	
	public static double sum(double[] ds) {
		// TODO Auto-generated method stub
		double sum = 0;
		for(int i = 0; i < ds.length; i++)
			sum+=ds[i];
		return sum;
	}

	//returns the pdf of a gaussian with mean u, std s, evaluated at v.
	public static double normal(double u, double s, double v)
	{
		return (1/Math.sqrt(2*Math.PI*s*s))*Math.pow(Math.E,-Math.pow((u-v)/(2*s*s),2));
	}
	
	public static double maximum(double[] s)
	{
		double max = Double.NEGATIVE_INFINITY;
		for(int i = 0; i < s.length; i++)
		{
			if(s[i] > max)
				max=s[i];
		}
		return max;
	}
	
	public static int maxIndex(double[] s)
	{
		double max = Double.NEGATIVE_INFINITY;
		int index = 0;
		for(int i = 0; i < s.length; i++)
		{
			if(s[i] > max)
			{
				max=s[i];
				index = i;
			}
		}
		return index;
	}
	
	public static double reflectWithin(double val, double min, double max)
	{
		if(val < min || val > max)
		{
			if(val < min) val = (min-val)+min;
			else if(val > max) val = max-(val-max);
//			System.err.println(val);
		}
		return Math.max(min,Math.min(val,max));
	}
	
	public static double minimum(double[] s)
	{
		double min = Double.POSITIVE_INFINITY;
		for(int i = 0; i < s.length; i++)
		{
			if(s[i] < min)
				min=s[i];
		}
		return min;
	}
	
	public static double average(double[] s)
	{
		double u = 0;
		for(int i = 0; i < s.length; i++)
		{
			u+=s[i];
		}
		return u/s.length;
	}

	public static double average(int[] s)
	{
		double u = 0;
		for(int i = 0; i < s.length; i++)
		{
			u+=s[i];
		}
		return u/s.length;
	}
	
	public static double average(ArrayList<Integer> s)
	{
		double u = 0;
		for(int i = 0; i < s.size(); i++)
		{
			u+=s.get(i);
		}
		return u/s.size();
	}
	
	
	
	public static double[] multiply(double[] s, double m)
	{
		for(int i = 0; i < s.length; i++)
			s[i]=s[i]*m;
		return s;
	}
	/**
	 * Returns a hashmap containing the angle values for twist roll and tilt for a segment of DNA indexed by capitalized trinucleotide. Ns treated as average
	 * @return
	 */
	public static HashMap<String,double[]> computePnucTable()
	{
        HashMap<String,double[]>pnucTable = new HashMap<String,double[]>();
        double[] angles1 = {Math.toRadians(0.0), Math.toRadians(34.3), Math.toRadians(3.3)};   //array of doubles corresponding to angular information (tilt, twist, roll)
        double[] angles2 = {Math.toRadians(0.0), Math.toRadians(34.3), Math.toRadians(5.4)};
        double[] angles3 = {Math.toRadians(0.0), Math.toRadians(34.3), Math.toRadians(7.5)};
        double[] angles4 = {Math.toRadians(0.0), Math.toRadians(34.3), Math.toRadians(5.8)};
        double[] angles5 = {Math.toRadians(0.0), Math.toRadians(34.3), Math.toRadians(0.0)};
        double[] angles6 = {Math.toRadians(0.0), Math.toRadians(34.3), Math.toRadians(5.2)};
        double[] angles7 = {Math.toRadians(0.0), Math.toRadians(34.3), Math.toRadians(3.7)};
        double[] angles8 = {Math.toRadians(0.0), Math.toRadians(34.3), Math.toRadians(0.7)};
        double[] angles9 = {Math.toRadians(0.0), Math.toRadians(34.3), Math.toRadians(3.8)};
        double[] angles10 = {Math.toRadians(0.0), Math.toRadians(34.3), Math.toRadians(6.0)};
        double[] angles11 = {Math.toRadians(0.0), Math.toRadians(34.3), Math.toRadians(10.0)};
        double[] angles12 = {Math.toRadians(0.0), Math.toRadians(34.3), Math.toRadians(3.0)};
        double[] angles13 = {Math.toRadians(0.0), Math.toRadians(34.3), Math.toRadians(5.3)};
        double[] angles14 = {Math.toRadians(0.0), Math.toRadians(34.3), Math.toRadians(8.3)};
        double[] angles15 = {Math.toRadians(0.0), Math.toRadians(34.3), Math.toRadians(4.7)};
        double[] angles16 = {Math.toRadians(0.0), Math.toRadians(34.3), Math.toRadians(4.2)};
        double[] angles17 = {Math.toRadians(0.0), Math.toRadians(34.3), Math.toRadians(6.5)};
        double[] angles18 = {Math.toRadians(0.0), Math.toRadians(34.3), Math.toRadians(6.7)};
        double[] angles19 = {Math.toRadians(0.0), Math.toRadians(34.3), Math.toRadians(2.0)};
        double[] angles20 = {Math.toRadians(0.0), Math.toRadians(34.3), Math.toRadians(2.2)};
        double[] angles21 = {Math.toRadians(0.0), Math.toRadians(34.3), Math.toRadians(2.8)};
        pnucTable.put( "AGA", angles1 );
        pnucTable.put( "TCT", angles1 );
        pnucTable.put( "CAA", angles1 );
        pnucTable.put( "TTG", angles1 );
        pnucTable.put( "AGG", angles2 );
        pnucTable.put( "CCT", angles2 );
        pnucTable.put( "GGT", angles2 );
        pnucTable.put( "ACC", angles2 );
        pnucTable.put( "GAG", angles2 );
        pnucTable.put( "CTC", angles2 );
        pnucTable.put( "GAC", angles2 );
        pnucTable.put( "GTC", angles2 );
        pnucTable.put( "CGT", angles2 );
        pnucTable.put( "ACG", angles2 );
        pnucTable.put( "TGA", angles2 );
        pnucTable.put( "TCA", angles2 );
        pnucTable.put( "TGG", angles2 );
        pnucTable.put( "CCA", angles2 );
        pnucTable.put( "AGC", angles3 );
        pnucTable.put( "GCT", angles3 );
        pnucTable.put( "CGC", angles3 );
        pnucTable.put( "GCG", angles3 );
        pnucTable.put( "AGT", angles4 );
        pnucTable.put( "ACT", angles4 );
        pnucTable.put( "AAA", angles5 );
        pnucTable.put( "TTT", angles5 );
        pnucTable.put( "AAG", angles6 );
        pnucTable.put( "CTT", angles6 );
        pnucTable.put( "TGT", angles6 );
        pnucTable.put( "ACA", angles6 );
        pnucTable.put( "AAC", angles7 );
        pnucTable.put( "GTT", angles7 );
        pnucTable.put( "TAC", angles7 );
        pnucTable.put( "GTA", angles7 );
        pnucTable.put( "AAT", angles8 );
        pnucTable.put( "ATT", angles8 );
        pnucTable.put( "GGA", angles9 );
        pnucTable.put( "TCC", angles9 );
        pnucTable.put( "GGG", angles10 );
        pnucTable.put( "CCC", angles10 );
        pnucTable.put( "TGC", angles10 );
        pnucTable.put( "GCA", angles10 );
        pnucTable.put( "GGC", angles11 );
        pnucTable.put( "GCC", angles11 );
        pnucTable.put( "GAA", angles12 );
        pnucTable.put( "TTC", angles12 );
        pnucTable.put( "GAT", angles13 );
        pnucTable.put( "ATC", angles13 );
        pnucTable.put( "CGA", angles14 );
        pnucTable.put( "TCG", angles14 );
        pnucTable.put( "CGG", angles15 );
        pnucTable.put( "CCG", angles15 );
        pnucTable.put( "CAG", angles16 );
        pnucTable.put( "CTG", angles16 );
        pnucTable.put( "CAC", angles17 );
        pnucTable.put( "GTG", angles17 );
        pnucTable.put( "CAT", angles18 );
        pnucTable.put( "ATG", angles18 );
        pnucTable.put( "TAA", angles19 );
        pnucTable.put( "TTA", angles19 );
        pnucTable.put( "TAG", angles20 );
        pnucTable.put( "CTA", angles20 );
        pnucTable.put( "TAT", angles21 );
        pnucTable.put( "ATA", angles21 );
        
        //all possible combinations involving Ns
        /*
         * NNA,NNC,NNT,NNG
         * NAN,NTN,NCN,NGN
         * ANN,TNN,CNN,GNN
         * NAA,NAC,NAG,NAT  
         * NCA,NCC,NCG,NCT 
         * NGA,NGC,NGG,NGT
         * NTA,NTC,NTG,NTT
         * ANA,ANC,ANT,ANG = TNT,GNT,ANT,CNT
         * CNA,CNC,CNT,CNG = TNG,GNG,CNG
         * GNA,GNC,GNT,GNG= TNC,GNC
         * TNA,TNC,TNT,TNG  = TNA
         * AAN,ACN,AGN,ATN (already covered)
         * CAN,CCN,CGN,CTN (already covered)
         * TAN,TCN,TGN,TTN (already covered)
         * GAN,GCN,GGN,GTN (already covered)
         * NNN
         */
        HashMap<String, double[]> temp = new HashMap<String,double[]>();
        temp.putAll(pnucTable);
        Iterator<String> itr = temp.keySet().iterator();
        while(itr.hasNext())
        {
        	String next = itr.next();
        	double angle = 0;
        	int count = 0;
        	if(next.substring(2,3).compareTo("A") == 0)
        	{
        		angle+= pnucTable.get(next)[2];
        		count++;
        	}
        	if(count > 0)
        	{
	        	double[] theAngle = {Math.toRadians(0.0), Math.toRadians(34.3),angle/count};
	        	pnucTable.put("NNA",theAngle);
	        	pnucTable.put("TNN",theAngle);
        	}
        }
        itr = temp.keySet().iterator();
        while(itr.hasNext())
        {
        	String next = itr.next();
        	double angle = 0;
        	int count = 0;
        	if(next.substring(2,3).compareTo("C") == 0)
        	{
        		angle+= pnucTable.get(next)[2];
        		count++;
        	}
        	if(count > 0)
        	{
	        	double[] theAngle = {Math.toRadians(0.0), Math.toRadians(34.3),angle/count};
	        	pnucTable.put("NNC",theAngle);
	        	pnucTable.put("GNN",theAngle);
        	}
        }
        itr = temp.keySet().iterator();
        while(itr.hasNext())
        {
        	String next = itr.next();
        	double angle = 0;
        	int count = 0;
        	if(next.substring(2,3).compareTo("T") == 0)
        	{
        		angle+= pnucTable.get(next)[2];
        		count++;
        	}
        	if(count > 0)
        	{
        		double[] theAngle = {Math.toRadians(0.0), Math.toRadians(34.3),angle/count};
        		pnucTable.put("NNT",theAngle);
        		pnucTable.put("ANN",theAngle);
        	}
        }
        itr = temp.keySet().iterator();
        while(itr.hasNext())
        {
        	String next = itr.next();
        	double angle = 0;
        	int count = 0;
        	if(next.substring(2,3).compareTo("G") == 0)
        	{
        		angle+= pnucTable.get(next)[2];
        		count++;
        	}
        	if(count > 0)
        	{

        	double[] theAngle = {Math.toRadians(0.0), Math.toRadians(34.3),angle/count};
        	pnucTable.put("NNG",theAngle);
        	pnucTable.put("CNN",theAngle);
        	}
        }
        itr = temp.keySet().iterator();
        while(itr.hasNext())
        {
        	String next = itr.next();
        	double angle = 0;
        	int count = 0;
        	if(next.substring(1,2).compareTo("A") == 0)
        	{
        		angle+= pnucTable.get(next)[2];
        		count++;
        	}
        	if(count > 0)
        	{

        	double[] theAngle = {Math.toRadians(0.0), Math.toRadians(34.3),angle/count};
        	pnucTable.put("NAN",theAngle);
        	}
        }
        itr = temp.keySet().iterator();
        while(itr.hasNext())
        {
        	String next = itr.next();
        	double angle = 0;
        	int count = 0;
        	if(next.substring(1,2).compareTo("C") == 0)
        	{
        		angle+= pnucTable.get(next)[2];
        		count++;
        	}
        	if(count > 0)
        	{

        	double[] theAngle = {Math.toRadians(0.0), Math.toRadians(34.3),angle/count};
        	pnucTable.put("NCN",theAngle);
        	}
        }
        itr = temp.keySet().iterator();
        while(itr.hasNext())
        {
        	String next = itr.next();
        	double angle = 0;
        	int count = 0;
        	if(next.substring(1,2).compareTo("T") == 0)
        	{
        		angle+= pnucTable.get(next)[2];
        		count++;
        	}
        	if(count > 0)
        	{

        	double[] theAngle = {Math.toRadians(0.0), Math.toRadians(34.3),angle/count};
        	pnucTable.put("NTN",theAngle);
        	}
        }
        itr = temp.keySet().iterator();
        while(itr.hasNext())
        {
        	String next = itr.next();
        	double angle = 0;
        	int count = 0;
        	if(next.substring(1,2).compareTo("G") == 0)
        	{
        		angle+= pnucTable.get(next)[2];
        		count++;
        	}
        	if(count > 0)
        	{
        		double[] theAngle = {Math.toRadians(0.0), Math.toRadians(34.3),angle/count};
        		pnucTable.put("NGN",theAngle);
        	}
        }
        itr = temp.keySet().iterator();
        while(itr.hasNext())
        {
        	String next = itr.next();
        	double angle = 0;
        	int count = 0;
        	if(next.substring(1,3).compareTo("AA") == 0)
        	{
        		angle+= pnucTable.get(next)[2];
        		count++;
        	}
        	if(count > 0)
        	{

        		double[] theAngle = {Math.toRadians(0.0), Math.toRadians(34.3),angle/count};
        		pnucTable.put("NAA",theAngle);
        		pnucTable.put("TTN",theAngle);
        	}
        }
        itr = temp.keySet().iterator();
        while(itr.hasNext())
        {
        	String next = itr.next();
        	double angle = 0;
        	int count = 0;
        	if(next.substring(1,3).compareTo("AC") == 0)
        	{
        		angle+= pnucTable.get(next)[2];
        		count++;
        	}
        	if(count > 0)
        	{

        	double[] theAngle = {Math.toRadians(0.0), Math.toRadians(34.3),angle/count};
        	pnucTable.put("NAC",theAngle);
        	pnucTable.put("GTN",theAngle);
        	}
        }
        itr = temp.keySet().iterator();
        while(itr.hasNext())
        {
        	String next = itr.next();
        	double angle = 0;
        	int count = 0;
        	if(next.substring(1,3).compareTo("AG") == 0)
        	{
        		angle+= pnucTable.get(next)[2];
        		count++;
        	}
        	if(count > 0)
        	{

        	double[] theAngle = {Math.toRadians(0.0), Math.toRadians(34.3),angle/count};
        	pnucTable.put("NAG",theAngle);
        	pnucTable.put("CTN",theAngle);
        	}
        }
        itr = temp.keySet().iterator();
        while(itr.hasNext())
        {
        	String next = itr.next();
        	double angle = 0;
        	int count = 0;
        	if(next.substring(1,3).compareTo("AT") == 0)
        	{
        		angle+= pnucTable.get(next)[2];
        		count++;
        	}
        	if(count > 0)
        	{

        	double[] theAngle = {Math.toRadians(0.0), Math.toRadians(34.3),angle/count};
        	pnucTable.put("NAT",theAngle);
        	pnucTable.put("ATN",theAngle);
        	}
        }
        itr = temp.keySet().iterator();
        while(itr.hasNext())
        {
        	String next = itr.next();
        	double angle = 0;
        	int count = 0;
        	if(next.substring(1,3).compareTo("CA") == 0)
        	{
        		angle+= pnucTable.get(next)[2];
        		count++;
        	}
        	if(count > 0)
        	{
        		double[] theAngle = {Math.toRadians(0.0), Math.toRadians(34.3),angle/count};
        		pnucTable.put("NCA",theAngle);
        		pnucTable.put("TGN",theAngle);
        	}
        }
        itr = temp.keySet().iterator();
        while(itr.hasNext())
        {
        	String next = itr.next();
        	double angle = 0;
        	int count = 0;
        	if(next.substring(1,3).compareTo("CC") == 0)
        	{
        		angle+= pnucTable.get(next)[2];
        		count++;
        	}
        	if(count > 0)
        	{

	        	double[] theAngle = {Math.toRadians(0.0), Math.toRadians(34.3),angle/count};
	        	pnucTable.put("NCC",theAngle);
	        	pnucTable.put("GGN",theAngle);
        	}
        }
        itr = temp.keySet().iterator();
        while(itr.hasNext())
        {
        	String next = itr.next();
        	double angle = 0;
        	int count = 0;
        	if(next.substring(1,3).compareTo("CG") == 0)
        	{
        		angle+= pnucTable.get(next)[2];
        		count++;
        	}
        	if(count > 0)
        	{

        	double[] theAngle = {Math.toRadians(0.0), Math.toRadians(34.3),angle/count};
        	pnucTable.put("NCG",theAngle);
        	pnucTable.put("CGN",theAngle);
        	}
        }
        itr = temp.keySet().iterator();
        while(itr.hasNext())
        {
        	String next = itr.next();
        	double angle = 0;
        	int count = 0;
        	if(next.substring(1,3).compareTo("CT") == 0)
        	{
        		angle+= pnucTable.get(next)[2];
        		count++;
        	}
        	if(count > 0)
        	{

        	double[] theAngle = {Math.toRadians(0.0), Math.toRadians(34.3),angle/count};
        	pnucTable.put("NCT",theAngle);
        	pnucTable.put("AGN",theAngle);
        	}
        }
        itr = temp.keySet().iterator();
        while(itr.hasNext())
        {
        	String next = itr.next();
        	double angle = 0;
        	int count = 0;
        	if(next.substring(1,3).compareTo("GA") == 0)
        	{
        		angle+= pnucTable.get(next)[2];
        		count++;
        	}
        	if(count > 0)
        	{

        	double[] theAngle = {Math.toRadians(0.0), Math.toRadians(34.3),angle/count};
        	pnucTable.put("NGA",theAngle);
        	pnucTable.put("TCN",theAngle);
        	}
        }
        itr = temp.keySet().iterator();
        while(itr.hasNext())
        {
        	String next = itr.next();
        	double angle = 0;
        	int count = 0;
        	if(next.substring(1,3).compareTo("GC") == 0)
        	{
        		angle+= pnucTable.get(next)[2];
        		count++;
        	}
        	if(count > 0)
        	{

        	double[] theAngle = {Math.toRadians(0.0), Math.toRadians(34.3),angle/count};
        	pnucTable.put("NGC",theAngle);
        	pnucTable.put("GCN",theAngle);
        	}
        }
        itr = temp.keySet().iterator();
        while(itr.hasNext())
        {
        	String next = itr.next();
        	double angle = 0;
        	int count = 0;
        	if(next.substring(1,3).compareTo("GG") == 0)
        	{
        		angle+= pnucTable.get(next)[2];
        		count++;
        	}
        	if(count > 0)
        	{

        	double[] theAngle = {Math.toRadians(0.0), Math.toRadians(34.3),angle/count};
        	pnucTable.put("NGG",theAngle);
        	pnucTable.put("CCN",theAngle);
        	}
        }
        itr = temp.keySet().iterator();
        while(itr.hasNext())
        {
        	String next = itr.next();
        	double angle = 0;
        	int count = 0;
        	if(next.substring(1,3).compareTo("GT") == 0)
        	{
        		angle+= pnucTable.get(next)[2];
        		count++;
        	}
        	if(count > 0)
        	{

        	double[] theAngle = {Math.toRadians(0.0), Math.toRadians(34.3),angle/count};
        	pnucTable.put("NGT",theAngle);
        	pnucTable.put("ACN",theAngle);
        	}
        }
        itr = temp.keySet().iterator();
        while(itr.hasNext())
        {
        	String next = itr.next();
        	double angle = 0;
        	int count = 0;
        	if(next.substring(1,3).compareTo("TA") == 0)
        	{
        		angle+= pnucTable.get(next)[2];
        		count++;
        	}
        	if(count > 0)
        	{

        	double[] theAngle = {Math.toRadians(0.0), Math.toRadians(34.3),angle/count};
        	pnucTable.put("NTA",theAngle);
        	pnucTable.put("TAN",theAngle);
        	}
        }
        itr = temp.keySet().iterator();
        while(itr.hasNext())
        {
        	String next = itr.next();
        	double angle = 0;
        	int count = 0;
        	if(next.substring(1,3).compareTo("TC") == 0)
        	{
        		angle+= pnucTable.get(next)[2];
        		count++;
        	}
        	if(count > 0)
        	{

        	double[] theAngle = {Math.toRadians(0.0), Math.toRadians(34.3),angle/count};
        	pnucTable.put("NTC",theAngle);
        	pnucTable.put("GAN",theAngle);
        	}
        }
        itr = temp.keySet().iterator();
        while(itr.hasNext())
        {
        	String next = itr.next();
        	double angle = 0;
        	int count = 0;
        	if(next.substring(1,3).compareTo("TG") == 0)
        	{
        		angle+= pnucTable.get(next)[2];
        		count++;
        	}
        	if(count > 0)
        	{

        	double[] theAngle = {Math.toRadians(0.0), Math.toRadians(34.3),angle/count};
        	pnucTable.put("NTG",theAngle);
        	pnucTable.put("CAN",theAngle);
        	}
        }
        itr = temp.keySet().iterator();
        while(itr.hasNext())
        {
        	String next = itr.next();
        	double angle = 0;
        	int count = 0;
        	if(next.substring(1,3).compareTo("TT") == 0)
        	{
        		angle+= pnucTable.get(next)[2];
        		count++;
        	}
        	if(count > 0)
        	{

        	double[] theAngle = {Math.toRadians(0.0), Math.toRadians(34.3),angle/count};
        	pnucTable.put("NTT",theAngle);
        	pnucTable.put("AAN",theAngle);
        	}
        }
        itr = temp.keySet().iterator();
        while(itr.hasNext())
        {
        	String next = itr.next();
        	double angle = 0;
        	int count = 0;
        	if(next.charAt(0)=='A'&& next.charAt(2)=='A')
        	{
        		angle+= pnucTable.get(next)[2];
        		count++;
        	}
        	if(count > 0)
        	{

        	double[] theAngle = {Math.toRadians(0.0), Math.toRadians(34.3),angle/count};
        	pnucTable.put("ANA",theAngle);
        	pnucTable.put("TNT",theAngle);
        	}
        }
        itr = temp.keySet().iterator();
        while(itr.hasNext())
        {
        	String next = itr.next();
        	double angle = 0;
        	int count = 0;
        	if(next.charAt(0)=='A'&& next.charAt(2)=='C')
        	{
        		angle+= pnucTable.get(next)[2];
        		count++;
        	}
        	if(count > 0)
        	{

        	double[] theAngle = {Math.toRadians(0.0), Math.toRadians(34.3),angle/count};
        	pnucTable.put("ANC",theAngle);
        	pnucTable.put("GNT",theAngle);
        	}
        }
        itr = temp.keySet().iterator();
        while(itr.hasNext())
        {
        	String next = itr.next();
        	double angle = 0;
        	int count = 0;
        	if(next.charAt(0)=='A'&& next.charAt(2)=='G')
        	{
        		angle+= pnucTable.get(next)[2];
        		count++;
        	}
        	if(count > 0)
        	{

        	double[] theAngle = {Math.toRadians(0.0), Math.toRadians(34.3),angle/count};
        	pnucTable.put("ANG",theAngle);
        	pnucTable.put("CNT",theAngle);
        	}
        }
        itr = temp.keySet().iterator();
        while(itr.hasNext())
        {
        	String next = itr.next();
        	double angle = 0;
        	int count = 0;
        	if(next.charAt(0)=='A'&& next.charAt(2)=='T')
        	{
        		angle+= pnucTable.get(next)[2];
        		count++;
        	}
        	if(count > 0)
        	{

        	double[] theAngle = {Math.toRadians(0.0), Math.toRadians(34.3),angle/count};
        	pnucTable.put("ANT",theAngle);
        	pnucTable.put("ANT",theAngle);
        	}
        }
        itr = temp.keySet().iterator();
        while(itr.hasNext())
        {
        	String next = itr.next();
        	double angle = 0;
        	int count = 0;
        	if(next.charAt(0)=='C'&& next.charAt(2)=='A')
        	{
        		angle+= pnucTable.get(next)[2];
        		count++;
        	}
        	if(count > 0)
        	{

        	double[] theAngle = {Math.toRadians(0.0), Math.toRadians(34.3),angle/count};
        	pnucTable.put("CNA",theAngle);
        	pnucTable.put("TNG",theAngle);
        	}
        }
        itr = temp.keySet().iterator();
        while(itr.hasNext())
        {
        	String next = itr.next();
        	double angle = 0;
        	int count = 0;
        	if(next.charAt(0)=='C'&& next.charAt(2)=='C')
        	{
        		angle+= pnucTable.get(next)[2];
        		count++;
        	}
        	if(count > 0)
        	{

        	double[] theAngle = {Math.toRadians(0.0), Math.toRadians(34.3),angle/count};
        	pnucTable.put("CNC",theAngle);
        	pnucTable.put("GNG",theAngle);
        	}
        }
        itr = temp.keySet().iterator();
        while(itr.hasNext())
        {
        	String next = itr.next();
        	double angle = 0;
        	int count = 0;
        	if(next.charAt(0)=='C'&& next.charAt(2)=='G')
        	{
        		angle+= pnucTable.get(next)[2];
        		count++;
        	}
        	if(count > 0)
        	{

        	double[] theAngle = {Math.toRadians(0.0), Math.toRadians(34.3),angle/count};
        	pnucTable.put("CNG",theAngle);
        	pnucTable.put("CNG",theAngle);
        	}
        }
        itr = temp.keySet().iterator();
        while(itr.hasNext())
        {
        	String next = itr.next();
        	double angle = 0;
        	int count = 0;
        	if(next.charAt(0)=='G'&& next.charAt(2)=='A')
        	{
        		angle+= pnucTable.get(next)[2];
        		count++;
        	}
        	if(count > 0)
        	{

        	double[] theAngle = {Math.toRadians(0.0), Math.toRadians(34.3),angle/count};
        	pnucTable.put("GNA",theAngle);
        	pnucTable.put("TNC",theAngle);
        	}
        }
        itr = temp.keySet().iterator();
        while(itr.hasNext())
        {
        	String next = itr.next();
        	double angle = 0;
        	int count = 0;
        	if(next.charAt(0)=='G'&& next.charAt(2)=='C')
        	{
        		angle+= pnucTable.get(next)[2];
        		count++;
        	}
        	if(count > 0)
        	{

        	double[] theAngle = {Math.toRadians(0.0), Math.toRadians(34.3),angle/count};
        	pnucTable.put("GNC",theAngle);
        	}
        	//pnucTable.put("GNC",theAngle);
        }
        itr = temp.keySet().iterator();
        while(itr.hasNext())
        {
        	String next = itr.next();
        	double angle = 0;
        	int count = 0;
        	if(next.charAt(0)=='T'&& next.charAt(2)=='A')
        	{
        		angle+= pnucTable.get(next)[2];
        		count++;
        	}
        	if(count > 0)
        	{

        	double[] theAngle = {Math.toRadians(0.0), Math.toRadians(34.3),angle/count};
        	pnucTable.put("TNA",theAngle);
        	}
        	//pnucTable.put("TNA",theAngle);
        }
        itr = temp.keySet().iterator();
        while(itr.hasNext())
        {
        	String next = itr.next();
        	double angle = 0;
        	int count = 0;
    		angle+= pnucTable.get(next)[2];
    		count++;
        	double[] theAngle = {Math.toRadians(0.0), Math.toRadians(34.3),angle/count};
        	pnucTable.put("NNN",theAngle);
        }
        /*
        itr = pnucTable.keySet().iterator();
        while(itr.hasNext())
        {
        	String next = itr.next();
        	double[] theAngle = pnucTable.get(next);
        	double omega01 = theAngle[2];
        	double energy = ((((a1 + a2) / 2) / 2) * ((k * k) + (omega01 * omega01) + (omega02 * omega02)) + ((a3 / 2) * ((omega3 - omega03) * (omega3 - omega03))));
        	//System.err.println("The energy for "+next+" is "+energy);
        }
        */
       return pnucTable;

	}
	
	public static final String DATE_FORMAT_NOW = "yyyy-MM-dd HH-mm-ss";
	/**
	 * Returns the current date and time
	 * @return
	 */
	  public static String now() {
		    Calendar cal = Calendar.getInstance();
		    SimpleDateFormat sdf = new SimpleDateFormat(DATE_FORMAT_NOW);
		    return sdf.format(cal.getTime());
		  }
	
	  public static void pause(long mili)
	  {
		long time = System.nanoTime();
		while(System.nanoTime()-time < mili*1000000.0)
		{
			//wait
		}
	  }
	  
	/**
	 * This returns an array of doubles the size of the sequence with the first histone/2 and the last histone/2 values set to 0.
	 * Histone is the number of nts that you want to calculate across (73 is good)
	 * @param sequence
	 * @param histone
	 * @param pnucTable
	 * @return
	 */
	public static double[] computeNucleosomeBinding(char[] sequence, int histone, HashMap<String,double[]> pnucTable)
	{
		double R = 4.19;  //radius
	    double P = 2.52;  //pitch
	    //private static int position;
	    double k = ((2 * Math.PI) * (2 * Math.PI) * R) / (((2 * Math.PI * R) * (2 * Math.PI * R)) + (P * P));
	    double twist_turns = 10.3;
	    double omega3 = (2 * Math.PI) / twist_turns;
	    //private static double w = omega3 - ((2 * Math.PI * P) / ((P * P) + ((2 * Math.PI * R) * (2 * Math.PI * R))));
	    double omega01;
	    double omega02 = 0;
	    double omega03 = (2 * Math.PI) / twist_turns;
	    int tm = 1;  //melting temperature
	    double a1 = (tm * 50);  //stiffness constant A1
	    double a2 = (tm * 50);  //stiffness constant A2
	    double a3 = (tm * 75);  //stiffness contsant A3
		double[] energy = new double[sequence.length];
		
        for ( int i = histone/2; i <= (sequence.length-histone/2-2); i++) 
        {  //iterate through each trinucleotide in the sequence
            double nrg = 0;  //energy
            if(i == histone/2)
            {
                for ( int j = i-histone/2; j <= (i+histone/2-1); j++ ) { //scan for histone length of nucleotides
                    char[] tri_hist = { sequence[j], sequence[j+1], sequence[j+2] };  //store trinucleotide as array of characters
                    String trinuc = new String(tri_hist);  //convert into string
                    double[] angle = null;
                    if(pnucTable.get(trinuc) != null)
                    {
	                    angle = (double[])pnucTable.get(trinuc);  //get corresponding angular info from hashtable
                    }
                    else
                    {
                    	//System.err.println("\nUnrecognized trinucleotide: "+trinuc+" using NNN instead\n");
                    	angle = (double[])pnucTable.get("NNN");
                    }
                    omega01 = (double)angle[2];  //store roll angle in variable 'omega01'
                    double temp =((((a1 + a2) / 2) / 2) * ((k * k) + (omega01 * omega01) + (omega02 * omega02)) + ((a3 / 2) * ((omega3 - omega03) * (omega3 - omega03))));  //calculate energy cost for nucleosome occupation 
                    if(Double.isNaN(temp))  //as long as the energy is not infinite (this was the case early on)
                    {
                    	System.err.println("Energy turned out to be NaN! Omega1 was "+omega01+" and trinuc was "+trinuc+". Using nrg of current average instead.");
                    	nrg+= nrg/j;  //just use the current average
                    		
                    }
                    else
                    	nrg+=temp;  
                }
                energy[i] = nrg;  //the first histone/2 spots will be empty but thats more accurate. The last histone/2 spots will also be empty but thats also ok.
                //windowStart2[i-histone/2] = Math.pow(Math.E,nrg/avg_e);  //store calculated energy cost in 2-dimensional array, where x = file number and i = position
            }//if
            else
            {
            	//optimization
            	double energy_before = 0;
                char[] tri_hist = { sequence[i-1-histone/2], sequence[i-histone/2], sequence[i-histone/2+1] };  //store trinucleotide as array of characters
                String trinuc = new String(tri_hist);  //convert into string
                double[] angle = null;
                if(pnucTable.get(trinuc) != null)
                {
                    angle = (double[])pnucTable.get(trinuc);  //get corresponding angular info from hashtable
                }
                else
                {
                	//System.err.println("\nUnrecognized trinucleotide: "+trinuc+" using NNN instead\n");
                	angle = (double[])pnucTable.get("NNN");
                }
                omega01 = (double)angle[2];  //store roll angle in variable 'omega01'
                energy_before =((((a1 + a2) / 2) / 2) * ((k * k) + (omega01 * omega01) + (omega02 * omega02)) + ((a3 / 2) * ((omega3 - omega03) * (omega3 - omega03))));  //calculate energy cost for nucleosome occupation 
                double energy_after = 0;
                char[] tri_hist2 = { sequence[i+histone/2-1], sequence[i+histone/2], sequence[i+histone/2+1] };  //store trinucleotide as array of characters
                trinuc = new String(tri_hist2);  //convert into string
                angle = null;
                if(pnucTable.get(trinuc) != null)
                {
                    angle = (double[])pnucTable.get(trinuc);  //get corresponding angular info from hashtable
                }
                else
                {
                	//System.err.println("\nUnrecognized trinucleotide: "+trinuc+" using NNN instead\n");
                	angle = (double[])pnucTable.get("NNN");
                }
                omega01 = (double)angle[2];  //store roll angle in variable 'omega01'
                energy_after =((((a1 + a2) / 2) / 2) * ((k * k) + (omega01 * omega01) + (omega02 * omega02)) + ((a3 / 2) * ((omega3 - omega03) * (omega3 - omega03))));  //calculate energy cost for nucleosome occupation 
                energy[i] = energy[i-1]+energy_after-energy_before;
            }
        }
        return energy;
	}
	
	
	public static double sumArrayList(ArrayList<Double> list)
	{
		double sum = 0;
		for(int i = 0; i < list.size(); i++)
			sum+=list.get(i);
		return sum;
	}

	
	
	/**
	 * Returns the standard deviation of an array of doubles
	 * @param x
	 * @return
	 */
	public static double std(double[] x)
	{
		double std = 0;
		double u = average(x);
		for(int i = 0; i < x.length; i++)
		{
			std+= (u-x[i])*(u-x[i]);
		}
		return Math.sqrt(std/x.length);
	}
	
	
	
	
	
	
	
	
	
	/**
	 * Copies an array
	 * @param s
	 * @return
	 */
	
	public static double[] duplicate(double[] s)
	{
		double[] copy = new double[s.length];
		for(int i = 0; i < s.length; i++)
			copy[i] = s[i];
		return copy;
	}
	
	
	/**
	 * returns an array that holds the sum both arrays at each index.
	 * @param a1
	 * @param a2
	 * @return
	 */
	public static double[] sumArrays(double[] a1, double[] a2)
	{
		double[] sum = new double[Math.min(a1.length, a2.length)];
		for(int i = 0; i < sum.length; i++)
			sum[i]=a1[i]+a2[i];
		return sum;
	}
	
	/**
	 * Converts ArrayList of double to double array
	 * @param list
	 * @return
	 */
	public static double[] ArrayListToArray(ArrayList<Double> list)
	{
		double[] result = new double[list.size()];
		for(int i = 0; i < list.size(); i++)
			result[i] = list.get(i);
		return result;
	}
	
	
	/**
	 * Computes the pearson correlation of the two equal sized arrays of doubles a pearson of 1 means perfectly linearly correlated. A pearson of -1 means 
	 * perfect anticorrelation. 0 means no correlation.
	 * @param scores1
	 * @param scores2
	 * @return
	 */
    public static double getPearsonCorrelation(double[] scores1,double[] scores2){
        double result = 0;
        double sum_sq_x = 0;
        double sum_sq_y = 0;
        double sum_coproduct = 0;
        double mean_x = scores1[0];
        double mean_y = scores2[0];
        for(int i=2;i<scores1.length+1;i+=1){
            double sweep =Double.valueOf(i-1)/i;
            double delta_x = scores1[i-1]-mean_x;
            double delta_y = scores2[i-1]-mean_y;
            sum_sq_x += delta_x * delta_x * sweep;
            sum_sq_y += delta_y * delta_y * sweep;
            sum_coproduct += delta_x * delta_y * sweep;
            mean_x += delta_x / i;
            mean_y += delta_y / i;
        }
        double pop_sd_x = (double) Math.sqrt(sum_sq_x/scores1.length);
        double pop_sd_y = (double) Math.sqrt(sum_sq_y/scores1.length);
        double cov_x_y = sum_coproduct / scores1.length;
        result = cov_x_y / (pop_sd_x*pop_sd_y);
        return result;
    }

    /**
     * Gets the average sum of the difference when you subtract b from a
     * @param a
     * @param b
     * @return
     */
	public static double SumDistance(double[] a, double[] b) {
		double distance = 0;
		for(int i = 0; i < a.length; i++)
		{
			distance+=(a[i]-b[i]);
		}
		return distance/a.length;
	}

	public static double[] add(double[] seq, double d) {
		double[] result = new double[seq.length];
		for(int i = 0; i < seq.length; i++)
			result[i]=seq[i]+d;
		
		return result;
	}

	/**
	 * Sums all of the positive values only
	 * @param d
	 * @return
	 */
	public static double SumPositive(double[] d ){
		double sum = 0;
		for(int i = 0; i < d.length; i++)
		{
			if(d[i] >0 )
				sum+=d[i];
		}
		return sum;
	}
	
	/**
	 * Sums negative vals only
	 * @param d
	 * @return
	 */
	public static double SumNegative(double[] d ){
		double sum = 0;
		for(int i = 0; i < d.length; i++)
		{
			if(d[i] <0 )
				sum+=d[i];
		}
		return sum;
	}

	/**
	 * Returns the capitalized version of an array of chars
	 * @param chars
	 * @return
	 */
	public static char[] capitalize(char[] chars) {
		char[] buf = new char[chars.length];
		for(int i = 0; i < chars.length; i++)
		{
			if(chars[i] >= 97)
				buf[i] = (char) (chars[i]-32);
			else
				buf[i] = chars[i];
		}
		return buf;
	}

	public static ArrayList<ArrayList<double[]>>kmeansCluster(ArrayList<double[]> solutions2, int k) {
		double[][] solutions = new double[solutions2.size()][];
		for(int i = 0; i < solutions2.size(); i++)
			solutions[i]=solutions2.get(i);
		ArrayList<ArrayList<double[]>> result = null;
		double[][] centers,oldcenters;
		double[] averages = new double[solutions[0].length];
		centers = new double[k][solutions[0].length];
		oldcenters = new double[k][solutions[0].length];
		for(int i = 0; i < solutions.length; i++)
		{
			for(int j = 0; j < solutions[0].length; j++)
				averages[j]+=solutions[i][j];
		}
		for(int j = 0; j < solutions[0].length; j++)
			averages[j]=averages[j]/solutions.length;
		for(int i = 0; i < k; i++)
		{
			for(int j = 0; j < centers[0].length; j++)
			{
				if(averages[j]!=0)
				centers[i][j]=solutions[i][j];
			}
		}	

		
		boolean changing = true;
		while(changing)
		{
			changing = false;  //switch this later
			//assign points to nearest center
			int[] assignment = new int[solutions.length];
			for(int i = 0; i < solutions.length; i++)
			{
				double best = Double.POSITIVE_INFINITY;
				double d = 0;
				for(int j = 0; j < k; j++)
				{
					d=0;
					for(int m = 2; m < averages.length; m++)
					{
						if(averages[m] != 0)
							d+=Math.pow((centers[j][m]-solutions[i][m])/averages[m],2);
					}
					
					if(d < best)
					{
						best = d;
						assignment[i]=j;
						//System.err.println("Solution "+i+" assigned to cluster "+assignment[i]+" with a distance of "+d);
					}
					else{
						//System.err.println("Solution "+i+" was not assigned to cluster "+j+" with a distance of "+d);
					}
				}
				
			}
			for(int i = 0; i < k; i++)
			{
				for(int j = 0; j < averages.length; j++)
					centers[i][j]=0;  //centers reset to 0 to be computed in a bit
			}
			//compute centers based on assignments
			int[] counts = new int[k];
			for(int i = 0; i < assignment.length; i++)
			{
				//System.err.print(" "+i+","+assignment[i]);
				for(int j = 0; j < averages.length; j++)
				{
					centers[assignment[i]][j]+=solutions[i][j];
				}
				counts[assignment[i]]++;
			}
			for(int i = 0; i < k; i++)
			{
				for(int j = 0; j < averages.length; j++)
				{
					if(averages[j] != 0&&counts[i]!=0)
						centers[i][j]=centers[i][j]/(counts[i]);
				}
			}
			//check against old centers
			//System.err.println("----------------");
			for(int i = 0; i < centers.length; i++)
			{
				for(int j = 0; j < centers[i].length; j++)
				{
					if(centers[i][j] != oldcenters[i][j])
					{
						changing = true;
					}
					//System.err.println(oldcenters[i][j]+"/"+centers[i][j]);
					oldcenters[i][j] = centers[i][j];
				}
			}
			if(!changing)
			{
				result = new ArrayList<ArrayList<double[]>>();
				for(int i = 0; i < k; i++)
					result.add(new ArrayList<double[]>());
				for(int i = 0; i < assignment.length; i++)
				{
					double[] copy = new double[solutions[i].length];
					for(int j = 0; j < copy.length; j++)
						copy[j] = solutions[i][j];
					result.get(assignment[i]).add(copy);
				}
			}
		}
		return result;
	}

	public static double computeAverageSilhouette(
			ArrayList<ArrayList<double[]>> clusters, int nparams) {
		if(clusters == null || clusters.size()==0 || clusters.get(0).size()==0) 
		{
			if(clusters==null)System.err.println("Clusters were null probably due to one of the clusters being empty during clustering!");
			else if(clusters!=null && clusters.size()==0) System.err.println("Clusters were empty");
			else System.err.println("The first cluster was empty!");
			return -1;
		}
		if(clusters.size() == 1) return 0;  //can't tell
		//silhouette is a measure of how well a data point has been clustered. The average silhouette for a dataset tells us how good the clustering is overall
		double averageS = 0;
		double count = 0;
		for(int i = 0; i < clusters.size(); i++)
		{
			System.err.println("Finished average silhouette calculations for points in cluster "+(i+1));
			for(int j = 0; j < clusters.get(i).size(); j++)
			{
				//datapoint j in cluster i
				//lets compute a(i) and b(i)
				//a(i)
				double ai = 0;
				for(int k = 0; k < clusters.get(i).size(); k++)
				{
					if(k!=j)
					{
						double d = 0;
						double a[] = clusters.get(i).get(j);
						double b[] = clusters.get(i).get(k);
						for(int l = 2; l < nparams+2; l++)
						{
							if(b[l] > a[l])
							{
								d = (b[l]-a[l])-(a[l+nparams*2]+b[l+nparams]);
							}
							else
								d = (a[l]-b[l])-(b[l+nparams*2]+a[l+nparams]);
							if(d!=0)
							{
								if(a[l]+b[l] != 0)
								d = d/(a[l]+b[l]);
							}
							ai+=d;
						}

					}
				}
				ai = ai/(Math.max(1,clusters.get(i).size()-1));
				//now lets compute the b(i)
				double bi = Double.POSITIVE_INFINITY;
				for(int m = 0; m < clusters.size(); m++)
				{
					if(m != i)
					{
						double ci = 0;
						for(int k = 0; k < clusters.get(m).size(); k++)
						{
							double d = 0;
							double a[] = clusters.get(i).get(j);
							double b[] = clusters.get(m).get(k);
							for(int l = 2; l < nparams+2; l++)
							{
								if(b[l] > a[l])
								{
									d = (b[l]-a[l])-(a[l+nparams*2]+b[l+nparams]);
								}
								else
									d = (a[l]-b[l])-(b[l+nparams*2]+a[l+nparams]);
								if(d!=0)
								{
									if(a[l]+b[l] != 0)
									d = d/(a[l]+b[l]);
								}
								ci+=d;
							}

						}
						ci = ci /Math.max(1,clusters.get(m).size());
						if(ci < bi) bi = ci;
					}
				}
				if(ai != 0 && bi != 0)
				{
					averageS+=(bi-ai)/Math.max(Math.abs(ai),Math.abs(bi));
					count++;
				}
			}
		}
		return averageS/Math.max(1,count);
	}
	public static double computeAverageSilhouette(
			ArrayList<ArrayList<double[]>> clusters, boolean[] mask) {
		if(clusters == null || clusters.size()==0 || clusters.get(0).size()==0) 
		{
			if(clusters==null)System.err.println("Clusters were null probably due to one of the clusters being empty during clustering!");
			else if(clusters!=null && clusters.size()==0) System.err.println("Clusters were empty");
			else System.err.println("The first cluster was empty!");
			return -1;
		}
		double[] averages = new double[clusters.get(0).get(0).length];
		int count = 0;
		for(int i = 0; i < clusters.size(); i++)
		{
			for(int j = 0; j < clusters.get(i).size(); j++)
			{
				for(int k = 0; k < clusters.get(i).get(j).length; k++)
				{
					averages[k]+=clusters.get(i).get(j)[k];
					count++;
				}
			}
		}
		for(int i = 0; i < averages.length; i++)
			averages[i]=averages[i]/count;
		double averageS = 0;
		count = 0;
		for(int i = 0; i < clusters.size(); i++)
		{
			for(int j = 0; j < clusters.get(i).size(); j++)
			{
				double adiff = 0;
				double bdiff = Double.POSITIVE_INFINITY;
				for(int k = 0; k < clusters.get(i).size(); k++)
				{
					for(int l = 0; l <averages.length; l++)
					{
						if(averages[l] != 0 && mask[l])
							adiff+=Math.pow((clusters.get(i).get(j)[l]-clusters.get(i).get(k)[l])/averages[l],2);
					}
				}
				adiff=adiff/clusters.get(i).size();
				for(int m = 0; m < clusters.size(); m++)
				{
					if(m!=i)
					{
						double localb = 0;
						for(int k = 0; k < clusters.get(m).size(); k++)
						{
							for(int l = 0; l < averages.length; l++)
							{
								if(averages[l] != 0&&mask[l])
									localb+=Math.pow((clusters.get(i).get(j)[l]-clusters.get(m).get(k)[l])/averages[l],2);
							}
						}
						localb=localb/clusters.get(m).size();
						if(localb < bdiff)
							bdiff=localb;
					}
				}
				double silhouette = (bdiff-adiff)/Math.max(adiff, bdiff);
				averageS+=silhouette;
				count++;
			}
		}
		return averageS/count;
	}

	public static int findTallestPeak(double[] x) {
		if(x.length==0)
			return 0;
		int bestPeak = 0;
		double bestPeakHeight = 0;
		double min = x[0];
		for(int i = 1; i < x.length-1;i++)
		{
			if(x[i]>x[i-1]&&x[i+1]<x[i])
			{
				//peaked
				double peakh = x[i]-min;
				if(peakh > bestPeakHeight)
				{
					bestPeakHeight = peakh;
					bestPeak = i;
				}
			}
			else if(x[i] < x[i-1]&& x[i] < x[i+1])
			{
				min = x[i];
			}
		}
		return bestPeak;
	}

	public static ArrayList<ArrayList<double[]>> kmeansCluster(ArrayList<double[]> solutions2, int k, boolean[] columns)
	{
			if(columns.length < solutions2.get(0).length)
			{
				boolean[] temp = new boolean[solutions2.get(0).length];
				for(int i = 0; i < columns.length; i++)
				{
					temp[i]=columns[i];
				}
				columns = temp;
			}
			/*
			System.err.println("Columns: ");
			for(int i = 0; i < columns.length; i++)
			{
				System.err.print(columns[i]+" ");
			}
			System.err.println();
*/
			double[][] solutions = new double[solutions2.size()][solutions2.get(0).length];
			for(int i = 0; i < solutions2.size(); i++)
			{
				for(int j = 0; j < solutions2.get(i).length; j++)
					solutions[i][j]=solutions2.get(i)[j];
			}
			ArrayList<ArrayList<double[]>> result = null;
			double[][] centers,oldcenters;
			double[] averages = new double[solutions[0].length];
			centers = new double[k][solutions[0].length];
			oldcenters = new double[k][solutions[0].length];
			for(int i = 0; i < solutions.length; i++)
			{
				for(int j = 0; j < solutions[0].length; j++)
					averages[j]+=solutions[i][j];
			}
			for(int j = 0; j < solutions[0].length; j++)
				averages[j]=averages[j]/solutions.length;
			for(int i = 0; i < solutions.length; i++)
			{
				for(int j = 0; j < solutions[i].length; j++)
				{
					if(averages[j] != 0)
						solutions[i][j] = solutions[i][j]/averages[j];
				}
			}
			for(int i = 0; i < k; i++)
			{
				for(int j = 0; j < centers[0].length; j++)
				{
					centers[i][j]=solutions[i][j];
				}
			}	

			
			boolean changing = true;
			while(changing)
			{
				changing = false;  //switch this later
				//assign points to nearest center
				int[] assignment = new int[solutions.length];
				for(int i = 0; i < solutions.length; i++)
				{
					double best = Double.POSITIVE_INFINITY;
					double d = 0;
					for(int j = 0; j < k; j++)
					{
						d=0;
						for(int m = 0; m < averages.length; m++)
						{
							if(columns[m])
								d+=Math.sqrt(Math.pow((centers[j][m]-solutions[i][m]),2));
						}
						if(d < best)
						{
							best = d;
							assignment[i]=j;
							//System.err.println("Solution "+i+" assigned to cluster "+assignment[i]+" with a distance of "+d);
						}
						else{
							//System.err.println("Solution "+i+" was not assigned to cluster "+j+" with a distance of "+d);
						}
					}
					
				}
				for(int i = 0; i < k; i++)
				{
					for(int j = 0; j < averages.length; j++)
						centers[i][j]=0;  //centers reset to 0 to be computed in a bit
				}
				//compute centers based on assignments
				int[] counts = new int[k];
				for(int i = 0; i < assignment.length; i++)
				{
					//System.err.print(" "+i+","+assignment[i]);
					for(int j = 0; j < averages.length; j++)
					{
						centers[assignment[i]][j]+=solutions[i][j];
					}
					counts[assignment[i]]++;
				}
				for(int i = 0; i < k; i++)
				{
					if(counts[i]!=0)
					{
						for(int j = 0; j < averages.length; j++)
						{
							centers[i][j]=centers[i][j]/(counts[i]);
						}
					}
					else
					{
						//whoops
						System.err.println("One of the clusters didn't get assigned at all!");
						return null;
					}
				}
				//check against old centers
				//System.err.println("----------------");
				for(int i = 0; i < centers.length; i++)
				{
					for(int j = 0; j < centers[i].length; j++)
					{
						if(centers[i][j] != oldcenters[i][j])
						{
							changing = true;
						}
						//System.err.println(oldcenters[i][j]+"/"+centers[i][j]);
						oldcenters[i][j] = centers[i][j];
					}
				}
				if(!changing)
				{
					result = new ArrayList<ArrayList<double[]>>();
					for(int i = 0; i < k; i++)
						result.add(new ArrayList<double[]>());
					for(int i = 0; i < assignment.length; i++)
					{
						double[] copy = new double[solutions[i].length];
						for(int j = 0; j < copy.length; j++)
							copy[j] = solutions2.get(i)[j];
						result.get(assignment[i]).add(copy);
					}
				}
			}
			return result;
	}

	/**
	 * Returns an agglomerative clustering of the solutions with cluster count = clusters 
	 * @param solutions ArrayList of double[]
	 * @param columns a boolean array the size of the data indicating which indeces to cluster
	 * @return ArrayList of ArrayList of double[]
	 */
	public static ArrayList<ArrayList<ArrayList<double[]>>> heirarchalCluster(
			ArrayList<double[]> solutions, boolean[] cluster_mask) {
		//agglomerate until k
		ArrayList<ArrayList<ArrayList<double[]>>> clusterings = new ArrayList<ArrayList<ArrayList<double[]>>>();
		ArrayList<ArrayList<double[]>> clustering = new ArrayList<ArrayList<double[]>>();
		ArrayList<double[]> averages = new ArrayList<double[]>();  //average values for each cluster
		for(int j = 0; j < solutions.size(); j++)
		{
			clustering.add(new ArrayList<double[]>());
			clustering.get(clustering.size()-1).add(solutions.get(j));
			averages.add(Methods.duplicate(solutions.get(j)));
		}
		ArrayList<ArrayList<double[]>> copy2 = new ArrayList<ArrayList<double[]>>();
		copy2.addAll(clustering);
		clusterings.add(0,copy2);
		for(int i = 0; i < solutions.size()-1; i++)
		{
			System.err.println(""+((i*100)/(solutions.size())+"% done."));

			//find the two closest clusters using an appropriate distance measure
			int best_a, best_b;
			best_a=best_b = 0;
			double bestDistance = Double.POSITIVE_INFINITY;
			for(int j = 0; j < clustering.size();j ++)
			{
				for(int k = j+1; k < clustering.size(); k++ )
				{
					double dist  = 0;
					for(int l = 0; l < averages.get(j).length; l++)
					{
						double d = Math.abs(averages.get(j)[l]-averages.get(k)[l])/(averages.get(j)[l]+averages.get(k)[l]);
						if(averages.get(j)[l]+averages.get(k)[l] == 0) d = 0;
						if(cluster_mask[l])dist+=d;
					}
					if(dist < bestDistance)
					{
						best_a=j;
						best_b=k;
						bestDistance=dist;
					}
				}
			}
			//combine best_a and best_b
			ArrayList<double[]> nc = new ArrayList<double[]>();
			nc.addAll(clustering.get(best_a));
			nc.addAll(clustering.get(best_b));
			double[] newaverage = Methods.averageArrayElements(averages.get(best_a),averages.get(best_b),clustering.get(best_a).size(),clustering.get(best_b).size());
			clustering.remove(best_b);
			clustering.remove(best_a);
			averages.remove(best_b);
			averages.remove(best_a);
			clustering.add(nc);
			averages.add(newaverage);
			ArrayList<ArrayList<double[]>> copy = new ArrayList<ArrayList<double[]>>();
			copy.addAll(clustering);
			clusterings.add(0,copy);
		}//main loop
		System.err.println();
		return clusterings;
	}

	/**
	 *  Returns a double[] representing the average of each of the kth elements in a and b
	 * @param a
	 * @param b
	 * @return
	 */
	private static double[] averageArrayElements(double[] a, double[] b, int w1, int w2) {
		// TODO Auto-generated method stub
		double[] c = new double[a.length];
		for(int i = 0; i < c.length; i++)
		{
			c[i]=(a[i]*w1+b[i]*w2)/(w1+w2);
		}
		return c;
	}

	public static int minIndex(double[] s) {
		// TODO Auto-generated method stub
		double max = Double.POSITIVE_INFINITY;
		int index = 0;
		for(int i = 0; i < s.length; i++)
		{
			if(s[i] < max)
			{
				max=s[i];
				index = i;
			}
		}
		return index;
	}

	public static double factorial(int i) {
		double f = 1;
		for(int j = 2; j <= i; j++)
		{
			f=f*j;
		}
		return f;
	}
	
	public static double factorial2(int i) {
		if(i==2)return 2;
		if(i==3)return 6;
		if(i==4)return 24;
		if(i==5)return 120;
		if(i==6)return 720;
		if(i==7)return 5040;
		if(i==8)return 40320;
		if(i==9)return 362880;
		if(i==10)return 3628800;
		if(i==11)return 39916800;
		if(i==12)return 479001600;
		if(i==13)return 6227020800.0;
		if(i==14)return 87178291200.0;
		if(i==15)return 1307674368000.0;
		if(i==16)return 20922789888000.0;
		return 1;
	}

	public static int[] duplicate(int[] vals) {
		int[] copy = new int[vals.length];
		for(int i = 0; i < copy.length; i++)
			copy[i] = vals[i];
		return copy;
	}

	public static ArrayList<ArrayList<ArrayList<double[]>>> heirarchalClusterWithArms(
			ArrayList<double[]> solutions,int nparams) {
		//agglomerate until k
		ArrayList<ArrayList<ArrayList<double[]>>> clusterings = new ArrayList<ArrayList<ArrayList<double[]>>>();
		ArrayList<ArrayList<double[]>> clustering = new ArrayList<ArrayList<double[]>>();
		ArrayList<double[]> averages = new ArrayList<double[]>();  //average values for each cluster
		for(int j = 0; j < solutions.size(); j++)  //one solution per cluster to start
		{
			clustering.add(new ArrayList<double[]>());
			clustering.get(clustering.size()-1).add(solutions.get(j));
			averages.add(Methods.duplicate(solutions.get(j)));  
		}
		ArrayList<ArrayList<double[]>> copy2 = new ArrayList<ArrayList<double[]>>();
		copy2.addAll(clustering);
		clusterings.add(0,copy2);
		for(int i = 0; i < solutions.size()-1; i++)
		{
			System.err.println(""+((i*100)/(solutions.size())+"% done."));

			//find the two closest clusters using an appropriate distance measure
			int best_a, best_b;
			best_a=best_b = 0;
			double bestDistance = Double.POSITIVE_INFINITY;
			for(int j = 0; j < clustering.size();j ++)	//cluster j
			{
				for(int k = j+1; k < clustering.size(); k++ )  //cluster k
				{
					double dist  = 0;
					for(int l = 2; l < nparams+2; l++)  //parameters
					{
						double d = 0;
						if(averages.get(k)[l] > averages.get(j)[l])	//average value larger in k compared to j for param L
						{
							//distance d = (BIGGER-SMALLER)-(SMALLER_UP+LARGER_DOWN) = distance between error bars
							//d = (averages.get(k)[l]-averages.get(j)[l])/Math.exp(averages.get(j)[l+nparams*2]+averages.get(k)[l+nparams]);
							d = (Math.max(0,(averages.get(k)[l]+averages.get(k)[l+nparams])-(averages.get(j)[l]+averages.get(j)[l+nparams*2])))/(averages.get(k)[l]+averages.get(j)[l]);
						}
						else  //distance d = (BIGGER-SMALLER)-(SMALLER_UP+LARGER_DOWN) = distance between error bars
							d = (Math.max(0,(averages.get(j)[l]+averages.get(j)[l+nparams])-(averages.get(k)[l]+averages.get(k)[l+nparams*2])))/(averages.get(k)[l]+averages.get(j)[l]);
						dist+=d;  //distance gets the distance between these two clusterings along the Lth parameter 
					}
					if(dist < bestDistance)
					{
						best_a=j;  //should combine clustering j and clustering k
						best_b=k;
						bestDistance=dist;
					}
				}
			}
			//combine best_a and best_b
			ArrayList<double[]> nc = new ArrayList<double[]>();  //new cluster
			nc.addAll(clustering.get(best_a));	//add all from besta
			nc.addAll(clustering.get(best_b));   //add all from bestb
			double[] newaverage = new double[2+nparams*3];  //update average (should reflect
			for(int j = 2; j < nparams+2;j++)
			{
				for(int k = 0; k < clustering.get(best_a).size(); k++ )
					newaverage[j] += clustering.get(best_a).get(k)[j];
				for(int k = 0; k < clustering.get(best_b).size(); k++ )
					newaverage[j] += clustering.get(best_b).get(k)[j];
				newaverage[j] = newaverage[j]/(clustering.get(best_a).size()+clustering.get(best_b).size()); //divide by total
				if(averages.get(best_a)[j] > averages.get(best_b)[j])
				{
					newaverage[j+nparams] = averages.get(best_a)[j]-averages.get(best_a)[j+nparams]; //bad
				}
			}
			clustering.remove(best_b);
			clustering.remove(best_a);
			averages.remove(best_b);
			averages.remove(best_a);
			clustering.add(nc);
			averages.add(newaverage);
			ArrayList<ArrayList<double[]>> copy = new ArrayList<ArrayList<double[]>>();
			copy.addAll(clustering);
			clusterings.add(0,copy);
		}//main loop
		System.err.println();
		return clusterings;
	}
	
	

	
	/**
	 * Helper function for calculating the lognormal PDF quickly at some value x
	 * @param u
	 * @param s
	 * @param x
	 * @return
	 */
	public static double getLogNormalPDF(double u, double s, double x)
	{
		double s1 = 2.50662827;  //sqrt(2*pi)
		double a = (1.0/(x*s*s1));
		double b = Math.exp(-Math.pow(Math.log(x)-u,2)/(2*s*s));
		return a*b;
	}
	
	
	/**
	 * Calculated the cumulative LogNormal Probability Distribution between the values a and b
	 * @param mu
	 * @param sigma
	 * @param a
	 * @param b
	 * @return
	 */
	public static double getLogNormalCDF(double mu, double sigma, double a,
			double b) {
		if(a == 0 || b==0) return 0;
		double k = 0.140012289; //8*(Math.PI-3)/(3*Math.PI*(4-Math.PI)); 
		double r2 = 1.41421356;//Math.sqrt(2);
		double A = (Math.log(a)-mu)/(r2*sigma);
		double B = (Math.log(b)-mu)/(r2*sigma);
		double A2 = A*A;
		double B2 = B*B;
		double k2 = 1.27323954;//4.0/Math.PI;
		double left = (B/Math.abs(B))*Math.sqrt(1-Math.exp(-B2*(k2+k*B2)/(1+k*B2)));
		double right = (A/Math.abs(A))*Math.sqrt(1-Math.exp(-A2*(k2+k*A2)/(1+k*A2)));
		return 0.5*(left-right);
	}
	
	/**
	 * Computes the lognormal CDF up to some value b
	 * @param mu
	 * @param sigma
	 * @param b
	 * @return
	 */
	public static  double getLogNormalCDF(double mu, double sigma, 
			double b) {
		double k = 0.140012289; //8*(Math.PI-3)/(3*Math.PI*(4-Math.PI)); 
		double r2 = 1.41421356;//Math.sqrt(2);
		double B = (Math.log(b)-mu)/(r2*sigma);
		double B2  = B*B;
		double k2 = 1.27323954;//4.0/Math.PI;
		double left = (B/Math.abs(B))*Math.sqrt(1-Math.exp(-B2*(k2+k*B2)/(1+k*B2)));
		return 0.5*(1+left);
	}
	
	/**
	 * Computes the gaussian cdf at the value a
	 * @param mu
	 * @param sigma
	 * @param a
	 * @return
	 */
	public static  double getGaussianCDF(double mu, double sigma, double a) {
		double k = 0.140012289; //8*(Math.PI-3)/(3*Math.PI*(4-Math.PI)); 
		double r2 = 1.41421356;//Math.sqrt(2);
		double A = (a-mu)/(r2*sigma);
		double A2 = A*A;
		double k2 = 1.27323954;//4.0/Math.PI;
		double right = (A/Math.abs(A))*Math.sqrt(1-Math.exp(-A2*(k2+k*A2)/(1+k*A2)));
		return 0.5*(1+right);
	}
	
	
	/**
	 * Returns the gaussian cdf between a and b
	 * @param mu
	 * @param sigma
	 * @param a
	 * @param b
	 * @return
	 */
	public static  double getGaussianCDF(double mu, double sigma, double a,
			double b) {
		double k = 0.140012289; //8*(Math.PI-3)/(3*Math.PI*(4-Math.PI)); 
		double r2 = 1.41421356;//Math.sqrt(2);
		double A = (a-mu)/(r2*sigma);
		double B = (b-mu)/(r2*sigma);
		double A2 = A*A;
		double B2 = B*B;
		double k2 = 1.27323954;//4.0/Math.PI;
		double left = (B/Math.abs(B))*Math.sqrt(1-Math.exp(-B2*(k2+k*B2)/(1+k*B2)));
		double right = (A/Math.abs(A))*Math.sqrt(1-Math.exp(-A2*(k2+k*A2)/(1+k*A2)));
		return 0.5*(left-right);
	}
	
	/**
	 * Computes the gaussian pdf at value x
	 * @param mu
	 * @param sigma
	 * @param x
	 * @return
	 */
	public static  double getGaussianPDF(double mu, double sigma, double x)
	{
		double s1 = 2.50662827;  //sqrt(2*pi)
		double k = 1/(sigma*s1);
		double e = Math.exp(-Math.pow(x-mu,2)/(2*sigma*sigma));
		return e*k;
	}
	

	/**
	 * New and improved computeCytons with pFs affecting both division and death (PCyton Model)
	 * @param params
	 * @return
	 */
	public static  double[][][] computeCytons(double[] params, int steps, int Generations, int T)
	{
		//lets itterate over all steps and then output
		double[][] NumDiv = new double[Generations][steps];
		double[][] NumDie = new double[Generations][steps];
		double[] divPDF0 = new double[steps];
		double[] deathPDF0 = new double[steps];
		double[] divPDFblast = new double[steps];
		double[] deathPDFblast = new double[steps];
		//double[] divCDF0 = new double[steps];
		//double[] deathCDF0 = new double[steps];
		//double[] divCDFblast  = new double[steps];
		//double[] deathCDFblast = new double[steps];
		double[] mechDeathPDF = new double[steps];
		double[] mechDeathCDF = new double[steps];
		double[][] LiveMatrix = new double[Generations+1][steps+1];
		double[][] DeadMatrix = new double[Generations+1][steps+1];
		double[] pF = new double[Generations+1];
		double dt =(((double)T)/steps);
		double pF0 = params[10];  //fit this as well
		try
		{
			double[] ddcurve = new double[Generations+1];
			ddcurve[0] = 1-pF0;
			double[]ddcdf = new double[Generations+1];
			ddcdf[0] = 1-pF0;

			for(int j = 1; j <= Generations; j++)
			{
				ddcurve[j] = pF0*(getGaussianCDF(params[8],params[9],j-1,j)/(1-getGaussianCDF(params[8],params[9],0)));
				ddcdf[j] = ddcdf[j-1]+ddcurve[j];
			}
			//ddcdf[Generations] = ddcdf[Generations-1]+pF0*(getGaussianCDF(params[8],params[9],Generations-1,Generations)/(1-getGaussianCDF(params[8],params[9],1)));
			pF[0] = params[10];
			for(int i = 1; i < Generations; i++)
			{
				if((1-ddcdf[i])/(1-ddcdf[i-1]) > 0)
					pF[i] = (1-ddcdf[i])/(1-ddcdf[i-1]);
				else
					pF[i] = 0;
			}
			for(int i = 1; i < steps; i ++)
			{
				divPDF0[i]=getLogNormalCDF(params[0], params[1], (i-1)*dt,(i)*dt);
				//divCDF0[i]=getLogNormalCDF(params[0], params[1],((i)*dt));
				divPDFblast[i]=getLogNormalCDF(params[2], params[3], (i-1)*dt,(i)*dt);
				//divCDFblast[i]=getLogNormalCDF(params[2], params[3],((i)*dt));
				deathPDF0[i]=getLogNormalCDF(params[4], params[5], (i-1)*dt,(i)*dt);
				//deathCDF0[i]=getLogNormalCDF(params[4], params[5],(i)*dt);
				deathPDFblast[i]=getLogNormalCDF(params[6], params[7], (i-1)*dt,(i)*dt);
				//deathCDFblast[i]=getLogNormalCDF(params[6], params[7],((i)*dt));			
				//mechDeathPDF[i]=(1-Math.exp(-params[13]*(i)*dt))-(1-Math.exp(-params[13]*((i-1)*dt)));  //Mechanical death is not used and hence not even calculated
				//mechDeathCDF[i] = (1-Math.exp(-params[13]*(i)*dt));
			}
		}
		catch(Exception e)
		{
			e.printStackTrace();
		}
		double sumdied = 0;
		//double sumdived = 0;
		for(int i = 1 ;i < steps; i++)
		{
			sumdied+=deathPDF0[i];  //up to AND including (some already die at time 1 )
//			if(i > 1)  //up to but not including (none have divided yet at time 1 since you would have had to divide at time 0)
//				sumdived+=divPDF0[i-1];
			NumDiv[0][i] = pF[0]*params[11]*(1-params[12]*mechDeathCDF[i])*divPDF0[i];
			NumDie[0][i] = (1-pF[0])*params[11]*((1-params[12]*mechDeathCDF[i])*deathPDF0[i]+(1-(1-pF[0])*sumdied)*params[12]*mechDeathPDF[i]);  //haven't divided, and died in one of two ways: mechanical death or normal lognormal death
		}
		for(int i = 1; i < Generations; i++)
		{		
			for(int j = 1; j < steps ; j++)
			{
				double totalDividedFromPreviousGen = 0;
				for(int jprime = 0; jprime <= j-1; jprime++)
				{
					totalDividedFromPreviousGen=NumDiv[i-1][jprime];  //the cohort from previous generation that divided at the same time
					//double probNotDead = 1-(1-pF[i])*deathCDFblast[j-jprime];  //probability that that cohort did not die up until this time (low for lower jprime)
					//double probNotDiv = 1;//probability that that cohort did not divide up until 2 time steps ago (could have divided last time but this is already accounted for)
					//if(j-jprime >= 1)  //if enough time has passed to make this possible
					//	probNotDiv=1-pF[i]*divCDFblast[j-1-jprime];  //subtract the fraction expected to divide
					//double survivingFractionOfCohort = probNotDead; //probability that this cohort survives still
					double chanceOfCohortDividingRightNow = divPDFblast[j-jprime];  //for this cohort it is now time j-jprime
					NumDiv[i][j] += 2*pF[i]*totalDividedFromPreviousGen*chanceOfCohortDividingRightNow; //division at this time for generation i
					//double chanceOfCohortStillUndivided = probNotDiv; //probability that this cohort hasn't divided yet
					double chanceOfCohortDyingRightNow = (1-pF[i])*deathPDFblast[j-jprime];  //the cohort's age is j-jprime
					NumDie[i][j] += 2*totalDividedFromPreviousGen*chanceOfCohortDyingRightNow;  //death right now for this cohort added to the total pool of dead at this time for this gen
				}			
			}
		}
		//lets populate the alive and death matrices
		for(int i = 0; i < Generations; i++)
		{
			if(i==0)
				LiveMatrix[i][0] = params[11];
			else
				LiveMatrix[i][0] = 0;  //not necessary really
			double sumDiei = 0;
			sumDiei += NumDie[i][0];
			//TODO Fix dead matrix such that you don't have negative cells registering as dead
			DeadMatrix[i][0] = NumDie[i][0];  //first round of deaths
			for(int j = 1; j < steps; j++)
			{
				if(i==0) LiveMatrix[i][j] = LiveMatrix[i][j-1]-NumDiv[i][j]-NumDie[i][j];
				else LiveMatrix[i][j] = LiveMatrix[i][j-1]+2*NumDiv[i-1][j]-NumDie[i][j]-NumDiv[i][j];

				if(LiveMatrix[i][j] < 0) LiveMatrix[i][j] = 0;
				DeadMatrix[i][j] = sumDiei;
			}
		
	
		}
		double[][][] results = new double[4][Generations][steps];
		results[0] = LiveMatrix;
		results[1] = DeadMatrix;
		results[2] = NumDiv;
		results[3] = NumDie;
		return results;
	}
	
	/**
	 * Computes the Cyton Model solutions (as published by Hawkins and Hodgkin)
	 * @param params
	 * @param steps
	 * @param Generations
	 * @param T
	 * @return
	 */
	public static  double[][][] computeCytonsOld(double[] params, int steps, int Generations, int T)
	{
		//lets itterate over all steps and then output
		double[][] NumDiv = new double[Generations][steps];
		double[][] NumDie = new double[Generations][steps];
		double[] divPDF0 = new double[steps];
		double[] deathPDF0 = new double[steps];
		double[] divPDFblast = new double[steps];
		double[] deathPDFblast = new double[steps];
		//double[] divCDF0 = new double[steps];
		//double[] deathCDF0 = new double[steps];
		double[] divCDFblast  = new double[steps];
		double[] deathCDFblast = new double[steps];
		double[][] LiveMatrix = new double[Generations+1][steps+1];
		double[][] DeadMatrix = new double[Generations+1][steps+1];
		//double[][] NumIdle = new double[Generations][steps];
		double[] pF = new double[Generations];
		double dt =(((double)T)/steps);
		double pF0 = params[10];  //fit this as well
		try
		{
			double[] ddcurve = new double[Generations+1];
			ddcurve[0] = 1-pF0;
			double[]ddcdf = new double[Generations+1];
			ddcdf[0] = 1-pF0;
			for(int j = 1; j <= Generations; j++)
			{
				ddcurve[j] = pF0*(getGaussianCDF(params[8],params[9],j,j+1)/(1-getGaussianCDF(params[8],params[9],1)));
				ddcdf[j] = ddcdf[j-1]+ddcurve[j];
			}
			//ddcdf[Generations] = ddcdf[Generations-1]+pF0*(getGaussianCDF(params[8],params[9],Generations-1,Generations)/(1-getGaussianCDF(params[8],params[9],1)));
			pF[0] = params[10];
			for(int i = 1; i < Generations; i++)
			{
				if((1-ddcdf[i+1])/(1-ddcdf[i]) > 0)
					pF[i] = (1-ddcdf[i+1])/(1-ddcdf[i]);
				else
					pF[i] = 0;
			}
			
			for(int i = 1; i < steps; i ++)
			{
				divPDF0[i]=getLogNormalCDF(params[0], params[1], (i-1)*dt,(i)*dt);
				//divCDF0[i]=getLogNormalCDF(params[0], params[1],((i)*dt));
				divPDFblast[i]=getLogNormalCDF(params[2], params[3], (i-1)*dt,(i)*dt);
				divCDFblast[i]=getLogNormalCDF(params[2], params[3],((i)*dt));
				deathPDF0[i]=getLogNormalCDF(params[4], params[5], (i-1)*dt,(i)*dt);
				//deathCDF0[i]=getLogNormalCDF(params[4], params[5],(i)*dt);
				deathPDFblast[i]=getLogNormalCDF(params[6], params[7], (i-1)*dt,(i)*dt);
				deathCDFblast[i]=getLogNormalCDF(params[6], params[7],((i)*dt));			
				//mechDeathPDF[i]=(1-Math.exp(-params[13]*(i)*dt))-(1-Math.exp(-params[13]*((i-1)*dt)));  //No mechanical death so its not even calculated
				//mechDeathCDF[i] = (1-Math.exp(-params[13]*(i)*dt));
			}
		}
		catch(Exception e)
		{
			e.printStackTrace();
		}
		double sumdied = 0;
		double sumdived = 0;
		for(int i = 0 ;i < steps; i++)
		{
			sumdied+=deathPDF0[i];  //up to AND including (some already die at time 0 )
			if(i > 0)  //up to but not including (none have divided yet at time 0 since you would have had to divide at time -1)
				sumdived+=divPDF0[i-1];
			NumDiv[0][i] = pF[0]*params[11]*(1-sumdied)*divPDF0[i];
			NumDie[0][i] = params[11]*(1-pF[0]*sumdived)*deathPDF0[i];
		}
		for(int i = 1; i < Generations; i++)
		{		
			for(int j = 1; j < steps ; j++)
			{
				double totalDividedFromPreviousGen = 0;
				for(int jprime = 0; jprime <= j-1; jprime++)
				{
					totalDividedFromPreviousGen=NumDiv[i-1][jprime];  //the cohort from previous generation that divided at the same time
					double probNotDead = 1-deathCDFblast[j-1-jprime];  //probability that that cohort did not die up until this time (low for lower jprime)
					double probNotDiv = 1;//probability that that cohort did not divide up until 2 time steps ago (could have divided last time but this is already accounted for)
					if(j-jprime >= 2)  //if enough time has passed to make this possible
						probNotDiv-=pF[i]*divCDFblast[j-2-jprime];  //subtract the fraction expected to divide
					double survivingFractionOfCohort = probNotDead; //probability that this cohort survives still
					double chanceOfCohortDividingRightNow = divPDFblast[j-jprime];  //for this cohort it is now time j-jprime
					NumDiv[i][j] += 2*pF[i]*totalDividedFromPreviousGen*survivingFractionOfCohort*chanceOfCohortDividingRightNow; //division at this time for generation i
					double chanceOfCohortStillUndivided = probNotDiv; //probability that this cohort hasn't divided yet
					double chanceOfCohortDyingRightNow = deathPDFblast[j-jprime];  //the cohort's age is j-jprime
					NumDie[i][j] += 2*totalDividedFromPreviousGen*chanceOfCohortStillUndivided*chanceOfCohortDyingRightNow;  //death right now for this cohort added to the total pool of dead at this time for this gen
				}			
			}
		}
		//lets populate the alive and death matrices
		for(int i = 0; i < Generations; i++)
		{
			if(i==0)
				LiveMatrix[i][0] = params[11];
			else
				LiveMatrix[i][0] = 0;  //not necessary really
			double sumDiei = 0;
			sumDiei += NumDie[i][0];
			//TODO Fix dead matrix such that you don't have negative cells registering as dead
			DeadMatrix[i][0] = NumDie[i][0];  //first round of deaths
			for(int j = 1; j < steps; j++)
			{
				if(i==0) LiveMatrix[i][j] = LiveMatrix[i][j-1]-NumDiv[i][j]-NumDie[i][j];
				else LiveMatrix[i][j] = LiveMatrix[i][j-1]+2*NumDiv[i-1][j]-NumDie[i][j]-NumDiv[i][j];
				if(LiveMatrix[i][j] < 0) LiveMatrix[i][j] = 0;
				DeadMatrix[i][j] = sumDiei;
			}
		
	
		}
		double[][][] results = new double[4][Generations][steps];
		results[0] = LiveMatrix;
		results[1] = DeadMatrix;
		results[2] = NumDiv;
		results[3] = NumDie;
		return results;
	}

	
	public static double[] getLogParams(double[] params2, int nparams)
	{
		for(int i = 0; i < params2.length; i++)
		{
			if(Double.isNaN(params2[i])) params2[i] = 0;
		}
		double[] params = new double[params2.length];
		if(params.length == nparams)
		{
			for(int i = 0; i <8; i+=2)  //convert to log scale params (u and s)
			{
				double log = Math.log(1+(params2[i+1]*params2[i+1])/(params2[i]*params2[i]));
				double sigma = Math.sqrt(log);
				double mean = Math.log(params2[i])-0.5*log;
				params[i] = mean;
				params[i+1] = sigma;
			}
			for(int i = 8; i < nparams;i++)
				params[i] = params2[i];
		}
		else if (params.length == 3*nparams)
		{
			for(int i = 0; i <8; i+=2)  //convert to log scale params (u and s)
			{
				double downmean = params2[i]-params2[i+nparams];
				double downsd =  params2[i+1]-params2[i+1+nparams];
				double upmean = params2[i]+params2[i+nparams*2];
				double upsd =  params2[i+1]+params2[i+1+nparams*2];
				double downlog = Math.log(1+(downsd*downsd)/(downmean*downmean));
				double downsigma = Math.sqrt(downlog);
				double downmu = Math.log(downmean)-0.5*downlog;
				double uplog = Math.log(1+(upsd*upsd)/(upmean*upmean));
				double upsigma = Math.sqrt(uplog);
				double upmu = Math.log(upmean)-0.5*uplog;
				double log = Math.log(1+(params2[i+1]*params2[i+1])/(params2[i]*params2[i]));
				double sigma = Math.sqrt(log);
				double mu = Math.log(params2[i])-0.5*log;
				params[i] = mu;
				params[i+1] = sigma;
				params[i+nparams] = mu-downmu;
				params[i+1+nparams] = sigma-downsigma;
				params[i+nparams*2] = upmu-mu;
				params[i+1+nparams*2] = upsigma-sigma;
			}
			for(int i = 8; i < nparams;i++)
			{
				params[i] = params2[i];
				params[i+nparams] = params2[i+nparams];
				params[i+2*nparams] = params2[i+nparams];
			}
		}
		return params;
	}

	public static double[] generateGaussians(double mu, double sigma, int i) {
		double[] answer = new double[i];
		Random gen = new Random();
		for(int j = 0; j < i; j++)
		{
			answer[j]=gen.nextGaussian()*sigma+mu;
		}
		return answer;
	}
	
	public static double[] generateLogNormals(double mu, double sigma, int i) {
		double[] answer = new double[i];
		Random gen = new Random();
		for(int j = 0; j < i; j++)
		{
			answer[j]=Math.exp(mu+sigma*gen.nextGaussian());
		}
		return answer;
	}

	public static double std(int[] x) {
		double std = 0;
		double u = average(x);
		for(int i = 0; i < x.length; i++)
		{
			std+= (u-x[i])*(u-x[i]);
		}
		return (Math.sqrt(std/x.length));
	}
}
