package common;

import java.awt.BorderLayout;
import java.awt.Component;
import java.awt.Dimension;
import java.awt.FlowLayout;
import java.awt.Font;
import java.awt.GridLayout;
import java.awt.Toolkit;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.awt.event.MouseMotionListener;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;

import javax.swing.ImageIcon;
import javax.swing.JButton;
import javax.swing.JComponent;
import javax.swing.JMenuItem;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JPopupMenu;

import common.*;

public class Plotter extends JPanel{

	
	protected TreeNode node = null;
	public int mode = 0; //
	public double clickx;
	public double clicky;
	protected HashMap<String,JButton> buttons = new HashMap<String,JButton>();

	protected JCheckButton xlog;
	protected JCheckButton ylog;
	protected Graph grf;
	protected Plotter self;
	protected JPopupMenu xmenu,ymenu;
	protected JMenuItem[] xmenus;
	protected JMenuItem[] ymenus;
	protected Gallery parent;
	protected JPanel bottom;
	protected StatsBox stats;
	
	private ArrayList<PlotterListener> listeners = new ArrayList<PlotterListener>();

	public Graph getGrf() {
		return grf;
	}
	
	protected void initializeGraph()
	{
		grf = new Graph(node, this);
	}

	

	public TreeNode getNode() {
		return node;
	}

	public void setNode(TreeNode node) {
		this.node = node;
	}

	public JCheckButton getXlog() {
		return xlog;
	}

	public JCheckButton getYlog() {
		return ylog;
	}

	
	JPopupMenu gmenu;
	JMenuItem copyMI;
	JMenuItem copyMI2;
	

	public void registerPlotterListener(PlotterListener pl)
	{
		listeners.add(pl);

	}
	
    private void updateListeners() {
		for(int i = 0; i < listeners.size(); i++)
			listeners.get(i).plotterUpdated(this);
		parent.refreshAllStats();
	}
    
    protected void pmouseDragged(MouseEvent evt)
    {
    	//System.err.println("mode "+mode+" grf.currentGate "+grf.currentGate);
		DoubleDouble dd = node.getTranslatedPoint(evt.getX(), evt.getY(), grf);
		double pointx = dd.x;
		double pointy = dd.y;
		// TODO Auto-generated method stub
		if((mode==1||mode==3||mode==4) && evt.getButton() == 0)
		{
			node.zoomx2 = evt.getX();
			node.zoomy2 = evt.getY();
			grf.updateUI();
		}
    	else if(mode==5 && grf.currentGate != null)//move the current vertex
    	{
    		grf.currentGate.adjust(grf.currentVertex, pointx, pointy);
    		grf.updateUI();
    	}
    	else if(mode==0 && grf.currentGate != null)// move the whole gate
    	{
    		if(clickx != 0 && clicky != 0)
    			grf.currentGate.translate(pointx-clickx, pointy-clicky);
    		clickx = pointx;
    		clicky = pointy;
    		grf.updateUI();
    	}
    }
    
    protected void pmouseMoved(MouseEvent evt)
    {
		// TODO Auto-generated method stub
		DoubleDouble dd = node.getTranslatedPoint(evt.getX(), evt.getY(), grf);
		double pointx = dd.x;
		double pointy = dd.y;    		
    	if(mode == 2 && grf.currentGate != null)
    	{
    		grf.currentGate.updateLastPoint(pointx,pointy);
    		grf.updateUI();
    	}
    	if(evt.isControlDown() && evt.getButton()==MouseEvent.BUTTON1)
    	{
			// TODO Auto-generated method stub
			if((mode==1||mode==3||mode==4) && evt.getButton() == 0)
			{
				node.zoomx2 = evt.getX();
				node.zoomy2 = evt.getY();
				grf.updateUI();
			}
	    	else if(mode==5 && grf.currentGate != null)
	    	{
	    		grf.currentGate.p.adjust(grf.currentVertex, pointx, pointy);
	    		grf.updateUI();
	    	}
	    	else if(mode==0 && grf.currentGate != null)
	    	{
	    		if(clickx != 0 && clicky != 0)
	    			grf.currentGate.translate(pointx-clickx, pointy-clicky);
	    		clickx = pointx;
	    		clicky = pointy;
	    		grf.updateUI();
	    	}
    	}
   	}
    
    protected void pmousePressed(MouseEvent evt)
    {
		requestFocusInWindow();
		parent.setCurrent_plot(self);
		DoubleDouble converted =node.getTranslatedPoint(evt.getX(), evt.getY(),self.grf);
		double pointx =converted.x;
		double pointy = converted.y;
		int mouse_button = -1;
		if(evt.getButton()==MouseEvent.BUTTON1)
		{
			mouse_button = 1;
		}
		else if(evt.getButton() == MouseEvent.BUTTON3)
		{
			mouse_button = 3;
		}
		else if(evt.isControlDown())
		{
			//assume mac
			mouse_button = 3;
		}
		//System.err.println("Clicked: "+evt.getX()+","+evt.getY()+" converted to: "+pointx+","+pointy+" mode is: "+mode+" button is "+evt.getButton()+" xmin="+node.xmin+" xmax="+node.xmax+" ymin="+node.ymin+" ymax="+node.ymax);
		if(mode == 0 && mouse_button==1) //left click
		{
			if(Math.abs(evt.getX()-7) <= 5 &&Math.abs(evt.getY()-5) <= 5)
			{
				//kill the window
				//parent.current_plot = null;
				parent.removeGraph(node,false);  // just want to close this one
				return;
			}
			if(node!= null && node.leaves.size() > 0) //check clicking on a gate
			{
				//System.err.println("Testing if we clicked on a gate");
				//there are some child nodes that could potentially be editable
				Iterator<TreeNode> itr = node.leaves.iterator();
				double d = Double.POSITIVE_INFINITY;
				TreeNode best = null;
				while(itr.hasNext())
				{
					TreeNode next = itr.next();
					
					double dist = 1;
					if(node.yindex> -1)dist=next.getGate().getDistanceToGate(pointx, pointy, node.xmax, node.ymax, node.xmin, node.ymin);
					else dist=next.getGate().getDistanceToGate(pointx, pointy, node.xmax, grf.getHeight(), node.xmin, 0);
					if(dist < d)
					{
						d = dist;
						best = next;
					}
				}
///							double fw =((double)(node.xmax-node.xmin))/grf.getWidth();
//						double fh = ((double)(node.ymax-node.ymin))/grf.getHeight();
//						double spixel_size = fw*fw+fh*fh;
					
				double logmult = 1;
				if(node.xlog) logmult*=3;
				if(node.ylog && node.gate != null && node.gate.gatetype <2) logmult*=3;
				System.err.println(" distance to gate was: "+d+" threshold: "+(logmult*5.0/Math.max(grf.getWidth(),grf.getHeight())));
				if(best != null && d < logmult*5.0/Math.max(grf.getWidth(),grf.getHeight()) )
				{
					grf.currentGate = best.getGate();
					
					int closestVertex = 0;
					if(node.yindex>-1) closestVertex=grf.currentGate.getClosestVertexWithinD(pointx, pointy, node.xmax, node.ymax, node.xmin, node.ymin,0.001);
					else closestVertex=grf.currentGate.getClosestVertexWithinD(pointx, pointy, node.xmax, grf.getHeight(), node.xmin, 0,0.001);
					System.err.println("Clicked on a gate! Identified closest vertex: "+closestVertex);
					if(closestVertex != -1)
					{
						grf.currentVertex = closestVertex;
						mode = 5;
					}
					grf.updateUI();
				}
				else
				{
					grf.currentGate = null;
					grf.currentVertex = -1;
					mode = 0;
					grf.updateUI();
				}
			}
		}
		else if(mode == 1 && mouse_button==1) //zooming
    	{
    		node.zoomx1 = evt.getX();
    		node.zoomy1 = evt.getY();
    	}
    	else if(mode == 2 && mouse_button==1)  //creating a polygon gate
    	{
    		if(grf.currentGate == null)
    			grf.currentGate = new Gate(node.xindex,node.yindex,node.xlog,node.ylog,"NewGate",0);
    		//lets figure out where x is hitting
    		if(grf.currentGate.p.getNpoints() > 2)
    		{
    			//check if we are back at the beginning
    			double dx = 0;
    			double dy = 0;
    			double rx = 0;
				double ry = 0;
				if(node.xlog)
				{
					dx =Math.log10( grf.currentGate.p.xpoints[0])-Math.log10(pointx);
					rx = 4*(Math.log10(node.xmax)-Math.log10(node.xmin))/(getWidth()-node.margin);
				}
				else
				{
					dx = grf.currentGate.p.xpoints[0]-pointx;
					rx =4*(node.xmax-node.xmin)/(getWidth()-node.margin);
				}
				if(node.ylog)
				{
					dy = Math.log10(grf.currentGate.p.ypoints[0])-Math.log10(pointy);
					ry = 4*(Math.log10(node.ymax)-Math.log10(node.ymin))/(getHeight()-node.margin);
				}
				else
				{
					dy = grf.currentGate.p.ypoints[0]-pointy;
					ry =4*(node.ymax-node.ymin)/(getHeight()-node.margin);
				}
    			if(dx*dx+dy*dy < rx*rx+ry*ry)
    			{
    				//we hit the begining again
    				//ask for a name
    				String gate_name = null;
    				while(gate_name == null)
    				{
    					gate_name =JOptionPane.showInputDialog("Gate name: ","Gate "+(node.leaves.size()+1));
    				}
    				grf.currentGate.name = gate_name;
//		    				if(dx*dx+dy*dy != 0)
//		    					currentGate.addPoint(currentGate.p.xpoints[0],currentGate.p.ypoints[0]);
    				grf.currentGate.updateLastPoint(grf.currentGate.p.xpoints[0],grf.currentGate.p.ypoints[0]);
    				
//		    				System.err.println("Done... last point "+currentGate.p.xpoints[currentGate.p.getNpoints()-1]+" "+currentGate.p.xpoints[currentGate.p.getNpoints()-1]);
    				TreeNode newNode = node.addAnd(grf.currentGate);
    				newNode.xindex=node.xindex;
    				newNode.yindex=node.yindex;
    				newNode.xlog = node.xlog;
    				newNode.ylog = node.ylog;
    				newNode.xmax = node.xmax;
    				newNode.xmin = node.xmin;
    				newNode.ymax = node.ymax;
    				newNode.ymin = node.ymin;
    				//newNode.inverse = node.inverse;
    				grf.currentGate = null;
    				mode = 0;
    				buttons.get("rgate").setEnabled(true);
    				buttons.get("gate").setEnabled(true);
    				buttons.get("1Dgate").setEnabled(true);
    				updateListeners();
    				return;
    			}
    			else //its just another point
    			{
    				if(grf.currentGate.p.xpoints.length > 0 && grf.currentGate.p.xpoints[grf.currentGate.p.xpoints.length-1]==pointx)
    				{
    					grf.currentGate.p.xpoints[grf.currentGate.p.xpoints.length-1]+=0.1;
    				//	System.err.println("Changed pointx");
    				}
    				if(grf.currentGate.p.getNpoints() == 0)
    					grf.currentGate.addPoint(pointx,pointy);
		    		grf.currentGate.addPoint(pointx,pointy);
    			}
    		}
    		else //if we only have 1 point
    		{
				if(grf.currentGate.p.xpoints.length > 0 && grf.currentGate.p.xpoints[grf.currentGate.p.xpoints.length-1]==pointx)
				{
					grf.currentGate.p.xpoints[grf.currentGate.p.xpoints.length-1]+=0.5;
					System.err.println("Changed pointx");
				}
				if(grf.currentGate.p.getNpoints() == 0)
					grf.currentGate.addPoint(pointx,pointy);
	    		grf.currentGate.addPoint(pointx,pointy);
    		}
    		grf.updateUI();
    	}
    	else if(mode == 3 && mouse_button==1)  //creating a rectangle gate
    	{
    		if(grf.currentGate == null)
    		{
    			grf.currentGate = new Gate(node.xindex,node.yindex,node.xlog,node.ylog,"NewGate",0);
    			grf.currentGate.setGateType(1); //rectangle gate
    			node.zoomx1 = evt.getX();  //temprarily remember where we started clicking
    			node.zoomy1 = evt.getY();
    		}
    		grf.updateUI();
    	}
    	else if(mode == 4 && mouse_button==1) //creating a 1D gate
    	{
    		if(grf.currentGate == null)
    		{
    			grf.currentGate = new Gate(node.xindex,node.xlog,pointx,pointy,"New1DGate");
    			grf.currentGate.setGateType(2);  // not necessary since this is automatically assumed given the constructor
    			node.zoomx1 = evt.getX();
    			node.zoomy1 = evt.getY();
    		}
    	}
   	}
    
    protected void pmouseReleased(MouseEvent evt)
    {
 		int mouse_button = -1;
		if(evt.getButton()==MouseEvent.BUTTON1)
		{
			mouse_button = 1;
		}
		else if(evt.getButton() == MouseEvent.BUTTON3)
		{
			mouse_button = 3;
		}
		else if(evt.isControlDown())
		{
			//assume mac
			mouse_button = 3;
		}
    	if(mode == 1 && mouse_button==1) //zoom finished
    	{
    		//set new bounds
    		mode = 0;
    		buttons.get("zoom").setEnabled(true);
    		//need to recompute new bounds for the plot
    		if(node.zoomx1> node.zoomx2)
    		{
    			int temp =node.zoomx2;
    			node.zoomx2 = node.zoomx1;
    			node.zoomx1=temp;
    		}
    		if(node.zoomy1< node.zoomy2)
    		{
    			int temp = node.zoomy2;
    			node.zoomy2 = node.zoomy1;
    			node.zoomy1=temp;
    		}
    		if(node.xlog)
    		{
    			double minpower = Math.log10(node.xmin); 
    			double maxpower = Math.log10(node.xmax);
    			double f1 = ((double)node.zoomx1-node.margin)/(grf.getWidth()-node.margin);
    			double f2 = ((double)node.zoomx2-node.margin)/(grf.getWidth()-node.margin);
    			f1 = Math.min(1,Math.max(0,f1));
    			f2 = Math.min(1,Math.max(0,f2));
    			node.xmin = (int)Math.pow(10,f1*(maxpower-minpower)+minpower);
    			node.xmax = (int)Math.pow(10,f2*(maxpower-minpower)+minpower);
    		}
    		else
    		{
    			double minpower =node.xmin; 
    			double maxpower = node.xmax;
    			double f1 = ((double)node.zoomx1-node.margin)/(grf.getWidth()-node.margin);
    			double f2 = ((double)node.zoomx2-node.margin)/(grf.getWidth()-node.margin);
    			f1 = Math.min(1,Math.max(0,f1));
    			f2 = Math.min(1,Math.max(0,f2));
    			node.xmin = (int)(f1*(maxpower-minpower)+minpower);
    			node.xmax = (int)(f2*(maxpower-minpower)+minpower);
    		}
    		if(node.ylog)
    		{
    			double minpower = Math.log10(node.ymin); 
    			double maxpower = Math.log10(node.ymax);
    			double f1 = ((double)(grf.getHeight()-node.margin-node.zoomy1))/(grf.getHeight()-node.margin);
    			double f2 = ((double)(grf.getHeight()-node.margin-node.zoomy2))/(grf.getHeight()-node.margin);
    			f1 = Math.min(1,Math.max(0,f1));
    			f2 = Math.min(1,Math.max(0,f2));
    			node.ymin = (int)Math.pow(10,f1*(maxpower-minpower)+minpower);
    			node.ymax = (int)Math.pow(10,f2*(maxpower-minpower)+minpower);
    		}
    		else
    		{
    			double minpower =node.ymin; 
    			double maxpower = node.ymax;
    			double f1 = ((double)(grf.getHeight()-node.margin-node.zoomy1))/(grf.getHeight()-node.margin);
    			double f2 = ((double)(grf.getHeight()-node.margin-node.zoomy2))/(grf.getHeight()-node.margin);
    			f1 = Math.min(1,Math.max(0,f1));
    			f2 = Math.min(1,Math.max(0,f2));
    			node.ymin = (int)(f1*(maxpower-minpower)+minpower);
    			node.ymax = (int)(f2*(maxpower-minpower)+minpower);
    		}
    		node.zoomx1=node.zoomx2=node.zoomy1=node.zoomy2=-1;
    		grf.redraw_needed = true;
    		updateUI();
    	}
    	else if (mode == 3)
    	{
    		//actually create the current rectangle gate (populate it with points).
			//ask for a name
			String gate_name = null;
			while(gate_name == null)
			{
				gate_name =JOptionPane.showInputDialog("Gate name: ","Rectangle "+(node.leaves.size()+1));
			}
			grf.currentGate.name = gate_name;
			DoubleDouble start =node.getTranslatedPoint(node.zoomx1, node.zoomy1,self.grf);
			double x =start.x;
			double y = start.y;
			DoubleDouble end =node.getTranslatedPoint(evt.getX(), evt.getY(),self.grf);
			double x2 =end.x;
			double y2 = end.y;
			
			
			grf.currentGate.addPoint(x, y);
    		grf.currentGate.addPoint(x, y2);
    		grf.currentGate.addPoint(x2, y2);
    		grf.currentGate.addPoint(x2, y);
    		grf.currentGate.addPoint(x, y);

			node.zoomx1 = -1;
			node.zoomx2 = -1;
			node.zoomy1 = -1;
			node.zoomy2 = -1;
			TreeNode newNode = node.addAnd(grf.currentGate);
			newNode.xindex=node.xindex;
			newNode.yindex=node.yindex;
			newNode.xlog = node.xlog;
			newNode.ylog = node.ylog;
			newNode.xmax = node.xmax;
			newNode.xmin = node.xmin;
			newNode.ymax = node.ymax;
			newNode.ymin = node.ymin;
			updateGateListeners(grf.currentGate);
			grf.currentGate = null;
			mode = 0;
			buttons.get("rgate").setEnabled(true);
			buttons.get("gate").setEnabled(true);
			buttons.get("1Dgate").setEnabled(true);
			grf.currentGate = null;	    
			//grf.redraw_needed = true;
			grf.updateUI();
			updateListeners();			    		
    	}
    	else if(mode==4)  //finished the 1D gate!
    	{
			//ask for a name
			String gate_name = null;
			while(gate_name == null)
			{
				gate_name =JOptionPane.showInputDialog("Gate name: ","Gate "+(node.leaves.size()+1));
			}
			grf.currentGate.name = gate_name;
    		DoubleDouble end =node.getTranslatedPoint(evt.getX(), evt.getY(),self.grf);
    		grf.currentGate.update1DMax(end.x);

			TreeNode newNode = node.addAnd(grf.currentGate);
			newNode.xindex=node.xindex;
			newNode.yindex=node.yindex;
			newNode.xlog = node.xlog;
			newNode.ylog = node.ylog;
			newNode.xmax = node.xmax;
			newNode.xmin = node.xmin;
			newNode.ymax = node.ymax;
			newNode.ymin = node.ymin;
			node.zoomx1 = -1;
			node.zoomx2 = -1;
			node.zoomy1 = -1;
			node.zoomy2 = -1;
			mode = 0;
			buttons.get("rgate").setEnabled(true);
			buttons.get("gate").setEnabled(true);
			buttons.get("1Dgate").setEnabled(true);
			grf.currentGate = null;
				    				
			grf.updateUI();
			updateListeners();
    	}
    	else if(mode==5) //editing the gate finished
    	{
    		mode = 0;
    		grf.node.getSourceFCS().propagateColors();
    		grf.node.propagateFromRoot();
    		grf.currentGate = null;
    		grf.currentVertex =  -1;
    		grf.redraw_needed = true;
    		updateListeners();
    		parent.updateAll();
    		
    	}
    	else if(mode ==0 && mouse_button==1) //maybe end editing the gate
    	{
    		grf.currentGate = null;
    		grf.currentVertex =  -1;
    		clickx = 0;
    		clicky = 0;
    		grf.node.getSourceFCS().propagateColors();
    		grf.node.propagateFromRoot();
    		grf.redraw_needed = true;
    		updateListeners();
    		parent.updateAll();
    	}
    	if ((evt.isPopupTrigger()||evt.isControlDown()) && mode == 0 && node != null) {
	          if(evt.getY() > grf.getHeight()-node.margin && evt.getX() >node.margin)
	        	  xmenu.show(evt.getComponent(), evt.getX(), evt.getY());
	          else if(evt.getY() <  grf.getHeight()-node.margin && evt.getX() < node.margin)
	        	  ymenu.show(evt.getComponent(), evt.getX(), evt.getY());
	          else
	        	  gmenu.show(evt.getComponent(),evt.getX(),evt.getY());
	        }
    
	}
    
	
	public Plotter(final Gallery parent, TreeNode tn)
	{
		this.parent = parent;
		self = this;

		addMouseMotionListener(new MouseMotionListener(){

			public void mouseDragged(MouseEvent evt) {
				//System.err.print(".");
				pmouseDragged(evt);
			}

			public void mouseMoved(MouseEvent evt) {
				pmouseMoved(evt);
			}
			
		});
		addMouseListener(new MouseAdapter() {
		    public void mousePressed(MouseEvent evt) {
		    	pmousePressed(evt);
		    	
		    }

			public void mouseReleased(MouseEvent evt) {
		    	pmouseReleased(evt);
		    }
			public void mouseClicked(MouseEvent arg0) {
				// TODO Auto-generated method stub
				
			}
			public void mouseEntered(MouseEvent arg0) {
				// TODO Auto-generated method stub
				
			}
			public void mouseExited(MouseEvent arg0) {
				// TODO Auto-generated method stub
				
			}
		});
		JButton zoomBtn= new JButton(Methods.getImageIcon("Zoom.png"));
		zoomBtn.addActionListener(new ActionListener(){

			public void actionPerformed(ActionEvent arg0) {
				// TODO Auto-generated method stub
				if(mode == 0)
				{
					buttons.get("zoom").setEnabled(false);
					mode=1;
				}
			}
		});
		zoomBtn.setPreferredSize(new Dimension(40,30));
		JButton gateBtn = new JButton(Methods.getImageIcon("Gate.png"));
		gateBtn.addActionListener(new ActionListener(){

			public void actionPerformed(ActionEvent arg0) {
				if(mode == 0)
				{
					// TODO Auto-generated method stub
					buttons.get("gate").setEnabled(false);
					if(grf.currentGate != null)
					{
						grf.currentGate = null;
						grf.currentVertex = -1;
					}
					mode=2;		
				}
			}
		});
		gateBtn.setPreferredSize(new Dimension(40,30));
		if(tn != null && tn.yindex == -1) gateBtn.setVisible(false);
		JButton rgateBtn = new JButton(Methods.getImageIcon("rGate.png"));
		rgateBtn.addActionListener(new ActionListener(){

			public void actionPerformed(ActionEvent arg0) {
				if(mode == 0)
				{
					// TODO Auto-generated method stub
					buttons.get("rgate").setEnabled(false);
					if(grf.currentGate != null)
					{
						grf.currentGate = null;
					}
					mode=3;		
				}
			}
		});
		rgateBtn.setPreferredSize(new Dimension(40,30));
		if(tn != null && tn.yindex == -1) rgateBtn.setVisible(false);
		JButton oneDgateBtn = new JButton(Methods.getImageIcon("1Dgate.png"));
		oneDgateBtn.addActionListener(new ActionListener(){

			public void actionPerformed(ActionEvent arg0) {
				if(mode == 0)
				{
					// TODO Auto-generated method stub
					buttons.get("1Dgate").setEnabled(false);
					if(grf.currentGate != null)
					{
						grf.currentGate = null;
					}
					mode=4;		
				}
			}
		});
		oneDgateBtn.setPreferredSize(new Dimension(40,30));
		if(tn != null && tn.yindex != -1) oneDgateBtn.setVisible(false);
		JButton undoZoomBtn = new JButton(Methods.getImageIcon("ZoomOut.png"));
		undoZoomBtn.setEnabled(false);
		undoZoomBtn.addActionListener(new ActionListener(){

			public void actionPerformed(ActionEvent arg0) {
				// TODO Auto-generated method stub
				node.xmin = node.ymin = 1;
				node.changeXIndex(node.xindex);
				node.changeYIndex(node.yindex);
				grf.redraw_needed = true;;
				updateUI();
			}
		});
		undoZoomBtn.setPreferredSize(new Dimension(40,30));

		// Set the component to show the popup menu
		gmenu = new JPopupMenu();
		copyMI = new JMenuItem("Copy image to clipboard");
		Font calibri = new Font("Calibri",Font.PLAIN,11);
		copyMI.setFont(calibri);
		gmenu.add(copyMI);
		copyMI.addActionListener(new ActionListener() {
			
			public void actionPerformed(ActionEvent e) {
				// TODO Auto-generated method stub
					ImageSelection imgSel = new ImageSelection(grf.getCopyImage());
					Toolkit.getDefaultToolkit().getSystemClipboard().setContents(imgSel,null);
			}
		});
		copyMI2 = new JMenuItem("Copy histogram to clipboard");
		copyMI2.setFont(calibri);
		gmenu.add(copyMI2);
		copyMI2.addActionListener(new ActionListener() {
			
			public void actionPerformed(ActionEvent e) {
				// TODO Auto-generated method stub
					ImageSelection imgSel = new ImageSelection(grf.getHistogramOnlyImage());
					Toolkit.getDefaultToolkit().getSystemClipboard().setContents(imgSel,null);
			}
		});
		
		xmenu = new JPopupMenu();
		xmenus = new JMenuItem[tn.getSourceFCS().getNames().length];
		for(int i = 0; i < xmenus.length; i++)
		{
			xmenus[i] = new JMenuItem(tn.getSourceFCS().getNames()[i]);
			xmenus[i].setFont(calibri);
			xmenu.add(xmenus[i]);
			xmenus[i].addActionListener(new ActionListener(){
				public void actionPerformed(ActionEvent e) {
					node.changeXIndex(xmenu.getComponentIndex((Component) e.getSource()));
					grf.redraw_needed = true;
					
					updateUI();
					stats.refresh();
				}
			});
		}


		xmenu.setFont(new Font("Calibri",Font.PLAIN,8));
		ymenu = new JPopupMenu();
		ymenus = new JMenuItem[tn.getSourceFCS().getNames().length+1];
		for(int i = 0; i < ymenus.length; i++)
		{
			if(i != 0)
			{
				ymenus[i] = new JMenuItem(tn.getSourceFCS().getNames()[i-1]);
			}
			else
			{
				ymenus[i] = new JMenuItem("Histogram");

			}
			ymenus[i].setFont(calibri);
			ymenu.add(ymenus[i]);	
			ymenus[i].addActionListener(new ActionListener(){
				public void actionPerformed(ActionEvent e) {
					int index = ymenu.getComponentIndex((Component) e.getSource());
					if(index != 0)
					{
						node.changeYIndex(index-1);
	    				buttons.get("rgate").setEnabled(true);
	    				buttons.get("gate").setEnabled(true);
	    				buttons.get("1Dgate").setEnabled(false);
						ylog.setEnabled(true);
						if(mode == 0)
						{
		    				buttons.get("rgate").setEnabled(true);
		    				buttons.get("gate").setEnabled(true);
		    				buttons.get("1Dgate").setEnabled(true);
						}
					}
					else
					{
						node.changeYIndex(-1);
	    				buttons.get("rgate").setEnabled(false);
	    				buttons.get("gate").setEnabled(false);
	    				buttons.get("1Dgate").setEnabled(true);
						ylog.setEnabled(false);
						if(mode == 0)
						{
		    				buttons.get("rgate").setEnabled(true);
		    				buttons.get("gate").setEnabled(true);
		    				buttons.get("1Dgate").setEnabled(true);
						}
					}
					grf.redraw_needed = true;
					stats.refresh();
					updateDrawParams();
					updateUI();
				}
			});

		}
		

		ymenu.setFont(new Font("Calibri",Font.PLAIN,8));
		bottom = new JPanel();
		bottom.setLayout(new FlowLayout(FlowLayout.LEFT,1,1));
		bottom.add(gateBtn);
		bottom.add(rgateBtn);
		bottom.add(oneDgateBtn);
		bottom.add(zoomBtn);
		bottom.add(undoZoomBtn);
		buttons.put("gate",gateBtn);
		buttons.put("rgate",rgateBtn);
		buttons.put("1Dgate",oneDgateBtn);
		buttons.put("zoom",zoomBtn);
		buttons.put("undozoom",undoZoomBtn);
		Font f = calibri;
		gateBtn.setFont(f);
		zoomBtn.setFont(f);
		undoZoomBtn.setFont(f);
		xlog= new JCheckButton(Methods.getImageIcon("xlogu.png"),Methods.getImageIcon("xlogc.png"));
		ylog = new JCheckButton(Methods.getImageIcon("ylogu.png"),Methods.getImageIcon("ylogc.png"));
		xlog.setEnabled(false);
		ylog.setEnabled(false);
		gateBtn.setEnabled(false);
		rgateBtn.setEnabled(false);
		zoomBtn.setEnabled(false);
		bottom.add(xlog);
		xlog.addActionListener(new ActionListener(){

			@Override
			public void actionPerformed(ActionEvent arg0) {
				// TODO Auto-generated method stub
				if(xlog != null && node != null)
				{
					xlog.setSelected(!xlog.isSelected());
					node.xlog = xlog.isSelected();
					grf.redraw_needed=true;
					updateUI();
				}
			}


		});
		xlog.setPreferredSize(new Dimension(40,30));
		ylog.addActionListener(new ActionListener(){

			@Override
			public void actionPerformed(ActionEvent arg0) {
				// TODO Auto-generated method stub
				if(ylog != null && node != null)
				{
					ylog.setSelected(!ylog.isSelected());
					node.ylog = ylog.isSelected();
					grf.redraw_needed=true;
					updateUI();
				}
			}


		});
		ylog.setPreferredSize(new Dimension(40,30));
		bottom.add(ylog);

		setLayout(new BorderLayout(1,1));
		//add(bottom,BorderLayout.SOUTH);
		JPanel bigbottom = new JPanel(new GridLayout(2,1));
		
		stats = new StatsBox(tn);
		bigbottom.add(stats);
		bigbottom.add(bottom);
		add(bigbottom,BorderLayout.SOUTH);
		node = tn;
		initializeGraph();
		add(grf,BorderLayout.CENTER);
	}
	


	protected void updateGateListeners(Gate currentGate) {
		// TODO Auto-generated method stub
		
	}

	/**
	 * Method that should be called to updated the properties of the menus and buttons to match that of the selected TreeNode
	 */
	public void updateDrawParams() {
		// TODO Auto-generated method stub
		xlog.setSelected(node.xlog);
		ylog.setSelected(node.ylog);
		xlog.setEnabled(true);
		ylog.setEnabled(true);
		for(int i = 0; i < xmenus.length; i++){
			xmenus[i].setText(node.getSourceFCS().getNames()[i]);
			ymenus[i+1].setText(node.getSourceFCS().getNames()[i]);
		}
		enabledAllButtons();
		if(node.yindex==-1)
		{
			getButton("1Dgate").setVisible(true);
			ylog.setVisible(false);
			getButton("gate").setVisible(false);
			getButton("rgate").setVisible(false);
		
		}
		else
		{
			ylog.setVisible(true);
			getButton("1Dgate").setVisible(false);
			getButton("gate").setVisible(true);
			getButton("rgate").setVisible(true);
		}
		grf.node = node;
		grf.redraw_needed = true;
//		updateUI();
		grf.updateUI();
	}
	
	public JButton getButton(String name)
	{
		if(buttons.get(name) != null)
			return buttons.get(name);
		else
			return new JButton();
	}
	
	public void enabledAllButtons()
	{
		Iterator<JButton> itr = buttons.values().iterator();
		while(itr.hasNext())
			itr.next().setEnabled(true);
	}
	
	public void disableAllButtons()
	{
		Iterator<JButton> itr = buttons.values().iterator();
		while(itr.hasNext())
			itr.next().setEnabled(false);
	}
}
