package common;

import java.io.FileNotFoundException;
import java.io.PrintWriter;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.Scanner;

public class TreeNode {
	protected LinkedList<TreeNode> leaves = new LinkedList<TreeNode>();
	protected Gate gate;
	public boolean recalculatePoints = true; // only needs to change if this
												// node or its parent(s) changed
												// their gate or compensation
												// matrix
	protected LinkedList<Point> points;
	public boolean xlog = false;
	public boolean ylog = false;
	public int xindex = 0;
	public int yindex = 1;
	public int xmin = 1;
	public int ymin = 1;
	public int xmax = 2;
	public int ymax = 2;
	public int margin = 30;
	public int zoomx1 = -1;
	public int zoomy1 = -1;
	public int zoomx2 = -1;
	public int zoomy2 = -1;
	protected TreeNode parent;

	protected FCS sourceFCS;

	public int countChildren() {
		int count = 0;
		Iterator<TreeNode> itr = leaves.iterator();
		while (itr.hasNext())
			count += 1 + itr.next().countChildren();
		return count;
	}

	public void setParent(TreeNode parent) {
		this.parent = parent;
	}

	public boolean isRecalculatePoints() {
		return recalculatePoints;
	}

	public void setRecalculatePoints(boolean recalculatePoints) {
		this.recalculatePoints = recalculatePoints;
	}

	public boolean isXlog() {
		return xlog;
	}

	public void setXlog(boolean xlog) {
		this.xlog = xlog;
	}

	public boolean isYlog() {
		return ylog;
	}

	public void setYlog(boolean ylog) {
		this.ylog = ylog;
	}

	public int getXindex() {
		return xindex;
	}

	public void setXindex(int xindex) {
		this.xindex = xindex;
	}

	public int getYindex() {
		return yindex;
	}

	public void setYindex(int yindex) {
		this.yindex = yindex;
	}

	public int getMargin() {
		return margin;
	}

	public void setMargin(int margin) {
		this.margin = margin;
	}

	public int getZoomx1() {
		return zoomx1;
	}

	public void setZoomx1(int zoomx1) {
		this.zoomx1 = zoomx1;
	}

	public int getZoomy1() {
		return zoomy1;
	}

	public void setZoomy1(int zoomy1) {
		this.zoomy1 = zoomy1;
	}

	public int getZoomx2() {
		return zoomx2;
	}

	public void setZoomx2(int zoomx2) {
		this.zoomx2 = zoomx2;
	}

	public int getZoomy2() {
		return zoomy2;
	}

	public void setZoomy2(int zoomy2) {
		this.zoomy2 = zoomy2;
	}

	public LinkedList<TreeNode> getLeaves() {
		return leaves;
	}

	public Gate getGate() {
		return gate;
	}

	public int getXmin() {
		return xmin;
	}

	public int getYmin() {
		return ymin;
	}

	public int getXmax() {
		return xmax;
	}

	public int getYmax() {
		return ymax;
	}

	public TreeNode getParent() {
		return parent;
	}

	public void setSourceFCS(FCS sourceFCS) {
		this.sourceFCS = sourceFCS;
	}

	public TreeNode clone() {
		TreeNode clone = new TreeNode(gate.clone());
		clone.recalculatePoints = true;
		clone.xlog = xlog;
		clone.ylog = ylog;
		clone.xindex = xindex;
		clone.yindex = yindex;
		clone.xmin = xmin;
		clone.ymin = ymin;
		clone.margin = margin;
		clone.parent = parent;
		clone.sourceFCS = sourceFCS;
		clone.xmax = xmax;
		clone.ymax = ymax;
		clone.changeXIndex(xindex);
		clone.changeYIndex(yindex);
		Iterator<TreeNode> itr = leaves.iterator();
		while (itr.hasNext()) {
			clone.leaves.add(itr.next().clone());
		}
		return clone;

	}

	public void reloadDataFromFile() {

		if (this.parent == null) {
			this.sourceFCS.reloadFromFCS();
			this.propagateFromRoot();
		} else
			parent.reloadDataFromFile();
	}

	public void reloadDataFromFCS() {
		if (this.parent == null) {
			points = new LinkedList<Point>();
			try{
				points.addAll(sourceFCS.getCompensatedPoints());
			}catch(Exception e)
			{
				//exceptioNJ!
			}
			this.propagateFromRoot();
		} else
			parent.reloadDataFromFCS();
	}

	public void clearAllData() {
		// TODO Auto-generated method stub
		if (this.parent == null) {
			this.points = null;
			Iterator<TreeNode> itr = leaves.iterator();
			while (itr.hasNext()) {
				TreeNode n = itr.next();
				n.clearAllData2();
			}
		} else
			parent.clearAllData();

	}

	private void clearAllData2() {
		this.points = null;
		Iterator<TreeNode> itr = leaves.iterator();
		while (itr.hasNext()) {
			TreeNode n = itr.next();
			n.clearAllData2();
		}
	}

	public void propagateNewParent(TreeNode newParent) {
		this.parent = newParent;
		this.sourceFCS = newParent.sourceFCS;
		Iterator<TreeNode> itr = leaves.iterator();
		while (itr.hasNext()) {
			TreeNode n = itr.next();
			n.propagateNewParent(this);
		}
	}

	public FCS getSourceFCS() {
		return sourceFCS;
	}

	public TreeNode(LinkedList<Point> points, FCS source) {
		this.points = points;
		this.parent = null;
		sourceFCS = source;
		changeXIndex(xindex);
		changeYIndex(yindex);
	}

	protected TreeNode() {
		super();
	}

	public TreeNode(Gate g) {
		gate = g;
	}

	public TreeNode addAnd(Gate g) {
		TreeNode newNode = new TreeNode(g);
		newNode.initialize(this);
		leaves.add(newNode);
		return newNode;
	}

	private void changeGate(Gate newGate) {
		this.gate = newGate;
		propagateChangeFlag();
	}

	protected void initialize(TreeNode tn) {
		this.parent = tn;
		this.sourceFCS = tn.sourceFCS;
		this.propagateChangeFlag();
		this.margin = tn.margin;
		this.xindex = tn.xindex;
		this.yindex = tn.yindex;
		this.xlog = tn.xlog;
		this.ylog = tn.ylog;
		this.xmin = tn.xmin;
		this.xmax = tn.xmax;
		this.ymin = tn.ymin;
		this.ymax = tn.ymax;
		changeXIndex(xindex);
		changeYIndex(yindex);
		this.zoomx1 = tn.zoomx1;
		this.zoomx2 = tn.zoomx2;
		this.zoomy1 = tn.zoomy1;
		this.zoomy2 = tn.zoomy2;
	}

	public LinkedList<Point> getPoints() {
		//System.err.println("GET points");
		if (recalculatePoints || points == null) {
			recalculatePoints = false;
			if (parent != null) {
				LinkedList<Point> ppoints = parent.getPoints();
				if (ppoints == null) {
					//System.err.println("Reloading points");
					if (sourceFCS.getSourceFile() != null)
						reloadDataFromFile();
					else
						reloadDataFromFCS();
					ppoints = parent.getPoints();
				}
				if (gate != null)
					points = gate.getGatedPoints(ppoints);
				else
					points = ppoints;
				return points;
			} else {
				//System.err.println("No parent available!");
				points = sourceFCS.getCompensatedPoints();
				return points;
			}
		} else {
			/*
			if (points.size() > 0)
				System.err.println("Returning current points: "
						+ points.get(0).ps[0] + " recalculate = "
						+ recalculatePoints + " col " + points.get(0).color
						+ " for node: " + this);
						*/
			return points;
		}
	}

	public void propagateFromRoot() {
		if (parent == null) {
			propagateChangeFlag();
		} else
			parent.propagateFromRoot();
	}

	private void propagateChangeFlag() {
		recalculatePoints = true;
		for (int i = 0; i < leaves.size(); i++)
			leaves.get(i).propagateChangeFlag();
	}

	public void saveSelf(PrintWriter pw) {
		// TODO Auto-generated method stub
		pw.println("TREENODE:");
		if (gate != null) {
			pw.println("GATE:" + gate.name + "," + gate.x + "," + gate.y + ","
					+ gate.logx + "," + gate.logy + "," + gate.c.getRGB()
					+ ",#");
			for (int i = 0; i < gate.p.xpoints.length; i++)
				pw.println(gate.p.xpoints[i] + "," + gate.p.ypoints[i]);
			pw.println("ENDGATE");
		}

		pw.println(xlog + "," + ylog + "," + xindex + "," + yindex + "," + xmin
				+ "," + ymin + "," + xmax + "," + ymax + "," + margin);
		pw.println("ENDTREENODE:" + leaves.size());
		Iterator<TreeNode> itr = leaves.iterator();
		while (itr.hasNext())
			itr.next().saveSelf(pw);
	}

	public TreeNode(Scanner s, FCS source) {
		this.sourceFCS = source;
		xmax = source.getMaxes()[0];
		ymax = source.getMaxes()[1];
		if (s.hasNextLine()) {
			String next = s.nextLine();
			if (!next.startsWith("TREENODE:"))
				return;
			System.err.println("Loading a new treenode");
			String gates = s.nextLine();
			if (gates.startsWith("GATE:")) {
				String[] vals = gates.substring(5).split(",");
				gate = new Gate(Integer.parseInt(vals[1]),
						Integer.parseInt(vals[2]), new Boolean(vals[3]),
						new Boolean(vals[4]), vals[0],
						Integer.parseInt(vals[5]));
				System.err.println("Gate loaded: " + gate.name);
				next = s.nextLine();
				int points_added = 0;
				while (!next.startsWith("ENDGATE")) {
					String[] spliter = next.split(",");
					if (vals.length == 6
							|| vals.length == 7
							|| (vals.length == 5 && gate.logx == false && gate.logy == false))
					{
							gate.addPoint(Double.parseDouble(spliter[0]),
									Double.parseDouble(spliter[1]));
							points_added++;
					}
					else if (vals.length == 5 && gate.logx == true
							&& gate.logy == false)
					{
						gate.addPoint(
								Math.pow(10, Double.parseDouble(spliter[0])),
								Double.parseDouble(spliter[1]));
						points_added++;
					}
					else if (vals.length == 5 && gate.logx == false
							&& gate.logy == true)
					{
						gate.addPoint(Double.parseDouble(spliter[0]),
								Math.pow(10, Double.parseDouble(spliter[1])));
						points_added++;
					}
					else if (vals.length == 5 && gate.logx == true
							&& gate.logy == true)
					{
						gate.addPoint(
								Math.pow(10, Double.parseDouble(spliter[0])),
								Math.pow(10, Double.parseDouble(spliter[1])));
						points_added++;
					}

					// System.err.println(spliter[0]+","+spliter[1]);
					next = s.nextLine();
				}
				if(points_added == 1) //this is actually a 1D gate!
				{
					gate.gatetype  =2;
				}
			} else
				next = gates;

			{
				System.err.println("Endgate");
				String details = next;
				if (details.compareTo("ENDGATE") == 0)
					details = s.nextLine(); // otherwise details is already
											// loaded since we didn't have gate
											// information.
				// if we are dealing with the old way of saving (matrix for each
				// TreeNode), then we have to check if the matrix is non
				// identity and if so assign to FCS... otherwise skip
				if (details.startsWith("INVERSE:")) {
					String[] es = details.substring(9).split(" ");
					int dim = (int) Math.sqrt(es.length);
					double[][] arr = new double[source.getNames().length][source
							.getNames().length];
					for (int i = 0; i < source.getNames().length; i++)
						arr[i][i] = 1.0;
					boolean identity = true;
					for (int i = 0; i < es.length; i++) {
						double value = Double.parseDouble(es[i]);
						int row = i / dim
								+ Math.min(source.getNames().length - dim, 2);
						int col = i % dim
								+ Math.min(source.getNames().length - dim, 2);
						arr[row][col] = value;
						if ((value != 1 && row == col)
								|| (value != 0 && row != col))
							identity = false;
					}
					if (!identity)
						source.setInverse(new Matrix(arr));
					details = s.nextLine();
				}
				String[] split = details.split(",");
				if(split.length==10)
				{
					Integer.parseInt(split[0]);
					xlog = new Boolean(split[1]);
					ylog = new Boolean(split[2]);
					xindex = Integer.parseInt(split[3]);
					yindex = Integer.parseInt(split[4]);
					xmin = Integer.parseInt(split[5]);
					ymin = Integer.parseInt(split[6]);
					xmax = Integer.parseInt(split[7]); 
					ymax = Integer.parseInt(split[8]);
					margin = Integer.parseInt(split[9]);
				}
				else
				{
					xlog = new Boolean(split[0]);
					ylog = new Boolean(split[1]);
					xindex = Integer.parseInt(split[2]);
					yindex = Integer.parseInt(split[3]);
					xmin = Integer.parseInt(split[4]);
					ymin = Integer.parseInt(split[5]);
					xmax = Integer.parseInt(split[6]); 
					ymax = Integer.parseInt(split[7]);
					margin = Integer.parseInt(split[8]);
				}

				String children = s.nextLine();
				if (children.startsWith("ENDTREENODE:")) {
					int count = Integer.parseInt(children.substring(12));
					System.err.println("Loading " + count + " children");
					for (int i = 0; i < count; i++) {
						TreeNode ntn = new TreeNode(s, source);
						ntn.parent = this;
						leaves.add(ntn);
					}
				} else {
					System.err.println("NO ENDTREENODE");
				}
			}
		}
	}

	public void changeXIndex(int newindex) {
		xindex = newindex;
		if (newindex > -1) {
			if (sourceFCS.getMaxes() != null)
				xmax = sourceFCS.getMaxes()[newindex];
		}
	}

	public void changeYIndex(int newindex) {
		yindex = newindex;
		if (newindex > -1) {
			if (sourceFCS.getMaxes() != null)
				ymax = sourceFCS.getMaxes()[newindex];
		}
	}

	public DoubleDouble getTranslatedPoint(double clickx, double clicky,
			Graph grf) {
		double pointx = 0;
		double pointy = 0;
		if (xlog) {
			double minpower = Math.log10(xmin);
			double maxpower = Math.log10(xmax);
			double f1 = (clickx - grf.ymargin) / (grf.getWidth() - grf.ymargin);
			f1 = Math.min(1, Math.max(0, f1));
			pointx = Math.pow(10, f1 * (maxpower - minpower) + minpower);
		} else {
			double minpower = xmin;
			double maxpower = xmax;
			double f1 = ((double) clickx - grf.ymargin) / (grf.getWidth() - grf.ymargin);
			f1 = Math.min(1, Math.max(0, f1));
			pointx = (f1 * (maxpower - minpower) + minpower);
		}
		if(yindex!=-1)
		{
			if (ylog) {
				double minpower = Math.log10(ymin);
				double maxpower = Math.log10(ymax);
				double f1 = ((double) grf.getHeight() - clicky - grf.xmargin)
						/ (grf.getHeight() - grf.xmargin);
				f1 = Math.min(1, Math.max(0, f1));
				pointy = Math.pow(10, f1 * (maxpower - minpower) + minpower);
			} else {
				double minpower = ymin;
				double maxpower = ymax;
				double f1 = ((double) grf.getHeight() - clicky - grf.xmargin)
						/ (grf.getHeight() - grf.xmargin);
				f1 = Math.min(1, Math.max(0, f1));
				pointy = (f1 * (maxpower - minpower) + minpower);
			}
		}
		else
			pointy=clicky;
		return new DoubleDouble(pointx, pointy);
	}

	public double getAverageX() {
		LinkedList<Point> pts = getPoints();
		double average = 0;
		for(Point p : pts)
		{
			average+=p.ps[xindex];
		}
		return average/pts.size();
	}
	
	public double getAverageY() {
		LinkedList<Point> pts = getPoints();
		double average = 0;
		for(Point p : pts)
		{
			average+=p.ps[yindex];
		}
		return average/pts.size();
	}
	
	public double getVarX() {
		double mean = getAverageX();
		LinkedList<Point> pts = getPoints();
		double var = 0;
		for(Point p : pts)
		{
			var+=(mean-p.ps[xindex])*(mean-p.ps[xindex]);
		}
		return var/pts.size();
	}
	
	public double getSdX()
	{
		return Math.sqrt(getVarX());
	}

}
